/*____________________________________________________________________________
		Copyright (C) 2002 PGP Corporation
        All rights reserved.

        $Id: CPGPdiskFormatter.cpp,v 1.16 2002/09/26 01:32:11 wjb Exp $
____________________________________________________________________________*/

#include "pgpClassesConfig.h"
#include <shlobj.h>

#include "CMutex.h"
#include "UMath.h"
#include "UTime.h"

#include "CPath.h"
#include "USecurity.h"
#include "UUnicode.h"
#include "UWinVersion.h"

#include "CPGPdiskFormatter.h"

_USING_PGP

_UNNAMED_BEGIN

// Constants

const PGPUInt32	kMaxBlocksPerClear	= 100;
const PGPUInt32	kMaxLockRetries		= 20;
const PGPUInt32	kMaxVolNameLength	= 11;
const PGPUInt32	kLockRetryMsDelay	= 500;

const char	*kFat12Or16FSString	= "FAT";

const char	*kDefaultOEMName	= "PGPDISK ";
const char	*kDefaultVolLabel	= "UNNAMED    ";

const char	*kInvalidVolNameChars		= "\"&*+./:;<=>?[\\]^|";
const char	kInvalidVolCharReplacement	= '_';

_UNNAMED_END


// Class CPGPdiskFormatter member variables

NewDiskStatusFuncType	CPGPdiskFormatter::mUserCallback;
void					*CPGPdiskFormatter::mUserValue;
CComboError				CPGPdiskFormatter::mError;


// Class CPGPdiskFormatter member functions

void 
CPGPdiskFormatter::Format(
	const CPGPdiskContext&	context, 
	const char				*root, 
	const char				*volName, 
	FileSys::Type			fsType, 
	NewDiskStatusFuncType	userCallback, 
	void					*userValue)
{
	pgpAssertStrValid(root);
	pgpAssertStrValid(volName);
	pgpAssertAddrValid(userCallback, FormatDiskStatusFuncType);

	// Root must refer to a mounted PGPdisk volume.
	CPGPdiskDiskSet		diskSet(context);
	CPGPdiskDiskIter	diskIter(diskSet);
	CPGPdiskDisk		disk;

	if (!diskIter.SearchOnRoot(root, disk))
		THROW_PGPERROR(kPGPError_BadParams);

	// Fixup volume name.
	CString	cvolName(volName);
	CanonicalizeVolumeName(cvolName);

	// Perform the conversion.
	if (UWinVersion::IsWin95Compatible() || 
		(UWinVersion::IsWinNT4Compatible() && 
		 !USecurity::IsCurrentUserAdmin()))
	{
		FormatCustomFat(disk, root, cvolName, fsType, userCallback, 
			userValue);
	}
	else if (UWinVersion::IsWinNT4Compatible())
	{
		FormatWinNT(disk, root, cvolName, fsType, userCallback, userValue);
	}
	else
	{
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);
	}

	// Make sure volume label is set.
	for (PGPUInt32 i = 0; i < 10; i++)
	{
		if (SetVolumeLabel(root, cvolName))
			break;

		Sleep(250);
	}
}

void 
CPGPdiskFormatter::GetValidFilesystems(
	PGPUInt64				blocksDisk, 
	CArray<FileSysInfo>&	fsInfos, 
	PGPUInt32&				numFsInfos)
{
	PGPUInt64	megsDisk	= (blocksDisk*kPGPdiskBlockSize) / 
		PFLConstants::kBytesPerMb;

	fsInfos.Resize(3);
	numFsInfos = 0;

	if (UWinVersion::IsWin95Compatible())
	{
		if (UFileSys::IsFileSysValidForSize(FileSys::kFat12Or16FileSys, 
			megsDisk))
		{
			fsInfos[numFsInfos].fsType = FileSys::kFat12Or16FileSys;
			fsInfos[numFsInfos].fsString = GetFSStringForFormat(
				FileSys::kFat12Or16FileSys);

			numFsInfos++;
		}

		if (UFileSys::IsFileSysValidForSize(FileSys::kFat32FileSys, megsDisk) && 
			UWinVersion::IsWin95OSR2Compatible())
		{
			fsInfos[numFsInfos].fsType = FileSys::kFat32FileSys;
			fsInfos[numFsInfos].fsString = GetFSStringForFormat(
				FileSys::kFat32FileSys);

			numFsInfos++;
		}
	}
	else if (UWinVersion::IsWinNT4Compatible())
	{
		if (UFileSys::IsFileSysValidForSize(FileSys::kFat12Or16FileSys, 
			megsDisk))
		{
			fsInfos[numFsInfos].fsType = FileSys::kFat12Or16FileSys;
			fsInfos[numFsInfos].fsString = GetFSStringForFormat(
				FileSys::kFat12Or16FileSys);

			numFsInfos++;
		}

		if (UFileSys::IsFileSysValidForSize(FileSys::kFat32FileSys, megsDisk) && 
			UWinVersion::IsWin2000Compatible())
		{
			// Ugly departure from Nick's pretty (complicated :-) coding
			// style. W2K can't format FAT32 volumes over 32 GB. 
			// See MSKB Q253726. No reasons why given. The world 
			// may never know..  -wjb
			if(megsDisk<32768)
			{
				fsInfos[numFsInfos].fsType = FileSys::kFat32FileSys;
				fsInfos[numFsInfos].fsString = GetFSStringForFormat(
					FileSys::kFat32FileSys);

				numFsInfos++;
			}
		}
		
		if (UFileSys::IsFileSysValidForSize(FileSys::kNTFSFileSys, 
			megsDisk) && 
			USecurity::IsCurrentUserAdmin() &&
			!(UWinVersion::IsWin2000Compatible() && 
			  USecurity::IsRemoteSession()))
		{
			fsInfos[numFsInfos].fsType = FileSys::kNTFSFileSys;
			fsInfos[numFsInfos].fsString = GetFSStringForFormat(
				FileSys::kNTFSFileSys);

			numFsInfos++;
		}
	}
}

const char * 
CPGPdiskFormatter::GetFSStringForFormat(FileSys::Type fsType)
{
	switch (fsType)
	{
	case FileSys::kFat12FileSys:
	case FileSys::kFat16FileSys:
	case FileSys::kFat12Or16FileSys:
		return kFat12Or16FSString;

	case FileSys::kFat32FileSys:
		return FileSys::kFat32IdStr;

	case FileSys::kNTFSFileSys:
		return FileSys::kNTFSIdStr;

	default:
		pgpAssert(FALSE);
		return "";
	}
}

void 
CPGPdiskFormatter::CanonicalizeVolumeName(CString& volName) const
{
	if (volName.Length() > kMaxVolNameLength)
	{
		volName.Resize(kMaxVolNameLength + 1);
		volName.SetAt(kMaxVolNameLength, '\0');
	}

	PGPUInt32	lengthStr	= volName.Length();
	PGPUInt32	lengthInv	= strlen(kInvalidVolNameChars);

	for (PGPUInt32 i = 0; i < lengthStr; i++)
	{
		for (PGPUInt32 j = 0; j < lengthInv; j++)
		{
			if (volName[i] == kInvalidVolNameChars[j])
				volName.SetAt(i, kInvalidVolCharReplacement);
		}
	}
}

void 
CPGPdiskFormatter::ClearBlocks(
	CPGPdiskDisk&			disk, 
	PGPUInt64				startBlock, 
	PGPUInt32				numBlocks, 
	NewDiskStatusFuncType	userCallback, 
	void					*userValue) const
{
	PGPUInt32	blocksPerClear	= pgpMin(numBlocks/10, kMaxBlocksPerClear);

	CArray<PGPUInt8>	blanks(blocksPerClear * kPGPdiskBlockSize);
	blanks.Wipe();

	// Calculate information used for progress updates.
	PGPUInt32	totalWrites	= static_cast<PGPUInt32>(
		UMath::CeilDiv<PGPUInt64>(numBlocks, blocksPerClear));

	PGPUInt32	writesDone		= 0;
	PGPUInt32	writesPerUpdate	= UMath::CeilDiv<PGPUInt32>(numBlocks, 100);

	for (PGPUInt64 i = startBlock; i < numBlocks; i += blocksPerClear)
	{
		disk.WriteVolume(blanks.Get(), i, blocksPerClear);

		PGPBoolean	continueFormat	= TRUE;

		if (writesDone++ % writesPerUpdate == 0)
		{
			continueFormat = userCallback(TRUE, writesDone*100/totalWrites, 
				userValue);
		}

		if (!continueFormat)
			THROW_PGPERROR(kPGPError_UserAbort);
	}
}

void 
CPGPdiskFormatter::InitBootSectorFat12(
	PGPUInt32					blocksDisk, 
	const UFileSys::FatData&	fatData, 
	FileSys::BootSectorFat12&	bs12) const
{
	pgpClearMemory(&bs12, sizeof(bs12));

	PGPUInt64	megsDisk	= static_cast<PGPUInt32>(
		UFileSys::CalcMegsDisk(blocksDisk, kPGPdiskBlockSize));

	PGPUInt32	secsPerTrack, tracksPerCyl;
	PGPUInt64	cylinders;

	UFileSys::CalcFakeFatGeom(kPGPdiskBlockSize, tracksPerCyl, cylinders, 
		secsPerTrack);
	
	bs12.bsJump[0]	= 0xEB;		// fill in jmp instruction to look real
	bs12.bsJump[1]	= 0x3C;
	bs12.bsJump[2]	= 0x90;

	strncpy(bs12.bsOemName, kDefaultOEMName, sizeof(bs12.bsOemName));

	bs12.bsBytesPerSec		= kPGPdiskBlockSize;		// bytes per sector
	bs12.bsSecPerClust		= 1;						// sectors per cluster
	bs12.bsResSectors		= fatData.fdReservedSecs;	// reserved sectors
	bs12.bsFats				= fatData.fdFatCount;		// number of FATs
	bs12.bsRootDirEnts		= fatData.fdRootDirEnts;	// entries in root dir
	bs12.bsSectors			= blocksDisk;
	bs12.bsMedia			= FileSys::kFatMediaByte;	// a hard disk
	bs12.bsFatSecs			= static_cast<PGPUInt16>(fatData.fdFatSize);
	bs12.bsSecPerTrack		= static_cast<PGPUInt16>(secsPerTrack);
	bs12.bsHeads			= static_cast<PGPUInt16>(tracksPerCyl);
	bs12.bsHiddenSecs		= 0;						// no hidden sectors
	bs12.bsHugeSectors		= 0;
	bs12.bsDriveNumber		= FileSys::kFatHardDriveId;	// a hard drive
	bs12.bsBootSignature	= FileSys::kFatFirstBootSig;
	bs12.bsVolumeId			= static_cast<PGPUInt32>(UTime::GetSystemTicks());
	bs12.bsSignature		= FileSys::kFatSecondBootSig;

	strncpy(bs12.bsVolumeLabel, kDefaultVolLabel, 
		sizeof(bs12.bsVolumeLabel));
	memset(bs12.bsFileSysType, ' ', sizeof(bs12.bsFileSysType));

	strncpy(bs12.bsFileSysType, FileSys::kFat12IdStr, 
		strlen(FileSys::kFat12IdStr));
}

void 
CPGPdiskFormatter::InitBootSectorFat16(
	PGPUInt32					blocksDisk, 
	const UFileSys::FatData&	fatData, 
	FileSys::BootSectorFat16&	bs16) const
{
	pgpClearMemory(&bs16, sizeof(bs16));

	PGPUInt32	megsDisk	= static_cast<PGPUInt32>(
		UFileSys::CalcMegsDisk(blocksDisk, kPGPdiskBlockSize));

	PGPUInt32	secsPerTrack, tracksPerCyl;
	PGPUInt64	cylinders;

	UFileSys::CalcFakeFatGeom(kPGPdiskBlockSize, tracksPerCyl, cylinders, 
		secsPerTrack);

	bs16.bsJump[0]	= 0xEB;		// fill in jmp instruction to look real
	bs16.bsJump[1]	= 0x3C;
	bs16.bsJump[2]	= 0x90;

	strncpy(bs16.bsOemName, kDefaultOEMName, sizeof(bs16.bsOemName));

	bs16.bsBytesPerSec		= kPGPdiskBlockSize;		// bytes per sector
	bs16.bsSecPerClust		= static_cast<PGPUInt8>(fatData.fdSpc);
	bs16.bsResSectors		= fatData.fdReservedSecs;	// reserved sectors
	bs16.bsFats				= fatData.fdFatCount;		// number of FATs
	bs16.bsRootDirEnts		= fatData.fdRootDirEnts;	// entries in root dir
	bs16.bsSectors			= (megsDisk < 32 ? blocksDisk : 0);
	bs16.bsMedia			= FileSys::kFatMediaByte;	// a hard disk
	bs16.bsFatSecs			= static_cast<PGPUInt16>(fatData.fdFatSize);
	bs16.bsSecPerTrack		= static_cast<PGPUInt16>(secsPerTrack);
	bs16.bsHeads			= static_cast<PGPUInt16>(tracksPerCyl);
	bs16.bsHiddenSecs		= 0;						// no hidden sectors
	bs16.bsHugeSectors		= (megsDisk >= 32 ? blocksDisk : 0);
	bs16.bsDriveNumber		= FileSys::kFatHardDriveId;	// a hard drive
	bs16.bsBootSignature	= FileSys::kFatFirstBootSig;
	bs16.bsVolumeId			= static_cast<PGPUInt32>(UTime::GetSystemTicks());
	bs16.bsSignature		= FileSys::kFatSecondBootSig;

	strncpy(bs16.bsVolumeLabel, kDefaultVolLabel, 
		sizeof(bs16.bsVolumeLabel));

	memset(bs16.bsFileSysType, ' ', sizeof(bs16.bsFileSysType));
	strncpy(bs16.bsFileSysType, FileSys::kFat16IdStr, 
		strlen(FileSys::kFat16IdStr));
}

void 
CPGPdiskFormatter::InitBootSectorFat32(
	PGPUInt32					blocksDisk, 
	const UFileSys::FatData&	fatData, 
	FileSys::BootSectorFat32&	bs32, 
	FileSys::BigFatBootFSInfo&	bfInfo) const
{
	pgpClearMemory(&bs32, sizeof(bs32));
	pgpClearMemory(&bfInfo, sizeof(bfInfo));

	PGPUInt32	megsDisk	= static_cast<PGPUInt32>(
		UFileSys::CalcMegsDisk(blocksDisk, kPGPdiskBlockSize));
	PGPUInt64	bytesDisk	= blocksDisk * kPGPdiskBlockSize;

	PGPUInt32	secsPerTrack, tracksPerCyl;
	PGPUInt64	cylinders;

	UFileSys::CalcFakeFatGeom(kPGPdiskBlockSize, tracksPerCyl, cylinders, 
		secsPerTrack);

	bs32.bsJump[0]	= 0xEB;		// fill in jmp instruction to look real
	bs32.bsJump[1]	= 0x3C;
	bs32.bsJump[2]	= 0x90;

	strncpy(bs32.bsOemName, kDefaultOEMName, sizeof(bs32.bsOemName));

	bs32.bsBytesPerSec		= kPGPdiskBlockSize;		// bytes per sector
	bs32.bsSecPerClust		= static_cast<PGPUInt8>(fatData.fdSpc);
	bs32.bsResSectors		= fatData.fdReservedSecs;	// reserved sectors
	bs32.bsFats				= fatData.fdFatCount;		// number of FATs
	bs32.bsRootDirEnts		= 0;
	bs32.bsSectors			= 0;
	bs32.bsMedia			= FileSys::kFatMediaByte;	// a hard disk
	bs32.bsFatSecs			= 0;
	bs32.bsSecPerTrack		= static_cast<PGPUInt32>(secsPerTrack);
	bs32.bsHeads			= static_cast<PGPUInt32>(tracksPerCyl);
	bs32.bsHiddenSecs		= 0;						// no hidden sectors
	bs32.bsHugeSectors		= blocksDisk;

	bs32.bsBigSectorsPerFat	= fatData.fdFatSize;		// total secs per FAT
	bs32.bsExtFlags			= 0;						// extended flags
	bs32.bsFS_Version		= 0;						// filesystem version
	bs32.bsRootDirStrtClus	= FileSys::kTypicalRootDirStart;
	bs32.bsFsInfoSec		= FileSys::kTypicalBigFatStart;
	bs32.bsBkUpBootSec		= -1;						// no backup boot sec

	bs32.bsDriveNumber		= FileSys::kFatHardDriveId;	// a hard drive
	bs32.bsBootSignature	= FileSys::kFatFirstBootSig;
	bs32.bsVolumeId			= static_cast<PGPUInt32>(UTime::GetSystemTicks());
	bs32.bsSignature		= FileSys::kFatSecondBootSig;

	strncpy(bs32.bsVolumeLabel, kDefaultVolLabel, 
		sizeof(bs32.bsVolumeLabel));

	memset(bs32.bsFileSysType, ' ', sizeof(bs32.bsFileSysType));
	strncpy(bs32.bsFileSysType, FileSys::kFat32IdStr, 
		strlen(FileSys::kFat32IdStr));

	// Finally initialize the BigFatBootInfo structure.
	bfInfo.bfSecSig					= FileSys::kBigFatSecSig;
	bfInfo.bfFSInf_Sig				= FileSys::kBigFatSig;
	bfInfo.bfFSInf_next_free_clus	= FileSys::kTypicalRootDirStart;
	bfInfo.bsSignature				= FileSys::kFatSecondBootSig;

	bfInfo.bfFSInf_free_clus_cnt	= static_cast<PGPUInt32>(
		bytesDisk / (fatData.fdSpc * kPGPdiskBlockSize) + 2);
}

void 
CPGPdiskFormatter::FormatCustomFat(
	CPGPdiskDisk&			disk, 
	const char				*root, 
	const char				*volName, 
	FileSys::Type			fsType, 
	NewDiskStatusFuncType	userCallback, 
	void					*userValue)
{
	if (!UFileSys::IsFatFileSys(fsType))
		THROW_PGPERROR(kPGPError_BadParams);

	// Get number of blocks on disk.
	PGPUInt64	bigBlocksDisk	= disk.GetBlocksDisk();

	if (UMath::GetHighDWord(bigBlocksDisk) != 0)
		THROW_PGPERROR(kPGPError_BadParams);

	PGPUInt32	blocksDisk	= UMath::GetLowDWord(bigBlocksDisk);
	PGPUInt32	megsDisk	= static_cast<PGPUInt32>(
		UFileSys::CalcMegsDisk(blocksDisk, kPGPdiskBlockSize));

	if (!UFileSys::IsFileSysValidForSize(fsType, megsDisk))
		THROW_PGPERROR(kPGPError_BadParams);

	// Initialize some FAT data.
	if (fsType == FileSys::kFat12Or16FileSys)
	{
		fsType = ((megsDisk < FileSys::kMinFat16Megs) ? 
			FileSys::kFat12FileSys : FileSys::kFat16FileSys);
	}

	UFileSys::FatData	fatData;
	UFileSys::InitFatData(blocksDisk, fsType, fatData);

	// Lock the volume for format. Retry since volume may have just been
	// mounted and may not be 'ready' yet.

	PGPUInt32	retryCount	= 0;

	while (TRUE)
	{
		try
		{
			disk.LockVolume(TRUE);
			break;
		}
		catch (CComboError& caughtErr)
		{
			Sleep(kLockRetryMsDelay);

			if (retryCount++ > kMaxLockRetries)
			{
				// Give up.
				throw caughtErr;
			}
		}
	}

	// Clear filesystem blocks.
	ClearBlocks(disk, 0, fatData.fdFirstSecData, userCallback, userValue);

	// Initialize and write out the FAT data structures.
	CArray<PGPUInt8>	blockBuf(kPGPdiskBlockSize);
	blockBuf.Wipe();

	PGPUInt32	*bufIn32	= reinterpret_cast<PGPUInt32 *>(blockBuf.Get());

	if (fsType == FileSys::kFat12FileSys)
	{
		FileSys::BootSectorFat12	bs12;

		InitBootSectorFat12(static_cast<PGPUInt32>(blocksDisk), fatData, 
			bs12);

		// Write boot sector.
		pgpCopyMemory(&bs12, blockBuf.Get(), sizeof(bs12));
		disk.WriteVolume(blockBuf.Get(), 0, 1);

		// Prepare fat ID sector.
		blockBuf.Wipe();
		bufIn32[0] = FileSys::kFat12Sig;
	}
	else if (fsType == FileSys::kFat16FileSys)
	{
		FileSys::BootSectorFat16	bs16;

		InitBootSectorFat16(static_cast<PGPUInt32>(blocksDisk), fatData, 
			bs16);

		// Write boot sector.
		pgpCopyMemory(&bs16, blockBuf.Get(), sizeof(bs16));
		disk.WriteVolume(blockBuf.Get(), 0, 1);

		// Prepare fat ID sector.
		blockBuf.Wipe();
		bufIn32[0] = FileSys::kFat16Sig;
	}
	else if (fsType == FileSys::kFat32FileSys)
	{
		FileSys::BigFatBootFSInfo	bfInfo;
		FileSys::BootSectorFat32	bs32;

		InitBootSectorFat32(static_cast<PGPUInt32>(blocksDisk), fatData, 
			bs32, bfInfo);

		// Write boot sector.
		pgpCopyMemory(&bs32, blockBuf.Get(), sizeof(bs32));
		disk.WriteVolume(blockBuf.Get(), 0, 1);

		// Write big fat boot sector info.
		blockBuf.Wipe();

		pgpCopyMemory(&bfInfo, blockBuf.Get(), sizeof(bfInfo));
		disk.WriteVolume(blockBuf.Get(), bs32.bsFsInfoSec, 1);

		// Prepare fat ID sector.
		blockBuf.Wipe();
		bufIn32[0] = FileSys::kFat32Clust1;
		bufIn32[1] = FileSys::kFat32Clust2;
		bufIn32[2] = FileSys::kFat32Clust3;
	}

	// Write FAT ids we just prepared;
	PGPUInt64	fatSec	= fatData.fdReservedSecs;

	for (PGPUInt32 i = 0; i < fatData.fdFatCount; i++)
	{
		disk.WriteVolume(blockBuf.Get(), fatSec, 1);
		fatSec += fatData.fdFatSize;
	}

	// Unlock the volume.
	disk.UnlockVolume();
}

BOOLEAN 
_stdcall 
CPGPdiskFormatter::FormatExCallback(
	UDynLink::FormatExCallbackCommand	command, 
	DWORD								subAction, 
	PVOID								actionInfo)
{
	PGPUInt32	percent;

	switch (command)
	{
	case UDynLink::kFMIFS_ProgressCmd:
		percent = static_cast<PGPUInt32>(
			*static_cast<PGPUInt32 *>(actionInfo) * 0.90);

		if (!mUserCallback(TRUE, percent, mUserValue))
			mError.pgpErr = kPGPError_UserAbort;
		break;

	case UDynLink::kFMIFS_Done:
		if (!*reinterpret_cast<PGPBoolean *>(actionInfo))
			mError.pgpErr = kPGPError_VolumeOpFailed; // NETABUG format error?
		else
			mUserCallback(TRUE, 100, mUserValue);
		break;

	case UDynLink::kFMIFS_InsufficientRights:
		mError.pgpErr = kPGPError_VolumeOpFailed; // NETABUG format error?
		break;

	default:
		break;
	}

	return mError.IsntError();
}

void 
CPGPdiskFormatter::FormatWinNT(
	CPGPdiskDisk&			disk, 
	const char				*root, 
	const char				*volName, 
	FileSys::Type			fsType, 
	NewDiskStatusFuncType	userCallback, 
	void					*userValue)
{
	pgpAssert(UWinVersion::IsWinNT4Compatible());

	// Get number of blocks on disk.
	PGPUInt64	blocksDisk	= disk.GetBlocksDisk();
	PGPUInt64	megsDisk	= UFileSys::CalcMegsDisk(
		blocksDisk, kPGPdiskBlockSize);

	if (!UFileSys::IsFileSysValidForSize(fsType, megsDisk))
		THROW_PGPERROR(kPGPError_BadParams);

	// get the appropriate root
	CPath	croot(root);

	if (UWinVersion::IsWin2000Compatible())
	{
		croot.EndInSlash();

		PGPBoolean	succeeded	= 
			UDynLink::Win2k_GetVolumeNameForVolumeMountPoint(croot, 
			croot.GetBuffer(kPGPdiskMaxPathLength), kPGPdiskMaxPathLength);
		croot.ReleaseBuffer();

		if (!succeeded)
			THROW_PGPERROR(kPGPError_VolumeOpFailed);

		croot.DontEndInSlash();
	}

	// convert to unicode
	CArray<WCHAR>	uniFSString, uniRoot, uniVolName;

	UUnicode::AsciiToUni(GetFSStringForFormat(fsType), uniFSString);
	UUnicode::AsciiToUni(croot, uniRoot);
	UUnicode::AsciiToUni(volName, uniVolName);

	// Format the drive.
	mUserCallback = userCallback;
	mUserValue = userValue;
	mError = CComboError();
	
	UDynLink::WinNT_FormatEx(uniRoot.Get(), 
		UDynLink::kFMIFS_HardDiskType, uniFSString.Get(), 
		uniVolName.Get(), FALSE, 0, FormatExCallback);

	THROW_IF_ERROR(mError);
}
