////////////////////////////////////////////////////////////////////////////////
//  File : CommandFactory.cpp
//
//  Copyright (C) 2002 PGP Corporation
//
//  ABSTRACT
//		Create CommandFactory object.
// 
//  Author: Satya S. Das
//
////////////////////////////////////////////////////////////////////////////////
// include files 
#include <crtdbg.h>
#include "pgpdebug.h"
#include "windows.h"
#include "shlwapi.h"
#include "gwdll.h"
#include "gwoapi.h"
#include "gwc3po.h"
#include "gwc3cmd.h"
#include "CommandFactory.h"
#include "GwCommand.h"
#include "util.h"
#include "ofcmndr.h"
#include "globals.h"
#include "resource.h"
#include "globals.h"
#include "globdefs.h"
#include "pgpclientprefs.h"

// Globals 
int	AcceptFlag=0;

//external functions
extern HRESULT Publish(LPCSTR szTokenBuf, BSTR *pbstrResult);

// external data 
extern	IC3POManager *m_pIMgr;

// internal functions 
IUnknown * BuildIUnkDispatch(REFCLSID libid, REFCLSID iid, IUnknown *pIUnk, void *pIFac);


// private functions 
//////////////////////////////////////////////////////////////////
//  Name:CCommandFactory::CCommandFactory				
//
//  Description:CCommandFactory constructor 
//
//  In:none
//
//  Out:none
//
//  Comments:		
//////////////////////////////////////////////////////////////////
CCommandFactory::CCommandFactory(IUnknown *pUnk)
{
	PgpGwTrace("CCommandFactory::CCommandFactory\n");
	m_pUnkOuter = pUnk;
	m_pIUnkDispFact = NULL;
}

//////////////////////////////////////////////////////////////////
//  Name:CCommandFactory::~CCommandFactory				
//
//  Description:CCommandFactory destructor	
//
//  In:none
//
//  Out:none
//
//  Comments:		
//////////////////////////////////////////////////////////////////
CCommandFactory::~CCommandFactory()
{
	PgpGwTrace("CCommandFactory::~CCommandFactory\n");
	if(m_pIUnkDispFact)
	{
		m_pIUnkDispFact->Release();
		m_pIUnkDispFact=NULL;
	}
}

//////////////////////////////////////////////////////////////////
//  Name:CCommandFactory::QueryInterface				
//
//  Description:QueryInterface for CCommandFactory
//
//  In:REFIID riid -  Reference id
//
//  Out:LPVOID FAR* ppv - 
//
//  Comments:		
////////////////////////////////////////////////////////////////////
STDMETHODIMP CCommandFactory::QueryInterface(REFIID riid, LPVOID FAR* ppv)
{
	PgpGwTrace("CCommandFactory::QueryInterface\n");
	if(NULL == ppv)
		return E_INVALIDARG;

    *ppv=NULL;


	if((IID_ICommandFactory == riid ) || (IID_IUnknown == riid))
	{
		*ppv  = this;
		((LPUNKNOWN)*ppv)->AddRef();
		return S_OK;
	}

	if(IID_IDispatch == riid)
	{
		if (NULL == m_pIUnkDispFact)
		{
			m_pIUnkDispFact = BuildIUnkDispatch(LIBID_PgpGwTypeLibrary,
				IID_ICommandFactory, (IUnknown *)(ICommandFactory *)this,
				(void *)(ICommandFactory *)this);
		}

		if (NULL != m_pIUnkDispFact)
			return m_pIUnkDispFact->QueryInterface(IID_IDispatch, ppv);
	}

	return E_NOINTERFACE;
}

//////////////////////////////////////////////////////////////////
//  Name:			CCommandFactory::BuildCommand				
//
//  Description:This method is invoked to create a Command object. 
//				The C3po cannot assume that Command::WantComand 
//				has been called before this method is invoked.
//	
//
//  In:	BSTR bstrContext - This is the context for the command. The 
//		context string contains either the class of a GW data object
//		(e.g. GW.MESSAGE.MAIL.X) or the class of the specific user 
//		interface(e.g. GW.APPLICATION.BROWSER).
//
//		BSTR bstrPersistentID - This can be either a "pre-defined" ID 
//		or is an ID provided in a previous run of GW.
//
//		IDispatch * pIDispBaseCommand - The command object of the 
//		command being overridden.
//
//		IDispatch * pIDispParameters - The parameter list of the command.
//
//  Out:IDispatch * FAR* ppIDispGWCommand - Command object for persistentID.
//
//  Comments:Each time the BuildCommand is invoked, the C3po should 
//			 return a new instance of a GWCommand object. This is 
//			 neccessary because BuildCommand can be called recursively.
////////////////////////////////////////////////////////////////////
STDMETHODIMP CCommandFactory::BuildCommand(BSTR bstrContext, 
			BSTR bstrPersistentID, IDispatch * pIDispBaseCommand, 
			IDispatch * pIDispParameters, IDispatch * FAR* ppIDispGWCommand)
{
	PgpGwTrace("CCommandFactory::BuildCommand(context=%S, id=%S)\n",
					bstrContext, bstrPersistentID);

	pgpAssert(NULL != pIDispBaseCommand);
	pgpAssert(NULL != ppIDispGWCommand);

	pgpAssert(0 == _wcsicmp(bstrContext, MAILMESSAGECONTEXTW));//should be a mail message
	pgpAssert(0 == _wcsicmp(bstrPersistentID, wceGW_CMDID_OPEN));//should be the open command

	HRESULT hrRet=S_OK;
	IGWCommand *pigwBaseCmd=NULL;
	CGWCommand *pcgwCmdTemp=NULL;
	if((NULL == pIDispBaseCommand) || (NULL == ppIDispGWCommand))
	{
		hrRet=E_FAIL;
		goto cleanup;
	}

	//check if we have expired. if so we do not show any commands
	if(TRUE == g_bPlugInExpired)
	{
		hrRet=E_FAIL;
		goto cleanup;
	}

	try
	{
		//create a new command object. 
		pcgwCmdTemp= new CGWCommand(IDM_MAIL_OPEN, CUIO_IS_AN_OPERATION);
	}
	catch(...)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}
	pgpAssert(NULL != pcgwCmdTemp);
	if(NULL == pcgwCmdTemp)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	//request command interface pointer from the dispatch passed
	hrRet=pIDispBaseCommand->QueryInterface(IID_IGWCommand, (void**)&pigwBaseCmd);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pigwBaseCmd));
	if(FAILED(hrRet) || (NULL == pigwBaseCmd))
		goto cleanup;


	//put the base command dispatch in the object
	pcgwCmdTemp->Set_BaseCommand(pigwBaseCmd);

    //request a dispatch interface. GW frees up this later. evidently it does not
	//do addref and assumes it is already addrefed
	hrRet=pcgwCmdTemp->QueryInterface(IID_IDispatch, (void**)ppIDispGWCommand);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != *ppIDispGWCommand));
	if(FAILED(hrRet) || (NULL == *ppIDispGWCommand))
		goto cleanup;
	
cleanup:
	
	return S_OK;
}

//////////////////////////////////////////////////////////////////
//  Name:CCommandFactory::WantCommand				
//
//  Description:This method is used to query the C3po for it's 
//				intention to support a predefined command.
//
//  In:BSTR bstrContext This is the actual object class associates 
//						with the command. 
//	   BSTR bstrPersistentID This can be either a "pre-defined"
//          ID or is an ID provided in a previous run of GW. 
//
//  Out:VARIANT_BOOL FAR* pbChanged
//			TRUE - I do want to support this predefined command
//			FALSE - I do not want to support this predefined command
//
//  Comments:		Not Implemented		
////////////////////////////////////////////////////////////////////
STDMETHODIMP CCommandFactory::WantCommand(BSTR bstrContext, 
					BSTR bstrPersistentID, VARIANT_BOOL FAR* pbChanged)
{
	PgpGwTrace("CCommandFactory::WantCommand(context=%S, id=%S)\n", 
			bstrContext, bstrPersistentID);

	//if it is a mail or the attachment viewer being opened
	if(((0 != wcsstr(bstrContext, MAILMESSAGECONTEXTW)) ||
		(0 != wcsstr(bstrContext, ATTCHVIEWERCONTEXTW))) && 
		(0 == _wcsicmp(bstrPersistentID, wceGW_CMDID_OPEN)))
	{
		//check if auto-decrypt bit on open is set in pgp options
		PGPBoolean pgpbFlag=FALSE;
		BOOL bRet=PgpGwGetBoolPref(kPGPPrefAutoDecrypt, pgpbFlag);
		pgpAssert(TRUE == bRet);

		//we trap open only if the the bit is set
		*pbChanged=pgpbFlag;
	}
	else
		*pbChanged=FALSE;

	return S_OK;
}
   

//////////////////////////////////////////////////////////////////
//  Name:CCommandFactory::CustomizeContextMenu				
//
//  Description:Method to customize the menu.	
//
//  In:BSTR bstrContext The context string property is the class of the
//          containing window.  For example "GW.APPLICATION.BROWSER"
//     IDispatch * pIDispGWMenu - The main menu object being modified.
//
//  Out:TOleBool - The return value indicates whether the modifications 
//		to the menu were "volatile".  If the return value is TRUE, 
//		CustomizeMenu() will continue to be called each time the menu 
//		may need to be updated e.g. at each popup creation).  Otherwise, 
//		C3PO servers can safely assume that CustomizeMenu is called only 
//		once for any single instance of a menu.  This simplifies the 
//		resulting code because there is no need to check for commands 
//		that might already be on the menu. That is, the C3PO's command 
//		are guaranteed to be absent from the menu at the time of the 
//		first CustomizeMenu call.
//
//  Comments:
////////////////////////////////////////////////////////////////////
STDMETHODIMP CCommandFactory::CustomizeContextMenu(BSTR bstrContext, 
						IDispatch * pIDispGWMenu)
{
	PgpGwTrace("CCommandFactory::CustomizeContextMenu(%S)\n", bstrContext);
	IGWMenu	*pIMenu=NULL;
	IGWMenuItems *pIMenuItems=NULL;
	IDispatch *pIDisp=NULL;
	IGWClientState *pClientState=NULL;
	IGWMessageList *pMsgs=NULL;
	HRESULT	hrRet=S_OK;
	VARIANT	vr={0};
	BSTR bstrCaption=NULL;
	LONG lMsgCount=0;
	int iResult=0;
	BSTR bstrResult=NULL;
	VARIANT_BOOL bResult;
	TCHAR *pstr=NULL;

	if (0 != wcsstr(bstrContext, MAILMESSAGECONTEXTW))          // See if it is the correct context
	{
		//get client state to determine if we're being called 
		//from the main browser, not from inside a view.
		hrRet=m_pIMgr->get_ClientState(&pIDisp);
		hrRet=pIDisp->QueryInterface(IID_IClientState, (void **)&pClientState);
		pIDisp->Release();

		pIDisp = NULL;
		hrRet = pClientState->get_SelectedMessages (&pIDisp);
		hrRet = pIDisp->QueryInterface(IID_IGWMessageList, (void **)&pMsgs);
		pIDisp->Release ();
		pIDisp = NULL;
		
		hrRet = pMsgs->get_Count (&lMsgCount);

		pClientState->Release ();
		pClientState = NULL;

		if (0 == lMsgCount)
		{
			// We're in a view, now see if it's a compose view
			g_pigwcCommander->Execute (g_bstrValidateCmd, &bstrResult, &bResult);

			pstr = FROM_OLE_STRING(bstrResult);

			// Handle OnReady event
			if (strcmp(pstr, NEWMESSAGEMSGIDA) == 0)
			{
				hrRet = pIDispGWMenu->QueryInterface(IID_IGWMenu, (void**)&pIMenu);

				//Add menu item
				// Get Menu Items menu
				hrRet = pIMenu->get_MenuItems(&pIDisp);
				pIMenu->Release();

				// Get Cts Menu Items interfaces
				hrRet = pIDisp->QueryInterface(IID_IGWMenuItems, (void**)&pIMenuItems);
				pIDisp->Release();

				// Create new GWCommand object for menu item
				CGWCommand *Cmd01 = new CGWCommand(IDC_ODMA_ATTACH, 0);
				Cmd01->QueryInterface(IID_IDispatch, (void**)&pIDisp);
				bstrCaption = SysAllocString(L"Attach ODMA Document...");

				VariantInit(&vr);
				vr.lVal = 3;
				vr.vt = VT_I4;

				hrRet = pIMenuItems->Add(bstrCaption, pIDisp, vr, &pIDisp );          //Add menu item to menu
				SysFreeString(bstrCaption);
				Cmd01->m_bstrLongPrompt = SysAllocString(L"Add attachments from an ODMA Document Management System");          // Set Long Prompt for menu item

				Cmd01->Release();
				pIDisp->Release();
				pIMenuItems->Release();
			}
		}
	}

	return S_OK;
}

//////////////////////////////////////////////////////////////////
//  Name:CCommandFactory::CustomizeMenu				
//
//  Description:Method to customise the menu.	
//
//  In:BSTR bstrContext - The context string property is the class 
//			of the containing window.  For example "GW.APPLICATION.BROWSER"
//	   IDispatch * pIDispGWMenu - The main menu object being modified.
//
//  Out:VARIANT_BOOL FAR* pbChanged - The return value indicates 
//		whether the modifications to the menu were "volatile".  If 
//		the return value is TRUE, CustomizeMenu() will continue to 
//		be called each time the menu may need to be updated e.g. at 
//		each popup creation).  Otherwise, C3PO servers can safely
//		assume that CustomizeMenu is called only once for any single 
//		instance of a menu.  This simplifies the resulting code because 
//		there is no need to check for commands that might already be on 
//		the menu. That is, the C3PO's command are guaranteed to be 
//		absent from the menu at the time of the first CustomizeMenu 
//		call.
//
//  Comments:
////////////////////////////////////////////////////////////////////
STDMETHODIMP CCommandFactory::CustomizeMenu(BSTR bstrContext,
						IDispatch * pIDispGWMenu, VARIANT_BOOL FAR* pbChanged)
{
	PgpGwTrace("CCommandFactory::CustomizeMenu(%S)\n", bstrContext);
	IGWMenu *pIMenu=NULL;
	IGWMenuItems *pIMenuItems=NULL;
	IDispatch *pIDispTemp=NULL, *pIDispTempTwo=NULL;
	HRESULT hrRet=S_OK;
	BOOL bRet=TRUE;
	VARIANT vTmpOne={0};
	VARIANT vTmpTwo={0};
	char *pszPrependMenuString=NULL, *pszMainMenuString=NULL;
	BSTR bstrCaption=NULL;
	BSTR bstrMessageId=NULL;
	CGWCommand *pcgwCmdTemp=NULL;
	DWORD dwCmdUiOptions=CUIO_IS_MENUBAR_ITEM;

	//if it is the main browser window
	if(0 != wcsstr(bstrContext, MAINWINDOWCONTEXTW))
		dwCmdUiOptions |= CUIO_MAINWND_CONTEXT;
	//or if it is a mail message
	else if (0 != wcsstr(bstrContext, MAILMESSAGECONTEXTW))
	{
		dwCmdUiOptions |= CUIO_MAIL_CONTEXT;

		//additionally if it is a discussion post
		if(0 != wcsstr(bstrContext, DISCUSSIONCONTEXTW))
			dwCmdUiOptions |= CUIO_DISCUSSION_CONTEXT;

		//discussion post context is unfortunately a subclass of mail context
		//so this check needs to be done so we know later that it is a discussion
		//rather than normal mail. the "post" of discussion will kick our tph
		//which would ignore any operations upon detecting this flag
	}
	//or if it is the attachment window
	else if(0 != wcsstr(bstrContext, ATTCHVIEWERCONTEXTW))
		dwCmdUiOptions |= CUIO_ATTVIEW_CONTEXT;
	else//if it is neither a mail message nor the main window nor
	{	//the attachment viewer window
		pgpAssert(FALSE);//erroneous install/faulty hooking
		goto cleanup;//bail out (we are done)
	}

	//request IGWMenu interface of the menu passed
	pgpAssert((NULL != pIDispGWMenu) && (NULL == pIMenu));
	hrRet = pIDispGWMenu->QueryInterface(IID_IGWMenu, (void**)&pIMenu);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIMenu));
	if(FAILED(hrRet) || (NULL == pIMenu))
		goto cleanup;
	TraceRefCount(pIMenu);//***


	//query for the read only property MenuItems (of the menu object)
	pgpAssert(NULL == pIDispTemp);
	hrRet = pIMenu->get_MenuItems(&pIDispTemp);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIDispTemp));
	if(FAILED(hrRet) || (NULL == pIDispTemp))
		goto cleanup;
	TraceRefCount(pIMenu);//***
	TraceRefCount(pIDispTemp);//***

	//ask that disp interface to give us IGWMenuItems iface if we go 
	//directly ask for this interface from the original disp passed 
	//to us there will be E_NOINTERFACE (ssd)
	pgpAssert(NULL == pIMenuItems);
	hrRet = pIDispTemp->QueryInterface(IID_IGWMenuItems, (void**)&pIMenuItems);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIMenuItems));
	if(FAILED(hrRet) || (NULL == pIMenuItems))
		goto cleanup;
	TraceRefCount(pIMenu);//***
	TraceRefCount(pIDispTemp);//***
	TraceRefCount(pIMenuItems);//***
	
	//we dont need the disp iface anymore
	if(NULL != pIDispTemp)
	{
		pIDispTemp->Release();
		pIDispTemp=NULL;
	}

	if(dwCmdUiOptions&CUIO_MAIL_CONTEXT)
	{
		//find whether send is enabled in the menu somewhere. if it is so this is 
		//an unsent mail otherwise this is a recieved mail
		VariantInit(&vTmpOne);
		V_VT(&vTmpOne) = VT_I4;
		V_I4(&vTmpOne) = BFTKN_SEND;
		hrRet = pIMenuItems->FindByID(vTmpOne, &pIDispTemp);
		if(SUCCEEDED(hrRet) && (NULL != pIDispTemp))//if we did find the send menu
			dwCmdUiOptions |= CUIO_NEWMAIL_CONTEXT;//this must be treated as a new mail context
		else if(SUCCEEDED(hrRet) && (NULL == pIDispTemp))
			dwCmdUiOptions |= CUIO_OLDMAIL_CONTEXT;//this must be treated as old mail being opened

		//free this up for the next operation (if any)
		if(NULL != pIDispTemp)
		{
			pIDispTemp->Release();
			pIDispTemp = NULL;
		}
	}

	//get the string for the main menu before which we put our menu
	//this string is "Tools" in english version of GroupWise
	pgpAssert(NULL == pszPrependMenuString);
	bRet=SafeLoadString(IDS_PREPEND_GWMAINMENU_CMD, &pszPrependMenuString);
	if((FALSE == bRet) || (NULL == pszPrependMenuString))
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	//we intend to put "PGP" menu just before the tools menu so prepare a 
	//variant for the "Tools" menu string 
	VariantInit(&vTmpOne);
	VariantInit(&vTmpTwo);
	V_VT(&vTmpOne) = VT_BSTR;
	V_BSTR(&vTmpOne) = A2WBSTR(pszPrependMenuString);
	if(NULL == V_BSTR(&vTmpOne))
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	//get the string for the groupwise plug-in main menu. this string shall be
	//[ surprise :) ] "PGP" in english version of GroupWise
	pgpAssert(NULL == pszMainMenuString);
	bRet=SafeLoadString(IDS_GWPLUGIN_MAIN_MENU_NAME, &pszMainMenuString);
	if((FALSE == bRet) || (NULL == pszMainMenuString))
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	//allocate for the new main menu string which is "PGP" in english
	pgpAssert(NULL == bstrCaption);
	bstrCaption = A2WBSTR(pszMainMenuString);
	pgpAssert(NULL != bstrCaption);
	if(NULL == bstrCaption)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	//add PGP menu
	pgpAssert(NULL == pIDispTemp);
	hrRet=pIMenuItems->AddMenu(bstrCaption, vTmpTwo,//no command object is relevant here(ssd)
				vTmpOne, &pIDispTemp);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIDispTemp));
	if(FAILED(hrRet) || (NULL == pIDispTemp))
		goto cleanup;
	TraceRefCount(pIDispTemp);//***
	TraceRefCount(pIMenuItems);//***


	//free up the stuff that we needed only for this call
	if(NULL != bstrCaption)
	{
		SysFreeString(bstrCaption);
		bstrCaption=NULL;
	}
	SysFreeString(V_BSTR(&vTmpOne));
	VariantInit(&vTmpOne);


	//free up before overwriting
	if(NULL != pIMenu)
	{
		pIMenu->Release();
		pIMenu=NULL;
	}

	//get the menu iface of this newly added menu
	pgpAssert(NULL == pIMenu);
	hrRet=pIDispTemp->QueryInterface(IID_IGWMenu, (void**)&pIMenu);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIMenu));
	if(FAILED(hrRet) || (NULL == pIMenu))
		goto cleanup;
	TraceRefCount(pIDispTemp);//***
	TraceRefCount(pIMenu);//***

	//free up before overwriting
	if(NULL != pIDispTemp)
	{
		pIDispTemp->Release();
		pIDispTemp=NULL;
	}

	//query for the read only property MenuItems (of the menu object)
	pgpAssert(NULL == pIDispTemp);
	hrRet = pIMenu->get_MenuItems(&pIDispTemp);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIDispTemp));
	if(FAILED(hrRet) || (NULL == pIDispTemp))
		goto cleanup;
	TraceRefCount(pIDispTemp);//***
	TraceRefCount(pIMenu);//***


	//free up before overwriting
	if(NULL != pIMenuItems)
	{
		pIMenuItems->Release();
		pIMenuItems=NULL;
	}

	//ask that disp interface to give us IGWMenuItems iface 
	pgpAssert(NULL == pIMenuItems);
	hrRet = pIDispTemp->QueryInterface(IID_IGWMenuItems, (void**)&pIMenuItems);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIMenuItems));
	if(FAILED(hrRet) || (NULL == pIMenuItems))
		goto cleanup;
	TraceRefCount(pIDispTemp);//***
	TraceRefCount(pIMenuItems);//***

		
	if(NULL != pIDispTemp)
	{
		pIDispTemp->Release();
		pIDispTemp=NULL;
	}
		
	//add the decrypt/verify menu
	hrRet = InternalAddMenuItem(IDS_COMMAND_DECRYPT, pIMenuItems, 
			IDM_DECRYPT_VERIFY, 
			dwCmdUiOptions|CUIO_IS_AN_OPERATION,
			IDS_HLPCMD_DECRYPT);
	pgpAssert(SUCCEEDED(hrRet));
	if(FAILED(hrRet))
		goto cleanup;
	TraceRefCount(pIMenuItems);//***
	
	if(dwCmdUiOptions&CUIO_MAIL_CONTEXT)//a message window
	{
		//add the show original menu
		hrRet = InternalAddMenuItem(IDS_COMMAND_SHOWORIG, pIMenuItems, 
						IDM_SHOW_ORIGINAL, 
						dwCmdUiOptions|CUIO_IS_AN_OPERATION,
						IDS_HLPCMD_SHOWORIG);
		pgpAssert(SUCCEEDED(hrRet));
		if(FAILED(hrRet))
			goto cleanup;

		//add the encrypt on send menu
		hrRet = InternalAddMenuItem(IDS_COMMAND_ENCRYPT_ONSEND, pIMenuItems, 
						IDM_AUTO_ENCRYPT_ONSEND, 
						dwCmdUiOptions|CUIO_IS_AN_OPTION|CUIO_PREPEND_MENU_SEPARATOR,
						IDS_HLPCMD_ENCRYPT_ONSEND);
		pgpAssert(SUCCEEDED(hrRet));
		if(FAILED(hrRet))
			goto cleanup;

		//add the sign on send menu
		hrRet = InternalAddMenuItem(IDS_COMMAND_SIGN_ONSEND, pIMenuItems, 
						IDM_AUTO_SIGN_ONSEND, 
						dwCmdUiOptions|CUIO_IS_AN_OPTION,
						IDS_HLPCMD_SIGN_ONSEND);
		pgpAssert(SUCCEEDED(hrRet));
		if(FAILED(hrRet))
			goto cleanup;

		//we do not support anything other than plain text in the mail body
		//uncomment these when we do the enhancements for RTF and/or HTML
		//add the convert to plain text menu
		//hrRet = InternalAddMenuItem(IDS_COMMAND_CONVTO_PLAINTXT, pIMenuItems, 
		//				IDM_CONVERT_TO_PLAIN_TEXT, 
		//				dwCmdUiOptions|CUIO_IS_AN_OPTION,
		//				IDS_HLPCMD_CONVTO_PLAINTXT);
		//pgpAssert(SUCCEEDED(hrRet));
		//if(FAILED(hrRet))
		//	goto cleanup;
	}

	//add the pgpkeys launch menu
	hrRet = InternalAddMenuItem(IDS_COMMAND_LAUNCH_PGPKEYS, pIMenuItems, 
			IDM_LAUNCH_PGPKEYS, dwCmdUiOptions|CUIO_IS_AN_OPERATION|CUIO_PREPEND_MENU_SEPARATOR,
			IDS_HLPCMD_LAUNCH_PGPKEYS);
	pgpAssert(SUCCEEDED(hrRet));
	if(FAILED(hrRet))
		goto cleanup;

	//add the options menu
	hrRet = InternalAddMenuItem(IDS_COMMAND_PGP_OPTIONS, pIMenuItems, 
			IDM_SHOW_PREFS, dwCmdUiOptions|CUIO_IS_AN_OPERATION,
			IDS_HLPCMD_PGP_OPTIONS);
	pgpAssert(SUCCEEDED(hrRet));
	if(FAILED(hrRet))
		goto cleanup;

cleanup:
	if(NULL != pIDispTemp)
	{
		pIDispTemp->Release();
		pIDispTemp=NULL;
	}
	if(NULL != pIDispTempTwo)
	{
		pIDispTempTwo->Release();
		pIDispTempTwo=NULL;
	}

	if(NULL != pIMenuItems)
	{
		pIMenuItems->Release();
		pIMenuItems=NULL;
	}
	if(NULL != pIMenu)
	{
		pIMenu->Release();
		pIMenu=NULL;
	}
	if(NULL != bstrCaption)
	{
		SysFreeString(bstrCaption);
		bstrCaption=NULL;
	}
	if(NULL != pszPrependMenuString)
	{
		free(pszPrependMenuString);
		pszPrependMenuString=NULL;
	}
	if(NULL != pszMainMenuString)
	{
		free(pszMainMenuString);
		pszMainMenuString=NULL;
	}
	return hrRet;
}

//////////////////////////////////////////////////////////////////
//  Name:CCommandFactory::CustomizeToolbar				
//
//  Description:Method to customize the toolbar. The C3po must assume 
//		that its commands are absebt from the toolbar. For example, 
//		when a toolbar has been saved and restored in another session 
//		of GroupWise, the commands can be put back onto the toolbar 
//		w/out any intervening calls to CustomizeToolbar. For that 
//		reason, the C3po should query, the toolbar first before adding 
//		anything to the toolbar window.
//	
//
//  In:BSTR bstrContext - Contains the class of the containing window.
//	   IDispatch * pIDispGWToolbar - Contains the GWToolbar object to 
//			be modified.
//
//  Out:VARIANT_BOOL FAR* pbChanged - Returning TRUE from this method 
//				indicates that the toolbar modifications were 'volatile'. 
//				In that case, each time the toolbar is validated in the 
//				UI, the	CommandFactory::CustomizeToolbar method will be 
//				called.
//
//
//  Comments:
////////////////////////////////////////////////////////////////////
STDMETHODIMP CCommandFactory::CustomizeToolbar(BSTR bstrContext,
						IDispatch * pIDispGWToolbar,
						VARIANT_BOOL FAR* pbChanged)
{
	PgpGwTrace("CCommandFactory::CustomizeToolbar(%S)\n", bstrContext);
	WORD wSize=0;
	IGWToolbar *pIToolbar=NULL;
	IGWToolbarItems *pIToolbarItems=NULL;
	IDispatch *pIDisp=NULL;
	
	HRESULT hrRet=S_OK;
	VARIANT vrTmp={0};
	DWORD dwCmdUiOptions=CUIO_IS_TOOLBAR_ITEM;
	HWND hWndReBar=NULL, hWndTBar=NULL;
	LPSUBCLASSINFO lpSubClassInfo=NULL;

	//determine the correct window context for this toolbar item
	//if it is the attachment viewer
	if (0 != wcsstr(bstrContext, ATTCHVIEWERCONTEXTW))
		dwCmdUiOptions |= CUIO_ATTVIEW_CONTEXT;
	else if(0 != wcsstr(bstrContext, MAINWINDOWCONTEXTW))//if it is the main browser window
		dwCmdUiOptions |= CUIO_MAINWND_CONTEXT;
	else if(0 != wcsstr(bstrContext, MAILMESSAGECONTEXTW))//or if it is a mail message
	{
		dwCmdUiOptions |= CUIO_MAIL_CONTEXT;

		//additionally if it is a discussion post
		if(0 != wcsstr(bstrContext, DISCUSSIONCONTEXTW))
			dwCmdUiOptions |= CUIO_DISCUSSION_CONTEXT;
	}
	else
	{
		pgpAssert(FALSE);//incorrect hooking ?
		hrRet=E_FAIL;
		goto cleanup;
	}

    //request the GWToolbar iface from dispatch
	pgpAssert(NULL == pIToolbar);
	hrRet = pIDispGWToolbar->QueryInterface(IID_IGWToolbar, (void**)&pIToolbar);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIToolbar));
	if(FAILED(hrRet) || (NULL == pIToolbar))
		goto cleanup;

    //get the read only ToolbarItems property of the toolbar
	pgpAssert(NULL == pIDisp);
    hrRet = pIToolbar->get_ToolbarItems(&pIDisp);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIDisp));
	if(FAILED(hrRet) || (NULL == pIDisp))
		goto cleanup;

    //request the GWToolbarItems iface of the dispatch
	pgpAssert(NULL == pIToolbarItems);
	hrRet = pIDisp->QueryInterface(IID_IGWToolbarItems, (void**)&pIToolbarItems);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIToolbarItems));
	if(FAILED(hrRet) || (NULL == pIToolbarItems))
		goto cleanup;

    if((CUIO_MAINWND_CONTEXT&dwCmdUiOptions) ||
	   (CUIO_ATTVIEW_CONTEXT&dwCmdUiOptions))
	{
		//add the decrypt message button
		hrRet = InternalAddToolbarItem(pIToolbarItems,"IDB_ALT_DECRYPT", 
				IDS_HLPCMD_DECRYPT, dwCmdUiOptions|CUIO_IS_AN_OPERATION, 
				IDM_DECRYPT_VERIFY);

		//add the pgp keys launch button
		hrRet = InternalAddToolbarItem(pIToolbarItems,"IDB_PGPKEYS", 
				IDS_HLPCMD_LAUNCH_PGPKEYS, dwCmdUiOptions|CUIO_IS_AN_OPERATION,
				IDM_LAUNCH_PGPKEYS);
	}
	else
	{
		BOOL bAddButtons=TRUE;//whether to add buttons. this is necessary to fix a bug when HTML
							  //is the default compose view, our icons dont show in the toolbar
		BOOL bOpenMailView=FALSE;//whether this view is a mail open view as against a compose view
		int iIndex=0, iBtnCount=0;
		TBBUTTON tbbInfo={0};

		//NOTE:we reuse pIDisp here
		if(NULL != pIDisp)
		{
			pIDisp->Release();
			pIDisp=NULL;
		}
		
		//search for send button in the toolbar
		VariantInit(&vrTmp);
		V_VT(&vrTmp) = VT_I4;
		V_I4(&vrTmp) = BFTKN_SEND;
		hrRet = pIToolbarItems->FindByID(vrTmp, &pIDisp);
		if(SUCCEEDED(hrRet) && (NULL == pIDisp))//if we did not find the send menu
		{
			bOpenMailView = TRUE;//this must be treated as old mail being opened
			goto addbuttons; //let's add buttons since this is not a compose view
		}

		//get the rebar window
		hWndReBar=GetWindow(GWO_CURRENTREBAR_WINDOW);
		pgpAssert(NULL != hWndReBar);
		if(NULL == hWndReBar)
			goto addbuttons;//if we cant figure out things let's add the buttons anyway

		//the control comes here only if a mail is being composed.
		pgpAssert(FALSE == bOpenMailView);

		//for all the toolbars under the rebar
		while(NULL != (hWndTBar=FindWindowEx(hWndReBar, hWndTBar, TOOLBARCLASSNAME, NULL)))
		{
			//get the property associated with this toolbar
			LONG lOurData=(LONG)GetProp(hWndTBar, MODULENAME);

			//if we have never been called to modify this toolbar before
			if(0 == lOurData)
			{
				DWORD dwRet=0, dwHtmlComposeDisabled=0;
				DWORD dwBufSize=sizeof(DWORD);

				//check what compose view is enabled for this user
				dwRet=SHGetValue(HKEY_CURRENT_USER, "Software\\Novell\\GroupWise\\Client\\Setup",
					"DoNotCreateHTMLMessages", NULL, &dwHtmlComposeDisabled, &dwBufSize);
				pgpAssert(ERROR_SUCCESS == dwRet);

				//we store no of times we have been called in low word and the view option
				//in the high word of the property long value. so we remember we have been
				//called once and this was the compose view option when we were called.
				_VERIFY(0 != SetProp(hWndTBar, MODULENAME, 
						(HANDLE)MAKELONG(1, ((WORD)dwHtmlComposeDisabled))));

				//if we could not find the value, as far as we know it is safe to add the buttons 
				if(ERROR_SUCCESS != dwRet) 
					goto addbuttons; 

				//if the default compose view is html
				if(0 == dwHtmlComposeDisabled)
					bAddButtons=FALSE;//dont do anything. we will be called later
			}
			else if(1 == LOWORD(lOurData))//if we are being called for the second time
			{
				//remember that we have been called for the second time
				_VERIFY(0 != SetProp(hWndTBar, MODULENAME, 
						(HANDLE)MAKELONG(2, HIWORD(lOurData))));
			}
		}

		addbuttons:	
		if(TRUE == bAddButtons)
		{
			//remember the determination as to whether this was a compose or open view
			dwCmdUiOptions |= (bOpenMailView?CUIO_OLDMAIL_CONTEXT:CUIO_NEWMAIL_CONTEXT);

			//decrypt/verify message
			hrRet = InternalAddToolbarItem(pIToolbarItems,"IDB_ALT_DECRYPT", 
					IDS_HLPCMD_DECRYPT_MSG, dwCmdUiOptions|CUIO_IS_AN_OPERATION, 
					IDM_DECRYPT_VERIFY);

			//encrypt message before sending
			hrRet = InternalAddToolbarItem(pIToolbarItems,"IDB_ALT_ENCRYPT", 
					IDS_HLPCMD_ENCRYPT_ONSEND, dwCmdUiOptions|CUIO_IS_AN_OPTION, 
					IDM_AUTO_ENCRYPT_ONSEND);

			//sign message before sending
			hrRet = InternalAddToolbarItem(pIToolbarItems,"IDB_ALT_SIGN", 
					IDS_HLPCMD_SIGN_ONSEND, dwCmdUiOptions|CUIO_IS_AN_OPTION, 
					IDM_AUTO_SIGN_ONSEND);

			//launch PGP keys
			hrRet = InternalAddToolbarItem(pIToolbarItems,"IDB_PGPKEYS", 
					IDS_HLPCMD_LAUNCH_PGPKEYS, dwCmdUiOptions|CUIO_IS_AN_OPERATION, 
					IDM_LAUNCH_PGPKEYS);
		}
	}
		
cleanup:

	if(NULL != pIDisp)
	{
		pIDisp->Release();
		pIDisp=NULL;
	}
	if(NULL != pIToolbar)
	{
		pIToolbar->Release();
		pIToolbar=NULL;
	}
	if(NULL != pIToolbarItems)
	{
		pIToolbarItems->Release();
		pIToolbarItems=NULL;
	}

	return hrRet;
}

//////////////////////////////////////////////////////////////////
//  Name:CCommandFactory::Init				
//
//  Description:This method is the first method of Command::Factory 
//		called. It is used to tell GroupWise what UI items will be 
//		modified.	
//
//  In:	long lcid - This is the locale ID of the application driving 
//					the C3po.
//
//  Out: long FAR* plCmdFlags - The C3po server returns any combination 
//			of the following bits:
//
//       eGW_CMDINIT_MENUS  The C3po intends to modify the menus 
//							(CustomizeMenu Interface)
//       eGW_CMDINIT_TOOLBARS The C3po intends to modify the toolbar 
//							(CustomizeMenu Interface)
//       eGW_CMDINIT_CONTEXT_MENUS - The C3po intends to modify the 
//							context menus (CustomizeMenu Interface)
//       eGW_CMDINIT_NO_PREDEFINED - Optimization flag that indicates 
//							the C3po will never respond to predefined 
//							commands. Returning this flag will supress 
//							calls to the WantCommand method for predefined 
//							commands.
//	Comments:		
////////////////////////////////////////////////////////////////////
STDMETHODIMP CCommandFactory::Init(long lcid, long FAR* plCmdFlags)
{
    PgpGwTrace("CCommandFactory::Init\n");
	*plCmdFlags = eGW_CMDINIT_MENUS|eGW_CMDINIT_TOOLBARS/*|eGW_CMDINIT_CONTEXT_MENUS*/;

	return S_OK;
}


STDMETHODIMP CCommandFactory::InternalAddMenuItem(UINT uiMenuItemStringId,
			 IGWMenuItems *pigwmParentMenuItems, int iMenuCommandId,
			 DWORD dwMenuUIOptions, UINT uiMenuItemHlpStringId/*=NULL*/)
{
	IDispatch *pIDispTemp=NULL, *pIDispTempTwo=NULL, *pIDispTempThree=NULL;
	CGWCommand *pcgwCmdTemp=NULL;
	char *pszMenuItemString=NULL;
	char *pszMenuItemHelpString=NULL;
	BSTR bstrCaption=NULL;
	HRESULT hrRet=S_OK;
	BOOL bRet=TRUE;
	VARIANT vTmp={0};

	//check parameters
	pgpAssert(0 != uiMenuItemStringId);//menu string id cannot be zero
	pgpAssert(NULL != pigwmParentMenuItems);//the parent interface must be passed
	if((0 == uiMenuItemStringId) || (NULL == pigwmParentMenuItems))
	{
		hrRet=E_INVALIDARG;//bad parameters passed to us
		goto cleanup;
	}

	//get the menu string from the string table
	bRet=SafeLoadString(uiMenuItemStringId, &pszMenuItemString);
	pgpAssert((TRUE == bRet) && (NULL != pszMenuItemString));
	if(FALSE == bRet)//if we could not get the string 
	{
		hrRet=E_FAIL;//may be due to low memory or bad resource id passed
		goto cleanup;
	}
	
	//if a menu help string resource id was passed to us
	if(0 != uiMenuItemHlpStringId)
	{
		//get the menu help string from string table
		bRet=SafeLoadString(uiMenuItemHlpStringId, &pszMenuItemHelpString);
		pgpAssert((TRUE == bRet) && (NULL != pszMenuItemHelpString));
		if(FALSE == bRet)//if we could not get the menu help string 
		{
			hrRet=E_FAIL;//low memory or bad resource id ??
			goto cleanup;
		}
	}

	try
	{
		//create new GWCommand object for menu item
		pcgwCmdTemp = new CGWCommand(iMenuCommandId, dwMenuUIOptions);
	}
	catch(...)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	if(NULL == pcgwCmdTemp)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	//request the dispatch interface of the command object
	pgpAssert(NULL == pIDispTemp);
	hrRet = pcgwCmdTemp->QueryInterface(IID_IDispatch, (void**)&pIDispTemp);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIDispTemp));
	if(FAILED(hrRet) || (NULL == pIDispTemp))
		goto cleanup;

	//prepare the menu text
	pgpAssert(NULL == bstrCaption);
	bstrCaption = A2WBSTR(pszMenuItemString);
	pgpAssert(NULL != bstrCaption);
	if(NULL == bstrCaption)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}
	
	//set long prompt for menu item. if the help string was passed
	//blank we set the help string same as the menu string
	pcgwCmdTemp->m_bstrLongPrompt = (NULL != pszMenuItemHelpString)?
		A2WBSTR(pszMenuItemHelpString):A2WBSTR(pszMenuItemString);
	pgpAssert(NULL != pcgwCmdTemp->m_bstrLongPrompt);
	if(NULL == pcgwCmdTemp->m_bstrLongPrompt)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	//add the menu item. we do this at the absolute end for the cleanup
	//code to be simpler - we dont have to remove the menu for any reason
	pgpAssert(NULL != pIDispTemp);
	pgpAssert(NULL == pIDispTempTwo);
	hrRet = pigwmParentMenuItems->Add(bstrCaption, pIDispTemp, vTmp, &pIDispTempTwo);          
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIDispTempTwo));
	if(FAILED(hrRet) || (NULL == pIDispTempTwo))
		goto cleanup;

	//add a separator before this item if we are called upon to do so
	if(dwMenuUIOptions & CUIO_PREPEND_MENU_SEPARATOR)
	{
		pgpAssert(NULL != pIDispTempTwo);
		pgpAssert(NULL == pIDispTempThree);
		pgpAssert(VT_EMPTY == vTmp.vt);
		vTmp.vt = VT_BSTR;
		V_BSTR(&vTmp)=A2WBSTR(pszMenuItemString);
		hrRet = pigwmParentMenuItems->AddSeparator(vTmp, &pIDispTempThree);
		pgpAssert(SUCCEEDED(hrRet) && (NULL != pIDispTempThree));
		if(FAILED(hrRet) || (NULL == pIDispTempThree))
			goto cleanup;
	}
	

cleanup:

	if(NULL != pszMenuItemString)
	{
		free(pszMenuItemString);
		pszMenuItemString=NULL;
	}
	if(NULL != pszMenuItemHelpString)
	{
		free(pszMenuItemHelpString);
		pszMenuItemHelpString=NULL;
	}
	if(NULL != bstrCaption)
	{
		SysFreeString(bstrCaption);
		bstrCaption=NULL;
	}
	if(NULL != pIDispTemp)
	{
		pIDispTemp->Release();
		pIDispTemp=NULL;
	}
	if(NULL != pIDispTempTwo)
	{
		pIDispTempTwo->Release();
		pIDispTempTwo=NULL;
	}
	if(NULL != pIDispTempThree)
	{
		pIDispTempThree->Release();
		pIDispTempThree=NULL;
	}
	if((NULL != vTmp.bstrVal) && (VT_BSTR == vTmp.vt))
	{
		SysFreeString(vTmp.bstrVal);
		vTmp.bstrVal=NULL;
		vTmp.vt=VT_EMPTY;
	}

	if(FAILED(hrRet))
	{
		if(NULL != pcgwCmdTemp)
		{
			pcgwCmdTemp->Release();
			pcgwCmdTemp=NULL;
		}
	}

	return hrRet;
}

STDMETHODIMP CCommandFactory::InternalAddToolbarItem(IGWToolbarItems *pigwtToolBarItems,
			char *pszIconIdString, UINT uiToolHlpStrId/*char *pszToolHelpString*/, DWORD dwToolUIOptions,
			int iToolCommandId)
{
	BSTR bstrItemHelpString=NULL, bstrFilename=NULL, bstrIconId=NULL;
	CGWCommand *pcgwCmdTemp=NULL;
	IGWToolbarItem *pIToolbarItem=NULL;
	IDispatch *pIDisp=NULL, *pIDispCmd=NULL;
	VARIANT vrTmp={0};
	char *pszIconFileName=NULL;
	char *pszToolHelpString=NULL;
	HRESULT hrRet=S_OK;
	BOOL bRet=TRUE;
	DWORD dwRet=0;

	//check parameters passed to us
	pgpAssert(NULL != pigwtToolBarItems);//must have the toolbaritems iface passed to us
	pgpAssert(NULL != pszIconIdString);//must have the icon resource string id
	pgpAssert(0 != uiToolHlpStrId);
	if((NULL == pigwtToolBarItems) || 
	   (NULL == pszIconIdString) || (0 == uiToolHlpStrId))
	{
		hrRet=E_INVALIDARG;
		goto cleanup;
	}

	//if tool help string id was passed to us
	if(0 != uiToolHlpStrId)
	{
		//get the toolbar item help string from string table
		bRet=SafeLoadString(uiToolHlpStrId, &pszToolHelpString);
		pgpAssert((TRUE == bRet) && (NULL != pszToolHelpString));
		if(FALSE == bRet)//if we could not get the tool help string 
		{
			hrRet=E_FAIL;//low memory or bad resource id ??
			goto cleanup;
		}
	}

	//add to Toolbar a new button
    bstrItemHelpString = A2WBSTR(pszToolHelpString);
	pgpAssert(NULL != bstrItemHelpString);
	if(NULL == bstrItemHelpString)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

    //create new GWCommand
	try
	{
		pcgwCmdTemp = new CGWCommand(iToolCommandId, dwToolUIOptions);
	}
	catch(...)
	{
		pgpAssert(FALSE);
		hrRet=E_OUTOFMEMORY;
	}

	pgpAssert(NULL != pcgwCmdTemp);
	if(NULL == pcgwCmdTemp)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

    //request dispatch interface to the command object
	pgpAssert(NULL == pIDispCmd);
	hrRet = pcgwCmdTemp->QueryInterface(IID_IDispatch, (void**)&pIDispCmd);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIDispCmd));
	if(FAILED(hrRet) || (NULL == pIDispCmd))
		goto cleanup;

	VariantInit(&vrTmp);//we want the toolbar button at the end

	//add the toolbar item
	pgpAssert(NULL == pIDisp);
	hrRet = pigwtToolBarItems->Add(bstrItemHelpString, pIDispCmd, vrTmp, &pIDisp);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIDisp));
	if(FAILED(hrRet) || (NULL == pIDisp))
		goto cleanup;

    if(NULL != bstrItemHelpString)
	{
		SysFreeString(bstrItemHelpString);
		bstrItemHelpString=NULL;
	}
	
	//get the icon file name
	pszIconFileName = GetIconsFilePath();
	pgpAssert(NULL != pszIconFileName);
	if(NULL == pszIconFileName)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	bstrFilename = A2WBSTR(pszIconFileName);
	pgpAssert(NULL != bstrFilename);
	if(NULL == bstrFilename)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

    //prepare name of bitmap resource
	bstrIconId = A2WBSTR(pszIconIdString);
	pgpAssert(NULL != bstrIconId);
	if(NULL == bstrIconId)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}

	//get GWToolbarItem iface of the button we just added
	pgpAssert(NULL == pIToolbarItem);
    hrRet = pIDisp->QueryInterface(IID_IGWToolbarItem, (void**)&pIToolbarItem);
	pgpAssert(SUCCEEDED(hrRet) && (NULL != pIToolbarItem));

	//if we got the interface to the toolbar item
	if(SUCCEEDED(hrRet) && (NULL != pIToolbarItem))
	{
		//set bitmap of the toolbar item
		VariantInit(&vrTmp);
		vrTmp.vt = VT_BSTR;
		V_BSTR(&vrTmp)=bstrIconId;
		hrRet = pIToolbarItem->SetBitmap(bstrFilename, vrTmp);
		pgpAssert(SUCCEEDED(hrRet));
	}
	//NOTE: we did not do cleanup under failure condition of not being able 
	//to get the interface to the item we just now added. we tolerate failures
	//after this point to keep the cleanup much simpler(ssd)
	hrRet=S_OK;

	//set the tooltip for the toolbar item
	pcgwCmdTemp->m_bstrToolTip = A2WBSTR(pszToolHelpString);
	pgpAssert(NULL != pcgwCmdTemp->m_bstrToolTip);

cleanup:

	if(NULL != pszToolHelpString)
	{
		free(pszToolHelpString);
		pszToolHelpString=NULL;
	}

    if(NULL != bstrIconId)
	{
		SysFreeString(bstrIconId);
		bstrIconId=NULL;
	}

	if(NULL != pszIconFileName)
	{
		free(pszIconFileName);
		pszIconFileName=NULL;
	}

    if(NULL != bstrFilename)
	{
		SysFreeString(bstrFilename);
		bstrFilename=NULL;
	}

    if(NULL != bstrItemHelpString)
	{
		SysFreeString(bstrItemHelpString);
		bstrItemHelpString=NULL;
	}

    if(NULL != pIToolbarItem)
	{
		pIToolbarItem->Release();
		pIToolbarItem=NULL;
	}

	if(NULL != pIDispCmd)
	{
		pIDispCmd->Release();
		pIDispCmd=NULL;
	}
	if(NULL != pIDisp)
	{
		pIDisp->Release();
		pIDisp=NULL;
	}

	if(NULL != pcgwCmdTemp)
	{
		pcgwCmdTemp->Release();
		pcgwCmdTemp=NULL;
	}

	return hrRet;
}