////////////////////////////////////////////////////////////////////////////////
//
//  File : Util.cpp
//
//  Copyright (C) 2002 PGP Corporation
//
//  ABSTRACT
//		Utility routines
// 
//  Author: Satya S. Das
//
////////////////////////////////////////////////////////////////////////////////

#include "util.h"
#include "globals.h"
#include "globdefs.h"
#include <stdio.h>
#include <crtdbg.h>
#include "pgpdebug.h"
#include "pgplnlib.h"
#include "resource.h"
#include "windows.h"

//function to convert wide string to ansi
//NOTE: caller must test for null return value. no null
//terminated empty wide string is returned from this fn
char* ConvertToAnsi(OLECHAR *szWideStringToConvert)
{
	//locals
	int iRet=0;
	ULONG ulCnt=0;
	char *pStr=NULL;

	//check parameters
	pgpAssert(NULL != szWideStringToConvert);
	if(NULL == szWideStringToConvert)
		return NULL;

	//find out how long of a buffer we need to convert. the return
	//length includes the terminating null character
	ulCnt = WideCharToMultiByte(CP_ACP, 0, szWideStringToConvert, -1, 
				NULL, 0, NULL, NULL);

	//allocate the buffer
	try
	{
		pStr = new char[ulCnt];
	}
	catch(...)
	{
	}

	//if allocation failed we return null
	if(NULL == pStr)
		return NULL;

	//fill it up with the converted string
	iRet=WideCharToMultiByte(CP_ACP, 0, szWideStringToConvert, -1, 
			pStr, ulCnt, NULL, NULL);
	pgpAssert(0 != iRet);//successful translation
	pgpAssert(iRet == (int)ulCnt);//counts are ok

	return pStr;
} 

//function to convert ansi string to wide
//NOTE: caller must test for null return value. no null
//terminated empty ansi string is returned from this fn
OLECHAR* ConvertToUnicode(char *szAnsiString)
{
	//locals
	int iRet=0;
	ULONG ulCnt=0;
	OLECHAR	*pOleStr=0;

	//check parameters
	pgpAssert(NULL != szAnsiString);
	if(NULL == szAnsiString)
		return NULL;

	//see how big the unicode string will be
	ulCnt=MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, szAnsiString, -1, NULL, 0)+1;

	//allocate a buffer for the unicode string
	try
	{
		pOleStr = new OLECHAR[ulCnt];
	}
	catch(...)
	{
	}

	//if allocation failed we return null
	if(NULL == pOleStr)
		return NULL;

	//fill the buffer with the converted string
	iRet=MultiByteToWideChar(CP_ACP, 0, szAnsiString, -1, pOleStr, ulCnt);
	pgpAssert(0 != iRet);//successful translation
	pgpAssert(iRet+1 == (int)ulCnt);//counts are good

	return pOleStr;
}


void ErrorMessage(LPSTR lpsz)
{
	MessageBox(NULL, lpsz, "Error", MB_OK|MB_ICONINFORMATION);
}


//tree item information retrieval
BOOL TV_GetString(HWND hwndTV, HTREEITEM hItem, TCHAR *szText, UINT cchTextMax)
{
    TV_ITEM	tvItem;


	if(IsWindow(hwndTV) && hItem) 
	{
		tvItem.mask = TVIF_TEXT;
		tvItem.hItem = hItem;
		tvItem.pszText = szText;
		tvItem.cchTextMax = cchTextMax;
		return TreeView_GetItem(hwndTV, &tvItem);
	}
	else
		return FALSE;
}

BSTR A2WBSTR(LPCSTR lpszAnsiString, int nLen/*= -1*/)
{
	BSTR bstrOutputString = NULL;//buffer for output string
	WCHAR *pwszWideString=NULL;
	int iRet=0, iConvertedLen=0;

	//check parameters
	pgpAssert(NULL != lpszAnsiString);//the ansi string parameter cannot be null
	if(NULL == lpszAnsiString)
		return NULL;

	//if length of the ansi string was passed to us
	if(-1 != nLen)
	{
		//make sure the buffer is readable
		pgpAssert(TRUE != IsBadReadPtr(lpszAnsiString, nLen));
		if(TRUE == IsBadReadPtr(lpszAnsiString, nLen))
			return NULL;
	}
	
	//find out the buffer requirements for wide char conversion
	iRet = MultiByteToWideChar(CP_ACP, 0, lpszAnsiString, 
			nLen, NULL, NULL);
	pgpAssert(0 != iRet);
	if(0 == iRet)//failure encountered during conversion
	{
		//TODO:do an error string lookup and debug output string
		return NULL;
	}

	//if we were told to find out the length requirements on our own
	if (-1 == nLen)
		iConvertedLen = iRet+1;

	//allocate the required length
	pwszWideString=(WCHAR *)calloc(iConvertedLen, sizeof(WCHAR));
	pgpAssert(NULL != pwszWideString);
	if (NULL != pwszWideString)
	{
		iRet=MultiByteToWideChar(CP_ACP, 0, lpszAnsiString, -1, 
				pwszWideString, iConvertedLen);
		pgpAssert(0 != iRet);//should succeed
		if(0 == iRet)//if failed to convert
		{
			//TODO:do an error string lookup and debug output string
			//free up the allocated buffer
			free(pwszWideString);
			pwszWideString=NULL;
			return NULL;//bail out
		}

		bstrOutputString=SysAllocString(pwszWideString);
		pgpAssert(NULL != bstrOutputString);
	}

	if(NULL != pwszWideString)
	{
		free(pwszWideString);
		pwszWideString=NULL;
	}

	return bstrOutputString;
} 

char * GetIconsFilePath()
{
	char *pszTemp=NULL, *pszIconsFilePath=NULL;
	int iLenAlloced=MAX_PATH;//to keep track of length allocated
							 //we jump it by MAX_PATH everytime
							 //we think truncation happened
	int iLenReqd=0;
	int iIndex=-1;
	DWORD dwRet=0;

	while(TRUE)
	{
		//free up the earlier buffer allocations if any
		if(NULL != pszIconsFilePath)
		{
			free(pszIconsFilePath);
			pszIconsFilePath=NULL;
		}
		
		//allcate again
		pgpAssert(NULL == pszIconsFilePath);
		pszIconsFilePath = (char *)calloc(iLenAlloced, 1);
		pgpAssert(NULL != pszIconsFilePath);

		//if we failed to allocate
		if(NULL == pszIconsFilePath)
		{
			pgpAssert(FALSE);//ran out of memory ??
			return NULL;
		}

		//attempt to get the path again
		dwRet=GetModuleFileName(g_hInstance, pszIconsFilePath, iLenAlloced-1);
		if((dwRet > 0) && ((int)dwRet < iLenAlloced))//return value does not include null
		{
#ifdef _DEBUG
			//check if our dll name is at the end of path returned
			pszTemp=strrchr(pszIconsFilePath, '\\');
			pgpAssert(NULL != pszTemp);
			if(NULL != pszTemp)
				pgpAssert(0 == stricmp(pszTemp+1, MODULEFILENAME));
#endif
			break;//we have successfully retrieved the path
		}
		
		//bump the alloc length by max_path and lets hope that is enough
		iLenAlloced+=MAX_PATH;
	}


	PgpGwTrace("GetIconsFilePath returning %s\n", pszIconsFilePath);
	return pszIconsFilePath;
}

void DisplayMessage(HWND hWnd, UINT uiResourceID, ...)
{
	char *pszTitle=NULL;
	char *pszBuffer=NULL;
	int iCurCharInd=0;
	int iNoOfParamsExp=0;
	va_list valArgList;
		
	BOOL bRet=SafeLoadString(IDS_TITLE, &pszTitle);
	pgpAssert(TRUE == bRet);//string is not found or some error condition
	pgpAssert(NULL != pszTitle);//potential low memory
	if((FALSE == bRet) || (NULL == pszTitle))
		goto cleanup;

	bRet=SafeLoadString(uiResourceID, &pszBuffer);
	pgpAssert(TRUE == bRet);//string is not found or some error condition
	pgpAssert(NULL != pszBuffer);//potential low memory
	if((FALSE == bRet) || (NULL == pszBuffer))
		goto cleanup;

	//if there is at least one format specification in there
	if(NULL != strchr(pszBuffer, '%'))
	{
		//we take the args and format the error string
		char szMessage[512]={0};
		va_start(valArgList, uiResourceID);
		_vsnprintf(szMessage, sizeof(szMessage), pszBuffer, valArgList);
		MessageBox(hWnd, szMessage, pszTitle, MB_ICONEXCLAMATION);
	}
	else
	{
		//there are no format specifications to be taken care of
		MessageBox(hWnd, pszBuffer, pszTitle, MB_ICONEXCLAMATION);
	}

	cleanup:
	if(NULL != pszTitle)
	{
		free(pszTitle);
		pszTitle=NULL;
	}

	if(NULL != pszBuffer)
	{
		free(pszBuffer);
		pszBuffer=NULL;
	}

	return;
}

#if (WINVER < 0x0500)

//copied from winuser.h. quick fix for compilation failures
typedef struct tagGUITHREADINFO
{
    DWORD   cbSize;
    DWORD   flags;
    HWND    hwndActive;
    HWND    hwndFocus;
    HWND    hwndCapture;
    HWND    hwndMenuOwner;
    HWND    hwndMoveSize;
    HWND    hwndCaret;
    RECT    rcCaret;
} GUITHREADINFO, *PGUITHREADINFO;

#endif //WINVER < 0x0500

/*WINUSERAPI BOOL WINAPI
GetGUIThreadInfo(
    DWORD          idThread,
    PGUITHREADINFO pgui);
*/
typedef WINUSERAPI BOOL WINAPI GETGUITHREADINFO
	(DWORD idThread, PGUITHREADINFO pgui);
typedef GETGUITHREADINFO * LPGETGUITHREADINFO;


HWND GetWindow(DWORD dwWhichWindow)
{
	DWORD dwRet=0;
	BOOL bRet=TRUE;
	char szClassName[255]={0};
	GUITHREADINFO gtiThreadInfo={0};
	gtiThreadInfo.cbSize=sizeof(GUITHREADINFO);
	HWND hWndRet=NULL;

	HMODULE hModUser32 = LoadLibrary("user32");
	LPGETGUITHREADINFO pGetGuiThreadInfo=(LPGETGUITHREADINFO)
			GetProcAddress(hModUser32, "GetGUIThreadInfo");
	pgpAssert(NULL != pGetGuiThreadInfo);

	if(NULL == pGetGuiThreadInfo)
		goto method_2;//95 does not support this method

	bRet=(*pGetGuiThreadInfo)(GetCurrentThreadId(), &gtiThreadInfo);
	pgpAssert(bRet != FALSE);
	if(FALSE == bRet)
	{
		dwRet=GetLastError();
		PgpGwTrace("GetGuiThreadInfo failed - because of error %d\n", dwRet);
		goto method_2;
	}
	else
	{	
		//we succeeded in getting the thread info
		//PgpGwTrace("gtiThreadInfo.hwndActive=%#x\n", gtiThreadInfo.hwndActive);
		//PgpGwTrace("              hwndFocus=%#x\n", gtiThreadInfo.hwndFocus);
		//PgpGwTrace("              hwndCapture=%#x\n", gtiThreadInfo.hwndCapture);
		//PgpGwTrace("              hwndMenuOwner=%#x\n", gtiThreadInfo.hwndMenuOwner);
		//PgpGwTrace("              hwndMoveSize=%#x\n", gtiThreadInfo.hwndMoveSize);
		//PgpGwTrace("              hwndCaret=%#x\n", gtiThreadInfo.hwndCaret);
	}

	hWndRet=gtiThreadInfo.hwndActive;
	if(NULL == hWndRet)//activewindow is null
		hWndRet=gtiThreadInfo.hwndFocus;

	if(NULL == hWndRet)//active and focus windows are null
		hWndRet=gtiThreadInfo.hwndCaret;

	method_2:
	if(NULL == hWndRet)//active, focus and caret windows are null
	{
		hWndRet=GetForegroundWindow();//get the foregroundwindow
		PgpGwTrace("GetForegroundWindow() returned %#x\n", hWndRet);
	}

	if((GWO_CURRENTMAIL_WINDOW == dwWhichWindow) || 
	   (GWO_CURRENTREBAR_WINDOW == dwWhichWindow))
	{
		pgpAssert(NULL != hWndRet);
		if(NULL != hWndRet)//if we are returning a valid window handle
		{
			//make sure it is a mail window

			int iRet=0;

			do
			{
				//clean up the buf
				ZeroMemory(szClassName, sizeof(szClassName));

				//get the window class name
				pgpAssert(NULL != hWndRet);
				iRet=GetClassName(hWndRet, szClassName, 
								sizeof(szClassName)-sizeof(char));
				pgpAssert(iRet > 0);
			}
			while((0 != strcmp(szClassName, MAILVIEWWINDOWCLASSNAMEA)) &&//until we get a mail window
				  (NULL != (hWndRet=GetParent(hWndRet))));//we switch to the parent of the window
		}
		
		if(GWO_CURRENTMAIL_WINDOW == dwWhichWindow)
		{
			if(NULL == hWndRet)
				PgpGwTrace("WARNING: current mail window not found !\n");
			return hWndRet;
		}

		//if we have not got the current mail window
		if(NULL == hWndRet)
		{
			//it may be that the mail window is not active yet- as in the
			//case of us being called during addition of toolbar items
			
			//we try to find all the mail windows under desktop window
			hWndRet=FindWindowEx(GetDesktopWindow(), NULL, 
								MAILVIEWWINDOWCLASSNAMEA, NULL);

			//if there is an invisible mail window - that is the mail window
			//we are looking for !
			while(NULL != hWndRet)
			{
				if(!IsWindowVisible(hWndRet))//if it is invisible
					break;//we have found the soon to be current mail window

				//get the next mail window
				hWndRet=FindWindowEx(GetDesktopWindow(), hWndRet, 
								MAILVIEWWINDOWCLASSNAMEA, NULL);
			}
		}

		//get the rebar window child of that window
		if(NULL != hWndRet)
			hWndRet=FindWindowEx(hWndRet, NULL, REBARCLASSNAME, NULL);

		pgpAssert(NULL != hWndRet);
		return hWndRet;
	}
	else if(GWO_CURRENTATTVIEWER_WINDOW == dwWhichWindow)
	{
		pgpAssert(NULL != hWndRet);
		if(NULL != hWndRet)//if we are returning a valid window handle
		{
			//make sure it is a attachment viewer window

			int iRet=0;
			do
			{
				//clean up the buf
				ZeroMemory(szClassName, sizeof(szClassName));

				//get the window class name
				pgpAssert(NULL != hWndRet);
				iRet=GetClassName(hWndRet, szClassName, 
								sizeof(szClassName)-sizeof(char));
				pgpAssert(iRet > 0);
			}
			while((0 != strcmp(szClassName, ATTACHMENTVIEWERWINDOWCLASSNAMEA)) &&//until we get the attachment viewer window
				  (NULL != (hWndRet=GetParent(hWndRet))));//we switch to the parent of the window
		}

		return hWndRet;
	}
	else if(GWO_CURRENTTOPLEVEL_WINDOW == dwWhichWindow)
	{
		pgpAssert(NULL != hWndRet);
		if(NULL != hWndRet)//if we are returning a valid window handle
		{
			//make sure it is the top level window
			HWND hWndParent=NULL;

			while(NULL != (hWndParent=GetParent(hWndRet)))
				hWndRet=hWndParent;
		}
		pgpAssert(GetCurrentThreadId() == GetWindowThreadProcessId(hWndRet, NULL));//a process id check is more accurate
		return hWndRet;
	}
		

	return hWndRet;
}

BOOL SafeLoadString(UINT uiResourceId, char **ppszString)
{
	pgpAssert(NULL != uiResourceId);
	pgpAssert((NULL != ppszString) && (NULL == *ppszString));
	if((NULL == uiResourceId) || (NULL == ppszString))
	   return FALSE;

	DWORD dwAllocSize=255;//initial allocation size
	DWORD dwAllocIncrement=255;//allocation increment
	BOOL bRet=FALSE;
	UINT uiRet=0;
	void *pTmpBuf=NULL;

	*ppszString=NULL;//if the caller had a string in there, well bad luck !!

	*ppszString=(char *)calloc(dwAllocSize, sizeof(char));
	pgpAssert(NULL != *ppszString);
	if(NULL == *ppszString)
		return FALSE;//initial allocation failed !!

	while(TRUE)
	{
		uiRet=LoadString(g_hInstance, uiResourceId, *ppszString, dwAllocSize-sizeof(char));
		if(0 == uiRet)//string not found ?
		{
			pgpAssert(FALSE);
			PgpGwTrace("String %d not found in resources. LoadString failed returning error: %d\n",
				uiResourceId, GetLastError());
			break;
		}

		pgpAssert(uiRet <= (dwAllocSize-sizeof(char)-1));
		if(uiRet < (dwAllocSize-sizeof(char)-1))
		{
			bRet=TRUE;//we have got the whole string (success)
			break;
		}

		//string must have been as long as the max we passed
		pgpAssert(uiRet == (dwAllocSize-sizeof(char)-1));

		//increment the allocation size and allocate again
		dwAllocSize += dwAllocIncrement;
		pTmpBuf=realloc(*ppszString, dwAllocSize);
		if(NULL == pTmpBuf)//failed to allocate more
			break;

		*ppszString=(char *)pTmpBuf;//switch to the new buffer
	}
	
	//if we did not get the string in its entirety
	if((FALSE == bRet) && (NULL != *ppszString))
	{
		//free the buffer
		free(*ppszString);
		*ppszString=NULL;
	}

	return bRet;
}

BOOL PgpGwGetBoolPref(PGPPrefIndex ppiPrefToQuery, PGPBoolean& refpbPrefValue)
{
	PGPBoolean pgpbFlag=FALSE;
	PGPPrefRef pgpprPref=NULL;
	PGPError pgpErrRet=kPGPError_NoErr;

	//get a ref to the prefs from the client lib
	pgpErrRet=PGPclPeekClientLibPrefRefs(&pgpprPref, NULL);
	pgpAssert(!IsPGPError(pgpErrRet) && (NULL != pgpprPref));
	if (IsPGPError(pgpErrRet) || (NULL == pgpprPref))
		return FALSE; //failed

	//get the pref proper
	pgpErrRet=PGPGetPrefBoolean(pgpprPref, ppiPrefToQuery, &refpbPrefValue);
	pgpAssert(!IsPGPError(pgpErrRet));
	if(!IsPGPError(pgpErrRet))
		return TRUE;//succeeded

	return FALSE;//failed
}

//function to do extended license checking for this plugin in addition to
//license checking in shared code during encrypt/decrypt operations
BOOL PgpGwExLicenseChk()
{
	BOOL bRet = FALSE;
	PGPError pgpErrRet=kPGPError_NoErr;

	//check if the user has license to run us. we must have a valid license
	//with desktop flag turned on in the license number for pgp operations
	//to be carried out.
	if(FALSE == (bRet = PGPlnDesktop()))
	{
		//let her know of license requirements and lead to buy.
		pgpErrRet=PGPclNoLicense(NULL, kPGPclGroupwisePlugin);
		pgpAssert(IsntPGPError(pgpErrRet));
	}

	return bRet;
}


void PgpGwTrace(LPCTSTR lpszFormat, ...)
{
#ifdef _DEBUG 
	if (!g_dwModuleOptions&GWMO_TRACE)
		return;

	va_list args;
	va_start(args, lpszFormat);

	int nBuf;
	char szBuffer[512]={0};

	nBuf = _vsnprintf(szBuffer, sizeof(szBuffer), lpszFormat, args);

	// was there an error? was the expanded string too long?
	//pgpAssert(nBuf >= 0);
	if(nBuf >= 0)
		OutputDebugString(szBuffer);

	va_end(args);
#endif
}

void TraceRefCount(LPUNKNOWN lpUnkIntfPtr)
{
#ifdef _DEBUG
	if(NULL != lpUnkIntfPtr)
	{
		ULONG ulCount=lpUnkIntfPtr->AddRef();
		ulCount=lpUnkIntfPtr->Release();
		PgpGwTrace("Interface pointer %#x has ref count %d\n", lpUnkIntfPtr, ulCount);
	}
#endif
}

