////////////////////////////////////////////////////////////////////////////////
//  File : pgpgwServer.cpp
//
//  Copyright (C) 2002 PGP Corporation
//
//  ABSTRACT
//		Create and handle pgpgwServer Object
// 
//  Author: Satya S. Das
//
////////////////////////////////////////////////////////////////////////////////

// include files 
#include <stdio.h>
#include <ole2.h>
#include <shlwapi.h>

#include "gwc3po.h"
#include "gwc3cmd.h"
#include "pgpgwserver.h"
#include "CommandFactory.h"
#include "EventMonitor.h"
#include "IconFactory.h"
#include "classfactory.h"
#include "crtdbg.h"
#include "pgpdebug.h"
//#include "globals.h"
#include "globdefs.h"
#include "util.h"

// external functions 
extern IUnknown * BuildIUnkDispatch(REFCLSID libid, REFCLSID iid, 
									IUnknown *pIUnk, void *pIFac);
// private data 
//PgpNwClsFactory pncfPluginFatory;

// public functions 
IC3POManager *m_pIMgr;


///////////////////////////////////////////////////////////////////
//  Name:CPgpGwC3PO::CPgpGwC3PO	
// 
//  Description:C3po constructor. 
// 
//  In:none
//
//  Out:none
// 
//  Comments:Init objects and set reference count to zero.
///////////////////////////////////////////////////////////////////
CPgpGwC3PO::CPgpGwC3PO()
{
	PgpGwTrace("CPgpGwC3PO::CPgpGwC3PO\n");	
	m_pICmdFact=NULL;//init command factory to null
	m_pIEventMonitor=NULL;//init event monitor to null
	m_pIIconFactory=NULL;//init icon factory to null
	m_pIMgr = NULL;//init C3po manager to null
	m_pIUnkDispServ=NULL;//init C3po IDispatch to null
	m_cRef = 0;//set reference count to zero
}

///////////////////////////////////////////////////////////////////
//  Name:CPgpGwC3PO::~CPgpGwC3PO
// 
//  Description:CPgpGwC3PO destructor. See if we have IDispatch 
//				and release it if we do.delete command factory object.
// 
//  In:none
// 
//  Out:none
// 
//  Comments:
///////////////////////////////////////////////////////////////////
CPgpGwC3PO::~CPgpGwC3PO()
{
	PgpGwTrace("CPgpGwC3PO::~CPgpGwC3PO\n");	
	pgpAssert(NULL != m_pIUnkDispServ);
	if(NULL != m_pIUnkDispServ)
	{
		m_pIUnkDispServ->Release();
		m_pIUnkDispServ=NULL;
	}
	
	pgpAssert(NULL != m_pICmdFact);
	if(NULL != m_pICmdFact)
	{
		delete m_pICmdFact;
		m_pICmdFact=NULL;
	}

	pgpAssert(NULL != m_pIEventMonitor);
	if(NULL != m_pIEventMonitor)
	{
		delete m_pIEventMonitor;
		m_pIEventMonitor=NULL;
	}

	pgpAssert(NULL != m_pIIconFactory);
	if(NULL != m_pIIconFactory)
	{
		delete m_pIIconFactory;
		m_pIIconFactory=NULL;
	}
}

///////////////////////////////////////////////////////////////////
//  Name:CPgpGwC3PO::Create
// 
//  Description:CPgpGwC3PO create. Create CommandFactory and EventMonitor. 
// 
//  In:none
//
//  Out:return success of create.
// 
//  Comments:
/////////////////////////////////////////////////////////////////////
BOOL CPgpGwC3PO::Create()
{
	PgpGwTrace("CPgpGwC3PO::Create\n");	
	BOOL bRet=TRUE;
	try
	{
		//allocate the member objects

		m_pICmdFact = new CCommandFactory(this);
		if(NULL == m_pICmdFact)
		{
			pgpAssert(FALSE);
			bRet=FALSE;
			throw;
		}

		m_pIEventMonitor = new EventMonitor(this);
		if(NULL == m_pIEventMonitor)
		{
			pgpAssert(FALSE);
			bRet=FALSE;
			throw;
		}

		m_pIIconFactory = new IconFactory(this);
		if(NULL == m_pIIconFactory)
		{
			pgpAssert(FALSE);
			bRet=FALSE;
			throw;
		}
	}
	catch(...)
	{
		bRet=FALSE;
	}

	return bRet;
}



///////////////////////////////////////////////////////////////////
//
//  Name:CPgpGwC3PO::QueryInterface	
// 
//  Description:QueryInterface for CPgpGwC3PO object. 
// 
//  In:REFIID riid - reference id.
//	   LPVOID FAR* ppv - return reference looked for.
// 
//  Out:return Interface, CommandFactory, or build IDispatch object.
// 		If riid not found return no interface found.
//
//  Comments:
///////////////////////////////////////////////////////////////////
STDMETHODIMP CPgpGwC3PO::QueryInterface(REFIID riid, LPVOID FAR* ppv)
{
	PgpGwTrace("CPgpGwC3PO::QueryInterface\n");	
	if(NULL == ppv)
		return E_INVALIDARG;

    *ppv=NULL;

	if((IID_IUnknown == riid) || (IID_IC3POServer == riid))
	{
		*ppv  = this;
		((LPUNKNOWN)*ppv)->AddRef();
		return S_OK;
	}

	if((IID_ICommandFactory == riid) && (NULL != m_pICmdFact))
	{
		*ppv  = m_pICmdFact;
		((LPUNKNOWN)*ppv)->AddRef();
		return S_OK;
	}

	if((IID_IEventMonitor == riid) && (NULL != m_pIEventMonitor))
	{
		*ppv  = m_pIEventMonitor;
		((LPUNKNOWN)*ppv)->AddRef();
		return S_OK;
	}

	if((IID_IIconFactory == riid) && (NULL != m_pIIconFactory))
	{
		*ppv  = m_pIIconFactory;
		((LPUNKNOWN)*ppv)->AddRef();
		return S_OK;
	}

	if(IID_IDispatch == riid)
	{
		if (NULL == m_pIUnkDispServ)
		{
			m_pIUnkDispServ = BuildIUnkDispatch( 
							LIBID_PgpGwTypeLibrary,
							IID_IC3POServer,
							(IUnknown *)(IC3POServer *)this,
							(void *)(IC3POServer *)this);
		}

		pgpAssert(NULL != m_pIUnkDispServ);
		if (NULL != m_pIUnkDispServ)
			return m_pIUnkDispServ->QueryInterface(IID_IDispatch, ppv);
	}

	return E_NOINTERFACE;
}

///////////////////////////////////////////////////////////////////
//  Name:CPgpGwC3PO::AddRef	
//
//  Description:Bump reference count.
//
//  In:none
//
//  Out:reference count
//
//  Comments:
/////////////////////////////////////////////////////////////////////
STDMETHODIMP_(ULONG) CPgpGwC3PO::AddRef()
{
	PgpGwTrace("CPgpGwC3PO::AddRef\n");	
	return ++m_cRef;
}

///////////////////////////////////////////////////////////////////
//  Name:			CPgpGwC3PO::Release	
//
//  Description:	decrement reference count. If count goest to zero delete object.
//
//  In:			none
//
//  Out:			reference count
//
//  Comments:
/////////////////////////////////////////////////////////////////////
STDMETHODIMP_(ULONG) CPgpGwC3PO::Release()
{
	PgpGwTrace("CPgpGwC3PO::Release\n");	
	ULONG cRef = --m_cRef;
	if(0 == m_cRef)
	{
		delete this;
	}
	return cRef;
}


// IC3POServer methods

///////////////////////////////////////////////////////////////////
//  Name:CPgpGwC3PO::get_CommandFactory	
//
//  Description:This property returns the CommandFactory for the 
//				C3PO. The C3PO server may return NULL.
//
//  In:	IDispatch * FAR* ppIDispCommandFactory - return the IDispatch 
//		pointer
//
//  Out:CommandFactory object or null in not supported
//
//  Comments:If CommandFactory has not been created return S_FALSE.
//
///////////////////////////////////////////////////////////////////
STDMETHODIMP CPgpGwC3PO::get_CommandFactory(IDispatch * FAR* ppIDispCommandFactory)
{
	PgpGwTrace("CPgpGwC3PO::get_CommandFactory\n");	

	//check parameters passed
	pgpAssert(NULL != ppIDispCommandFactory);
	if(NULL == ppIDispCommandFactory)
		return 	E_INVALIDARG;

	//nullify the param before writing anything
	*ppIDispCommandFactory=NULL;

	if(NULL != m_pICmdFact)//if we have inited the command factory
	{
		//return the dispatch interface to the object
		return m_pICmdFact->QueryInterface(IID_IDispatch, 
					(void**)ppIDispCommandFactory);
	}

	return S_FALSE;
}

///////////////////////////////////////////////////////////////////
//
//  Name:CPgpGwC3PO::get_Description	
//
//  Description:Returns a human readable description of the C3PO server.
//
//  In:BSTR FAR* pbstrDescription - pointer to description.
//
//  Out:return string description
//
//  Comments:
/////////////////////////////////////////////////////////////////////
STDMETHODIMP CPgpGwC3PO::get_Description(BSTR FAR* pbstrDescription)
{
	PgpGwTrace("CPgpGwC3PO::get_Description\n");		

	//check parameters passed to us
	pgpAssert(NULL != pbstrDescription);
	if(NULL == pbstrDescription)
		return E_INVALIDARG;
	
	*pbstrDescription = SysAllocString(L"PGP plugin for GroupWise");
	if(NULL == *pbstrDescription)
		return E_OUTOFMEMORY;

	pgpAssert(NULL != *pbstrDescription);
	return S_OK;
}

///////////////////////////////////////////////////////////////////
//  Name:CPgpGwC3PO::get_EventMonitor	
//
//  Description:This property returns the event monitor for the C3PO.  
//				The C3PO server may return NULL. 
//
//  In:IDispatch * FAR* ppIDispEventMonitor
//
//  Out:EventMonitor object or null if not supported
//
//  Comments:We do not support event monitor at this time, so 
//			 return S_FALSE
/////////////////////////////////////////////////////////////////////
STDMETHODIMP CPgpGwC3PO::get_EventMonitor(IDispatch * FAR* ppIDispEventMonitor)
{
	PgpGwTrace("CPgpGwC3PO::get_EventMonitor\n");		

	//check parameters passed
	pgpAssert(NULL != ppIDispEventMonitor);
	if(NULL == ppIDispEventMonitor)
		return 	E_INVALIDARG;

	//nullify the param before writing anything
	*ppIDispEventMonitor=NULL;

	if(NULL != m_pIEventMonitor)//if we have inited an event monitor object
	{
		//return the dispatch interface to the object
		return m_pIEventMonitor->QueryInterface(IID_IDispatch, (void**)ppIDispEventMonitor);
	}

	return S_FALSE;
}

///////////////////////////////////////////////////////////////////
//
//  Name:CPgpGwC3PO::get_IconFactory	
//
//  Description:This property returns the IconFactory for the C3PO.  
//				The C3PO server may return NULL.
//
//  In:IDispatch * FAR* ppIDispIconFactory
//
//  Out:IconFactory object or null if not supported
//
//  Comments:
///////////////////////////////////////////////////////////////////
STDMETHODIMP CPgpGwC3PO::get_IconFactory(IDispatch * FAR* ppIDispIconFactory)
{
	PgpGwTrace("CPgpGwC3PO::get_IconFactory\n");

	//check parameters passed
	pgpAssert(NULL != ppIDispIconFactory);
	if(NULL == ppIDispIconFactory)
		return 	E_INVALIDARG;

	//nullify the param before writing anything
	*ppIDispIconFactory = NULL;

	if(NULL != m_pIIconFactory)//if we have inited an icon factory object
	{
		//pass the dispatch interface to the object
		return m_pIIconFactory->QueryInterface(IID_IDispatch, (void**)ppIDispIconFactory);
	}

	return S_FALSE;
}

///////////////////////////////////////////////////////////////////
//  Name:			CPgpGwC3PO::CanShutdown	
//
//  Description:This method is invoked to query whether the C3PO server can
//  shutdown.  Typically, this is used when the C3PO server has a window open 
//  on the UI screen and is not in a state where shutdown is possible.  Each 
//  C3PO server is guaranteed to have CanShutdown queried at least once before 
//  the C3PO system shuts down.

//  Semantically, CanShutdown is a request from the C3POManager to a given C3PO.
//  It's not a request for shutting down a C3PO as much as it is a request to 
//  shutdown the C3POManager and consequently disconnect the C3PO.  Typically, a 
//  C3PO would shutdown along with the C3POManager, but this is not strictly required.
//
//  No particular C3PO order can be relied on for querying the CanShutdown property.
//
//
//  In:VARIANT_BOOL FAR* pbCanShutdown
//
//  Out:TRUE if it is OK to shutdown
//      FALSE if it is not OK to shutdown
//  Comments:
///////////////////////////////////////////////////////////////////
STDMETHODIMP CPgpGwC3PO::CanShutdown(VARIANT_BOOL FAR* pbCanShutdown)
{
	PgpGwTrace("CPgpGwC3PO::CanShutdown\n");

	//check parameters
	pgpAssert(NULL != pbCanShutdown);
	if(NULL == pbCanShutdown)
		return E_INVALIDARG;

	//always allow shutdown since we do not have any situations
	//as of now when we want GW not to shutdown. return false here
	//for GW to suspend shutdown if required.
	*pbCanShutdown = VARIANT_TRUE;

	return S_OK;
}

///////////////////////////////////////////////////////////////////
//  Name:			CPgpGwC3PO::Init	
//
//  Description:  This method is the first method invoked in the C3POServer object
//                when loading a C3PO server.  If the server fails this call
//                (via the HRESULT), the C3PO server is unloaded.
//
//                One Init call will be issued to the C3POServer for each C3PO
//                Client using the C3PO system.  C3POs that wish to be capable of
//                loading into multiple Clients (irrespective of process boundaries)
//                should be multiple-instance OLE servers.  That is, a new C3POServer
//                object should be created for each C3POManager that wished to use
//                the services of the C3PO.  By tracking the Manager pointer for
//                each C3POServer, the C3PO can sort out which requests are being
//                issued from which clients.
//
//                The Manager object is valid until a future DeInit() call.
//                That is, the C3PO does not need to AddRef() this object but can
//                simply store it for the life of the C3PO (until DeInit is called).
//
//
//  In:			IDispatch * pIDispManager - C3POManager object
//
//  Out:
//
//  Comments:
///////////////////////////////////////////////////////////////////
STDMETHODIMP CPgpGwC3PO::Init(IDispatch * pIDispManager)
{
	PgpGwTrace("CPgpGwC3PO::Init\n");

	//check parameters
	pgpAssert(NULL != pIDispManager);
	if(NULL == pIDispManager)
		return E_POINTER;

	//cache the manager
	pgpAssert(NULL == m_pIMgr);//should not have been inited already
	HRESULT hrRes = pIDispManager->QueryInterface
				(IID_IC3POManager, (void**)&m_pIMgr);
	if(FAILED(hrRes))
	{
		pgpAssert(FALSE);//request for manager iface failed !!
		return hrRes;
	}

	return S_OK;
}

///////////////////////////////////////////////////////////////////
//  Name:				CPgpGwC3PO::DeInit	
//
//  Description:Terminates the relationship of the C3PO Manager with 
//				the C3PO server. Note that this is a separate issue 
//				from the shutdown sequence (including shutdown events).  
//				For example, a C3PO may be unloaded from memory as a 
//				runtime optimization.  In that scenario, the C3PO 
//				server first receives CanShutdown() calls followed by 
//				DeInit() the C3PO is unloaded, but the client application 
//				has not necessarilly terminated.
//
//  In:none
//
//  Out:none			
//
//  Comments:The C3POManager pointer passed in to C3POServer::Init() is still 
//			 valid during this call.  However, when DeInit() returns, the 
//			 C3POManager pointer is not guaranteed to be valid.  The C3POServer 
//			 must release all holds to the C3POManager during the DeInit call.
//
//			 One DeInit call will be issued to the C3POServer for each C3PO Client 
//			 using the C3PO system.  C3POs that wish to be capable of loading into 
//			 multiple Clients (irrespective of process boundaries) should be 
//			 multiple-instance OLE servers.  That is, a new C3POServer object 
//			 should be created for each C3POManager that wished to use the 
//			 services of the C3PO.  By tracking the Manager pointer for each 
//			 C3POServer, the C3PO can sort out which requests are being issued 
//			 from which clients.
//	
/////////////////////////////////////////////////////////////////////
STDMETHODIMP CPgpGwC3PO::DeInit()
{
	PgpGwTrace("CPgpGwC3PO::DeInit\n");
	if(NULL != m_pIMgr)
	{
		m_pIMgr->Release(); 
		m_pIMgr=NULL;
	}

	return S_OK;
}

///////////////////////////////////////////////////////////////////
//  Name:InsertUsInTPHChain
//
//  Description:Inserts ourselves onto the TPH chain. If another 
//  pgpgw.dll is found in the chain the path is updated to this path.
//  otherwise we insert ourselves at the end of the chain.
//
//  In: pszModulePath - the path of this module
//
//  Out:none
//
//  Comments:
///////////////////////////////////////////////////////////////////
BOOL InsertUsInTPHChain(char *pszModulePath)
{
	LONG lRet=0;
	HKEY hKey=NULL;
	BOOL bRet=FALSE, bFoundOurselves=FALSE;
	UINT uiIndex=0;
	DWORD dwNoOfValues=0, dwTempBufLen=0, dwMaxValueLen=0;
	char szTmpBuffer[100]={0};
	char *pszValue=NULL, *pszFileName=NULL;

	//check parameters passed to us
	pgpAssert(NULL != pszModulePath);
	if(NULL == pszModulePath)
		return FALSE;

	//make the TPH related entries. open the key at HKCU\Software\Novell\GroupWise\Client\Third Party
	//if it does not exist create the key
	pgpAssert(NULL == hKey);
	lRet=RegCreateKeyEx(HKEY_CURRENT_USER, TPHKEY, 0, NULL, NULL, KEY_ALL_ACCESS,
			NULL, &hKey, NULL);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hKey));
	if((ERROR_SUCCESS != lRet) || (NULL == hKey))
		goto cleanup;

	//get the number and max length of values involved
	pgpAssert(NULL != hKey);
	lRet=RegQueryInfoKey(hKey, NULL, NULL, NULL, NULL, NULL, NULL,
			&dwNoOfValues, NULL, &dwMaxValueLen, NULL, NULL);
	pgpAssert(ERROR_SUCCESS == lRet);
	if(ERROR_SUCCESS != lRet)
		goto cleanup;

	//allocate the value buffer
	pgpAssert(NULL == pszValue);
	pszValue=(char *)calloc(dwMaxValueLen+1, 1);
	pgpAssert(NULL != pszValue);
	if(NULL == pszValue)
		goto cleanup;

	//check if we already appear in the chain
	//for each of the DLLx values under the registry key
	for(uiIndex=1; uiIndex <= dwNoOfValues; uiIndex++)
	{
		//prepare the value name
		ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
		sprintf(szTmpBuffer, "DLL%d", uiIndex);

		//read the value data
		pgpAssert(NULL != hKey);
		dwTempBufLen=dwMaxValueLen;
		lRet = RegQueryValueEx(hKey, szTmpBuffer, 0, NULL, 
						(LPBYTE)pszValue, &dwTempBufLen);
		pgpAssert((ERROR_SUCCESS == lRet) && (dwTempBufLen > 0));
		pgpAssert(dwTempBufLen < dwMaxValueLen);//damaged TPH chain ?

		//get at the file name from the path
		pszFileName=strrchr(pszValue, '\\');//find the last slash
		pgpAssert(NULL != pszFileName);
		if(NULL == pszFileName)
			continue;//junk entry in chain. skip it.
		pszFileName++;

		//if the entry is an entry for us
		if(0 == strcmpi(pszFileName, MODULEFILENAME))
		{
			//we will update the path instead of inserting into the chain again
			//this puts to rest all problems because of double registration
			lRet = RegSetValueEx(hKey, szTmpBuffer, 0, REG_SZ, (const BYTE*)pszModulePath, 
						lstrlen(pszModulePath));
			pgpAssert(ERROR_SUCCESS == lRet);
			if(ERROR_SUCCESS == lRet)
				bRet=TRUE;//success

			goto cleanup;//we are done
		}
	}

	//the control comes here only if we could not find ourselves in the
	//TPH chain. so let us set ourselves as the last entry in the chain
	ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	sprintf(szTmpBuffer, "DLL%d", dwNoOfValues+1);
    lRet = RegSetValueEx(hKey, szTmpBuffer, 0, REG_SZ, (const BYTE*)pszModulePath, lstrlen(pszModulePath));
	pgpAssert(ERROR_SUCCESS == lRet);
	if(ERROR_SUCCESS == lRet)
		bRet=TRUE;//success

cleanup:
	if(NULL != pszValue)
	{
		free(pszValue);
		pszValue=NULL;
	}

	if(NULL != hKey)
	{
		RegCloseKey(hKey);
		hKey=NULL;
	}

	return bRet;
}

///////////////////////////////////////////////////////////////////
//  Name:RegServer
//
//  Description:Sets up the system registery for the plugin
//
//  In:none
//
//  Out:none
//
//  Comments:
///////////////////////////////////////////////////////////////////
void RegServer()
{
	char szTmpBuffer[512]={0};
	char szTmpBuffer2[100]={0};
	HMODULE hModPgpGw=NULL;
	DWORD dwNoOfValues=0;
	
	HKEY hKey=NULL, hSubKey=NULL, hSubKey1=NULL;
	LONG lRet=ERROR_SUCCESS;

	//open/create the key at HKLM\Software\Novell\GroupWise\5.0\C3PO\DataTypes
	pgpAssert(NULL == hKey);
	lRet = RegCreateKey(HKEY_LOCAL_MACHINE, C3POKEY, &hKey);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hKey));

	//create the modulename\objects hierarchy under GW.MESSAGE.MAIL subkey
	pgpAssert(NULL == hSubKey);
	sprintf(szTmpBuffer, "%s\\%s\\Objects", MAILMESSAGECONTEXTA, MODULENAME);
	lRet = RegCreateKey(hKey, szTmpBuffer, &hSubKey);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hSubKey));

    //create a "CommandFactory" value of null value
	pgpAssert(NULL != hSubKey);
	lRet = RegSetValueEx(hSubKey, "CommandFactory", 0, REG_SZ,(const BYTE*)"", 1);
	pgpAssert(ERROR_SUCCESS == lRet);
    RegCloseKey(hSubKey);
	hSubKey=NULL;

   //create the modulename\events hierarchy under GW.MESSAGE.MAIL subkey
	//NOTE: we do not create any values under this newly created hierarchy
	ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	sprintf(szTmpBuffer, "%s\\%s\\Events", MAILMESSAGECONTEXTA, MODULENAME);
	pgpAssert((NULL != hKey) && (NULL == hSubKey));
	lRet = RegCreateKey(hKey, szTmpBuffer, &hSubKey);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hSubKey));

	//create a value to trap the open event (GW#C#OPEN)
	pgpAssert(NULL != hSubKey);
	lRet = RegSetValueEx(hSubKey, eGW_CMDID_OPEN, 0, REG_SZ,(const BYTE*)"", 1);
	pgpAssert(ERROR_SUCCESS == lRet);
    RegCloseKey(hSubKey);
	hSubKey=NULL;
 
	//create the modulename\objects hierarchy under GW.CLIENT.WINDOW.BROWSER subkey
	pgpAssert(NULL == hSubKey);
	sprintf(szTmpBuffer, "%s\\%s\\Objects", MAINWINDOWCONTEXTA, MODULENAME);
	lRet = RegCreateKey(hKey, szTmpBuffer, &hSubKey);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hSubKey));

    //create a "CommandFactory" value of null value
	pgpAssert(NULL != hSubKey);
	lRet = RegSetValueEx(hSubKey, "CommandFactory", 0, REG_SZ,(const BYTE*)"", 1);
	pgpAssert(ERROR_SUCCESS == lRet);
    RegCloseKey(hSubKey);
	hSubKey=NULL;

	//create the modulename\objects hierarchy under GW.CLIENT.WINDOW.ATTACHVIEWER subkey
	//pgpAssert(NULL == hSubKey);
	//sprintf(szTmpBuffer, "%s\\%s\\Objects", ATTCHVIEWERCONTEXTA, MODULENAME);
	//lRet = RegCreateKey(hKey, szTmpBuffer, &hSubKey);
	//pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hSubKey));

    //create a "CommandFactory" value of null value
	//pgpAssert(NULL != hSubKey);
	//lRet = RegSetValueEx(hSubKey, "CommandFactory", 0, REG_SZ,(const BYTE*)"", 1);
	//pgpAssert(ERROR_SUCCESS == lRet);
    //RegCloseKey(hSubKey);
	//hSubKey=NULL;

    //create the modulename\objects hierarchy under GW.CLIENT subkey
	ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	sprintf(szTmpBuffer, "%s\\%s\\Objects", CLIENTCONTEXTA, MODULENAME);
	pgpAssert((NULL != hKey) && (NULL == hSubKey));
	lRet = RegCreateKey(hKey, szTmpBuffer, &hSubKey);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hSubKey));

    //create an EventMonitor value of null string in that key
	lRet = RegSetValueEx(hSubKey, "EventMonitor", 0, REG_SZ,(const BYTE*)"", 1);
	pgpAssert(ERROR_SUCCESS == lRet);
    RegCloseKey(hSubKey);
	hSubKey=NULL;
    
   	//create the modulename\events hierarchy under GW.CLIENT subkey
	ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	sprintf(szTmpBuffer, "%s\\%s\\Events", CLIENTCONTEXTA, MODULENAME);
	pgpAssert((NULL != hKey) && (NULL == hSubKey));
	lRet = RegCreateKey(hKey, szTmpBuffer, &hSubKey);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hSubKey));

    //create "OnReady" and "OnShutdown" values under that key
	lRet = RegSetValueEx(hSubKey, "OnReady", 0, REG_SZ, (const BYTE*)"", 1);
	pgpAssert(ERROR_SUCCESS == lRet);
    lRet = RegSetValueEx(hSubKey, "OnShutdown", 0, REG_SZ, (const BYTE*)"", 1);
	pgpAssert(ERROR_SUCCESS == lRet);
    RegCloseKey(hSubKey);
	hSubKey=NULL;

    
	RegCloseKey(hKey);
	hKey=NULL;

	//prepare the path of this server from the location
	//where it lies at the point of registration
	ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	hModPgpGw=GetModuleHandle(MODULENAME);
	pgpAssert(NULL != hModPgpGw);
	_VERIFY(0 != GetModuleFileName(hModPgpGw, szTmpBuffer, sizeof(szTmpBuffer)));
	
	//put ourselves in the TPH chain
	_VERIFY(TRUE == InsertUsInTPHChain(szTmpBuffer));

	//setup the COM infrastructure for ourselves

    //create the module name key under classes root 
	pgpAssert(NULL == hSubKey);
	lRet = RegCreateKey(HKEY_CLASSES_ROOT, MODULENAME, &hSubKey);
	pgpAssert((ERROR_SUCCESS == lRet) &&(NULL != hSubKey));

    //put the description as the unnamed(default) value of the key
	char *p = MODULEDESCRIPTION;
    lRet = RegSetValueEx(hSubKey, NULL, 0, REG_SZ, (const BYTE*)p,lstrlen(p));
	pgpAssert(ERROR_SUCCESS == lRet);

    //create a clsid key under the module name key
	lRet = RegCreateKey(hSubKey,"CLSID", &hSubKey1);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hSubKey1));
	RegCloseKey(hSubKey);
	hSubKey=NULL;

    //set the CLSID string as the default value of the key
	p = MODULECLSID;
    lRet = RegSetValueEx(hSubKey1, NULL, 0, REG_SZ, (const BYTE*)p, lstrlen(p));
	pgpAssert(ERROR_SUCCESS == lRet);
    RegCloseKey(hSubKey1);
	hSubKey1=NULL;
    
    //set up the reverse lookup infrastructure. so lets open the clsid key
	//under class root hive
	lRet = RegOpenKey(HKEY_CLASSES_ROOT,"CLSID", &hKey);
	pgpAssert((ERROR_SUCCESS == lRet) &&(NULL != hKey));

	//create the clsid string under that as a key
    lRet = RegCreateKey(hKey, MODULECLSID, &hSubKey);
	pgpAssert((ERROR_SUCCESS == lRet) &&(NULL != hSubKey));
	RegCloseKey(hKey);
	hKey=NULL;

    //set the description as the default value of the key
	p = MODULEDESCRIPTION;
    lRet = RegSetValueEx(hSubKey, NULL, 0, REG_SZ, (const BYTE*)p,lstrlen(p));
	pgpAssert(ERROR_SUCCESS == lRet);

    //create ProgId key under that
	lRet = RegCreateKey(hSubKey,"ProgID", &hSubKey1);
	pgpAssert((ERROR_SUCCESS == lRet) &&(NULL != hSubKey1));

	//put the module name as the default value of the key
	//NOTE:the version info is normally put at the end as a dotted suffix
    p = MODULENAME;
    lRet = RegSetValueEx(hSubKey1, NULL, 0, REG_SZ, (const BYTE*)p,lstrlen(p));
	pgpAssert(ERROR_SUCCESS == lRet);
    RegCloseKey(hSubKey1);
	hSubKey1=NULL;

    //create inprocserver32 key under the module clsid key
	lRet = RegCreateKey(hSubKey,"InProcServer32", &hSubKey1);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hSubKey1));

	//set the default value of the key to the path of the server
    p = szTmpBuffer;
    lRet = RegSetValueEx(hSubKey1, NULL, 0, REG_SZ, (const BYTE*)p,lstrlen(p));
	pgpAssert(ERROR_SUCCESS == lRet);
    RegCloseKey(hSubKey1);
	hSubKey1=NULL;
}


///////////////////////////////////////////////////////////////////
//  Name:RemoveUsFromTPHChain
//
//  Description:Removes us from the TPH chain and makes sure that
//		 other TPH's behave well after we unregister
//
//  In:none
//
//  Out:none
//
//  Comments:
///////////////////////////////////////////////////////////////////
BOOL RemoveUsFromTPHChain()
{
	char szPgpGwPath[1024]={0};
	HKEY hKey=NULL;
	HMODULE hModPgpGw=NULL;
	LONG lRet=ERROR_SUCCESS;
	DWORD dwNoOfValues=0, dwMaxValueNameLen=0, dwMaxValueLen=0, dwTempBufLen=0;
	char *pszValueName=NULL, *pszValue=NULL;
	BOOL bRet=FALSE, bFoundOurselves=FALSE;
	unsigned int uiIndex=0;

	//get the full path of pgpgw module
	ZeroMemory(szPgpGwPath, sizeof(szPgpGwPath));
	hModPgpGw=GetModuleHandle(MODULENAME);
	pgpAssert(NULL != hModPgpGw);
	_VERIFY(0 != GetModuleFileName(hModPgpGw, szPgpGwPath, sizeof(szPgpGwPath)));
	
	//open the key at HKCU\Software\Novell\GroupWise\Client\Third Party
	pgpAssert(NULL == hKey);
	lRet=RegOpenKey(HKEY_CURRENT_USER, TPHKEY, &hKey);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hKey));

	//get the number of values, the max length of the value name and value data
	pgpAssert(NULL != hKey);
	lRet=RegQueryInfoKey(hKey, NULL, NULL, NULL, NULL, NULL, NULL,
			&dwNoOfValues, &dwMaxValueNameLen, &dwMaxValueLen, NULL, NULL);
	pgpAssert(ERROR_SUCCESS == lRet);

	//allocate appropriately for the strings used to store the value name and value data
	pszValueName=(char *)calloc(dwMaxValueNameLen+1, 1);
	pgpAssert(NULL != pszValueName);

	pszValue=(char *)calloc(dwMaxValueLen+1, 1);
	pgpAssert(NULL != pszValue);

	//if either of the allocations failed
	if((NULL == pszValueName) || (NULL == pszValue))
		goto cleanup;

	//for each of the DLLx values under the registry key
	for(uiIndex=1; uiIndex <= dwNoOfValues; uiIndex++)
	{
		//prepare the value name
		sprintf(pszValueName, "DLL%d", uiIndex);

		//read the value data
		pgpAssert(NULL != hKey);
		dwTempBufLen=dwMaxValueLen;
		lRet = RegQueryValueEx(hKey, pszValueName, 0, NULL, 
						(LPBYTE)pszValue, &dwTempBufLen);
		pgpAssert(ERROR_SUCCESS == lRet);

		if(FALSE == bFoundOurselves)//if we have not found ourselves yet
		{
			//check to see if it is us
			if(0 == strcmpi(pszValue, szPgpGwPath))
				bFoundOurselves=TRUE;
		}
		else//ie. we have found the entry for pgpgw.dll
		{
			//time to keep pushing back all the tph's registered after pgpgw

			//prepare the value name for the previous entry
			sprintf(pszValueName, "DLL%d", uiIndex-1);

			//set the previous value to this value
			lRet=RegSetValueEx(hKey, pszValueName, 0, REG_SZ, 
							(const BYTE*)pszValue, lstrlen(pszValue));
			pgpAssert(ERROR_SUCCESS == lRet);
		}
	}

	if(TRUE == bFoundOurselves)//if we did find ourselves
	{
		//we must remove the last DLLx entry as the data has been
		//put into the last but one entry
		sprintf(pszValueName, "DLL%d", dwNoOfValues);
		lRet=RegDeleteValue(hKey, pszValueName);
		pgpAssert(ERROR_SUCCESS == lRet);
	}

	bRet=TRUE;

cleanup:
	if(NULL != hKey)
	{
		RegCloseKey(hKey);
		hKey=NULL;
	}
	if(NULL != pszValueName)
	{
		free(pszValueName);
		pszValueName=NULL;
	}
	if(NULL != pszValue)
	{
		free(pszValue);
		pszValue=NULL;
	}
	return bRet;
}

///////////////////////////////////////////////////////////////////
//  Name:UnRegServer
//
//  Description:Unregisters the pgpgw plugin
//
//  In:none
//
//  Out:none
//
//  Comments:
///////////////////////////////////////////////////////////////////
void UnRegServer()
{
	char szTmpBuffer[1024]={0};
	HKEY hKey=NULL, hSubKey=NULL;
	LONG lRet=ERROR_SUCCESS;

	//open the key at HKLM\Software\Novell\GroupWise\5.0\C3PO\DataTypes
	lRet = RegOpenKey(HKEY_LOCAL_MACHINE, C3POKEY, &hKey);
	pgpAssert((ERROR_SUCCESS == lRet) && (NULL != hKey));

   	//delete the module name key under GW.CLIENT
	ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	sprintf(szTmpBuffer, "%s\\%s", CLIENTCONTEXTA, MODULENAME);
	lRet = SHDeleteKey(hKey, szTmpBuffer);
	pgpAssert(ERROR_SUCCESS == lRet);

   	//delete the module name key under GW.CLIENT.WINDOW.BROWSER
	ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	sprintf(szTmpBuffer, "%s\\%s", MAINWINDOWCONTEXTA, MODULENAME);
	lRet = SHDeleteKey(hKey, szTmpBuffer);
	pgpAssert(ERROR_SUCCESS == lRet);

   	//delete the module name key under GW.CLIENT.WINDOW.ATTACHVIEWER
	//ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	//sprintf(szTmpBuffer, "%s\\%s", ATTCHVIEWERCONTEXTA, MODULENAME);
	//lRet = SHDeleteKey(hKey, szTmpBuffer);
	//pgpAssert(ERROR_SUCCESS == lRet);
   	
	//delete the module name key under GW.MESSAGE.MAIL
	ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	sprintf(szTmpBuffer, "%s\\%s", MAILMESSAGECONTEXTA, MODULENAME);
	lRet = SHDeleteKey(hKey, szTmpBuffer);
	pgpAssert(ERROR_SUCCESS == lRet);

	RegCloseKey(hKey);
	hKey=NULL;
    
	//remove ourselves from the TPH chain
	_VERIFY(TRUE == RemoveUsFromTPHChain());

	//cleanup com entries

	//delete the HKEY_CLASSES_ROOT\<module name> entry
	lRet = SHDeleteKey(HKEY_CLASSES_ROOT, MODULENAME);
	pgpAssert(ERROR_SUCCESS == lRet);

	//delete the HKEY_CLASSES_ROOT\clsid\<module clsid> entry
	ZeroMemory(szTmpBuffer, sizeof(szTmpBuffer));
	sprintf(szTmpBuffer, "%s\\%s", "CLSID", MODULECLSID);
	lRet = SHDeleteKey(HKEY_CLASSES_ROOT, szTmpBuffer);
	pgpAssert(ERROR_SUCCESS == lRet);
}


