////////////////////////////////////////////////////////////////////////////////
//
//  File : pgpgwtph.cpp
//
//  Copyright (C) 2002 PGP Corporation
//
//  ABSTRACT
//		The interaction between GroupWise and third-party DLLs is as follows:
//  GroupWise loads third-party DLLs at startup, before any C3POs are loaded.
//  GroupWise calls Entry() method once initially.GroupWise passes tokens to 
//  third-party DLL throught the HandleToken() method. DLLs return values 
//	indicating:
//		a. Let a token pass to GroupWise.
// 		b. Pass a token unprocessed or use a token value as a trigger to process 
//		   a token, then pass it on, indicating it was unprocessed.
//		c. Generate and pass new tokens.
//		d. Block a token.
//  	e. Return a token, indicating it was processed. Processed tokens are 
//		   not passed to other DLLs or GroupWise. The DLL that processes a token 
//		   must provide required return values.
//      f. Modify a token.
//  
//	
// 	We register our dll at the location
//  
//		Key	"HKEY_CURRENT_USER\Software\Novell\GroupWise\Client\Third Party"
//		Value Name "DLLx" (where x is a number starting at 1.  DLL1 is always 
//							loaded before DLL2, etc.)
//  	Value Type	STRING
//		Value Value	<DLL path and filename>
//  
//
//  The method calls are by ordinal. so please do NOT change the ordinals. note
//  that we are using 1 byte alignment for the structures tossed around with gw
//  the rest of the structures are 8byte aligned as is default in VC6 and is nice
//  to Win32 and WinNT.
//
//  Author: Satya S. Das
//
////////////////////////////////////////////////////////////////////////////////

#include "windows.h"
#include <stdio.h>
//we want the params passed around in this file to be packed as is
//expected by groupwise
#if PGP_WIN32
#pragma pack(push, 1)
#endif
#include "gwdll.h"
#include "gwoapi.h"
#if PGP_WIN32
#pragma pack(pop)
#endif
#include <objbase.h>
#include <crtdbg.h>
#include "pgpdebug.h"
#include "ofcmndr.h"
#include "globdefs.h"
#include "globals.h"
#include "util.h"
#include "resource.h"
#include "encryptsign.h"
#include "decryptverify.h"
#include <direct.h>
//#include "autodecrypt.h"
#include <shlwapi.h>


HRESULT Publish(LPCSTR szTokenBuf, BSTR *pbstrResult);
void TraceHandleToken(LPTPH_RETURNVAL lpTokenData, HWND	hLinkWnd, WORD wMsg);
HRESULT IsMessageAMail(BSTR bstrMessageId, BOOL &bRet, int *piMessageType=NULL);
BOOL PrepareMsgRecipientsArray(BSTR bstrMessageId, char ***pppszRecipients, DWORD& dwNoOfRecips);
//CAutoDecrypt *g_cadFileWatchAndDecrypt=NULL;

////////////////////////////////////////////////////////////////////////////////
// Name: IsMessageAMail
// Description:	This method is not part of the API.  It is a utility method
//				to help determine if the message is a mail.
////////////////////////////////////////////////////////////////////////////////
HRESULT IsMessageAMail(BSTR bstrMessageId, BOOL &bRet, 
					   int *piMessageType/*=NULL*/)
{
	HRESULT hrRet=S_OK;
	char szTokenBuf[255]={0};
	BSTR bstrResult=NULL;

	//check parameters
	pgpAssert(NULL != bstrMessageId);//message id must be passed
	if(NULL == bstrMessageId)
	{
		hrRet=E_POINTER;
		goto cleanup;
	}

	//if messagetype output ptr was passed to us
	if(NULL != piMessageType)
	{
		//make sure the pointer is ok
		pgpAssert(!IsBadWritePtr(piMessageType, sizeof(int)));
		if(IsBadWritePtr(piMessageType, sizeof(int)))
		{
			hrRet=E_POINTER;
			goto cleanup;
		}
	}

	//get the type of this message being sent
	sprintf(szTokenBuf, "ItemGetType(\"%S\")", bstrMessageId);
	hrRet=Publish(szTokenBuf, &bstrResult);
	if(FAILED(hrRet))
	{
		pgpAssert(FALSE);
		goto cleanup;
	}

	//check if the message being sent is a mail. mail messages are of type "2"
	bRet = (0 == wcscmp(bstrResult, MAILMESSAGETYPEW))?TRUE:FALSE;

	//if we are supposed to give the integer message type additionally
	if(NULL != piMessageType)
	{
		*piMessageType=_wtol(bstrResult);
		pgpAssert(0 != *piMessageType);//no error in conversion
		pgpAssert((0 < *piMessageType) && (6 > *piMessageType));//only 5 possible values in GW5.2
	}

	cleanup:
	if(NULL != bstrResult)
	{
		SysFreeString(bstrResult);
		bstrResult=NULL;
	}
	
	return hrRet;
}

HRESULT FillTokenInfo(GWTOKENINFO &gtiTokenInfo, LPTPH_RETURNVAL lpTokenData, HWND hLinkWnd)
{
	pgpAssert(NULL != lpTokenData);
	pgpAssert((NULL != lpTokenData) && (NULL != lpTokenData->lpToken));
	
	HRESULT hrRet=S_OK;

	//copy the token id into the structure
	if((NULL != lpTokenData) && (NULL != lpTokenData->lpToken))
		gtiTokenInfo.wTokenID=lpTokenData->lpToken->wTokenId;
	else
		gtiTokenInfo.wTokenID=BFTKN_INVALID_TOKEN;

	//copy the window handle
	gtiTokenInfo.hLinkWnd=hLinkWnd;
	
	//get the message id involved
	hrRet=Publish("ItemMessageIDFromView()", &(gtiTokenInfo.bstrMessageId));
	if(FAILED(hrRet))
	{
		pgpAssert(FALSE);//failed to get the message id
		goto cleanup;
	}

	//check if the message is a mail
	gtiTokenInfo.bIsMailMsg=FALSE;//reset
	hrRet=IsMessageAMail(gtiTokenInfo.bstrMessageId, gtiTokenInfo.bIsMailMsg);
	if(FAILED(hrRet))
	{
		pgpAssert(FALSE);//failed to obtain the message type !!
		goto cleanup;
	}
	
	cleanup:
	return hrRet;
}

////////////////////////////////////////////////////////////////////////////////
// Name: Publish
// Description:	This method is not part of the API.  It is a utility method
//				to help publish tokens through the Token Commander.
////////////////////////////////////////////////////////////////////////////////
HRESULT Publish(LPCSTR szTokenBuf, BSTR *pbstrResult)
{
	HRESULT hrRet=S_OK;
	BSTR bstrTokenBuf=NULL;
	VARIANT_BOOL vtbResult=TRUE;

	//check parameters
	pgpAssert(NULL != szTokenBuf);
	pgpAssert(NULL != pbstrResult);
	if((NULL == szTokenBuf) || (NULL == pbstrResult))
	{
		hrRet=E_POINTER;
		goto cleanup;
	}

	//the token buffer must have not null contents
	if(NULL == *szTokenBuf)
	{
		hrRet=E_POINTER;
		goto cleanup;
	}

	//check if we have the commander interface handy
	pgpAssert(NULL != g_pigwcCommander);
	if(NULL == g_pigwcCommander)
	{
		hrRet=E_FAIL;
		goto cleanup;
	}
	
	//free the result buffer if anything is in there
	if(NULL != *pbstrResult)
	{
		SysFreeString(*pbstrResult);
		*pbstrResult=NULL;
	}

	//make a bstr of the token string
	pgpAssert(NULL == bstrTokenBuf);
	bstrTokenBuf=A2WBSTR(szTokenBuf);
	pgpAssert(NULL != bstrTokenBuf);
	if(NULL == bstrTokenBuf)
	{
		hrRet=E_OUTOFMEMORY;
		goto cleanup;
	}


	//attempt to publish the token
	hrRet=g_pigwcCommander->Execute(bstrTokenBuf, pbstrResult, &vtbResult);
	if(FAILED(hrRet) || (FALSE == vtbResult) || (NULL == *pbstrResult))
	{
		pgpAssert(FALSE);
		PgpGwTrace("Execute on %S returned error %S\n", 
				bstrTokenBuf, *pbstrResult);
		hrRet=E_FAIL;
		goto cleanup;
	}

	cleanup:
	if(NULL != bstrTokenBuf)
	{
		SysFreeString(bstrTokenBuf);
		bstrTokenBuf=NULL;
	}

	if(FAILED(hrRet))
	{
		if(NULL != *pbstrResult)
		{
			SysFreeString(*pbstrResult);
			*pbstrResult=NULL;
		}
	}
	return hrRet;

 	/*IGWCommander *pGWC=NULL;
  	VARIANT_BOOL vSuccess;
	BSTR bsReturn=NULL;
	LPTSTR lpReturn=NULL;
	WORD wSize = 256;

	if (SUCCEEDED(CoCreateInstance(CLSID_GWCommander, NULL,
			CLSCTX_ALL,	IID_IGWCommander, (LPVOID *) &pGWC)))
	{
		pGWC->Execute(bsToken, &bsReturn, &vSuccess);
		pGWC->Release();

		if (TRUE == vSuccess)
		{
			if (NULL != bsReturn)
			{
				lpReturn = new TCHAR[wSize];
				if (lpReturn)
				{	
					WideCharToMultiByte(CP_ACP, 0, bsReturn, -1, lpReturn, wSize, NULL, NULL);  
				}
			}
			return lpReturn;
		}
	}

	return NULL;*/
}


////////////////////////////////////////////////////////////////////////////////
// Name: TPHVersion
// Descrip:	This method is reserved for later use.  Even though it is currently
//			not called by GroupWise, it must be included.
////////////////////////////////////////////////////////////////////////////////
DWORD WINAPI TPHVersion (void) 
{
	return 0;
}


////////////////////////////////////////////////////////////////////////////////
//
// Name: Compatibility
// Description: This method lets a DLL specify the GroupWise version to load it 
//		with and ensures compatibility between GroupWise and third-party DLLs. 
//		We can verify major and minor GW versions, then return	GroupWise a value 
//		indicating whether to load ourselves.  A Novell application name and 
//		GroupWise version are passed as parameters.
//				
//		The GroupWise AppName is GROUPWISE. The AppVersion is a word value with 
//		the low-order byte containing the GroupWise major version, and the 
//		high-order byte containing the minor version. 
//
//		currently we are compatible with GW 5.2 and later. The LOBYTE of AppVersion 
//		will be 5, and the HIBYTE will 2 gor GW 5.2. One can check these values to 
//		see which version of GW is running.
//
// Returns:To indicate that you want the DLL to be loaded, return a value of 
//		(DWORD) TRUE.  To indicate that you don't want it to be	loaded, return a 
//		value of (DWORD) FALSE.
//
////////////////////////////////////////////////////////////////////////////////
DWORD WINAPI Compatibility(ATOM AppAtom, WORD AppVersion) 
{
	if ((MINIMUMGWMAJORVERSUPPORTED <= LOBYTE(AppVersion)) && 
		(MINIMUMGWMINORVERSUPPORTED <= HIBYTE(AppVersion)))
	{
		//we want to be loaded
		return (DWORD) TRUE;
	}
	else 
	{
		//incompatible, lets not get loaded
		return (DWORD) FALSE;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Name: TimeStamp
// Descrip:	This method is reserved for later use.  Even though it is currently
//			not called by GroupWise, it must be included.
//					
// Returns: 0
////////////////////////////////////////////////////////////////////////////////
DWORD WINAPI TimeStamp(void) 
{
	 return 0;
}


////////////////////////////////////////////////////////////////////////////////
// Name:Entry
// Descrip:	This is where the third party DLL sets up its necessary data 
//			structures, memory allocations, initializations, etc.  Also, it tells 
//			us the language under which GW is running. If we do not support this 
//			language, we can bail out and not load ourselves.
//
//			For example, if we support the language in the US, the LOBYTE will be 
//			'U', and HIBYTE will be 'S'.
//
// Returns: After examining the language code, to load our DLL,	return a (DWORD)1.  
//			If not, return (DWORD) 0.
////////////////////////////////////////////////////////////////////////////////
WORD WINAPI Entry(WORD wLanguage) 
{
	char *pszLangSupported=NULL;

	//get the language string from string table.For US English it is "US".
	BOOL bRet=SafeLoadString(IDS_PLUGIN_LANGUAGE, &pszLangSupported);
	pgpAssert(TRUE == bRet);
	pgpAssert(NULL != pszLangSupported);
	if((FALSE == bRet) || (NULL == pszLangSupported))
	{
		bRet=FALSE;//dont load ourselves
		goto cleanup;
	}
	
	//compare the language this module is supporting and the
	//language word passed to us by GW
	pgpAssert(2 == strlen(pszLangSupported));//should be a 2 character string
	if ((pszLangSupported[0] == (char)LOBYTE(wLanguage)) && 
		(pszLangSupported[1] == (char) HIBYTE(wLanguage))) 
	{
		bRet=TRUE;//load ourselves
		goto cleanup;
	}
	else 
	{
		bRet=FALSE;//dont load ourselves
		goto cleanup;
	}

	cleanup:
	if(NULL != pszLangSupported)
	{
		free(pszLangSupported);
		pszLangSupported=NULL;
	}

	return bRet;
}


////////////////////////////////////////////////////////////////////////////////
// Name: Exit
// Description:The exit method is where your third party DLL frees up memory,
//		closes files, or does any other activities to prepare for closing down.  
//		Simply return a (WORD) 1 when finished.
//
// Returns: (WORD) 1
////////////////////////////////////////////////////////////////////////////////
WORD WINAPI Exit(void) 
{
	return (WORD) 1;
}


////////////////////////////////////////////////////////////////////////////////
// Name: HandleToken
// Description:	The HandleToken method is where GroupWise tokens are routed
//				to your DLL for processing.  You have the ability to process
//				the token and have GroupWise process it, or prevent GroupWise
//				from processing, or even create new tokens to be routed through
//				to GroupWise.
//
//				In this example, we will demonstrate two of these concepts
//				by capturing the token associated with bring up the About Dialog
//				Box (Select Help | About GroupWise, from the menu).
//
//				A dialog box will come up allowing you to either ignore the token
//				and let GroupWise process it, or process it instead of GroupWise.
//				Thus, you'll be able to see how your DLL can interact with and
//				influence how GroupWise processes requests.
//
//
// Returns: The following return values are possible:
//		DLL_HAN_NOT_HANDLED  The token was not processed.
//		DLL_HAN_NO_ERROR	 The token was processed. This value 
//							 should be returned to block a token.
//		DLL_HAN_NOT_FOUND	 The token resulted in a GroupWise "not found" 
//							 condition.
//		DLL_HAN_CANCEL		 The user cancelled the method.
//		DLL_HAN_TOKEN_ERROR  The token was not valid.
//		DLL_HAN_PARM_ERROR	 One or more of the parameters were invalid.
//
////////////////////////////////////////////////////////////////////////////////
int WINAPI HandleToken(LPTPH_RETURNVAL lpTokenData, //received token
					   HWND	hLinkWnd, //handle to the link window
					   WORD	wMsg)     //link window execute message.
{
	LPTSTR lpReturn=NULL;//used with the Publish method
	TraceHandleToken(lpTokenData, hLinkWnd, wMsg);

	switch(lpTokenData->lpToken->wTokenId)
	{
		case BFTKN_SEND_STDMAIL://new mail item
			return DLL_HAN_NOT_HANDLED;
			break;

		case BFTKN_READ:
			/*{
				HRESULT hrRet=FillTokenInfo(g_gtiLastTokenInfo, lpTokenData, hLinkWnd);
				if(FAILED(hrRet))
				{
					pgpAssert(FALSE);//some error in getting basic information about this message
					goto send_stdmail_cleanup;
				}
				
				if(FALSE == g_gtiLastTokenInfo.bIsMailMsg)
				{
					PgpGwTrace("Skipping one non-mail item opened.\n");
					return DLL_HAN_NOT_HANDLED;//its ok for the client to go ahead and do its stuff
					goto send_stdmail_cleanup;
				}
				else
				{
					//must be a new message
					pgpAssert(0 == wcscmp(g_gtiLastTokenInfo.bstrMessageId, NEWMESSAGEMSGIDW));
				}

				send_stdmail_cleanup:
				return DLL_HAN_NOT_HANDLED;
			}
			break;*/
			return DLL_HAN_NOT_HANDLED;
			break;

		case BFTKN_VIEW_ATTACH:
			//start watching file creation in the directory
			//if(NULL == g_cadFileWatchAndDecrypt)
			//{
			//	g_cadFileWatchAndDecrypt = new CAutoDecrypt;
			//	pgpAssert(NULL != g_cadFileWatchAndDecrypt);
			//}

			//if(NULL != g_cadFileWatchAndDecrypt)
			//	g_cadFileWatchAndDecrypt->Start();
			return DLL_HAN_NOT_HANDLED;
			break;
		case BFTKN_SEND:
			{
				HRESULT hrRet=S_OK;
				HWND hCtxtWindow=NULL;
				BSTR bstrResult=NULL;
				DWORD dwNoOfRecips=0;
				int iHandleTokenRet=DLL_HAN_NO_ERROR;
				char szTokenBuf[200]={0};
				int iPlainTxtMsgLen=0;
				BOOL bSkip = FALSE, bEncrypt=FALSE, bSign=FALSE;
				int iNoOfAtchMents=0, iAtchMentInd=0, iAtchMentType=0;

				//PGP structs
				PGPclRecipientDialogStruct *prdsRecipDlgStruct=NULL;
				PGPError pgpErrRet=kPGPError_NoErr;
				PGPtlsContextRef tlsContext=NULL;
				PGPOptionListRef polOptions=NULL;
				PGPSize pgpsBufOutLen=0;
				PGPUInt32 pgpuiNumOfKeys=0;
				PGPKeySetRef pgpksrKeySet=NULL;
				char *pOutBuf=NULL;
				char *pFinalBuf=NULL;
				char *pAnsiInputBuf=NULL;
				char *pszTitle=NULL;
				char **ppszRecipients=NULL;
				PMSGINFO lpMsgInfo=NULL;
				int iRet=0;
				DWORD dwRet=0;
				BOOL bRet=TRUE;
				BOOL bAbort=FALSE;
				
				//check if we have expired. just abort silently - since the user would 
				//have already got a popup for expiry at the point of first check
				if(TRUE == g_bPlugInExpired)
				{
					PgpGwTrace("Skipping one item since we are not licensed.\n");
					iHandleTokenRet=DLL_HAN_NOT_HANDLED;
					goto send_cleanup;
				}

				hrRet=FillTokenInfo(g_gtiLastTokenInfo, lpTokenData, hLinkWnd);
				if(FAILED(hrRet))
				{
					pgpAssert(FALSE);//some error in getting basic information about this message
					DisplayMessage(GetActiveWindow(), IDS_E_FATAL);
					goto send_cleanup;
				}
				
				if(FALSE == g_gtiLastTokenInfo.bIsMailMsg)
				{
					PgpGwTrace("Skipping one non-mail item sent.\n");
					iHandleTokenRet=DLL_HAN_NOT_HANDLED;//its ok for the client to go ahead and do its stuff
					goto send_cleanup;
				}

				//get the current mail window handle
				hCtxtWindow=GetWindow(GWO_CURRENTMAIL_WINDOW);
				pgpAssert(NULL != hCtxtWindow);
				if(NULL == hCtxtWindow)
				{
					hrRet=E_FAIL;
					DisplayMessage(GetActiveWindow(), IDS_E_CURRENT_MAILWND_NOT_FOUND);
					goto send_cleanup;
				}

				//the hash must have been setup already
				pgpAssert(NULL != g_pccihCmdHash);
				if(NULL == g_pccihCmdHash)
				{
					hrRet=E_FAIL;
					DisplayMessage(hCtxtWindow, IDS_E_FATAL);
					goto send_cleanup;
				}

				//get the context info on this message
				lpMsgInfo=g_pccihCmdHash->Get((ULONG)hCtxtWindow);
				pgpAssert(NULL != lpMsgInfo);
				if(NULL == lpMsgInfo)
				{
					hrRet=E_FAIL;
					DisplayMessage(hCtxtWindow, IDS_E_INTERNAL);
					goto send_cleanup;
				}

				//check the options set for this message
				g_pccihCmdHash->Lock();
				bSkip=((lpMsgInfo->dwState)&MIS_SKIP_OPERATION)?TRUE:FALSE;//discussion notes are skipped
				bEncrypt=((lpMsgInfo->dwState)&MIS_AUTO_ENCRYPT_ON_SEND)?TRUE:FALSE;
				bSign=((lpMsgInfo->dwState)&MIS_AUTO_SIGN_ON_SEND)?TRUE:FALSE;
				g_pccihCmdHash->UnLock();

				//if we are to skip or being told to do neither encrypt nor sign
				if((TRUE == bSkip) || ((FALSE == bEncrypt) && (FALSE == bSign)))
				{
					iHandleTokenRet=DLL_HAN_NOT_HANDLED;//let the client do its stuff
					goto send_cleanup;
				}

				//check if the user has the appropriate license to run us. she must have
				//a desktop license or enterprise license to run groupwise plugin
				if(FALSE == PgpGwExLicenseChk())
					goto send_cleanup;


				//check if there is any attachments in the mail
				ZeroMemory(szTokenBuf, sizeof(szTokenBuf));
				sprintf(szTokenBuf, "ItemAttachmentGetCount(\"%S\")", 
					g_gtiLastTokenInfo.bstrMessageId);
				hrRet=Publish(szTokenBuf, &bstrResult);
				if(FAILED(hrRet))
				{
					pgpAssert(FALSE);
					DisplayMessage(hCtxtWindow, IDS_E_ATT_OPERATION);
					goto send_cleanup;
				}

				//convert the attachment count string returned to number
				iNoOfAtchMents=wcstol(bstrResult, L'\0', 10);
				pgpAssert((LONG_MAX != iNoOfAtchMents) && (LONG_MIN != iNoOfAtchMents));

				//for each attachment of the mail
				for(iAtchMentInd=iNoOfAtchMents-1; iAtchMentInd >= 0; iAtchMentInd--)
				{
					//check what type of attachment it is
					ZeroMemory(szTokenBuf, sizeof(szTokenBuf));
					sprintf(szTokenBuf, "ItemAttachmentGetClass(\"%S\";%d)", 
								g_gtiLastTokenInfo.bstrMessageId, iAtchMentInd);
					hrRet=Publish(szTokenBuf, &bstrResult);
					if(FAILED(hrRet))
					{
						pgpAssert(FALSE);
						DisplayMessage(hCtxtWindow, IDS_E_ATT_OPERATION);
						goto send_cleanup;
					}

					iAtchMentType=wcstol(bstrResult, L'\0', 10);
					pgpAssert((LONG_MAX != iAtchMentType) && (LONG_MIN != iAtchMentType));

					if(egwFile != iAtchMentType)//if it is not a file attachment
					{
						hrRet=E_FAIL;
						//we do not support encryption of ole and other type of attachments
						//there is no documented way to do this. ItemAttachmentUpdate() fails
						//with E_FAIL for any attachment other than attached files. When the
						//NDK supports other types this code may be changed to cover them. Or
						//there might be a hack to do this :) We give a message and bail out.
						DisplayMessage(hCtxtWindow, IDS_E_CANTHANDLEATTS);
						goto send_cleanup;
					}
				}

				//allocate for the recipient dialog struct.
				prdsRecipDlgStruct=(PGPclRecipientDialogStruct *)
							calloc(sizeof(PGPclRecipientDialogStruct), 1);
				pgpAssert(NULL != prdsRecipDlgStruct);
				if(NULL == prdsRecipDlgStruct)
				{
					hrRet=E_OUTOFMEMORY;//we do not show user any error lest we might crash
					goto send_cleanup;
				}
				//attempt to open the default key rings
				pgpAssert(NULL != g_pgpContext);//we should have inited clientlib already
				pgpErrRet=PGPclOpenDefaultKeyrings(g_pgpContext, kPGPOpenKeyDBFileOptions_Mutable, 
								&(prdsRecipDlgStruct->keydbOriginal));
				pgpAssert(!IsPGPError(pgpErrRet));
				if(IsPGPError(pgpErrRet))
				{
					pgpErrRet=PGPclOpenDefaultKeyrings(g_pgpContext, kPGPOpenKeyDBFileOptions_None, 
								&(prdsRecipDlgStruct->keydbOriginal));
					pgpAssert(!IsPGPError(pgpErrRet));
					if (IsPGPError(pgpErrRet))
					{
						hrRet=E_FAIL;					
						DisplayMessage(hCtxtWindow, IDS_E_OPEN_DEF_KEYRING, pgpErrRet);
						goto send_cleanup;
					}
				}
				
				//get a tls context for ourselves
				pgpErrRet=PGPNewTLSContext(g_pgpContext, &tlsContext);
				if(IsPGPError(pgpErrRet) || (NULL == tlsContext))
				{
					pgpAssert(FALSE);
					hrRet=E_FAIL;
					DisplayMessage(hCtxtWindow, IDS_E_INTERNAL);
					goto send_cleanup;
				}

				//get the recipient selection dialog title from resource
				bRet=SafeLoadString(IDS_RECIPIENTTITLE, &pszTitle);
				pgpAssert(TRUE == bRet);//string is not found or some error condition
				pgpAssert(NULL != pszTitle);//low memory ?
				if((FALSE == bRet) || (NULL == pszTitle))
				{
					hrRet=E_FAIL;
					DisplayMessage(hCtxtWindow, IDS_E_INTERNAL);
					goto send_cleanup;
				}

				//let us fill in the struct now that the keyring is opened
				prdsRecipDlgStruct->Version=kPGPCurrentRecipVersion;//DUKE ?? what the hell ?! -ssd
				prdsRecipDlgStruct->hwndParent=hCtxtWindow;
				prdsRecipDlgStruct->szTitle=pszTitle;
				prdsRecipDlgStruct->context=g_pgpContext;//we inited this in InitPGP()
				prdsRecipDlgStruct->tlsContext=tlsContext;
				prdsRecipDlgStruct->dwOptions=kPGPclASCIIArmor;
				prdsRecipDlgStruct->dwDisableFlags=kPGPclDisableWipeOriginal|
						kPGPclDisableASCIIArmor|kPGPclDisableSelfDecryptingArchive;


				//if we are encrypting to the recipients we need to find out more about 
				//the mail recipients. if we are only signing we dont care about the 
				//recipients
				if(FALSE == bEncrypt)//clear signing
					goto after_recipients_dialog;

				//get a single array that has all the recipients of this message
				bRet=PrepareMsgRecipientsArray(g_gtiLastTokenInfo.bstrMessageId, &ppszRecipients, dwNoOfRecips);
				pgpAssert((TRUE == bRet) && (NULL != ppszRecipients) && (0 != dwNoOfRecips));
				if((FALSE == bRet) || (NULL == ppszRecipients) || (0 == dwNoOfRecips))
				{
					iHandleTokenRet=DLL_HAN_NOT_HANDLED;//let the client show the error
					goto send_cleanup;
				}
				
				//if we could not find the to recipient. NOTE: groupwise does not give us
				//invalid or crossed out recipients by this method. 
				//if(0 == bstrToRecipients[0])
				//{
				//		//we hope the send will fail as the recipient is missing. let
				//		//groupwise prompt its own error message for this.
				//		iHandleTokenRet=DLL_HAN_NOT_HANDLED;
				//		goto send_cleanup;
				//}
												
				//fill the recipients
				prdsRecipDlgStruct->dwNumRecipients=dwNoOfRecips;
				prdsRecipDlgStruct->szRecipientArray=ppszRecipients;

				//if shift is pressed, force the dialog to pop.
				if (GetAsyncKeyState(VK_SHIFT)&0x8000)
					prdsRecipDlgStruct->dwDisableFlags|=kPGPclDisableAutoMode;

				//finally call the client lib to do the dialog stuff-are we tired yet ? ;) 
				bRet=PGPclRecipientDialog(prdsRecipDlgStruct);
				if(FALSE == bRet)//the user cancelled the operation
				{
					PgpGwTrace("User aborted the recipients dialog\n");
					hrRet=E_ABORT;
					goto send_cleanup;
				}

				//if keys for any recipients were imported from key server
				//in the recipient dialog
				if (NULL != prdsRecipDlgStruct->keydbAdded)
				{
					//create a new keyset from the added keys
					pgpErrRet=PGPNewKeySet(prdsRecipDlgStruct->keydbAdded, &pgpksrKeySet);
					pgpAssert(!IsPGPError(pgpErrRet));
					pgpAssert(NULL != pgpksrKeySet);
					if(IsPGPError(pgpErrRet) || (NULL == pgpksrKeySet))
					{
						hrRet=E_FAIL;
						DisplayMessage(hCtxtWindow, IDS_E_INTERNAL);
						goto send_cleanup;
					}

					//count the no of keys in the keyset
					pgpErrRet=PGPCountKeys(pgpksrKeySet, &pgpuiNumOfKeys);
					pgpAssert(!IsPGPError(pgpErrRet));
					pgpAssert(0 < pgpuiNumOfKeys);
					if(IsPGPError(pgpErrRet) || (0 == pgpuiNumOfKeys))//NOTE:zero keys may not be an error condition so change accordingly
					{
						hrRet=E_FAIL;
						DisplayMessage(hCtxtWindow, IDS_E_INTERNAL);
						goto send_cleanup;
					}

					//import the keyset
					pgpErrRet=PGPclImportKeys(g_pgpContext, tlsContext, hCtxtWindow, 
						pgpksrKeySet, prdsRecipDlgStruct->keydbOriginal, 
						kPGPclNoDialogForValidSingletons);
					
					//if there was an error and it was not because of the user aborting the import
					//by choosing to cancel the operation
					pgpAssert((kPGPError_UserAbort == pgpErrRet)?TRUE:!IsPGPError(pgpErrRet));
					if((kPGPError_UserAbort != pgpErrRet) && (IsPGPError(pgpErrRet)))
					{
						hrRet=E_FAIL;
						DisplayMessage(hCtxtWindow, IDS_E_INTERNAL);
						goto send_cleanup;
					}

				}//end if keys were downloaded in the recipients dialog

				after_recipients_dialog:

				//get the message text
				ZeroMemory(szTokenBuf, sizeof(szTokenBuf));
				sprintf(szTokenBuf, "ItemGetText(\"%S\";10)", g_gtiLastTokenInfo.bstrMessageId); 
				hrRet=Publish(szTokenBuf, &bstrResult);
				if(FAILED(hrRet))
				{
					pgpAssert(FALSE);
					DisplayMessage(hCtxtWindow, IDS_E_GET_MESSAGE_TXT, hrRet);
					goto send_cleanup;
				}
				
				//ansi convert the message text
				pAnsiInputBuf=ConvertToAnsi(bstrResult);
				pgpAssert(NULL != pAnsiInputBuf);
				if(NULL == pAnsiInputBuf)
				{
					hrRet=E_OUTOFMEMORY;//we do not show user any error lest we might crash
					goto send_cleanup;
				}

				//calculate the message length in bytes
				iPlainTxtMsgLen=strlen(pAnsiInputBuf);

				//encrypt and/or sign the message text as is required of us
				pgpAssert(NULL != prdsRecipDlgStruct);
				pgpAssert(!IsBadWritePtr(prdsRecipDlgStruct, sizeof(PGPclRecipientDialogStruct)));
				pgpErrRet=EncryptSignBuffer(g_hInstance, hCtxtWindow, g_pgpContext, 
					tlsContext, CLIENTNAME, MODULEFILENAME, pAnsiInputBuf, 
					iPlainTxtMsgLen, prdsRecipDlgStruct, NULL,
					&polOptions, (void **)&pOutBuf, &pgpsBufOutLen, bEncrypt, bSign, 
					FALSE, FALSE); 
				if(kPGPError_UserAbort == pgpErrRet)
				{
					PgpGwTrace("EncryptSignBuffer was aborted (license expired ?) !\n");
					hrRet=E_ABORT;
					iHandleTokenRet=DLL_HAN_NOT_HANDLED;//let the token go through
					goto send_cleanup;
				}

				pgpAssert(!IsPGPError(pgpErrRet));
				pgpAssert(NULL != pOutBuf);
				pgpAssert(0 < pgpsBufOutLen);
				if(IsPGPError(pgpErrRet) ||	(NULL == pOutBuf) || (0 >= pgpsBufOutLen))
				{
					hrRet=E_FAIL;
					DisplayMessage(hCtxtWindow, IDS_E_ENCRYPT_SIGN, pgpErrRet);
					goto send_cleanup;
				}
				
				pgpAssert(NULL != pOutBuf);
				pgpAssert(0 < pgpsBufOutLen);

				//allocate buffer for the token and parameters
				pFinalBuf=(char *)calloc(pgpsBufOutLen+sizeof(szTokenBuf), 1);
				pgpAssert(NULL != pFinalBuf);
				if(NULL == pFinalBuf)
				{
					hrRet=E_OUTOFMEMORY;//we do not show user any error lest we might crash
					goto send_cleanup;
				}

				//change the mail body to the encrypted stuff.
				sprintf(pFinalBuf, "ItemSetText(\"%S\";10;\"%s\")", 
								g_gtiLastTokenInfo.bstrMessageId, pOutBuf); 
				hrRet=Publish(pFinalBuf, &bstrResult);
				if(FAILED(hrRet))
				{
					pgpAssert(FALSE);
					DisplayMessage(hCtxtWindow, IDS_E_SET_ENCMSG_TXT, hrRet);
					goto send_cleanup;
				}

				//TODO:any failure after this point should notify about inconsistency 
				//in the mail - since the body is encrypted already

				//for each attachment of the mail. all attachments are expected to be of file kind
				for(iAtchMentInd=iNoOfAtchMents-1, bAbort=FALSE; iAtchMentInd >= 0; iAtchMentInd--)
				{
					PgpGwTrace("Processing attachment (att# %d)\n", iAtchMentInd);

					{
						char *pszAttachName=NULL;
						char szTmpPath[MAX_PATH]={0}, szTmpAtchSavePath[MAX_PATH+12]={0};
						char szEncFilePath[MAX_PATH+12]={0};

						//get the attachment display name
						ZeroMemory(szTokenBuf, sizeof(szTokenBuf));
						sprintf(szTokenBuf, "ItemAttachmentGetDisplayName(\"%S\";%d)", 
									g_gtiLastTokenInfo.bstrMessageId, iAtchMentInd);
						hrRet=Publish(szTokenBuf, &bstrResult);
						if(FAILED(hrRet))
						{
							pgpAssert(FALSE);
							bAbort=TRUE;
							DisplayMessage(hCtxtWindow, IDS_E_ATT_OPERATION);
							goto fileatchtype_cleanup;
						}
						
						//convert the wide chars to ansi name
						pszAttachName=ConvertToAnsi(bstrResult);
						if(NULL == pszAttachName)
						{
							pgpAssert(FALSE);
							bAbort=TRUE;
							DisplayMessage(hCtxtWindow, IDS_E_ATT_OPERATION);
							goto fileatchtype_cleanup;
						}

						//get the temporary file path that we will use for the attachment
						dwRet=GetTempPath(sizeof(szTmpPath), szTmpPath);
						pgpAssert(dwRet > 0);
						if(0 == dwRet)
						{
							bAbort=TRUE;
							DisplayMessage(hCtxtWindow, IDS_E_ATT_OPERATION);
							goto fileatchtype_cleanup;
						}

						_VERIFY(0 != GetTempFileName(szTmpPath, "55D", 0, szTmpAtchSavePath));
						if(0 == szTmpAtchSavePath[0])
						{
							bAbort=TRUE;
							DisplayMessage(hCtxtWindow, IDS_E_ATT_OPERATION);
							goto fileatchtype_cleanup;
						}
							
						_VERIFY(TRUE == DeleteFile(szTmpAtchSavePath));//delete the temp file created
																	   //cause we want to use that as a directory name

						//make the new temporary directory under temp directory
						bRet=CreateDirectory(szTmpAtchSavePath, NULL);
						if(FALSE == bRet)
						{
							dwRet=GetLastError();
							pgpAssert(FALSE);
							bAbort=TRUE;
							DisplayMessage(hCtxtWindow, IDS_E_ATT_OPERATION);
							goto fileatchtype_cleanup;
						}

						//the encrypted file path should keep the file name same as the attachment name
						//we add our extension to the file name so that our app opens when the attachment is opened
						sprintf(szEncFilePath, "%s\\%s%s", szTmpAtchSavePath, pszAttachName, PGPEXTENSIONA);
						
						//the temporary file name for attachment is the attachment name
						strcat(szTmpAtchSavePath, "\\");
						strcat(szTmpAtchSavePath, pszAttachName);
						
						//save the attachment in a file in the temp directory
						ZeroMemory(szTokenBuf, sizeof(szTokenBuf));
						sprintf(szTokenBuf, "ItemAttachmentSaveAs(\"%S\";%d;\"%s\")", 
									g_gtiLastTokenInfo.bstrMessageId, iAtchMentInd, szTmpAtchSavePath);
						hrRet=Publish(szTokenBuf, &bstrResult);
						if(FAILED(hrRet))
						{
							pgpAssert(FALSE);
							bAbort=TRUE;
							DisplayMessage(hCtxtWindow, IDS_E_ATT_OPERATION);
							goto fileatchtype_cleanup;
						}

						//encrypt/sign the attachment as is required to be done
						pgpErrRet=EncryptSignFile(g_hInstance, hCtxtWindow, g_pgpContext, 
								tlsContext, CLIENTNAME, MODULEFILENAME, szTmpAtchSavePath,
								prdsRecipDlgStruct,	NULL, &polOptions, szEncFilePath, 
								bEncrypt, bSign, TRUE/*binary*/);
						if(IsPGPError(pgpErrRet))
						{
							hrRet=E_FAIL;
							bAbort=TRUE;
							DisplayMessage(hCtxtWindow, IDS_E_ATT_ENCRYPT_SIGN, iAtchMentInd+1, pgpErrRet);
							goto fileatchtype_cleanup;
						}

						//replace the original file attachment with this file
						ZeroMemory(szTokenBuf, sizeof(szTokenBuf));
						sprintf(szTokenBuf, "ItemAttachmentUpdate(\"%S\";%d;\"%s\")", 
									g_gtiLastTokenInfo.bstrMessageId, iAtchMentInd, szEncFilePath);
						hrRet=Publish(szTokenBuf, &bstrResult);
						if(FAILED(hrRet))
						{
							pgpAssert(FALSE);
							bAbort=TRUE;
							DisplayMessage(hCtxtWindow, IDS_E_REPLACE_ATT, iAtchMentInd+1, hrRet);
							goto fileatchtype_cleanup;
						}

						fileatchtype_cleanup:
						if(NULL != pszAttachName)
						{
							delete pszAttachName;
							pszAttachName=NULL;
						}

						//delete the temporary file(s)
						if(0 != szTmpAtchSavePath[0])
							_VERIFY(TRUE == DeleteFile(szTmpAtchSavePath));

						//we delete the encrypted file only if we failed in some operation
						//this is necessary as groupwise standard handler would like to read
						//the contents of the file before sending it. so we just add the
						//file to our deferred file deletion list
						if((0 != szEncFilePath[0]) && FAILED(hrRet))
							_VERIFY(TRUE == DeleteFile(szEncFilePath));
						else
							_VERIFY(TRUE == g_pccihCmdHash->MarkFileForDeletion(szEncFilePath));
					}

					if(TRUE == bAbort)//if there was an error to make us abort
						goto send_cleanup;
				}

				//case egwMessage://encapsulated item
				//	PgpGwTrace("Processing a message attachment (att# %d)\n", iAtchMentInd);
				//	break;

				//case 3://embedded or linked object
				//	PgpGwTrace("Processing an OLE attachment (att# %d)\n", iAtchMentInd);
				//	break;

				//default:
				//	PgpGwTrace("Processing an unknown/unsupported attachment (att# %d)\n", iAtchMentInd);
				//	pgpAssert(FALSE);
				//	break;

				//toggle the encrypt/sign options that were chosen by the user for this mail. 
				//this is to prevent accidental reencryption/signing in case the send fails 
				//because of some other reason after we have done our stuff.
				g_pccihCmdHash->Lock();
				if(TRUE == bEncrypt)//if encrypt was chosen
					lpMsgInfo->dwState ^= MIS_AUTO_ENCRYPT_ON_SEND;//we reset the flag
				if(TRUE == bSign)//if sign was chosen
					lpMsgInfo->dwState ^= MIS_AUTO_SIGN_ON_SEND;
				g_pccihCmdHash->UnLock();

				//all success so let the token be processed further
				iHandleTokenRet=DLL_HAN_NOT_HANDLED;

				send_cleanup:
				if(NULL != bstrResult)
				{
					SysFreeString(bstrResult);
					bstrResult=NULL;
				}
				
				if(NULL != tlsContext)
				{
					PGPFreeTLSContext(tlsContext);
					tlsContext=NULL;
				}

				if (NULL != polOptions)
				{
					PGPFreeOptionList(polOptions);
					polOptions=NULL;
				}

				if(NULL != pszTitle)
				{
					free(pszTitle);
					pszTitle=NULL;
				}

				if(NULL != pOutBuf)
				{
					pgpAssert(pgpsBufOutLen > 0);
					PGPFreeData(pOutBuf);
					pOutBuf=NULL;
				}

				if(NULL != pFinalBuf)
				{
					free(pFinalBuf);
					pFinalBuf=NULL;
				}

				if(NULL != pAnsiInputBuf)
				{
					delete pAnsiInputBuf;
					pAnsiInputBuf=NULL;
				}

				if(NULL != pgpksrKeySet)
				{
					PGPFreeKeySet(pgpksrKeySet);
					pgpksrKeySet=NULL;
				}

				//if we allocated the dialog struct
				if(NULL != prdsRecipDlgStruct)
				{
					if(NULL != prdsRecipDlgStruct->szRecipientArray)
					{
						//free the pointer to the first token. that would free the
						//single buffer in which we keep all the recipient names
						pgpAssert(prdsRecipDlgStruct->dwNumRecipients > 0);
						pgpAssert(!IsBadWritePtr(prdsRecipDlgStruct->szRecipientArray, 
								prdsRecipDlgStruct->dwNumRecipients*sizeof(char *)));
						if(NULL != prdsRecipDlgStruct->szRecipientArray[0])
						{
							free(prdsRecipDlgStruct->szRecipientArray[0]);
							prdsRecipDlgStruct->szRecipientArray[0]=NULL;
						}
						
						//now free the pointer array
						free(prdsRecipDlgStruct->szRecipientArray);
						prdsRecipDlgStruct->szRecipientArray=NULL;
					}

					if (NULL != prdsRecipDlgStruct->keydbOriginal)
					{
						PGPFreeKeyDB(prdsRecipDlgStruct->keydbOriginal);
						prdsRecipDlgStruct->keydbOriginal=NULL;
					}

					if(NULL != prdsRecipDlgStruct->keydbAdded)
					{
						PGPFreeKeyDB(prdsRecipDlgStruct->keydbAdded);
						prdsRecipDlgStruct->keydbAdded=NULL;
					}

					if(NULL != prdsRecipDlgStruct->keydbSelected)
					{
						PGPFreeKeyDB(prdsRecipDlgStruct->keydbSelected);
						prdsRecipDlgStruct->keydbSelected=NULL;
					}

					//finally free the dialog struct
					free(prdsRecipDlgStruct);
					prdsRecipDlgStruct=NULL;
				}

				return iHandleTokenRet;
			}
			break;	
		default:
			return DLL_HAN_NOT_HANDLED;
			break;
	}

	// When the end user right-clicks on the toolbar and chooses "Properties"
	// the About dialog will display instead of the edit toolbar dialog.
	// Is the wTokenID the Toolbar edit token?
	/*if (DTKN_BTNBAR_EDIT == lpTokenData->lpToken->wTokenId) 
	{
		// Display the About Dialog instead by publishing a new token.
		lpReturn = Publish(L"AboutDlg()");

		// If lpReturn = NULL then there was an error,
		// otherwise it will be an allocated string that may or may not
		// contain a value.  You must release it when you are done.
		if ( lpReturn )
		{
			// Evaluate and process the return value here.
			// ...

			// Cleanup
			delete lpReturn;
			lpReturn = NULL;
		}

		//Return the value indicating we processed the token so the
		//GroupWise client doesn't have to.
		return DLL_HAN_NO_ERROR;
	}*/

	return DLL_HAN_NOT_HANDLED;
}


////////////////////////////////////////////////////////////////////////////////
// Name: ValidateToken
// Descrip:	This method is reserved for later use.  Even though it is currently
//			not called by GW, it must be included.
////////////////////////////////////////////////////////////////////////////////
DWORD WINAPI ValidateToken(LPMAC_TOKEN lpTokenData, LPVOID lpDataTypeInfo)
{
	return 0;
}


void TraceHandleToken(LPTPH_RETURNVAL lpTokenData, HWND	hLinkWnd, WORD wMsg)
{
#ifdef _DEBUG 
	char szBuffer[255]={0};

	PgpGwTrace(" Token Info \n");

	//trace the window details
	int iRet=GetWindowText(hLinkWnd, szBuffer, sizeof(szBuffer));
	pgpAssert(iRet > 0);
	PgpGwTrace("hLinkWnd=%#x\"%s\"\n", hLinkWnd, szBuffer);

	//trace the message details
	PgpGwTrace("wMsg=%#x\n", wMsg);

	//check if tokendata  ptr is null
	pgpAssert(NULL != lpTokenData);
	if(NULL == lpTokenData)
	{
		PgpGwTrace("lpTokenData=NULL\n");
		return;
	}

	//check if the token ptr in it is null
	pgpAssert(NULL != lpTokenData->lpToken);
	if(NULL == lpTokenData->lpToken)
	{
		PgpGwTrace("lpTokenData->lpToken=NULL\n");
		return;
	}

	//dump the details in the token itself
	PgpGwTrace("lpTokenData->lpToken->hszCommand=%#x\n", lpTokenData->lpToken->hszCommand);
	PgpGwTrace("                      Version=%#x\n", lpTokenData->lpToken->Version);
	PgpGwTrace("                      hszRequestor=%#x\n", lpTokenData->lpToken->hszRequestor);
	PgpGwTrace("                      dwMacroID=%#x\n", lpTokenData->lpToken->dwMacroID);
	PgpGwTrace("                      atomApp=%#x\n", lpTokenData->lpToken->atomApp);
	PgpGwTrace("                      wReserved=%#x\n", lpTokenData->lpToken->wReserved);
	PgpGwTrace("                      wTokenId=%d", lpTokenData->lpToken->wTokenId);
	switch(lpTokenData->lpToken->wTokenId)
	{
		case BFTKN_SEND_STDMAIL:
			PgpGwTrace("(BFTKN_SEND_STDMAIL)\n");
			break;
		case BFTKN_SEND:
			PgpGwTrace("(BFTKN_SEND)\n");
			break;
		case BFTKN_CLOSE_WINDOW:
			PgpGwTrace("(BFTKN_CLOSE_WINDOW)\n");
			break;
		case BFTKN_READ:
			PgpGwTrace("(BFTKN_READ)\n");
			break;
		case BFTKN_FORWARD:
			PgpGwTrace("(BFTKN_FORWARD)\n");
			break;
		case BFTKN_REPLY:
			PgpGwTrace("(BFTKN_REPLY)\n");
			break;
		case AFTKN_ITEM_MSGID_FROM_VIEW:
			PgpGwTrace("(AFTKN_ITEM_MSGID_FROM_VIEW)\n");
			break;
		case AFTKN_EVENT_NOTIFY:
			PgpGwTrace("(AFTKN_EVENT_NOTIFY)\n");
			break;
		case AFTKN_TYPE_CHAR:
			PgpGwTrace("(AFTKN_TYPE_CHAR)\n");
			break;
		case AFTKN_OPEN_VIEW_TYPE:
			PgpGwTrace("(AFTKN_OPEN_VIEW_TYPE)\n");
			break;
		case DTKN_ATTACH_FILE:
			PgpGwTrace("(DTKN_ATTACH_FILE)\n");
			break;
		case BFTKN_SPELLER:
			PgpGwTrace("(BFTKN_SPELLER)\n");
			break;
		case BFTKN_SHOW_LOOKER:
			PgpGwTrace("(BFTKN_SHOW_LOOKER)\n");
			break;
		case BFTKN_VIEW_ATTACH:
			PgpGwTrace("(BFTKN_VIEW_ATTACH)\n");
			break;
		case BFTKN_OPEN_ATTACH:
			PgpGwTrace("(BFTKN_OPEN_ATTACH)\n");
			break;
		case BFTKN_VIEW_HTML:
			PgpGwTrace("(BFTKN_VIEW_HTML)\n");
			break;
		case BFTKN_VIEW_RTF:
			PgpGwTrace("(BFTKN_VIEW_RTF)\n");
			break;
		default:
			PgpGwTrace("(unknown)\n");
			break;
	}
	PgpGwTrace("                      cParam=%#x\n", lpTokenData->lpToken->cParam);
	PgpGwTrace("                      dwFlags=%#x\n", lpTokenData->lpToken->dwFlags);
	PgpGwTrace("                      dwReserved=%#x\n", lpTokenData->lpToken->dwReserved);

	//if there are any parameters
	if(lpTokenData->lpToken->cParam > 0)
	{
		for(unsigned int uiParamCount=0; uiParamCount < lpTokenData->lpToken->cParam; uiParamCount++)
		{
			PgpGwTrace("\n                      param%d:", uiParamCount+1);	
			PgpGwTrace("flags=%#x:type=%#x,", lpTokenData->lpToken->rgParam[uiParamCount].eType,
						lpTokenData->lpToken->rgParam[uiParamCount].wFlags);
			switch(lpTokenData->lpToken->rgParam[uiParamCount].eType)
			{
				case eParmBoolean://'C' BOOL type
					PgpGwTrace("eParmBoolean,value=%#x",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.dwValue);
					break;
				case eParmByteSigned://signed 8-bit value
					PgpGwTrace("eParmByteSigned, value=%#x",
						*(LPBYTE)(lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr));
					break;
				case eParmByteUnsigned://unsigned 8-bit value
					PgpGwTrace("eParmByteUnsigned, value=%#x",
						*(LPBYTE)(lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr));
					break;
				case eParmCentimeters://double
					PgpGwTrace("eParmCentimeters, value=%f",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.drValue);
					break;
				case eParmDwordSigned://signed 32-bit value
					PgpGwTrace("eParmDwordSigned, value=%#x",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.dwValue);
					break;
				case eParmDwordUnsigned://unsigned 32-bit value
					PgpGwTrace("eParmDwordUnsigned, value=%#x",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.dwValue);
					break;
				case eParmDWPUSigned://signed 32-bit unit (WPU)
					PgpGwTrace("eParmDWPUSigned, value=%#x",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.dwValue);
					break;
				case eParmDWPUUnsigned://unsigned 32-bit unit (WPU)
					PgpGwTrace("eParmDWPUUnsigned, value=%#x",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.dwValue);
					break;
				case eParmEnumeration://enumeration 16-bit value
					PgpGwTrace("eParmEnumeration, value=%#x",
						*(LPWORD)&(lpTokenData->lpToken->rgParam[uiParamCount].uData.dwValue));
					break;
				case eParmFloat://double
					PgpGwTrace("eParmFloat, value=%f",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.drValue);
					break;
				case eParmInches://double
					PgpGwTrace("eParmInches, value=%f",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.drValue);
					break;
				case eParmMillimeters://double
					PgpGwTrace("eParmMillimeters, value=%f",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.drValue);
					break;
				case eParmPoints://double
					PgpGwTrace("eParmPoints, value=%f",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.drValue);
					break;
				case eParmStringAnsi://ANSI string (zero terminated)
					PgpGwTrace("eParmStringAnsi, value=%s",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr);
					break;
				case eParmStringLabel://WPCORP Word STring (zero terminated) containing a label reference.
					PgpGwTrace("eParmStringLabel, value=%s",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr);
					break;
				case eParmStringOem://OEM string (zero terminated)
					PgpGwTrace("eParmStringOem, value=%s",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr);
					break;
				case eParmStringWord://WPCORP Word String (zero terminated)
					PgpGwTrace("eParmStringWord, value=%s",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr);
					break;
				case eParmStringVariable://word string containing a variable name
					PgpGwTrace("eParmStringVariable, value=%s",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr);
					break;
				case eParmTokenID://unsigned 16-bit value
					PgpGwTrace("eParmTokenID, value=%#x",
						*(LPWORD)(lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr));
					break;
				case eParmWordSigned://signed 16-bit value
					PgpGwTrace("eParmWordSigned, value=%#x",
						*(LPWORD)&(lpTokenData->lpToken->rgParam[uiParamCount].uData.dwValue));
					break;
				case eParmWordUnsigned://unsigned 16-bit value
					PgpGwTrace("eParmWordUnsigned, value=%#x",
						*(LPWORD)&(lpTokenData->lpToken->rgParam[uiParamCount].uData.dwValue));
					break;
				case eParmWPFname://ANSI string (zero terminated) containing a file name.
					PgpGwTrace("eParmWPFname, value=%s",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr);
					break;
				case eParmWPUSigned:// signed 16-bit unit (WPU)
					PgpGwTrace("eParmWPUSigned, value=%#x",
						*(LPWORD)(lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr));
					break;
				case eParmWPUUnsigned://unsigned 16-bit unit (WPU)
					PgpGwTrace("eParmWPUUnsigned, value=%#x",
						*(LPWORD)(lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr));
					break;
				case eParmUserDialog://user dialog buffer (like a string)
					PgpGwTrace("eParmUserDialog, value=%s",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr);
					break;
				case eParmStringWide://Wide/Unicode (zero terminated)
					PgpGwTrace("eParmStringWide, value=%S",
						lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr);
					break;
				case eValArrayDefinition://definition of an array
					PgpGwTrace("eValArrayDefinition, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValBoolean://'C' BOOL type
					PgpGwTrace("eValBoolean, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValCentimeters://double
					PgpGwTrace("eValCentimeters, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValFloat://double
					PgpGwTrace("eValFloat, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValInches://double
					PgpGwTrace("eValInches, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValInteger://signed 32-bit value
					PgpGwTrace("eValInteger, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValMillimeters://double
					PgpGwTrace("eValMillimeters, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValOLEObject://???
					PgpGwTrace("eValOLEObject, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValPoints://double
					PgpGwTrace("eValPoints, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValString://WPCORP Word String (zero terminated)
					PgpGwTrace("eValString, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValStringAnsi://ANSI string (zero terminated)
					PgpGwTrace("eValStringAnsi, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValStringOem://OEM string (zero terminated)
					PgpGwTrace("eValStringOem, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValWPUnits://signed 32-bit unit (WPU)
					PgpGwTrace("eValWPUnits, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValAlias://used for reference parameters
					PgpGwTrace("eValAlias, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValRecord://definition of a record
					PgpGwTrace("eValRecord, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValDateTime://unsigned 32-bit value
					PgpGwTrace("eValDateTime, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValRawBinary://raw binary data type
					PgpGwTrace("eValRawBinary, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				case eValStringWide://Wide/Unicode (zero terminated)
					PgpGwTrace("eValStringWide, value=???"	
						/*lpTokenData->lpToken->rgParam[uiParamCount].uData.lpvPtr*/);
					break;
				default:
					PgpGwTrace("unknown/unexpected, value=???");
					break;
			}
		}
		PgpGwTrace("\n");	
	}

	PgpGwTrace("\n");
#endif
}


BOOL PrepareMsgRecipientsArray(BSTR bstrMessageId, char ***pppszRecipients, DWORD& dwNoOfRecips)
{
	char szTokenBuf[255]={0};
	UINT uiRecipIndex=0;
	UINT uiElemsAlloced=0;
	BSTR bstrRecipients=NULL;
	char *pszRecipients=NULL, **ppszRecipArray=NULL;
	char *pTmpToken=NULL;
	char *pszToRecips=NULL, *pszCcRecips=NULL, *pszBccRecips=NULL;
	char *pszRecipSeparator=NULL;
	DWORD dwLenReqd=0;
	HRESULT hrRet=S_OK;
	BOOL bRet=FALSE;

	//check parameters passed
	pgpAssert((NULL != bstrMessageId) && (0 != bstrMessageId[0]));//not null message id string
	pgpAssert(NULL != pppszRecipients);//should be not null pointer since we are gonna dereference it
	if((NULL == bstrMessageId) || (0 == bstrMessageId[0]) || 
	   (NULL == pppszRecipients))
		goto cleanup;

	//get the "TO" recipient(s)
	ZeroMemory(szTokenBuf, sizeof(szTokenBuf));
	sprintf(szTokenBuf, "ItemGetText(\"%S\";0)", bstrMessageId); 
	hrRet=Publish(szTokenBuf, &bstrRecipients);
	if(FAILED(hrRet))
	{
		pgpAssert(FALSE);
		goto cleanup;
	}
	
	//if there are invalid recipients in the TO list or no recipients at all
	if(L'\0' == bstrRecipients[0])
	{
		//groupwise does not allow mails without any to addresses.
		//invalid addresses always lead to prompting the user with 
		//the address book. so let us bail out of doing our stuff.
		goto cleanup;
	}

	//copy the recipients to an ansi string
	pszToRecips=ConvertToAnsi(bstrRecipients);
	pgpAssert(NULL != pszToRecips);

	//get the CC recipients
	ZeroMemory(szTokenBuf, sizeof(szTokenBuf));
	sprintf(szTokenBuf, "ItemGetText(\"%S\";2)", bstrMessageId); 
	hrRet=Publish(szTokenBuf, &bstrRecipients);
	if(FAILED(hrRet))
	{
		pgpAssert(FALSE);
		goto cleanup;
	}

	//copy the carbon copy recipients to an ansi string
	pszCcRecips=ConvertToAnsi(bstrRecipients);
	pgpAssert(NULL != pszCcRecips);

	//get the BCC recipients
	ZeroMemory(szTokenBuf, sizeof(szTokenBuf));
	sprintf(szTokenBuf, "ItemGetText(\"%S\";3)", bstrMessageId); 
	hrRet=Publish(szTokenBuf, &bstrRecipients);
	if(FAILED(hrRet))
	{
		pgpAssert(FALSE);
		goto cleanup;
	}

	//copy the blind carbon copy recipients to an ansi string
	pszBccRecips=ConvertToAnsi(bstrRecipients);
	pgpAssert(NULL != pszBccRecips);

	//figure out the length of the string that has all these recipients
	dwLenReqd += strlen(pszToRecips);
	dwLenReqd += strlen(pszCcRecips);
	dwLenReqd += strlen(pszBccRecips);

	//if we need to allocate
	if(dwLenReqd > 0)
	{
		//allocate buffer for the combined recipients string
		pszRecipients=(char *)calloc(dwLenReqd+10, 1);
		pgpAssert(NULL != pszRecipients);
		if(NULL == pszRecipients)
			goto cleanup;

		//get the character that separates the recipients
		_VERIFY(TRUE == SafeLoadString(IDS_RECIPIENT_SEPARATOR, &pszRecipSeparator));
		if(NULL == pszRecipSeparator)
			goto cleanup;

		//copy the strings to the buffer. 
		strcat(pszRecipients, pszToRecips);//add the to recipients

		if('\0' != pszCcRecips[0])//if there are cc recipients
		{
			//if there are any recipients in the master list already
			if((NULL != pszRecipients[0]) && 
			   (pszRecipSeparator[0] != pszRecipients[strlen(pszRecipients)]))//and there is no semicolon at the end
			   strcat(pszRecipients, pszRecipSeparator);

			//add the cc recipients
			strcat(pszRecipients, pszCcRecips);
		}

		if('\0' != pszBccRecips[0])//if there are bcc recipients
		{
			//if there are any recipients in the master list already
			if((NULL != pszRecipients[0]) && 
			   (pszRecipSeparator[0] != pszRecipients[strlen(pszRecipients)]))//and there is no semicolon at the end
			   strcat(pszRecipients, pszRecipSeparator);

			//add the bcc recipients
			strcat(pszRecipients, pszBccRecips);
		}

		//get the recipients as tokens from this string
		while(NULL != (pTmpToken=strtok(pTmpToken?NULL:pszRecipients, pszRecipSeparator)))
		{
			//if we are trying to write just beyond our current allocation
			if(uiRecipIndex == uiElemsAlloced)
			{
				UINT uiChunkSize=32;//everytime we run out of space we allocate space for 32 more pointers
				char **ppchTmpPtr=(char **)realloc(ppszRecipArray, (uiElemsAlloced+uiChunkSize)*sizeof(char *));
				pgpAssert(NULL != ppchTmpPtr);
				if(NULL == ppchTmpPtr)
					goto cleanup;//realloc failed !!
				
				//zero out the newly allocated bytes
				ZeroMemory(ppchTmpPtr+uiElemsAlloced, uiChunkSize*sizeof(char *));
				ppszRecipArray=ppchTmpPtr;
				uiElemsAlloced += uiChunkSize;
			}

			//trim the token of any leading or trailing spaces. our pgp stuff
			//gets fooled by spaces in the names.
			StrTrim(pTmpToken, " ");

			//remember the pointer 
			ppszRecipArray[uiRecipIndex++]=pTmpToken;
		}
	}

	//copy the array to the out param passed to us
	*pppszRecipients=ppszRecipArray;
	dwNoOfRecips=uiRecipIndex;
	bRet=TRUE;//all success

	cleanup:
	if(NULL != bstrRecipients)
	{
		SysFreeString(bstrRecipients);
		bstrRecipients=NULL;
	}
	if(NULL != pszToRecips)
	{
		delete pszToRecips;
		pszToRecips=NULL;
	}
	if(NULL != pszCcRecips)
	{
		delete pszCcRecips;
		pszCcRecips=NULL;
	}
	if(NULL != pszBccRecips)
	{
		delete pszBccRecips;
		pszBccRecips=NULL;
	}
	if(NULL != pszRecipSeparator)
	{
		free(pszRecipSeparator);
		pszRecipSeparator=NULL;
	}
	if(FALSE == bRet)//if we failed
	{
		//free the recipient array
		if(NULL != pszRecipients)
		{
			free(pszRecipients);
			pszRecipients=NULL;
		}

		if(NULL != ppszRecipArray)
		{
			free(ppszRecipArray);
			ppszRecipArray=NULL;
		}
	}
	return bRet;
}