/*______________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.


	$Id: nPGPNts.c,v 1.17 2002/11/21 22:28:19 sdas Exp $
______________________________________________________________________________*/

/*::: MODULE OVERVIEW :::::::::::::
Provides central management services for the back-end infrastructure of the 
Lotus Notes PGP Plug-In.

--- revision history --------
10/30/02 Version 1.2.2 Paul Ryan
+ logic enhancement to support gracefully user-cancelation during key-server 
  searches [creation of i_MatchPgpKeys(), i_PgpMatchSmtpAddresses(), 
  associated adjustment of i_PgpMatchNotesNames(), xs_PgpSecureMessage()]
+ listing format adjustment, token renaming, documentation improvement

9/6/02 Version 1.2.1 Paul Ryan
+ logic enhancemnt during message decoding to wrap PGP-encoded attachments in 
  messages not generated by the Notes plug-in so that special PGP handling can 
  be provided
+ logic enhancement to support high-ASCII (international) characters in the 
  message body and in attachment handling
+ removed unnecessary critical-section variable
+ explicitly set WINVER preprocessor token to 0x0400 (Win95/NT4 baseline)
+ much listing format adjustment, documentation adjustment, token renaming

7/2/01 Version 1.2 Paul Ryan
+ logic enhancement in support of robust plug-in version handling
+ listing format adjustment, documentation adjustment

11/16/00 Version 1.1.8 Paul Ryan
+ added work-around method for marking a document as read

11/14/00 Version 1.1.7 Paul Ryan
+ added support for resetting the view/folder context associated with a 
  document currenty being shown to the user in the Notes R5 user interface, see 
  xs_vetFolderCtx()
+ documentation adjustment

10/7/00 Version 1.1.6 Paul Ryan
+ Revamped the plug-in setup procedure. See xs_VetPluginReadiness().
+ bug fixes to es_SpecialDecryptAttachment()
+ documentation adjustment

9/16/00 Version 1.1.5 Paul Ryan
+ enhancements surrounding special handling of PGP-encoded attachments in a 
  Notes message
+ design improvement to internals controlling the checkmark toggling of the 
  PGP drop-down action-bar menu
+ documentation adjustment, token renaming

8/9/00 Version 1.1.2 Paul Ryan
+ manufacture of checkmark toggling for the PGP drop-down action-bar menu
+ work towards improved file-attachment handling (not enabled)
+ got rid of hack in i_PgpEncodeAttachments() for attaching encoded files 
  without Notes trying to sign them when sending the message in MIME format
+ documentation improvement

3/20/00 Version 1.1.1 Paul Ryan
+ adjustment to support Notes R5, Notes Mail R5 template and PGP 7.0
+ documentation adjustments, logic enhancements
+ Bug fix of behavior where a pre-existing encoded rich-text attachment was 
  not being removed during a new encoding. Fix relates primarily to message 
  forward and "reply with history" operations.

9/12/99 Version 1.1 Paul Ryan
+ PGP 6.5.1 compatibility
+ logic enhancements
   - support for reading into memory rich-text content resident in a 
     compressed attachment
   - default handling of Notes addresses in the user-friendlier abbreviated 
     format
   - improved Domino Directory name-lookup functionality to emulate better 
     Notes' normal algorithm, and thereby enhanced the plug-in's group-
	 resolution capability
+ documentation enhancements

1/29/99 Version 1.0 Paul Ryan
::::::::::::::::::::::::::::::::::::*/

#include "nPGPNts.h"


//global-scope declarations
HINSTANCE  eh_Instance;

//module-scope declarations
static const char  mpc_PGP_ARMOR_EXT_ASC[] = ".asc", 
					mpc_PGP_ARMOR_EXT_PGP[] = ".pgp", 
					mpc_ROOTFILENM_PGP_RTF[] = "pgp.rtf", 
					mpc_PROPNM_ORIG_WNDPROC[] = "OrigWndProc", 
					mpc_PROPNM_EXTR_DLG_CHOICE[] = "ActionChoice";
static const UINT  mui_LEN_ARMOR_EXT_ASC = sizeof( mpc_PGP_ARMOR_EXT_ASC) - 1, 
					mui_LEN_ARMOR_EXT_PGP = sizeof( mpc_PGP_ARMOR_EXT_PGP) - 1;
static const int  mi_CANCEL = 2, mi_SUCCESS = 3, 
					mi_MASK_DCRYPTD = 0x8000;

static char  mpc_fileNmPgpRtf[ sizeof( mpc_ROOTFILENM_PGP_RTF) + 
											sizeof( mpc_PGP_ARMOR_EXT_ASC) - 1];
static ResettableRtfInfo  mt_RtfResettable;
static HFONT  mh_fntFileCaption;
static OSVERSIONINFO  mt_Os = {sizeof( OSVERSIONINFO)};


/** xs_markRead( ***
Mark the specified note as read. Works around strange unread-marks behaviors 
seen with the Notes R5 client.

--- parameter & return ----
h_NOTE: handle to the note to be marked as read
RETURN:
	eus_SUCCESS if no errors occurred
	eus_ERR_INVLD_ARG if any passed-in parameter is obviously invalid
	the Notes API error code otherwise

--- revision history ------
9/6/02 PR: listing format adjustment, minor exception-handling adjustment
11/16/00 PR: created		*/
short xs_markRead( NOTEHANDLE  h_NOTE)	{
	static char  pc_userNm[ MAXUSERNAME + 1];
	static WORD  us_lenUserNm;

	NOTEID  nid;
	DBHANDLE  h_db;
	HANDLE  h_unrd, h_unrdOrig = NULL;
	BOOL  f_chngd;
	STATUS  us_err, us_error;

	if (!h_NOTE)
		return (short) eus_ERR_INVLD_ARG;

	//if necessary, get the username associated with the ID associated with the 
	//	execution going on here
	if (!us_lenUserNm)	{
		if (us_err = SECKFMGetUserName( pc_userNm))
			return us_err;
		us_lenUserNm = strlen( pc_userNm);
	}

	//derive the NoteID and the handle to the database containing the note
	NSFNoteGetInfo( h_NOTE, _NOTE_ID, &nid);
	NSFNoteGetInfo( h_NOTE, _NOTE_DB, &h_db);

	//get the current unread table
	if (us_err = NSFDbGetUnreadNoteTable( h_db, pc_userNm, us_lenUserNm, TRUE, 
																	&h_unrd))
		return us_err;
	if (us_err = NSFDbUpdateUnread( h_db, h_unrd))
		goto errJump;
	if (us_err = IDTableCopy( h_unrd, &h_unrdOrig))
		goto errJump;

	//if the NoteID of the specified note is present in the unread table, 
	//	delete that entry 
	if (us_err = IDDelete( h_unrd, nid, &f_chngd))
		goto errJump;

	//if a deletion did occur, update the operative unread table
	if (f_chngd)
		us_err = NSFDbSetUnreadNoteTable( h_db, pc_userNm, us_lenUserNm, FALSE, 
															h_unrdOrig, h_unrd);

errJump:
	//free allocated resources, as necessary
	if (h_unrdOrig)
		if (( us_error = IDDestroyTable( h_unrdOrig)) && !us_err)
			us_err = us_error;
	if (( us_error = IDDestroyTable( h_unrd)) && !us_err)
		us_err = us_error;

	return us_err;
} //xs_markRead(


/** xs_vetFolderCtx( ***
Handles the determining and resetting of the view/folder context associated 
with a document currently being shown to the user in the Notes R5 user 
interface. The view/folder context allows the @Commands "@Command( [Folder])" 
and "@Command( [EditClear])" to function with the parent view/folder from which 
the document was opened as well as with the document itself. In terms of 
immediately obvious effects, with @Command( [Folder]) this means that if the 
parent view/folder is a folder, the "Move" button will be enabled. And with 
@Command( [EditClear], it means that the view line associated with the document 
will be marked with the trash icon.

The ability to reset the view/folder context is useful when the document 
currently being shown to the user was opened programmatically, not directly by 
the user herself through the view/folder. Opening documents programmatically 
does not populate a view/folder context, so the @Commands noted above operate 
differently in that their functioning does not interact with a parent 
view/folder as described: the Move button in the folder dialog is never 
enabled, and no trash icon mark will be placed as the user might expect.

NOTE: The memory mapping and manipulation conducted by this procedure was 
arrived at purely by means of reverse-engineering the Notes user-interface DLL 
nnotesws.dll. Naturally, this approach is a "hack" that would have been avoided 
if it were possible to accomplish the aims of the procedure in another way. As 
with most any hack, the approach is prone to maintenance problems as product 
changes or upgrades occur. The procedure has been verified to work with Notes 
R5.0.2 and R5.0.4, and cannot be guaranteed to succeed with other (or future) 
versions of Notes.

--- parameters & return -----
pnid: Input & Output. Address of variable containing the NoteID of the parent 
	view/folder of the document currently being shown to the user. If variable 
	is null, it signifies that procedure should populate the variable with the 
	NoteID found within the workspace context maintained by the Notes system. 
	If variable is non-null, it's value is written to the workspace context 
	provided that the context currently has that value as null. That is, the 
	procedure will not allow a non-blank value to be overwritten.
pus: Input & Output: Address of variable containing the type-value of the 
	parent view/folder. If the view/folder NoteID is being written to the 
	workspace context, the type-value held by this variable will be written as 
	as well. If the view/folder NoteID is being output to the variable pointed 
	to by the pnid argument, the type-value will also be output.
RETURN: ms_VB_TRUE if procedure encountered no errors; ms_VB_FALSE otherwise

--- revision history --------
9/6/02 PR: listing format adjustment
11/14/00 PR: created			*/
short xs_vetFolderCtx( NOTEID *const pnid, 
						short *const  pus)	{
	//static for speed
	static const char  pc_CLASSNM_SBPRG_FRMST[] = "FramesetSubprog", 
						pc_CLASSNM_SBPRG_NOTES[] = "NotesSubprog";
	static const UINT  ui_OFST_T16_NID_VW = 0x98, ui_OFST_T16_VW_TYP = 0xDA;

	HWND  h;
	char  pc[ 21];
	BYTE * puc_tsbprg, * puc_t16;
	NOTEID  nid;

	//if parameters are obviously invalid, short-circuit with failure
	if (!( pnid && pus))
		goto errJump;

	//Get address of subprogram structure associated with the main message 
	//	window. Do so in a consistent way of navigating to the first parent 
	//	frameset subprogram, and then down to the first child Notes subprogram.
	if (!( h = GetFocus()))
		goto errJump;
	if (!( h = GetParent( h)))
		goto errJump;
	if (!( h = GetParent( h)))
		goto errJump;
	if (!GetClassName( h, pc, sizeof( pc)))
		goto errJump;
	if (strcmp( pc, pc_CLASSNM_SBPRG_FRMST) != ei_SAME)	{
		if (!( h = GetParent( h)))
			goto errJump;
		if (strcmp( pc, pc_CLASSNM_SBPRG_FRMST) != ei_SAME)
			goto errJump;
	}
	if (!( h = FindWindowEx( h, NULL, pc_CLASSNM_SBPRG_NOTES, NULL)))
		goto errJump;
	if (!( puc_tsbprg = (BYTE *) GetWindowLong( h, 0)))
		goto errJump;

	//if we've been asked to output the current view/folder context...
	puc_t16 = (BYTE *) *(long *) *(long *) (puc_tsbprg + 0x8E);
	nid = *(long *) (puc_t16 + ui_OFST_T16_NID_VW);
	if (!*pnid)	{
		//do so
		*pnid = nid;
		*pus = *(puc_t16 + ui_OFST_T16_VW_TYP);
	//else, provided that we're not overwriting a current context, populate the 
	//	current document context with the folder context we acquired in a prior 
	//	call
	}else if (!nid)	{
		*(NOTEID *) (puc_t16 + ui_OFST_T16_NID_VW) = *pnid;
		*(puc_t16 + ui_OFST_T16_VW_TYP) = (BYTE) *pus;
	} //if (!*pnid)

	return ms_VB_TRUE;

errJump:
	return ms_VB_FALSE;
} //xs_vetFolderCtx(


/** xs_DecryptVerifyNote( ***
Purpose is to PGP decrypt/verify the PGP encrypted/signed content present in 
the specified note.

--- parameters & return ----
h_NOTE: handle to the note to undergo PGP decoding
pc_ITMNM_RTF: pointer to a string telling the name of the rich-text field to be 
	used in the decryption/verification
ps_msgChanged: pointer to the variable of VisualBasic integer type in which to 
	return whether message content was changed (ms_VB_TRUE) or not (ms_VB_FALSE)
RETURN: eus_SUCCESS if no error occured. If the user canceled a PGP dialog, 
	ei_USER_ABORT. Otherwise !eus_SUCCESS or, if available, the Notes API error 
	code.

--- revision history -------
9/6/02 PR: listing format adjustment, minor documentation change, token renaming
11/14/00 PR: minor token renaming
9/12/99 PR: documentation adjustment
2/2/99 PR: created			*/
short xs_DecryptVerifyNote( NOTEHANDLE  h_NOTE, 
							char  pc_ITMNM_RTF[], 
							short *const  ps_msgChanged)	{
	RtfCursor  t_crsr;
	RtfContext  t_RtfContext;
	DWORD  ul_objId, ul_lenContent;
	BLOCKID  bid_item;
	STATUS  us_err;
	int  i_err, i_errNonFatal = NULL;

	if (!( h_NOTE && pc_ITMNM_RTF && ps_msgChanged))
		return !eus_SUCCESS;

	//default the message-changed flag to say that no change has occurred
	*ps_msgChanged = ms_VB_FALSE;

	//initialize resources associated with the rich-text handling we're going 
	//	to do
	if (us_err = eus_InitializeRtfContext( h_NOTE, pc_ITMNM_RTF, &t_crsr, 
																&t_RtfContext))
		return us_err;

	//if the special PGP-Armored file attachment containing Notes rich-text is 
	//	present on the note...
	if (us_err = us_LocatePgpRtfAttachment( t_crsr, &t_RtfContext, h_NOTE, 
										&ul_objId, &ul_lenContent, &bid_item))
		i_errNonFatal = us_err;
	if (bid_item.pool && !i_errNonFatal)	{
		//PGP decode the attachment's rich-text content
		i_err = i_PgpDecodeUsingRichTextObj( ul_objId, ul_lenContent, bid_item, 
												h_NOTE, pc_ITMNM_RTF, 
												&t_RtfContext, &i_errNonFatal);
		if (i_err && !i_errNonFatal)
			goto errJump;

		//if no error occurred, inform the caller that the message content 
		//	changed
		if (!i_errNonFatal)
			*ps_msgChanged = ms_VB_TRUE;
	} //if (bid_item.pool && !i_errNonFatal)

	//if no the special rich-text version exists or if a non-fatal error 
	//	occurred in decoding that version, we will decode the message's first 
	//	PGP-ASCII block
	if (!bid_item.pool || i_errNonFatal)	{
		BOOL  f_MsgChanged;
		if (i_err = i_PgpDecodeUsingAsciiText( h_NOTE, pc_ITMNM_RTF, t_crsr, 
												&t_RtfContext, &f_MsgChanged))
			goto errJump;

		//if it's true, inform the caller that the message content did change
		if (f_MsgChanged)
			*ps_msgChanged = ms_VB_TRUE;
	} //if (!bid_item.pool || i_errNonFatal)

	if (i_errNonFatal)	{
		const char *const  ppc_MSG_PART[] = {"An error occurred (", ") "
								"during the attempt to\nPGP decode a "
								"rich-text version of this message.\nThe "
								"plain-text version will be used instead."};
		char  pc_msg[ 255];

		sprintf( pc_msg, "%s%i%s", ppc_MSG_PART[0], i_errNonFatal, 
															ppc_MSG_PART[1]);
		MessageBox( eh_mainWnd, pc_msg, epc_APPNM, MB_OK | MB_ICONINFORMATION);
		i_err = us_err = eus_SUCCESS;
	} //if (i_errNonFatal)

errJump:
	ef_FreeRtfContext( &t_RtfContext);

	//if no errors occurred, return that procedure was successful; otherwise 
	//	return an appropriate error code
	return (short) (us_err + i_err);
} //xs_DecryptVerifyNote(


/** i_PgpDecodeUsingAsciiText( ***


--- parameters & return ----

RETURN: eus_SUCCESS if no error occured; ... otherwise

--- revision history -------
9/6/02 PR
+ added special attachment-handling for PGP-encoded attachments, causing slight 
  signature change
+ logic enhancement to support high-ASCII (international) characters in message 
  body
+ minor documentation adjustment, listing format adjustment, token renaming

9/12/99 PR: code adjustment associated with an improvement to handling of 
	public-key decoding
1/24/99 PR: created			*/
//DOC!!
static int i_PgpDecodeUsingAsciiText( NOTEHANDLE  h_NOTE, 
										char  pc_ITMNM[], 
										RtfCursor  t_crsr, 
										RtfContext *const  pt_rtCtx, 
										BOOL *const  pf_MsgChanged)	{
	DWORD  ul, ul_lenBlk;
	char * pc_blk, * pc_inp = NULL, * pc_inpOs = NULL, 
			* pc_outpOs = NULL, * pc_outp = NULL;
	STATUS  us_err = eus_SUCCESS;
	BOOL  f_fail;
	int  i_err = NULL;

	//Get the textual content of the rich-text field along with its length. If 
	//	no content exists, short-circuit with success. (Remember, 
	//	eul_GetRtfText() returns the size of the buffer allocated, not the 
	//	length of the null-terminated string.) The content needs to be 
	//	translated to the native character set because clear-signed text may 
	//	include extended characters.
	if (eul_ERR_FAILURE == (ul = eul_GetRtfText( pt_rtCtx, NULL, TRUE, 
																	&pc_inp)))
		return !eus_SUCCESS;
	if (ul <= 1)
		goto errJump;
	if (!(pc_inpOs = malloc( ul)))	{
		i_err = -11999;		//kPGPError_OutOfMemory
		goto errJump;
	}
	if (f_fail = !OSTranslate( OS_TRANSLATE_LMBCS_TO_NATIVE, pc_inp, (WORD) 
												(ul - 1), pc_inpOs, (WORD) ul))
		goto errJump;

	//PGP decode just the portion of the content containing the PGP-encoded 
	//	block
	if (i_err = ei_FindAndDecodePgpBlock( &pc_inpOs, ul, &pc_blk, &ul_lenBlk, 
															&pc_outpOs, NULL))
		goto errJump;

	//if output was generated by the decoding (none will be if the PGP-encoded 
	//	block is a key block)...
	if (pc_outpOs)	{
		//replace the PGP-encoded block in the rich-text field with the decoded 
		//	output
		const size_t  UI = strlen( pc_outpOs) + 1;
		if (!( pc_outp = malloc( UI)))
			goto errJump;
		if (f_fail = !OSTranslate( OS_TRANSLATE_NATIVE_TO_LMBCS, pc_outpOs, 
										(WORD) (UI - 1), pc_outp, (WORD) UI))
			goto errJump;
		if (us_err = us_ReplaceRichAsciiLines( pc_inpOs, pc_blk, ul_lenBlk, 
													pc_outp, &t_crsr, pt_rtCtx))
			goto errJump;

		//if the note contains any attachments, wrap PGP-encoded attachments 
		//	within special-handling hotspots
		if (NSFNoteHasObjects( h_NOTE, NULL))
			if (us_err = us_ActionWrapPgpAttachments( h_NOTE, t_crsr, pt_rtCtx))
				goto errJump;

		//commit to note the changes made to the rich-text field
		if (us_err = eus_CommitChangedRtf( h_NOTE, pc_ITMNM, pt_rtCtx))
			goto errJump;

		//inform the caller that the message content did change
		*pf_MsgChanged = TRUE;
	} //if (pc_outpOs)

errJump:
	//free any memory allocation made by eul_GetRtfText() and free all 
	//	resources devoted to this instance of rich-text handling
	if (pc_outp)
		free( pc_outp);
	if (pc_outpOs)
		e_FreePgpMem( pc_outpOs);
	if (pc_inpOs)
		free( pc_inpOs);
	if (pc_inp)
		free( pc_inp);

	return us_err + f_fail + i_err;
} //i_PgpDecodeUsingAsciiText(


/** us_ActionWrapPgpAttachments( ***
Wrap attachments that seem to be PGP-encoded with a rich-text action hotspot 
that enables special PGP handling.

--- suggested enhancement ----
9/6/02 PR
+ Add to subform QueryOpen code that notes in lt_pgp item whether and which 
  PGP-encoded attachments appear below-the-line in a message that does not 
  include the special PGP-encoded rich-text attachment. This state would be 
  ascertained in xus_SetupPgpAttachments() and noted with something like 
  'BL-file1.ext.pgp/file2.ext.asc...' and, for not-applicable or none 'BL-/' 
  (or instead of BL is there already another PGP flag that can be extended, 
  e.g. AR?). Then this procedure could be more efficient and not have to search 
  through the entire rich-text field to determine if it needs to wrap certain 
  below-the-line attachments.
+ get rid of PGP-development shortcuts

--- parameters & return ------
h_NOTE: handle to the note being operated on
t_crsr: rich-text cursor set to where the search should start for attachemnt 
	hotspots whose parent attachments should be encoded
pt_rtCtx: Input & Output. Address of the rich-text context structure 
	describing the rich-text field to be searched and updated with PGP-encoded 
	attachment-action hotspots. Structure may undergo change even if an error 
	occurs.
RETURN: eus_SUCCESS if no error occured; the Notes API error code otherwise

--- revision history ---------
9/6/02 PR: created			*/
static STATUS us_ActionWrapPgpAttachments( NOTEHANDLE  h_NOTE, 
											RtfCursor  t_crsr, 
											RtfContext *const  pt_rtCtx)	{
	char * pc_objNm, * pc_fileNm;
	RtfSpan  t_rtSpan;
	UINT  ui_lenFileNm;
	StringNode * pt_notBelow = NULL, * pt_nd = NULL;
	BLOCKID  bid_itm = ebid_NULLBLOCKID;
	char  pc[ mi_MAXLEN_ATTCH_OBJNM];
	WORD  us_typ;
	STATUS  us_err;
	BOOL  f_fail = FALSE;

	_ASSERTE( (t_crsr.puc_location ? t_crsr.us_recLength : TRUE) && pt_rtCtx);

	//note whether the note's PGP-state item contains information on the note's 
	//	below-the-line PGP-encoded attachments state

	//if a file-attachment hotspot is found at or beyond our starting point...
	if (!ef_CursorToAttachmentHotspot( &t_crsr, pt_rtCtx, FALSE, &t_rtSpan, 
														&pc_objNm, &pc_fileNm))
		return !eus_SUCCESS;
	_ASSERTE( t_crsr.puc_location ? pc_objNm && *pc_objNm && pc_fileNm && 
															*pc_fileNm : TRUE);

	//loop until no further file-attachment hotspots exist in the rich-text 
	//	field
	if (!*mpc_fileNmPgpRtf)
		strcat( strcpy( mpc_fileNmPgpRtf, mpc_ROOTFILENM_PGP_RTF), 
														mpc_PGP_ARMOR_EXT_ASC);
	while (t_crsr.puc_location)	{
		//if we don't know the note's full below-the-line PGP-encoded 
		//	attachments state, note this attachment's object name
if (TRUE)
			if (!ef_AddStringNodeFifo( pc_objNm, TRUE, &pt_nd))	{
				us_err = ERR_MEMORY;
				goto errJump;
			}else if (!pt_notBelow)
				pt_notBelow = pt_nd;

		//if the attachment seems to be PGP encoded and not the special 
		//	encapsulated rich-text attachment, wrap a PGP action hotspot around 
		//	the file-attachment hotspot
		if (stricmp( pc_fileNm, mpc_fileNmPgpRtf) != ei_SAME && (stricmp( 
										pc_fileNm + (ui_lenFileNm = strlen( 
										pc_fileNm)) - mui_LEN_ARMOR_EXT_ASC, 
										mpc_PGP_ARMOR_EXT_ASC) == ei_SAME || 
										stricmp( pc_fileNm + ui_lenFileNm - 
										mui_LEN_ARMOR_EXT_PGP, 
										mpc_PGP_ARMOR_EXT_PGP) == ei_SAME))
			if (us_err = us_ActionWrapAttachment( &t_rtSpan, pc_fileNm, 
											pc_objNm, h_NOTE, FALSE, pt_rtCtx))
				goto errJump;

		//advance the rich-text cursor to the next file-attachment hotspot
		if (f_fail = !ef_CursorToAttachmentHotspot( &t_crsr, pt_rtCtx, FALSE, 
											&t_rtSpan, &pc_objNm, &pc_fileNm))
			goto errJump;
	} //while (t_crsr.puc_location)

	//if we know the note's full below-the-line PGP-encoded attachments state...
if (FALSE)
		//if attachments are to be wrapped, wrap them
		;
	//else process any attachments beyond those wrapped above...
	else	{
		UINT  ui = NULL;

		//if beginning cursor did not start at the beginning of the rich-text 
		//	field...
			//for each attachment found from the beginning of the field up to 
			//	where the beginning cursor points, add it's object name to the 
			//	do-not-process list

//PGP development shortcut: assumption here that any attachments beyond those 
//	wrapped above are below-the-line attachments
		//for each seeming attachment associated with the note...
		if (us_err = eus_getAttachInfoNext( h_NOTE, NULL, &bid_itm, pc, NULL, 
													NULL, NULL, &us_typ, NULL))
			goto errJump;
		while (bid_itm.pool)	{
			//if the object is a normal attachment, and its name is not in the 
			//	do-not-process list, and it's not the special encapsulated 
			//	rich-text attachment...
			if (!(us_typ == HOST_STREAM || us_typ == HOST_OLELIB || 
											us_typ == HOST_BYTEARRAY_EXT || 
											us_typ == HOST_BYTEARRAY_PAGE) && 
											!( ef_ListContainsString( pc, 
											pt_notBelow, FALSE) || stricmp( pc, 
											mpc_fileNmPgpRtf) == ei_SAME))
				//Wrap a PGP action hotspot around a file-attachment hotspot 
				//	created for the attachment. If this is the first time 
				//	through for this display event, add a new paragraph first 
				//	in which to display the hotspot.
				if (us_err = us_ActionWrapAttachment( NULL, pc, pc, h_NOTE, 
												!ui++ ? TRUE : FALSE, pt_rtCtx))
					goto errJump;

			//move on to the next object
			if (us_err = eus_getAttachInfoNext( h_NOTE, NULL, &bid_itm, pc, 
											NULL, NULL, NULL, &us_typ, NULL))
				goto errJump;
		} //while (bid_itm.pool)
	} //if (

errJump:
	if (pt_notBelow)
		e_FreeList( &pt_notBelow, TRUE);

	return us_err + f_fail;
} //us_ActionWrapPgpAttachments(


/** i_PgpDecodeUsingRichTextObj( ***


--- parameters & return ----
ul_OBJID: Optional.
RETURN: 

--- revision history -------
10/7/00 PR: minor documentation adjustment

3/20/00 PR
+ accommodated change in signature to ei_PgpDecodeBuffer()
+ improved safety code

9/12/99 PR: accommodation of compressed objects (forced into this because the 
	Domino SMTP MTA may decide to compress)
12/15/98 PR: created		*/
//DOC!!
static int i_PgpDecodeUsingRichTextObj( const DWORD  ul_OBJID, 
										const DWORD  ul_LEN_OBJ, 
										const BLOCKID  bid_OBJ_ITEM, 
										NOTEHANDLE  h_NOTE, 
										char  pc_ITMNM_RTF[], 
										RtfContext *const  pt_rtfContext, 
										int *const  pi_errNonFatal)	{
	HANDLE  h;
	BYTE * puc;
	DWORD  ul;
	char * pc_output = NULL;
	STATUS  us_err;
	int  i_err;

	_ASSERTE( h_NOTE && ul_LEN_OBJ && bid_OBJ_ITEM.pool && pc_ITMNM_RTF && 
															pt_rtfContext);

	if (pi_errNonFatal)
		*pi_errNonFatal = FALSE;

	//read the special attachment into a memory buffer
	if (us_err = eus_getObjectContentsInBuffer( ul_OBJID, ul_OBJID ? 
											ebid_NULLBLOCKID : bid_OBJ_ITEM, 
											ul_LEN_OBJ, h_NOTE, NULL, &h))	{
		if (pi_errNonFatal)
			*pi_errNonFatal = us_err;
		return us_err;
	}

	//PGP decode the buffer, then, our work with it done, free the buffer
	puc = OSLockObject( h);
	i_err = ei_PgpDecodeBuffer( puc, ul_LEN_OBJ, TRUE, &pc_output, &ul);
	OSUnlockObject( h);
	OSMemFree( h);
	if (i_err)	{
		if (pi_errNonFatal && i_err != ei_USER_ABORT)
			*pi_errNonFatal = i_err;
		goto errJump;
	}

	//Replace the current rich-text field with the CD records in the 
	//	decrypted buffer. From here on all errors are fatal because we 
	//	will no longer know the state of the rich-text tracking context.
	if (us_err = eus_ReplaceRtfWithCdStream( pc_output, ul, pt_rtfContext))
		goto errJump;

	//commit to note the changes made to the rich-text field
	if (us_err = eus_CommitChangedRtf( h_NOTE, pc_ITMNM_RTF, pt_rtfContext))
		goto errJump;

	//remove the item used by the note to point to the special attachment
	us_err = NSFNoteDetachFile( h_NOTE, bid_OBJ_ITEM);

errJump:
	if (pc_output)
		e_FreePgpMem( pc_output);

	return us_err + i_err;
} //i_PgpDecodeUsingRichTextObj(


/** us_ReplaceRichAsciiLines( ***
Replaces the first matching text block in the given rich-text field with the 
given null-terminated replacement text block.

--- parameters & return ----
pc_CTNT: Address of a null-terminated ASCII-text version of the specified 
	rich-text field. This input allows procedure to find the correct occurrence 
	of pc_BLK within the rich-text field.
pc_BLK: Address of the beginning of the block of CRLF-delimited text lines in 
	the rich-text field that is to be replaced. (Address must therefore fall 
	within the pc_CTNT string.) The first line of the block is used to search 
	for a matching position within the rich-text field.
ul_LEN_BLK: Length of text to be replaced. Value must therefore not offset 
	beyond the end of the pc_CTNT string.
pc_REPL: Address of null-terminated text block to replace the block within the 
	rich-text field identified by pc_BLK. If necessary, a CRLF will be appended 
	to maintain the concept of replacing _lines_.
pt_crsr: Input & Output. Address of rich-text cursor from which to start 
	searching for the specified text block (pc_BLK) within the rich-text field. 
	On successful return from call, cursor is eset to point to the CD record 
	following the replacement.
pt_rtCtx: Input & Output. Address of opaque structure describing the current 
	state of the rich-text field involved in the operation.
RETURN: eus_SUCCESS if no error occured; the Notes API error code otherwise

--- revision history -------
9/6/02 PR
+ logic enhancement to allow output of rich-text position following the 
  replacement block
+ improved procedure name, much token renaming, improved call-safety checking
+ completed standard documentation
+ listing format adjustment

7/16/99 PR: changed default nullification of end-cursor member of the RtfSpan 
	structure passed to eus_ReplaceRtfSpanText() to use the standard 
	null-cursor flag
11/23/98 PR: created		*/
static STATUS us_ReplaceRichAsciiLines( const char *const  pc_CTNT, 
										const char *const  pc_BLK, 
										const DWORD  ul_LEN_BLK, 
										const char *const  pc_REPL, 
										RtfCursor *const  pt_crsr, 
										RtfContext *const  pt_rtCtx)	{
	long  l;
	const char * pc;
	char * pc_cmp = NULL, * pc_ = NULL;
	WORD  us_offst, us = NULL;
	UINT  ui;
	RtfSpan  t_rtfSpn;
	STATUS  us_err;
	BOOL  f_fail;
	int  i_err = NULL;

	_ASSERTE( pc_CTNT && pc_BLK && ul_LEN_BLK && pc_BLK >= pc_CTNT && pc_BLK + 
								ul_LEN_BLK <= pc_CTNT + strlen( pc_CTNT) && 
								pc_REPL && pt_crsr->puc_location && pt_rtCtx);

	//Note the first line of the text block to be replaced. This will be used 
	//	to find the start of this block within the actual rich-text field.
	if (f_fail = !(BOOL) (pc_cmp = malloc( l = strstr( pc_BLK, epc_CRLF) - 
																pc_BLK + 1)))
		goto errJump;
	memcpy( pc_cmp, pc_BLK, l - 1);
	pc_cmp[ l - 1] = NULL;

	//starting from the caller-specified position in the rich-text field, loop 
	//	to the start of the text block to be replaced...
	pc = pc_CTNT - 1;
	do	{
		//locate the next instance of the sought-for line in the rich-text field
		if (f_fail = !ef_CursorToStringStart( pc_cmp, TRUE, *pt_crsr, us, 
												pt_rtCtx, pt_crsr, &us_offst))
			goto errJump;
		if (f_fail = !(BOOL) pt_crsr->puc_location)
			goto errJump;

		//move to the corresponding line in the regular-text version of the 
		//	content
		if (f_fail = !(BOOL) (pc = strstr( pc + 1, pc_cmp)))
			goto errJump;

		//if the line we found isn't the actual beginning of the sought-for 
		//	text block, loop for a further pass
		us = us_offst + 1;
	} while (pc < pc_BLK);

	//if necessary, tack a CRLF onto the replacement text to maintain the 
	//	concept of replacing lines with lines
	if (strcmp( pc_REPL + (ui = strlen( pc_REPL)) - ei_LEN_CRLF, epc_CRLF) != 
																ei_SAME)	{
		pc_ = malloc( ui + ei_LEN_CRLF + 1);
		strcat( strcpy( pc_, pc_REPL), epc_CRLF);
	}

	//replace the ASCII text block identified within the rich-text field with 
	//	the replacement content
	t_rtfSpn.t_crsrBgin = *pt_crsr;
	t_rtfSpn.us_offstBgin = us_offst;
	t_rtfSpn.t_crsrEnd.puc_location = NULL;
	us_err = eus_ReplaceRtfSpanText( &t_rtfSpn, ul_LEN_BLK, TRUE, pc_ ? pc_ : 
															pc_REPL, pt_rtCtx);
	if (pc_)
		free( pc_);
	if (us_err)
		goto errJump;

	//note for caller the position of the CD record following the replacement 
	//	text
	*pt_crsr = t_rtfSpn.t_crsrEnd;
	us_err = eus_CursorAdvance( pt_crsr, pt_rtCtx);

errJump:
	if (pc_cmp)
		free( pc_cmp);

	return us_err ? us_err : i_err || f_fail ? !eus_SUCCESS : eus_SUCCESS;
} //us_ReplaceRichAsciiLines(


/** xus_TestNoteForPgpAscii( ***
This exported interface allows the note passed in to be inspected for 
PGP-encoded ASCII text in the message body (rich-text field).

--- parameters & return ----
h_NOTE: the handle to the note to be inspected for PGP-encoded ASCII
pc_ITMNM_RTF: the name of the message-body rich-text field
pus_hasPgpAscii: flag telling whether encoded ASCII was found (TRUE) or not 
	(FALSE)
RETURN:
	eus_SUCCESS if no errors occurred
	eus_ERR_INVLD_ARG if any passed-in parameter is obviously invalid
	the Notes API error code otherwise

--- revision history -------
9/6/02 PR
+ logic enhancement to support high-ASCII (international) characters in 
  attachment handling
+ minor documentation adjustment, listing format adjustment, token renaming, 
  exception-handling adjustment

11/18/98 PR: created		*/
STATUS xus_TestNoteForPgpAscii( NOTEHANDLE  h_NOTE, 
								char  pc_ITMNM_RTF[], 
								short *const  ps_hasPgpAscii)	{
	RtfCursor  t_crsr;
	RtfContext  t_rtCtx;
	char * pc_ctnt = NULL, * pc_ctntOs;
	ULONG  ul;
	BOOL  f_hasPgpAscii, f_failure;
	STATUS  us_err;
	int  i_err = NULL;

	if (!( h_NOTE && pc_ITMNM_RTF && *pc_ITMNM_RTF && ps_hasPgpAscii))
		return eus_ERR_INVLD_ARG;

	//set default that PGP-encoded ASCII content was not found
	*ps_hasPgpAscii = ms_VB_FALSE;

	//initialize resources associated with the rich-text handling we're going 
	//	to do
	if (us_err = eus_InitializeRtfContext( h_NOTE, pc_ITMNM_RTF, &t_crsr, 
																	&t_rtCtx))
		return us_err;

	//if we can't get the textual content of the rich-text field or if that 
	//	content is too short to contain a PGP block, short-circuit
	if ( (f_failure = eul_ERR_FAILURE == (ul = eul_GetRtfText( &t_rtCtx, 
											NULL, TRUE, &pc_ctnt))) || ul <= 1)
		goto errJump;

	//if PGP-encoded ASCII is included in the content, indicate this to caller
	if (!(pc_ctntOs = malloc( ul)))	{
		i_err = -11999;		//kPGPError_OutOfMemory
		goto errJump;
	}
	if (OSTranslate( OS_TRANSLATE_LMBCS_TO_NATIVE, pc_ctnt, (WORD) (ul - 1), 
													pc_ctntOs, (WORD) ul))	{
		if (!( i_err = ei_TestForPgpAscii( pc_ctntOs, &f_hasPgpAscii)) && 
																f_hasPgpAscii)
			*ps_hasPgpAscii = ms_VB_TRUE;
	}else
		i_err = !eus_SUCCESS;
	free( pc_ctntOs);

errJump:
	//as necessary, free resources allocated during this procedure
	if (pc_ctnt)
		free( pc_ctnt);
	ef_FreeRtfContext( &t_rtCtx);

	return f_failure || i_err ? !eus_SUCCESS : eus_SUCCESS;
} //xus_TestNoteForPgpAscii(


/** xus_TestNoteForPgpArmor( ***
Determine whether PGP-encoded file attachments occur in the specified rich-text 
field, providing differentiation between the special PGP-encoded rich-text 
object and regular encoded attachments.

--- parameters & return ----
h_NOTE: the handle to the note containing the specified rich-text field
pc_ITMNM_RTF: pointer to a string telling the name of the rich-text field to be 
	processed
ps_hasPgpAtchmt: Provisionally Optional Output. Address of the Visual-Basic 
	type flag in which to return whether any PGP-encoded file attachemnts were 
	found apart from the special rich-text attachment. If null, output will be 
	suppressed. At least one output must be requested within call.
ps_hasRtfObj: Provisionally Optional Output. Address of the Visual-Basic type 
	flag in which to return wehter the special PGP-encoded rich-text attachment 
	is present on the given note. If null, output will be suppressed. At least 
	one output must be requested within call.
RETURN:
	eus_ERR_INVLD_ARG if any input is obviously invalid
	!eus_SUCCESS if a rich-text cursoring error occurred
	eus_SUCCESS if no error occured
	the Notes API error code otherwise

--- revision history -------
9/6/02 PR
+ fixed variable-initialization bug
+ fixed bug that could give a false-positive when deciding if an attachment was 
  the special PGP-encoded rich-text attachment
+ listing format adjustment, token renaming

9/16/00 PR
+ added functionality to differentiate between the special rich-text object and 
  other PGP-encoded attachments; required a signature change
+ updated so that attachments with either PGP extension are supported
+ logic streamlining

11/12/98 PR: created		*/
STATUS xus_TestNoteForPgpArmor( const NOTEHANDLE  h_NOTE, 
								char  pc_ITMNM_RTF[], 
								short *const  ps_hasPgpAtchmt, 
								short *const  ps_hasRtfObj)	{
	static UINT  ui_lenFileNmPgpRtf;

	const WORD  us_LEN_ITEMNM = strlen( epc_ITEMNM_STANDARD_ATTACH);

	BLOCKID  bid_itm, bid_ctnt, bid_prvItm;
	const BYTE * puc;
	const FILEOBJECT * pflo;
	BOOL  f_hasRtfObj = FALSE, f_hasPgpAtchmt = FALSE, f_fail = FALSE;
	STATUS  us_err;

	if (!( h_NOTE && pc_ITMNM_RTF && (ps_hasPgpAtchmt || ps_hasRtfObj)))
		return eus_ERR_INVLD_ARG;

	if (ps_hasPgpAtchmt)
		*ps_hasPgpAtchmt = ms_VB_FALSE;
	if (ps_hasRtfObj)
		*ps_hasRtfObj = ms_VB_FALSE;

	//if it hasn't been done already, construct the filename of the special PGP 
	//	rich-text object
	if (!*mpc_fileNmPgpRtf)	{
		strcat( strcpy( mpc_fileNmPgpRtf, mpc_ROOTFILENM_PGP_RTF), 
														mpc_PGP_ARMOR_EXT_ASC);	
		ui_lenFileNmPgpRtf = strlen( mpc_fileNmPgpRtf);
	}

	//Get a handle on the first attachment item present in the note. If no such 
	//	item is present, we've determined that the answer to all the caller's 
	//	questions is negative, so short-circuit with success.
	if (us_err = NSFItemInfo( h_NOTE, epc_ITEMNM_STANDARD_ATTACH, 
													us_LEN_ITEMNM, &bid_itm, 
													NULL, &bid_ctnt, NULL))
		if (ERR( us_err) == ERR_ITEM_NOT_FOUND)
			return eus_SUCCESS;
		else
			return us_err;

	//cycle through the attachment items (and thus object names) until we've 
	//	gleaned as much information useful to our purpose as we can...
	do	{
		//if the object name given by the item matches the name of the special 
		//	PGP rich-text object...
		pflo = (FILEOBJECT *) (OSLockBlock( WORD, bid_ctnt) + 1);
		puc = (BYTE *) (pflo + 1);
		if (ui_lenFileNmPgpRtf == pflo->FileNameLength && strnicmp( puc, 
											mpc_fileNmPgpRtf, 
											ui_lenFileNmPgpRtf) == ei_SAME)	{
			//if the caller wants to be told this information...
			if (ps_hasRtfObj)	{
				//make a note of it
				f_hasRtfObj = TRUE;

				//if the caller doesn't need to be told whether a regular 
				//	PGP-encoded attachment is present or if it's already known 
				//	that one is present, we need go no further, so break out of 
				//	the loop
				if (!ps_hasPgpAtchmt || f_hasPgpAtchmt)	{
					OSUnlockBlock( bid_ctnt);
					break;
				}
			} //if (ps_hasRtfObj)
		//else if the caller wants to know whether a regular PGP-encoded 
		//	attachment is present and this is still not known and the 
		//	associated object name sports one of the PGP extensions...
		}else if (ps_hasPgpAtchmt && !f_hasPgpAtchmt)	{
			puc += pflo->FileNameLength;
			if (strnicmp( puc - mui_LEN_ARMOR_EXT_ASC, mpc_PGP_ARMOR_EXT_ASC, 
										mui_LEN_ARMOR_EXT_ASC) == ei_SAME || 
										strnicmp( puc - mui_LEN_ARMOR_EXT_PGP, 
										mpc_PGP_ARMOR_EXT_PGP, 
										mui_LEN_ARMOR_EXT_PGP) == ei_SAME)	{
				//make a note that such an attachment is present
				f_hasPgpAtchmt = TRUE;

				//if the caller doesn't need to be told whether the special 
				//	rich-text object is present or if it's already known that 
				//	it is present, we need go no further, so break out of the 
				//	loop
				if (!ps_hasRtfObj || f_hasRtfObj)	{
					OSUnlockBlock( bid_ctnt);
					break;
				}
			} //if (strnicmp( puc - mui_LEN_ARMOR_EXT_ASC,
		} //if (ui_lenFileNmPgpRtf == pflo->FileNameLength && strnicmp(

		//get a handle on the next attachment item in the note
		OSUnlockBlock( bid_ctnt);
		bid_prvItm = bid_itm;
		if ((us_err = NSFItemInfoNext( h_NOTE, bid_prvItm, 
											epc_ITEMNM_STANDARD_ATTACH, 
											us_LEN_ITEMNM, &bid_itm, NULL, 
											&bid_ctnt, NULL)) && 
											ERR( us_err) != ERR_ITEM_NOT_FOUND)
			return us_err;
	} while (!us_err);

	//if we didn't finish finding everything out by querying _object_ names, 
	//	traverse the rich-text field itself to check whether the _file_ names 
	//	(the ultimate word) in any attachment hotspots can add something...
	if (ps_hasPgpAtchmt && !f_hasPgpAtchmt || ps_hasRtfObj && !f_hasRtfObj)	{
		RtfContext  t_rtCtx;
		RtfCursor  t_crsr;
		const char * pc_fileNm, * pc;

		//initialize contextual information about the specified rich-text field
		if (us_err = eus_InitializeRtfContext( h_NOTE, pc_ITMNM_RTF, &t_crsr, 
																	&t_rtCtx))
			return us_err;

		//advance the rich-text cursor to the next file-attachment hotspot, 
		//	which (though unlikely) may be the first CD record 
		if (!ef_CursorToAttachmentHotspot( &t_crsr, &t_rtCtx, FALSE, NULL, 
														NULL, &pc_fileNm))	{
			ef_FreeRtfContext( &t_rtCtx);
			return !eus_SUCCESS;
		}

		//loop until either no further file-attachment hotspots exist in the 
		//	rich-text field (or until we've determined the information 
		//	requested by caller)
		while (t_crsr.puc_location)	{
			//if we didn't know of it before and just now hit on the special 
			//	rich-text object...
			if (ps_hasRtfObj && !f_hasRtfObj && stricmp( pc_fileNm, 
											mpc_fileNmPgpRtf) == ei_SAME)	{
				//if caller wants to know whether the object is present, note 
				//	that it is and break out of the loop if there's nothing 
				//	left to find out
				if (ps_hasRtfObj)	{
					f_hasRtfObj = TRUE;
					if (!ps_hasPgpAtchmt || f_hasPgpAtchmt)
						break;
				}
			//Else if caller wants to know whether regular PGP-encoded 
			//	attachments are present and none have been located yet...
			}else if (ps_hasPgpAtchmt && !f_hasPgpAtchmt)	{
				//if we've hit on one, note that it is and break out of the 
				//	loop if there's nothing left to find out
				pc = pc_fileNm + strlen( pc_fileNm);
				if (stricmp( pc - mui_LEN_ARMOR_EXT_ASC, 
										mpc_PGP_ARMOR_EXT_ASC) == ei_SAME || 
										stricmp( pc - mui_LEN_ARMOR_EXT_PGP, 
										mpc_PGP_ARMOR_EXT_PGP) == ei_SAME)	{
					f_hasPgpAtchmt = TRUE;
					if (!ps_hasRtfObj || f_hasRtfObj)
						break;
				} //if (stricmp( pc - mui_LEN_ARMOR_EXT_ASC
			} //if (stricmp( pc_fileNm, mpc_fileNmPgpRtf)

			//advance the rich-text cursor to the next file-attachment hotspot
			if (f_fail = !ef_CursorToAttachmentHotspot( &t_crsr, &t_rtCtx, 
												FALSE, NULL, NULL, &pc_fileNm))
				break;
		} //while (t_crsr.puc_location)

		//free the contextual information about the rich-text field
		ef_FreeRtfContext( &t_rtCtx);

		if (f_fail)
			return !eus_SUCCESS;
	} //if (ps_hasPgpAtchmt || ps_hasRtfObj &&

	//inform caller of any positive findings
	if (f_hasPgpAtchmt)
		*ps_hasPgpAtchmt = ms_VB_TRUE;
	if (f_hasRtfObj)
		*ps_hasRtfObj = ms_VB_TRUE;

	return eus_SUCCESS;
} //xus_TestNoteForPgpArmor(


/** us_LocatePgpRtfAttachment( ***


--- parameters & return ----

RETURN: 

--- revision history -------
9/6/02 PR: logic performance enhancement, minor token renaming, listing format 
	adjustment

9/16/00 PR: token renaming
9/12/99 PR: added smart differentiation between compressed & non-compressed 
	files
1/30/99 PR: created			*/
//DOC!!
static STATUS us_LocatePgpRtfAttachment( RtfCursor  t_crsr, 
											const RtfContext *const  pt_RT_CTX, 
											NOTEHANDLE  h_NOTE, 
											DWORD *const  pul_objId, 
											DWORD *const  pul_size, 
											BLOCKID *const  pbid_itm)	{
	const char * pc_objNm, * pc_fileNm;
	BOOL  f_Match = FALSE, f_compressed;
	DWORD  ul_objId;
	STATUS  us_err;

	_ASSERTE( t_crsr.puc_location && t_crsr.us_recLength && pul_objId && 
														pul_size && pbid_itm);

	*pul_objId = *pul_size = NULL;
	*pbid_itm = ebid_NULLBLOCKID;

	if (!NSFNoteHasObjects( h_NOTE, NULL))
		return eus_SUCCESS;

	//advance the rich-text cursor to the next file-attachment hotspot, which 
	//	(though unlikely) may be the first CD record 
	if (!ef_CursorToAttachmentHotspot( &t_crsr, pt_RT_CTX, FALSE, NULL, 
														&pc_objNm, &pc_fileNm))
		return !eus_SUCCESS;

	//Loop either until no further file-attachment hotspots exist in the 
	//	rich-text field or until the PGP rich-text attachment is found. We 
	//	search for the special attachment using this technique first in case 
	//	more than one version of the PGP rich-text attachment exists on the 
	//	document. If there is more than one, we should default to the one 
	//	actually present in the RTF because that one must be the most recent 
	//	version.
	if (!*mpc_fileNmPgpRtf)
		strcat( strcpy( mpc_fileNmPgpRtf, mpc_ROOTFILENM_PGP_RTF), 
														mpc_PGP_ARMOR_EXT_ASC);
	while (t_crsr.puc_location)	{
		//A file-attachment hotspot must have been found, so determine whether 
		//	the "original" name of the attachment as stored in the hotspot 
		//	matches the filename we're after. If we have found a match, break 
		//	out of this loop to move on to the next stage
		if (f_Match = stricmp( pc_fileNm, mpc_fileNmPgpRtf) == ei_SAME)
			break;

		//advance the rich-text cursor to the next file-attachment hotspot
		if (!ef_CursorToAttachmentHotspot( &t_crsr, pt_RT_CTX, FALSE, NULL, 
														&pc_objNm, &pc_fileNm))
			return !eus_SUCCESS;
	} //while (t_crsr.puc_location)

	//Get the necessary location info for the caller to use in accessing the 
	//	attachment. If we didn't find a corresponding attachment hotspot in the 
	//	rich-text field, we will accept a properly named attachment not 
	//	incorporated.
	if (!f_Match)
		pc_objNm = mpc_fileNmPgpRtf;
	if (us_err = eus_getAttachmentInfo( h_NOTE, NULL, pc_objNm, &ul_objId, 
													pul_size, &f_compressed, 
													NULL, pbid_itm, NULL))
		return us_err;

	//if the file is not compressed, set the object ID return variable, thus 
	//	signifying that the file is uncompressed (important later to 
	//	eus_getObjectContentsInBuffer()
	if (!f_compressed)
		*pul_objId = ul_objId;

	return us_err;
} //us_LocatePgpRtfAttachment(


/** xus_SetupPgpRtfObj( ***
Purpose is to work the special PGP rich-text attachment and its associated 
rich-text hotspot associated into a user- and system-friendly state. 
User-friendly by making it hidden and system-friendly by having it behave as 
needed (even if that means getting rid of it).

--- parameter & return ----
h_NOTE: the handle to the note containing the hotspots to be converted
pc_ITMNM_RTF: the name of the message-body rich-text field
RETURN:
	eus_SUCCESS if no errors occurred
	eus_ERR_INVLD_ARG if any input parameter is obviously invalid
	the Notes API error code otherwise

--- revision history ------
9/6/02 PR: minor logic adjustments for performance and in support of signature 
	change to eus_getAttachmentInfo, listing format adjustment, token renaming, 
	minor documentation adjustment

9/16/00 PR
+ updated so that attachments with either PGP extension are supported
+ changed name of function to better reflect its work
+ token renaming

1/29/99 PR: created			*/
STATUS xus_SetupPgpRtfObj( const NOTEHANDLE  h_NOTE, 
								char  pc_ITMNM_RTF[])	{
	static const char  pc_NOTES_MAIL_MATCH[] = "Lotus Notes Release";
	static char  pc_ITMNM_MAILER[] = "$Mailer";		//MAIL_MAILER_ITEM not 
													//	defined in 4.5.1 API

	RtfContext  t_rtCtx, t_rtCtxAttachPgrph;
	RtfCursor  t_crsr;
	RtfSpan  t_span;
	const char * pc_fileNm;
	char * pc_objNm, * pc;
	BOOL  f_utilRtfStarted = FALSE, f_mailedByNotes = TRUE, f_fail;
	BLOCKID  bid_ctnts, bid_itm;
	STATUS  us_err;

	if (!( h_NOTE && pc_ITMNM_RTF))
		return eus_ERR_INVLD_ARG;

	//if there aren't any attachments, short-circuit with success
	if (!NSFNoteHasObjects( h_NOTE, NULL))
		return eus_SUCCESS;

	//initialize resources associated with the rich-text handling we're going 
	//	to do
	if (us_err = eus_InitializeRtfContext( h_NOTE, pc_ITMNM_RTF, &t_crsr, 
																	&t_rtCtx))
		return us_err;

	//advance the rich-text cursor to the next file-attachment hotspot, which 
	//	(though unlikely) may be the first CD record
	if (f_fail = !ef_CursorToAttachmentHotspot( &t_crsr, &t_rtCtx, FALSE, 
												&t_span, &pc_objNm, &pc_fileNm))
		goto errJump;

	//search for a special PGP-encoded rich-text attachment hotspot in the 
	//	rich-text field
	if (!*mpc_fileNmPgpRtf)
		strcat( strcpy( mpc_fileNmPgpRtf, mpc_ROOTFILENM_PGP_RTF), 
														mpc_PGP_ARMOR_EXT_ASC);
	while (t_crsr.puc_location && stricmp( pc_fileNm, mpc_fileNmPgpRtf) != 
																		ei_SAME)
		if (f_fail = !ef_CursorToAttachmentHotspot( &t_crsr, &t_rtCtx, FALSE, 
												&t_span, &pc_objNm, &pc_fileNm))
			goto errJump;

	//determine whether the message was mailed by Notes
	if (us_err = eus_AttemptCopyTextItemContent( h_NOTE, pc_ITMNM_MAILER, 
															NULL, &pc, NULL))
		goto errJump;
	if (pc)	{
		f_mailedByNotes = strnicmp( pc, pc_NOTES_MAIL_MATCH, sizeof( 
										pc_NOTES_MAIL_MATCH) - 1) == ei_SAME;
		free( pc);
	}

	//if a rich-text hotspot was found associated with the special rich-text 
	//	attachment...
	if (t_crsr.puc_location)	{
		//remove current hotspot because one way or another it's got to go
		if (us_err = eus_RemoveRtSpan( &t_span, &t_rtCtx))
			goto errJump;

		//if the message wasn't mailed by Notes...
		if (!f_mailedByNotes)	{
			//delete the attachment associated with the removed 
			//	special-attachment hotspot since it's very unlikely that its 
			//	contents are valid (e.g. the attachment originated in a Notes 
			//	plug-in client-sent e-mail to a non-Notes e-mail client, then 
			//	the attachment was forwarded on to a plug-in-enabled Notes 
			//	client, so the attachment contains the contents of the original 
			//	message, not the forwarded message, not a good thing)
			if (us_err = eus_DeleteAttachment( h_NOTE, pc_objNm, NULL))
				if (ERR( us_err) != ERR_ITEM_NOT_FOUND)
					goto errJump;
				else
					us_err = eus_SUCCESS;
		//else eliminate the possibility of the special PGP rich-text 
		//	attachment being inherited into documents created via the @Command( 
		//	[MailForward]) mechanism or via forms having the "Inherit entire 
		//	selected document into the rich-text field ... as rich-text" 
		//	setting...
		}else	{
			//initialize the rich-text hidden paragraph we will use to store the 
			//	special object
			if (us_err = us_SetupRtfAttachParagraph( h_NOTE, 
												&t_rtCtxAttachPgrph, pc_objNm))
				goto errJump;
			f_utilRtfStarted = TRUE;

			//Append the special ending hidden-paragraph rich-text to the our 
			//	adjusted main rich-text field. Since both above-mentioned 
			//	mechanisms ignore whatever's in hidden paragraphs, this method 
			//	avoids the inheritance issue and is cosmetically attractive to boot.
			if (us_err = eus_AppendItemsToRtf( &t_rtCtxAttachPgrph, &t_rtCtx))
				goto errJump;
		} //if (f_mailedByNotes)

		//commit the updated rich-text field to the note, replacing the current 
		//	actual rich-text content
		us_err = eus_CommitChangedRtf( h_NOTE, pc_ITMNM_RTF, &t_rtCtx);
	//Else if the special PGP rich-text attachment is present without a 
	//	corresponding hotspot, the message probably came into the system via 
	//	the client POP3/IMAP4 "Direct to Internet" mechanism...
	}else if (us_err = eus_getAttachmentInfo( h_NOTE, NULL, mpc_fileNmPgpRtf, 
														NULL, NULL, NULL, NULL, 
														&bid_itm, &bid_ctnts))
		goto errJump;
	else if (bid_ctnts.pool)	{
		//if the message was put in the mail by Notes...
		if (f_mailedByNotes)	{
			//Change the "host type" of the special attachment to a type that 
			//	Notes will not inherit into documents created via the 
			//	mechanisms described above. This is an acceptable thing to do 
			//	for _received_ messages, since this message won't be going 
			//	through a SMTP MTA anymore, so we don't care about the fact 
			//	that the MTA will not forward attachments with the STREAM host 
			//	type. Another nice by-product of doing this is that the 
			//	attachment will not show up "below the line" on the message 
			//	(i.e. it'll be hidden)
			FILEOBJECT *const  ptfo = (FILEOBJECT *) (OSLockBlock( WORD, 
																bid_ctnts) + 1);

			ptfo->HostType = HOST_STREAM;
			OSUnlockBlock( bid_ctnts);
		//else delete the special attachment since it's very unlikely that 
		//	its contents remain valid (e.g. the attachment originated in a 
		//	Notes plug-in client-sent e-mail to a non-Notes e-mail client, then 
		//	the attachment was forwarded on to a plug-in-enabled Notes client, 
		//	so the attachment contains the contents of the original message, 
		//	not the forwarded message, not a good thing)
		}else
			us_err = NSFNoteDetachFile( h_NOTE, bid_itm);
	} //if (us_err = eus_getAttachmentInfo(

errJump:
	if ((us_err || f_fail) && f_utilRtfStarted)
		ef_FreeRtfContext( &t_rtCtxAttachPgrph);
	ef_FreeRtfContext( &t_rtCtx);

	return us_err + f_fail;
} //xus_SetupPgpRtfObj(


/** DllMain( ***
Win32 entry point for loading and unloading this DLL.  We use the unload 
event to free statically allocated resources.

--- parameters & return ---
h_DllInstance: the overall handle to this DLL's runtime information, settings
ul_reason: parameter tells why the function was called
pv_reserved: just that, reserved

--- revision history ------
3/20/00 PR
+ support of rich-text reset functionality, part of changes needed for 
  Notes Mail R5-template compatibility 
+ adjustment in support of PGP 7.0

11/18/98 PR: created		*/
BOOL WINAPI DllMain( HINSTANCE  h_DllInstance, 
						DWORD  ul_reason, 
						LPVOID  pv_reserved)	{
	switch (ul_reason)	{
		case DLL_PROCESS_ATTACH:
			//Set up debugging aids and the memory-leak detector in 
			//	particular.  Lines are preprocessed out in Release builds.
			_CrtSetReportMode( _CRT_WARN, _CRTDBG_MODE_DEBUG | 
														_CRTDBG_MODE_WNDW);
			_CrtSetDbgFlag( _CRTDBG_LEAK_CHECK_DF | 
										_CrtSetDbgFlag( _CRTDBG_REPORT_FLAG));

			//keep ahold of the instance handle
			eh_Instance = h_DllInstance;

			//for some efficiency gain, tell system there's no need to tell 
			//	our DLL about thread attachment & detachment
			if (!DisableThreadLibraryCalls( h_DllInstance))
				return FALSE;

			break;

		case DLL_PROCESS_DETACH:
			//free any resources still allocated to a copy of rich-text 
			//	content stored in memory
			if (mt_RtfResettable.h_content)
				OSMemFree( mt_RtfResettable.h_content);
	} //switch (ul_reason)

	return TRUE;
} //DllMain(


/** es_PreMailSend( ***
Purpose is to take any PGP action necessary before the message gets sent on 
to MAIL.BOX as the current mail-send event is carried through.

Although it probably would have been more elegant to place this procedure in 
the ndbPGP DLL, the debugging infrastructure doesn't like one DLL freeing 
memory allocated in another DLL, even if the DLLs are in the same process. I 
don't know whether this is a limitation of the Windows debugging 
infrastructure or if its just a no-no in general for reasons unknown to me.

--- parameter & return ----
ppv_ARGV: pointer to the series of arguments provided by the PGP @DbCommand 
	used to call this DLL
RETURN: eus_SUCCESS if successful; an appropriate error code otherwise

--- revision history ------
9/6/02 PR: minor documentation adjustment

3/20/00 PR
+ support of rich-text reset functionality, part of changes needed for 
  Notes Mail R5-template compatibility
+ token renaming

9/12/99 PR
+ logic adjustment to support enhancement of better emulation of Lotus Notes' 
  name lookup behavior
+ documentation adjustment, logic shortening

1/23/99 PR: created			*/
short es_PreMailSend( void * *const  ppv_ARGV)	{
	const int  i_ARG_HANDLE = 0, i_ARG_PGP_FLAGS = 1, 
				i_ARG_RECIPIENT_NMS = 2, i_ARG_LOOKUP_EXHAUST = 3, 
				i_ARG_LOOKUP_HOW = 4, i_ARG_ATTACHMENTS = 5, 
				i_ARG_RTF_FLDNM = 6;
	const char  pc_PGPFLG_SIGN[] = "SI", pc_PGPFLG_ENCRYPT[] = "EN";

	double  d;
	NOTEHANDLE  h_note;
	char * pc_ItmNmRtf = NULL, * pc_LookupHow = NULL, 
			* pc_RecipientsList = NULL;
	BOOL  f_PgpSign, f_PgpEncrypt, f_lookupExhaust = FALSE, f_failure = FALSE;
	int  i_Attachments = 0;
	STATUS  us_err;
	short  s_error = NULL;

	if (!( ppv_ARGV && *ppv_ARGV))
		return !eus_SUCCESS;

	//recover the handle to the calling note's session
	if (!ef_GetNumberListEntry( ppv_ARGV[ i_ARG_HANDLE], TRUE, 0, &d))
		return !eus_SUCCESS;
	h_note = (NOTEHANDLE) (DWORD) d;

	//determine what sort of PGP encoding is to be performed
	f_PgpSign = ef_TextListContainsEntry( ppv_ARGV[ i_ARG_PGP_FLAGS], TRUE, 
												pc_PGPFLG_SIGN, NULL, NULL);
	f_PgpEncrypt = ef_TextListContainsEntry( ppv_ARGV[ i_ARG_PGP_FLAGS], 
										TRUE, pc_PGPFLG_ENCRYPT, NULL, NULL);

	//get the recipient-list string
	if (us_err = eus_CreateTextListEntryCopy( 0, 
											ppv_ARGV[ i_ARG_RECIPIENT_NMS], 
											TRUE, &pc_RecipientsList))
		return us_err;
	if (!pc_RecipientsList && f_PgpEncrypt)
		return !eus_SUCCESS;

	//if PGP security has been specified to be performed and if the list of 
	//	recipient names contains someone...
	if ((f_PgpSign || f_PgpEncrypt) && pc_RecipientsList)	{
		//get the name of the rich-text field to process
		if ((us_err = eus_CreateTextListEntryCopy( 0, 
											ppv_ARGV[ i_ARG_RTF_FLDNM], TRUE, 
											&pc_ItmNmRtf)) || 
											(f_failure = !pc_ItmNmRtf))
			goto errJump;

		//if we're to encrypt...
		if (f_PgpEncrypt)	{
			//get the number of attachments associated with the note
			if (!ef_GetNumberListEntry( ppv_ARGV[ i_ARG_ATTACHMENTS], TRUE, 
																	0, &d))
				return !eus_SUCCESS;
			i_Attachments = (int) d;

			//get the name of the user's mail (home) server
			if (us_err = eus_CreateTextListEntryCopy( 0, 
												ppv_ARGV[ i_ARG_LOOKUP_HOW], 
												TRUE, &pc_LookupHow))
				goto errJump;
			if (!pc_LookupHow && !(pc_LookupHow = calloc( 1, sizeof( char))))
				goto errJump;

			//determine whether caller wants name lookups to be "exhaustive"
			if (f_failure = !ef_GetNumberListEntry( ppv_ARGV[ 
										i_ARG_LOOKUP_EXHAUST], TRUE, 0, &d))
				goto errJump;
			f_lookupExhaust = (BOOL) d;
		} //if (f_PgpEncrypt)

		//Initialize the information structure we use to hold onto a copy of 
		//	the PGP-encoded rich-text field we will be constructing. We may 
		//	need to reset the note's rich-text field to the PGP-encoded 
		//	version if the encoding was initiated by a user press of the 
		//	"Send and File" action button.
		if (mt_RtfResettable.h_note)	{
			OSMemFree( mt_RtfResettable.h_content);
			mt_RtfResettable.h_content = NULL;
		}
		mt_RtfResettable.h_note = h_note;

		//Prepare the message for sending in a PGP-secured manner. If failure 
		//	is encountered, flag that the structure used to hold a copy of 
		//	the PGP-encoded rich-text field is empty.
		if (s_error = xs_PgpSecureMessage( h_note, f_PgpSign, f_PgpEncrypt, 
										pc_RecipientsList, f_lookupExhaust, 
										pc_LookupHow, i_Attachments, 
										pc_ItmNmRtf, &mt_RtfResettable))
			mt_RtfResettable.h_note = NULL;
	} //if ((f_PgpSign || f_PgpEncrypt) && 

errJump:
	//free resources allocated via this procedure
	if (pc_RecipientsList)
		free( pc_RecipientsList);
	if (pc_ItmNmRtf)
		free( pc_ItmNmRtf);
	if (pc_LookupHow)
		free( pc_LookupHow);

    return (short) us_err + s_error + f_failure;
} //es_PreMailSend(


/** es_ReplaceMimeWithRtf( ***
Replaces the specified string of MIME-part fields with a previously stored 
Notes rich-text stream.

--- parameter & return ----
ppv_ARGV: pointer to the series of arguments provided by the PGP @DbCommand 
	used to call this DLL
RETURN:
	eus_SUCCESS if no errors occurred
	eus_ERR_INVLD_ARG if any input parameter is obviously invalid
	the Notes API error code otherwise

--- revision history ------
9/6/02 PR: listing format adjustment, minor exception-handling adjustment, 
	minor documentation adjustment
3/20/00 PR: created			*/
short es_ReplaceMimeWithRtf( void * *const  ppv_ARGV)	{
	const int  i_ARG_HANDLE = 0, i_ARG_RTF_FLDNM = 1;
	const WORD  us_TYPE_MIME_PART = 0x0019;

	NOTEHANDLE  h;
	double  d = NULL;
	char * pc;
	WORD  us;
	BOOL  f_RtfContextSet = FALSE;
	BYTE * puc;
	RtfContext  t;
	STATUS  us_err;

	if (!( ppv_ARGV && *ppv_ARGV && mt_RtfResettable.h_content))
		return eus_ERR_INVLD_ARG;

	//Recover the handle to the calling note's session. If it doesn't 
	//	correspond to the note associated with the previously stored rich-text 
	//	stream, short-circuit with failure.
	if (!( ef_GetNumberListEntry( ppv_ARGV[ i_ARG_HANDLE], TRUE, 0, &d) && d))
		return eus_ERR_INVLD_ARG;
	if ((h = (NOTEHANDLE) (DWORD) d) != mt_RtfResettable.h_note)
		return eus_ERR_INVLD_ARG;

	//obtain the name of the items to be replaced
	if (us_err = eus_CreateTextListEntryCopy( 0, ppv_ARGV[ i_ARG_RTF_FLDNM], 
																	TRUE, &pc))
		return (short) us_err;
	if (!*pc)	{
		us_err = eus_ERR_INVLD_ARG;
		goto errJump;
	}

	//if the first of the named items is not of type MIME Part, short-circuit 
	//	with failure
	if (us_err = NSFItemInfo( h, pc, (WORD) strlen( pc), NULL, &us, NULL, 
																	NULL))	{
		if (ERR( us_err) == ERR_ITEM_NOT_FOUND)
			us_err = eus_ERR_INVLD_ARG;
		goto errJump;
	}
	if (us != us_TYPE_MIME_PART)	{
		us_err = eus_ERR_INVLD_ARG;
		goto errJump;
	}

	//remove the MIME-part items
	if (us_err = eus_RemoveItem( h, pc, ebid_NULLBLOCKID))
		goto errJump;

	//set up a replacement field based on the already stored stream of 
	//	rich-text
	puc = OSLockObject( mt_RtfResettable.h_content);
	us_err = eus_ReplaceRtfWithCdStream( puc, mt_RtfResettable.ul_len, 
												ept_InitUtilityRtfContext( &t));
	OSUnlockObject( mt_RtfResettable.h_content);
	f_RtfContextSet = TRUE;
	if (us_err)
		goto errJump;

	//commit the rich-text to the note under the same name used by the 
	//	MIME-part items
	us_err = eus_CommitChangedRtf( h, pc, &t);

errJump:
	//If an error occurred, free resources allocated to the stored version, 
	//	and flag that this has been done. We don't free if no error occurred 
	//	because the rich-text stream may be resed in the note's post-save 
	//	(R5) or query-close (R4) events.
	if (us_err)	{
		OSMemFree( mt_RtfResettable.h_content);
		memset( &mt_RtfResettable, NULL, sizeof( ResettableRtfInfo));
	}

	if (f_RtfContextSet)
		ef_FreeRtfContext( &t);
	if (pc)
		free( pc);

	return (short) us_err;
} //es_ReplaceMimeWithRtf(


/** xs_VetPluginReadiness( ***
Determines (or assists the subform in determining) whether the calling plug-in 
subform is compatible with this DLL component of the plug-in.

--- parameter & return ----
pc_PLGN_VER_SUBFM: address of string stating the version of the calling PGP 
	Plug-In subform
pc_MIN_CPTBL_VER_DLL: Optional. Address of string stating what the subform 
	considers the minimum version of this DLL with which it is compatible. If 
	null, procedure will consider that version to be the subform's own version, 
	meaning that a downlevel DLL is incompatible.
ps_status: Optional Output. Address of variable in which to store the relative 
	state of the subform component's version to this DLL component's version. 
	The output value mirrors what we would expect from a "cmp" function: Result 
	is less than zero if subform < DLL, zero if subform = DLL, greater than 
	zero if subform > DLL.
pc_instldVer: Optional Output, but required if ps_status output is requested. 
	Address of string buffer in which to copy the version of this DLL in case 
	the subform component is interested (as it might well be if the subform 
	component is of later version than the DLL).
ps_unEngaged: Optional Output, but required if ps_status output is requested. 
	Address of VisualBasic flag variable in which to note whether this DLL has 
	been loaded by the Notes client on a session-long basis (i.e. whether it's 
	been loaded yet as a "database driver" extension).
ps_prvIncompatibilityFound: Optional Output, but required if ps_status output 
	is requested. Address of VisualBasic flag variable in which to note whether 
	incompatibility had been found in a previous compatibility check during 
	this DLL session.
pv_hDll: Optional Output, but required if ps_status output is requested. 
	Address of variable in which to store the module handle of this DLL in case 
	the subform wishes to force the unloading of the DLL in order to replace it 
	with a preferred DLL of earlier or later version.
RETURN:
	eus_ERR_INVLD_ARG if any input parameter is found to be invalid
	eus_SUCCESS if this DLL is compatible with the calling subform component
	ei_USER_ABORT it this DLL is incompatible with the calling subform component

--- revision history ------
10/25/02 PR: upgraded plug-in version string to 2.2.2
9/6/02 PR: upgraded plug-in version string to 2.2.1, minor exception-handling 
	improvement

7/2/01 PR
+ Logic enhancment in support of overhaul of plug-in initialization. Included 
  addition of five output parameters following the two input parameters (the 
  first original, the second new). Procedure still supports old versions of the 
  plug-in subform which pass just the one input parameter.
+ updated plug-in version string to 2.2; minor token renaming

4/24/01 PR: updated plug-in version string to 2.1.8
11/16/00 PR: updated plug-in version string to 2.1.7
11/14/00 PR: updated plug-in version string to 2.1.6
10/7/00 PR: created			*/
short xs_VetPluginReadiness( const char  pc_PLGN_VER_SUBFM[], 
								const char  pc_MIN_CPTBL_VER_DLL[], 
								short *const  ps_status, 
								char  pc_instldVer[], 
								short *const  ps_unEngaged, 
								short *const  ps_prvIncompatibilityFound, 
								HMODULE *const  pv_hDll)	{
	//static for speed
	static const char  pc_PLGN_VER_DLL[] = "2.2.2", 
						pc_MIN_CPTBL_VER_SUBFM[] = "2.1.6";

	static BOOL  f_incompatibilityFound;

	int  i_mainRslt, i_rslt, i_err;
	BOOL  f_compatible, f_downwardCptblPossibility = FALSE;

	if (!( pc_PLGN_VER_SUBFM && *pc_PLGN_VER_SUBFM))
		return eus_ERR_INVLD_ARG;

	//if the caller is using the latest version of this function which has 
	//	seven parameters instead of one...
	if (ps_status)	{
		//if any of the other "new" output parameters are obviously invalid...
		if (!( pc_instldVer && ps_unEngaged && ps_prvIncompatibilityFound && 
																	pv_hDll))
			//short-circuit with failure
			return eus_ERR_INVLD_ARG;

		//if the provided minimum-compatible DLL version was provided...
		if (pc_MIN_CPTBL_VER_DLL && *pc_MIN_CPTBL_VER_DLL)	{
			//if that DLL version is obviously invalid, short-circuit with 
			//	failure
			if (i_err = ei_verCmp( pc_MIN_CPTBL_VER_DLL, pc_PLGN_VER_SUBFM, 
																	&i_rslt))
				return (short) i_err;
			if (i_rslt > ei_SAME)
				return eus_ERR_INVLD_ARG;

			//if the subform component supports any downlevel versions of the 
			//	DLL component, take note of that
			if (i_rslt < ei_SAME)
				f_downwardCptblPossibility = TRUE;
		} //if (pc_MIN_CPTBL_VER_DLL &&

		//initialize all the "new" output parameters
		*pv_hDll = (HMODULE *) (*ps_status = *ps_unEngaged = 
							*ps_prvIncompatibilityFound = *pc_instldVer = NULL);
	} //if (ps_status)

	//if this DLL's version is the same as the subform component's...
	if (i_err = ei_verCmp( pc_PLGN_VER_SUBFM, pc_PLGN_VER_DLL, &i_mainRslt))
		return (short) i_err;
	if (i_mainRslt == ei_SAME)	{
		//short-circuit with success
		if (ps_status)	{
			*ps_status = i_mainRslt;
			*ps_unEngaged = ef_LoadedSessionLong ? ms_VB_FALSE : ms_VB_TRUE;
			*ps_prvIncompatibilityFound = f_incompatibilityFound ? 
													ms_VB_TRUE : ms_VB_FALSE;
			*pv_hDll = eh_Instance;
		}
		return eus_SUCCESS;
	} //if (i_rslt == ei_SAME)

	//if this DLL's version is greater than the subform component's...
	if (i_mainRslt < ei_SAME)	{
		//determine whether the subform component is still compatible with this 
		//	DLL
		if (i_err = ei_verCmp( pc_PLGN_VER_SUBFM, pc_MIN_CPTBL_VER_SUBFM, 
																	&i_rslt))
			return (short) i_err;
		f_compatible = i_rslt >= ei_SAME;
	//else if the calling subform supports certain downlevel DLLs...
	}else if (f_downwardCptblPossibility)	{
		//determine whether this DLL is still compatible with the subform 
		//	component
		if (i_err = ei_verCmp( pc_PLGN_VER_DLL, pc_MIN_CPTBL_VER_DLL, &i_rslt))
			return (short) i_err;
		f_compatible = i_rslt >= ei_SAME;
	//else declare the subform component incompatible since we can't know any 
	//	better
	}else
		f_compatible = FALSE;

	//if the caller is using the latest version of this function which has 
	//	seven parameters instead of one, fill out the expected outputs
	if (ps_status)	{
		*ps_status = i_mainRslt;
		strcpy( pc_instldVer, pc_PLGN_VER_DLL);
		*ps_unEngaged = ef_LoadedSessionLong ? ms_VB_FALSE : ms_VB_TRUE;
		*ps_prvIncompatibilityFound = f_incompatibilityFound ? 
													ms_VB_TRUE : ms_VB_FALSE;
		*pv_hDll = eh_Instance;
	} //if (ps_status)

	//if it hasn't been encountered already in a previous call, note whether 
	//	incompatibility has been found during this DLL session
	if (!f_incompatibilityFound)
		f_incompatibilityFound = !f_compatible;

	return f_compatible ? eus_SUCCESS : (short) ei_USER_ABORT;
} //xs_VetPluginReadiness(


/** xs_PgpSecureMessage( ***
PGP encode the specified Notes Mail message (meaning its Notes rich-text field 
and any associated file attachments) according to caller instructions.

--- parameters & return ----
h_NOTE: handle to the note to undergo PGP encoding
f_sign: flag telling whether the message should be PGP signed
f_encrypt: flag telling whether the message should be PGP encrypted
pc_recipList: Optional. Pointer to a string of user addresses, delimited by 
	semicolon, to be used in the encoding process. The "addresses" will be 
	fully resolved as needed against the Domino Directory. Required only if PGP 
	encryption is specified.
f_LOOKUP_EXHAUST: Optional. Flag telling whether the addresses provided will be 
	checked exhaustively for ambiguity against the Domino Directory 
	configuration currently employed by the caller. Required only if PGP 
	encryption is specified.
pc_LookupHow: Optional. Pointer to the name of the server holding the Domino 
	Directory to be used potentially in the resolution process. Required only 
	if PGP encryption is specified.
i_ATTCHMTS: Optional. The number of attachments included in the message. 
	Ignored if PGP encryption is not specified.
pc_ITMNM_RTF: pointer to a string telling the name of the rich-text field to be 
	used in the encoding
pt_RtfResettable: Optional Input & Output. Address of the information structure 
	in later resetting the message's rich-text field to the used encoded 
	version to be created by this procedure. The need for this functionality is 
	driven by peculiarities of the "Send and File" action button in the R5 
	version of the Notes Mail template and of the R5 client's MIME rendering 
	during mailing. The h_note property of the information structure is used as 
	a safety check on the execution of this functionality. In the context of 
	this procedure, it is used to ensure the caller is aware of what it's doing 
	in using this parameter by checking that its value equals that specified in 
	the primary h_NOTE parameter described above. The t_RtfContext property of 
	the structure will be filled upon successful completion of this procedure. 
	If parameter is null, the procedure understands that the rich-text field 
	reset functionality has not been engaged by the caller.
RETURN: eus_SUCCESS if no error occured; the Notes API error code otherwise

--- revision history -------
10/30/02 PR: logic enhancement to support gracefully user-cancelation during 
	key-server searches

9/6/02 PR: listing format adjustment, token renaming, minor documentation 
	adjustment

3/20/00 PR
+ support of rich-text reset functionality, part of changes needed for Notes 
  Mail R5-template compatibility
+ enhancement to delete the PGP-encoded rich-text attachment if error is 
  encountered after the attachment has been accomplished
+ completed standard documentation

9/12/99 PR: logic adjustments to support enhancement of better emulation of 
	Lotus Notes' name-lookup behavior, including removal of recipient-count 
	parameter

1/19/99 PR: created			*/
//					static FILE * Log;
STATUS xs_PgpSecureMessage( NOTEHANDLE  h_NOTE, 
							BOOL  f_sign, 
							BOOL  f_encrypt, 
							char  pc_recipList[], 
							const BOOL  f_LOOKUP_EXHAUST, 
							char  pc_LookupHow[], 
							const int  i_ATTCHMTS, 
							char  pc_ITMNM_RTF[], 
							ResettableRtfInfo *const  pt_RtfResettable)	{
	const BOOL  f_RESETTABLE = pt_RtfResettable && pt_RtfResettable->h_note;

	RtfCursor  t_crsr;
	RtfContext  t_rtCtx, t_rtCtxUtil;
	char * * ppc_recips = NULL, * pc_outp = NULL;
	UINT  ui_recips;
	PgpEncodeContext  t_encdCtx;
	BOOL  f_PgpEncdCtxSetUp = FALSE, f_RtfAttchPgraph = FALSE, f_SyncUnkKeys, 
			f_ambig, f_fail = FALSE;
	BLOCKID  bid_itmAttch;
	STATUS  us_err;
	int  i_err = NULL;

	_ASSERTE( h_NOTE && (f_sign || f_encrypt) && (f_encrypt ? (BOOL) 
									pc_recipList : TRUE) && pc_ITMNM_RTF && 
									(f_RESETTABLE ? pt_RtfResettable->h_note == 
									h_NOTE : TRUE));
/*if (!( Log = fopen( "nPGPLog.txt", "wtc")))
	return !eus_SUCCESS;
if (setvbuf( Log, NULL, _IONBF, NULL) != eus_SUCCESS)
	return !eus_SUCCESS;
fprintf( Log, "Started message log\n");
fclose( Log);
Log = fopen( "nPGPLog.txt", "atc");
setvbuf( Log, NULL, _IONBF, NULL);
*/
	//initialize resources associated with the rich-text handling we're going 
	//	to do
	if (us_err = eus_InitializeRtfContext( h_NOTE, pc_ITMNM_RTF, &t_crsr, 
																	&t_rtCtx))
		return us_err;

	//initialize the PGP context for encoding content
	if (i_err = ei_SetUpPgpEncodeContext( f_sign = !!f_sign, f_encrypt = 
									!!f_encrypt, &t_encdCtx, &f_SyncUnkKeys))
		goto errJump;
	f_PgpEncdCtxSetUp = TRUE;

	//if message is to be encrypted...
	if (f_encrypt)	{
		//as much as possible, resolve recipient addressses and create a 
		//	character-string array of the addresses, the format expected by the 
		//	PGP infrastructure
		if (us_err = us_CreateRecipientsArray( pc_recipList, 
											!!f_LOOKUP_EXHAUST, pc_LookupHow, 
											&ppc_recips, &ui_recips, &f_ambig))
			goto errJump;

		//match as much as possible the recipient addresses with PGP keys, 
		//	wherever those keys may reside, be the addresses Notes names or 
		//	SMTP names
		if ((i_err = i_MatchPgpKeys( f_SyncUnkKeys, ui_recips, ppc_recips, 
										&t_encdCtx)) && i_err != ei_USER_ABORT)
			goto errJump;

		//obtain the final list of encryption keys for use in the ensuing 
		//	encoding, using the standard PGP client infrastructure (e.g. the 
		//	encryption-keys dialog)
		if (i_err = ei_ResolveEncryptionKeys( ppc_recips, ui_recips, f_ambig, 
													f_SyncUnkKeys, &t_encdCtx))
			goto errJump;
	} //if (f_encrypt)

	//PGP-encode all the textual content of the rich-text field into a buffer
	if (i_err = i_PgpEncodeRtfText( &t_encdCtx, &t_rtCtx, &pc_outp))
		goto errJump;

	//if necessary, PGP Armor each attachment in the message
	if (i_ATTCHMTS && f_encrypt)
		if (i_err = i_PgpEncodeAttachments( &t_encdCtx, h_NOTE, t_crsr, 
																	&t_rtCtx))
			goto errJump;

	//all manipulation done, PGP-encode the rich-text field as a PGP-Armored 
	//	file and attach it to the note and reference it in an ending, hidden 
	//	paragraph
	f_RtfAttchPgraph = TRUE;
	if (i_err = i_PgpEncodeRtfAsAttachment( &t_encdCtx, h_NOTE, &t_rtCtx, 
												ept_InitUtilityRtfContext( 
												&t_rtCtxUtil), &bid_itmAttch))
		goto errJump;

	//completely replace the original rich-text field with the PGP-encoded 
	//	textual content, in Courier 10 font
	if (us_err = eus_ReplaceRtfWithStandardText( pc_outp, TRUE, &t_rtCtx))
		goto errJump;

	//append the special ending hidden-paragraph item to the new virtual 
	//	rich-text field
	if (us_err = eus_AppendItemsToRtf( &t_rtCtxUtil, &t_rtCtx))
		goto errJump;

	//Commit the updated rich-text field to the note, replacing the current 
	//	actual rich-text content
	if (us_err = eus_CommitChangedRtf( h_NOTE, pc_ITMNM_RTF, &t_rtCtx))
		ef_UnappendRtfItems( &t_rtCtxUtil, &t_rtCtx);

errJump:
	if (pc_outp)
		free( pc_outp);

	//if an error occurred and context for PGP's hidden rich-text paragraph has 
	//	been initialized...
	if ((i_err || us_err) && f_RtfAttchPgraph)	{
		//if the encoded version of the rich-text field has been encapsulated 
		//	as an attachment to the note, remove that attachment
		if (us_err)
			f_fail = NSFNoteDetachFile( h_NOTE, bid_itmAttch);

		//free any resources allocated to PGP's hidden rich-text paragraph
		ef_FreeRtfContext( &t_rtCtxUtil);
	} //if ((i_err || us_err) && f_RtfAttchPgraph)

	if (f_PgpEncdCtxSetUp)
		f_fail = !ef_FreePgpEncodeContext( &t_encdCtx);
	if (ppc_recips)
		e_FreeStringArray( &ppc_recips, ui_recips);
	_ASSERTE( !f_fail);

	//if caller wants a copy of the encoded version of the rich-text field 
	//	available for a potential reset of the field later, copy the version 
	//	content into memory
	if (f_RESETTABLE && !(us_err || i_err))
		us_err = eus_CopyRtfIntoBuffer( &t_rtCtx, &pt_RtfResettable->h_content, 
											&pt_RtfResettable->ul_len, NULL);

	//free all resources associated with the encoded rich-text field 
	//	constructed by this procedure
	ef_FreeRtfContext( &t_rtCtx);

	return us_err + (short) i_err;
} //xs_PgpSecureMessage(


/** i_PgpEncodeAttachments( ***
PGP encode the attachments referenced in a given rich-text field, and customize 
the associated attachment hotspots to give us control over the decoding 
process, at least when invoked from the Notes client.

--- parameters & return ----
pt_EncodeContext: Input & Output. Address of the PGP information structure 
	being used in the current round of encoding. The opaque structure may be 
	updated if user adjusts presented defaults.
h_NOTE: handle to the note to which the encoded rich-text field will be 
	attached as an attachment
t_cursor: rich-text cursor set to where the search should start for attachemnt 
	hotspots whose parent attachments should be encoded
pt_rtCtx: Input & Output. Address of the rich-text context structure describing 
	the rich-text field to be searched and updated with PGP-encoded 
	attachment-action hotspots. Structure may undergo change even if an error 
	occurs.
RETURN: eus_SUCCESS if successful; the Notes API or PGP SDK error code if 
	unsuccessful

--- suggested enhancment ---
3/20/00 PR: any way to improve safety by providing roll-back if something goes 
	wrong?

--- revision history -------
9/6/02 PR
+ logic enhancement to support high-ASCII (international) characters in 
  attachment handling
+ token renaming, listing format adjustment, minor buffer tightening

9/16/00 PR: completed enhancement associated with improved handling of 
  PGP-encoded attachments

8/9/00 PR
+ got rid of hack for attaching encoded files without Notes trying to sign them 
  when sending the message in MIME format
+ started enhancement associated with improved handling of PGP-encoded 
  attachments

5/30/00 PR
+ extension to wrap attachments within an action hotspot that allows for custom 
  handling
+ standard documentation, improved safety handling

3/20/00 PR: enhancement to prevent attachments already PGP encoded from being 
	encoded again; associated documentation adjustments, logic shortening
9/12/99 PR: documentation adjustment
1/30/99 PR: created			*/
static int i_PgpEncodeAttachments( PgpEncodeContext *const  pt_EncodeContext, 
									NOTEHANDLE  h_NOTE, 
									RtfCursor  t_crsr, 
									RtfContext *const  pt_rtCtx)	{
	const WORD  us_LEN_ITMNM_STANDARD_ATTACH = (WORD) strlen( 
													epc_ITEMNM_STANDARD_ATTACH);

	static char  pc_dirNmTemp[ MAXPATH - 2];
	char * pc_objNm, * pc_fileNm;
	RtfSpan  t_rtSpan;
	UINT  ui_lenFileNm;
	BLOCKID  bid;
	STATUS  us_err;
	int  i_err, i_error;

	_ASSERTE( pt_EncodeContext && pt_EncodeContext->f_Encrypt && h_NOTE && 
											t_crsr.puc_location && 
											t_crsr.us_recLength && 
											pt_rtCtx && NSFItemIsPresent( 
											h_NOTE, epc_ITEMNM_STANDARD_ATTACH, 
											us_LEN_ITMNM_STANDARD_ATTACH));

	//if this is the first time through, initialize the path to the user's 
	//	temporary directory
	if (!*pc_dirNmTemp)
		if (!epc_getTempDirNm( pc_dirNmTemp, MAXPATH - 2))
			return !eus_SUCCESS;

	//advance the rich-text cursor to the next file-attachment hotspot, which 
	//	may be the current CD record 
	if (!ef_CursorToAttachmentHotspot( &t_crsr, pt_rtCtx, FALSE, &t_rtSpan, 
														&pc_objNm, &pc_fileNm))
		return !eus_SUCCESS;

	//loop until no further file-attachment hotspots exist in the rich-text 
	//	field
	while (t_crsr.puc_location)	{
		//if the attachment hasn't already been PGP encoded, get the attachment 
		//	item that corresponds to the attachment referenced in the hotspot
		bid.pool = NULL;
		if (!( stricmp( pc_fileNm + (ui_lenFileNm = strlen( pc_fileNm)) - 
										mui_LEN_ARMOR_EXT_ASC, 
										mpc_PGP_ARMOR_EXT_ASC) == ei_SAME || 
										stricmp( pc_fileNm + ui_lenFileNm - 
										mui_LEN_ARMOR_EXT_PGP, 
										mpc_PGP_ARMOR_EXT_PGP) == ei_SAME))
			if (us_err = eus_getAttachmentInfo( h_NOTE, 
										epc_ITEMNM_STANDARD_ATTACH, pc_objNm, 
										NULL, NULL, NULL, NULL, &bid, NULL))
				return us_err;

		//if the attachment should be encoded...
		if (bid.pool)	{
			char  pc[ MAXPATH], * pc_fileNmNew, pc_extFileNmTmp[ MAXPATH], 
					pc_extFileNmNewOs[ MAXPATH], pc_extFileNmNew[ MAXPATH];

			//if the ultimate filename would exceed the maximum length allowed, 
			//	short-circuit with failure
			if (strlen( pc_fileNm) + mui_LEN_ARMOR_EXT_ASC + 1 > 
														mi_MAXLEN_ATTCH_OBJNM)
				return !eus_SUCCESS;

			//Extract the attachment to a file in the user's temporary 
			//	directory. Need to use the real file name so PGP will default 
			//	to this name when the file is ultimately decoded from the file 
			//	system by the user.
			if (us_err = NSFNoteExtractFile( h_NOTE, bid, strcat( strcpy( 
												pc_extFileNmTmp, pc_dirNmTemp), 
												pc_fileNm), NULL))	{
				_ASSERTE( !(ERR( us_err) == ERR_NOEXTRACT_ENCRYPTED || ERR( 
											us_err) == ERR_NOTE_BADATTSIGN));
				return us_err;
			}

			//translate our filename to the native character set
			if (!OSTranslate( OS_TRANSLATE_LMBCS_TO_NATIVE, pc_extFileNmTmp, 
											(WORD) strlen( pc_extFileNmTmp), 
											pc, sizeof( pc)))
				return !eus_SUCCESS;

			//PGP encode the file
			i_err = ei_PgpEncodeFile( pt_EncodeContext, pc, strcat( strcpy( 
														pc_extFileNmNewOs, pc), 
														mpc_PGP_ARMOR_EXT_ASC));

			//delete the original file copied to the temporary directory 
			if (i_error = ei_PgpWipeFile( pc))	{
				int i_err = remove( pc);
				_ASSERTE( i_err == eus_SUCCESS);
			}

			//if any error occurred during PGP encoding or deletion, 
			//	short-circuit with failure
			if (i_err)
				return i_err;
			if (i_error)
				return i_error;

			//write the encoded file back into the mail database as a file 
			//	object, replacing the file's unencoded counterpart
			if (us_err = NSFNoteDetachFile( h_NOTE, bid))
				return us_err;
//PGP SHORTCUT: not checking on unique attachment name!!
			strcat( strcpy( pc_extFileNmNew, pc_extFileNmTmp), 
														mpc_PGP_ARMOR_EXT_ASC);
			if (us_err = eus_AttachFileAsObject( pc_extFileNmNewOs, NULL, 
									h_NOTE, epc_ITEMNM_STANDARD_ATTACH, 
									pc_fileNmNew = strrchr( pc_extFileNmNew, 
									ec_PATH_SPECIFIER) + 1, HOST_MSDOS))
				return us_err;

			//Delete the encoded file written to the temporary directory. 
			//	Ignore any failure of this operation.
			if (ei_PgpWipeFile( pc_extFileNmNewOs) != eus_SUCCESS)	{
				i_error = remove( pc_extFileNmNewOs);
				_ASSERTE( i_error == eus_SUCCESS);
			}

			//wrap a PGP action hotspot around the file-attachment hotspot
			if (us_err = us_ActionWrapAttachment( &t_rtSpan, pc_fileNmNew, 
												NULL, h_NOTE, TRUE, pt_rtCtx))
				return us_err;
		} //if (bid.pool)

		//advance the rich-text cursor to the next file-attachment hotspot
		if (!ef_CursorToAttachmentHotspot( &t_crsr, pt_rtCtx, FALSE, &t_rtSpan, 
														&pc_objNm, &pc_fileNm))
			return !eus_SUCCESS;
	} //while (t_crsr.puc_location)

	return eus_SUCCESS;
} //i_PgpEncodeAttachments(


/** us_ActionWrapAttachment( ***
Wraps the specified Notes rich-text file-attachment hotspot with a custom 
action hotspot that enables custom PGP handling.

--- parameters & return ------
pt_RT_SPAN: Optional. Address of rich-text span structure giving the 
	cordinates of a file-attachment hotspot to be wrapped. Null address 
	signifies that a new generic file-attachment hotspot should be constructed 
	and wrapped at the end of the rich-text field (used for below-the-line 
	attachments that don't have asssociated hotspots in the rich-text).
pc_FILENM: address of string giving the filename of the attachment being wrapped
pc_OBJNM: Optional. Address of string giving the object name of the attachment 
	being wrapped. If null, the object name will be set in the default manner 
	used by our module handling rich-text processing.
h_NOTE: Required only if pc_OBJNM provided. Handle to the note being operated 
	upon.
f_RESET: If rich-text span input is provided, flag tells whether filenames 
	within the file-attachment hotspot should be reset to reflect the filename 
	given in pc_FILENM. If no rich-text span input is provided, flag tells 
	whether a new paragraph should be started before adding the attachment 
	hotspot.
pt_rtCtx: Input & Output. Address of the rich-text context structure 
	describing the rich-text field to be searched and updated with PGP-encoded 
	attachment-action hotspots. Structure may undergo change even if an error 
	occurs.
RETURN: eus_SUCCESS if no error occured; the Notes API error code otherwise

--- suggested enhancement ----
9/6/02 PR: get rid of PGP development shortcut

--- revision history ---------
9/6/02 PR: created			*/
static STATUS us_ActionWrapAttachment( const RtfSpan *const  pt_RT_SPAN, 
										const char  pc_FILENM[], 
										const char  pc_OBJNM[], 
										NOTEHANDLE  h_NOTE, 
										const BOOL  f_RESET, 
										RtfContext *const  pt_rtCtx)	{
	static const char *const  ppc_FORMULA[] = { "%s%s%s%s%s%s%s", 
								"@If( !@IsNumber( xn_hNote);\n\t@Command( "
								"[AttachmentProperties]) & @Return(\"\");"
								"\n\t\"\");\n\nv_result := @DbCommand( "
								"\"PGP\"; \"", "\" : \"", "\"; xn_hNote);"
								"\n\n@If( @IsError( v_result);\n\t@Prompt( "
								"[OK]; \"", "\"; \"An error occurred in "
								"handling the PGP encrypted attachment, \" + "
								"@Char(10) + \"translated by Notes as \\\"\" + "
								"@Text( v_result) + \"\\\"\" + @Repeat( "
								"@Char(10); 2) + \"The usual Notes attachment "
								"commands are still available.\") & @Command( "
								"[AttachmentProperties]);\n\t\"\")"}, 
						pc_ENCRYPTED_TAG[] = "<PGP_encrypted> ";

	static const BYTE * puc_gnericAttchBmp;
	static WORD  us_lenGnericAttchBmp;

	char  pc[ 0x400];
	RtfCursor  t_crsr, * pt_crsr;
	RtfContext  t_rtUtil;
	STATUS  us_err;
	BOOL  f_fail = FALSE;

	_ASSERTE( (pt_RT_SPAN ? pt_RT_SPAN->t_crsrBgin.puc_location && 
									pt_RT_SPAN->t_crsrBgin.us_recLength : 
									TRUE) && pt_rtCtx);

	pt_crsr = pt_RT_SPAN ? &pt_RT_SPAN->t_crsrBgin : memset( &t_crsr, NULL, 
															sizeof( RtfCursor));

	//if this is a below-the-line attachment...
	if (!pt_RT_SPAN)
		//initialize a utility rich-text context into which we will construct 
		//	what we need
		ept_InitUtilityRtfContext( &t_rtUtil);
	//else if the attachment hotspot is immediately preceded by the opening of 
	//	an action hotspot, we'll assume that it has already been wrapped with 
	//	the PGP-handling action hotspot (e.g. recursive decoding), so 
	//	short-circuit with success
	else	{
		BOOL  f;
		if (us_err = eus_ActionHotspotJustBefore( pt_crsr, pt_rtCtx, &f))
			return us_err;
		if (f)
			return eus_SUCCESS;
	} //if (!pt_RT_SPAN)

	//if caller wants an attachment hotspot constructed as well and has further 
	//	specified that a new paragraph should first be added, add a visible 
	//	paragraph
	if (!pt_RT_SPAN && f_RESET)
		if (us_err = eus_AppendRtParagraph( 
0xCBA8, 
										&mt_RT_PGRPH_NORMAL, &t_rtUtil, NULL))
			goto errJump;

	//write head & guts of wrapper action hotspot that will allow custom PGP 
	//	attachment handling
	sprintf( pc, ppc_FORMULA[0], ppc_FORMULA[1], pc_FILENM, ppc_FORMULA[2], 
													pc_FILENM, ppc_FORMULA[3], 
													epc_APPNM, ppc_FORMULA[4]);
	if (us_err = eus_StartFormulaActionHotspot( pc, *pt_crsr, pt_RT_SPAN ? 
														pt_rtCtx : &t_rtUtil))
		goto errJump;

	//if caller's having us construct the attachment hotspot as well...
	if (!pt_RT_SPAN)	{
		char  pc_objNm[ mi_MAXLEN_ATTCH_OBJNM] = {NULL};

		//if not done yet, load the resource containing the generic attachment 
		//	bitmap CD records we will need
		if (!puc_gnericAttchBmp)	{
			const HRSRC h_RSRC = FindResource( eh_Instance, MAKEINTRESOURCE( 
										mi_IDR_BLANK_ATTACH_BMP), RT_RCDATA);

			HGLOBAL  h;

			if (f_fail = !h_RSRC)
				goto errJump;
			if (f_fail = !( us_lenGnericAttchBmp = (WORD) SizeofResource( 
														eh_Instance, h_RSRC)))
				goto errJump;
			if (f_fail = !( h = LoadResource( eh_Instance, h_RSRC)))
				goto errJump;
			if (f_fail = !(puc_gnericAttchBmp = LockResource( h)))
				goto errJump;
		} //if (!puc_gnericAttchBmp)

		//Insert attachment hotspot at the end of the rich-text field, and, 
		//	like the Domoino SMTP mail-receiver does, insert the name of the 
		//	file in a string at the end of the hotspot.
		strcat( strcpy( pc, " - "), pc_FILENM);
		if (pc_OBJNM)
			strcpy( pc_objNm, pc_OBJNM);
		if (us_err = eus_InsertAttachHotspot( pt_crsr, pc_FILENM, pc_objNm, 
													h_NOTE, puc_gnericAttchBmp, 
													us_lenGnericAttchBmp, pc, 
													NULL, &t_rtUtil))
			goto errJump;
	//else if caller wants, reset the filenames held by the hotspot and add a 
	//	tag indicating the attachment is PGP encoded (because the hotspot 
	//	picture is not being changed and probably provides the non-PGP-encoded 
	//	filename)
	}else if (f_RESET)	{
//PGP development shortcut: not checking for unique filenames!
		if (us_err = eus_ResetAttachHotspotNames( pc_FILENM, pc_OBJNM && 
												*pc_OBJNM ? pc_OBJNM : 
												pc_FILENM, *pt_crsr, pt_rtCtx))
			return us_err;
		if (us_err = eus_InsertTextAtHotspotEnd( pc_ENCRYPTED_TAG, *pt_crsr, 
																	pt_rtCtx))
			return us_err;
	} //if (f_RESET)

	//write tail of wrapper action hotspot
	if (pt_RT_SPAN)	{
		t_crsr = pt_RT_SPAN->t_crsrEnd;
		pt_crsr = &t_crsr;
		if (us_err = eus_CursorAdvance( pt_crsr, pt_rtCtx))
			return us_err;
		if (us_err = eus_InsertHotspotEnd( *pt_crsr, pt_rtCtx))
			return us_err;
	}else if (us_err = eus_InsertHotspotEnd( *pt_crsr, &t_rtUtil))
		goto errJump;

	//if caller's having us construct the attachment hotspot as well, 
	//	incorporate that construction
	if (!pt_RT_SPAN)
		us_err = eus_AppendItemsToRtf( &t_rtUtil, pt_rtCtx);

errJump:
	if (us_err + f_fail && !pt_RT_SPAN)
		ef_FreeRtfContext( &t_rtUtil);

	return us_err + f_fail;
} //us_ActionWrapAttachment(


/** es_SpecialDecryptAttachment( ***
Carries out for a particular attachment the plug-in's special treatment of 
PGP-encrypted attachments.

--- suggested enhancement ---
10/7/00 PR: enable the <Esc> key for canceling the dialog!

--- side effect -------------
If the user executes the launch, not the extract, choice from the choice dialog 
shown, the extended filename of the launched file will be added to the 
associated text-list item on the note, thus notifying front-end code that the 
file should be deleted when the parent message is closed.

--- parameters & return -----
ppv_ARGV: Pointer to the array of arguments provided by the associated PGP 
	@DbCommand used to call this DLL. First element provides the object name 
	and the original filename of the PGP-encrypted attachment to be handled. 
	Second element provides the handle of the note currently being processed 
	and containing the target attachment.
RETURN:
	eus_SUCCESS if no errors occurred
	eus_ERR_INVLD_ARG if any input parameter is obviously invalid
	!eus_SUCCESS if a Windows API error occurs
	the Notes API error code otherwise

--- revision history -------
?/?/00 PR
+ logic enhancement to support high-ASCII (international) characters in 
  attachment handling
+ minor exception-handling adjustment, listing format adjustment, minor 
  documentation adjustment

10/7/00 PR: bug fixes, adjusted "launched" item's name
9/16/00 PR: completed code
8/9/00 PR: created			*/
short es_SpecialDecryptAttachment( void * ppv_ARGV[])	{
	static const char * ppc_OVR_WRT[] = { "%s%s%s", "The file \'", 
												"\' already exists.\n\nDo you "
												"wish to overwrite it?"};

	//static so last filename is maintained between calls
	static char  pc_extFileNm[ MAXPATH];

	BYTE * puc_names;
	double  dbl;
	NOTEHANDLE  h;
	char  * pc_objNm, * pc_fileNm = NULL, * pc, 
			* pc_outFileNmOs = NULL, pc_fileNmOs[ 0x60], 
			pc_LaunchExtFileNm[ MAXPATH] = {NULL}, pc_buf[ MAXPATH + 40];
	int  i_choice;
	DlgIoExtract  t_in = { NULL, pc_extFileNm, &i_choice};
	BLOCKID  bid;
	BOOL  f_extPgp;
	size_t  ui_len;
	STATUS  us_err;
	int  i_err;

	//validate environment, short-circuiting if we fail
	if (!(*(WORD *) (puc_names = ppv_ARGV[0]) == TYPE_TEXT_LIST && 
									ef_GetNumberListEntry( ppv_ARGV[1], TRUE, 
									0, &dbl) && (h = (NOTEHANDLE) (long) dbl)))
		return eus_ERR_INVLD_ARG;
	if (us_err = eus_CreateTextListEntryCopy( 0, puc_names, TRUE, &pc_objNm))
		return us_err;
	if (us_err = eus_CreateTextListEntryCopy( 1, puc_names, TRUE, &pc_fileNm))
		goto errJump;
	if (!( *pc_objNm && *pc_fileNm))	{
		us_err = eus_ERR_INVLD_ARG;
		goto errJump;
	}

	//after completing its inputs, show the extract-choice dialog
	if (!OSTranslate( OS_TRANSLATE_LMBCS_TO_NATIVE, pc_fileNm, (WORD) strlen( 
													pc_fileNm), pc_fileNmOs, 
													sizeof( pc_fileNmOs)))	{
		us_err = !eus_SUCCESS;
		goto errJump;
	}
	t_in.pc_fileNm = pc_fileNmOs;
	if ((i_err = DialogBoxParam( eh_Instance, MAKEINTRESOURCE( 
									mi_IDD_EXTRACT), eh_mainWnd, 
									f_DlgProcExtract, (long) &t_in)) <= 0)	{
		us_err = !eus_SUCCESS;
		goto errJump;
	}

	//if choice was to cancel, short-circuit with success
	if (i_err == mi_CANCEL)	{
		i_err = eus_SUCCESS;
		goto errJump;
	}

	//if our code intentionally threw an error...
	if (i_err != mi_SUCCESS)	{
		//short-circuit with failure
		us_err = !eus_SUCCESS;
		goto errJump;
	//else clear the error variable for future use
	}else
		i_err = eus_SUCCESS;

	//if the user chose to extract, not launch, the file...
	if (us_err = eus_getAttachmentInfo( h, NULL, pc_objNm, NULL, NULL, NULL, 
															NULL, &bid, NULL))
		goto errJump;
	if ((i_choice & mi_IDC_RD_EXTR) == mi_IDC_RD_EXTR)	{
		//ensure that a PGP extension is included for the extract location, as 
		//	the file will be extracted initially in its encoded state
		if (!( (f_extPgp = stricmp( (pc = pc_extFileNm) + (ui_len = 
								strlen( pc)) - mui_LEN_ARMOR_EXT_PGP, 
								mpc_PGP_ARMOR_EXT_PGP) == ei_SAME) || 
								stricmp( pc + ui_len - mui_LEN_ARMOR_EXT_ASC, 
								mpc_PGP_ARMOR_EXT_ASC) == ei_SAME))
			strcat( pc_extFileNm, mpc_PGP_ARMOR_EXT_ASC);
	//else the choice must be to decode & launch the file, so set the extract 
	//	location into the user's "temp" directory
	}else	{
		if (!epc_getTempDirNm( pc_LaunchExtFileNm, MAXPATH))
			goto errJump;
		strcat( pc = pc_LaunchExtFileNm, pc_fileNmOs);
	} //if ((i_choice & mi_IDC_RD_EXTR) ==

	//if a file of the same name already exists in the location just 
	//	determined, inform user, allowing her to abort
	if (ei_FileAccess( pc, ei_FILE_ACCS_EXIST, TRUE) == eus_SUCCESS)	{
		sprintf( pc_buf, ppc_OVR_WRT[0], ppc_OVR_WRT[1], pc, ppc_OVR_WRT[2]);
		if (IDNO == MessageBox( eh_mainWnd, pc_buf, epc_APPNM, MB_YESNO | 
															MB_ICONQUESTION))
			goto errJump;
	}

	//extract the file in its encoded state to the location just determined
	if (!OSTranslate( OS_TRANSLATE_NATIVE_TO_LMBCS, pc, (WORD) strlen( pc), 
												pc_buf, sizeof( pc_buf)))	{
		us_err = !eus_SUCCESS;
		goto errJump;
	}
	if (us_err = NSFNoteExtractFile( h, bid, pc_buf, NULL))			
		goto errJump;

	//if we need to do more than just extract the file in its encoded state...
	if (i_choice != mi_IDC_RD_EXTR)	{
		int  i_error;

		//decode the file just extracted
		i_err = ei_PgpDecodeFile( pc, &pc_outFileNmOs);

		//safe erase (wipe) the encoded file
		if (i_error = ei_PgpWipeFile( pc))	{
			int i_err = remove( pc);
			_ASSERTE( i_err == eus_SUCCESS);
		}

		//if any problem occurred during the decode or erase, short-circuit 
		//	with failure
		if (i_err)	{
			if (i_err == ei_USER_ABORT)
				i_err = eus_SUCCESS;
			goto errJump;
		}
		if (i_error)	{
			i_err = i_error;
			goto errJump;
		}

		//if choice was to just decrypt and the decoded file does not have the 
		//	same name as what the user specified in the dialog... (PGP includes 
		//	the original decoded filename when encoding *.pgp files and inists 
		//	on using it when decoding. Also, the "shared" codebase may adjust 
		//	filenames in another way as well, it seems. In any case, there 
		//	seems to be no foolproof way to specify the output filename.)
		if (i_choice & mi_MASK_DCRYPTD && strnicmp( pc_outFileNmOs, 
											pc_extFileNm, strlen( 
											pc_outFileNmOs)) != ei_SAME)	{
			//produce the extended filename the user asked for
			pc_extFileNm[ ui_len - (f_extPgp ? mui_LEN_ARMOR_EXT_PGP : 
												mui_LEN_ARMOR_EXT_ASC)] = NULL;

			//if a file of the same name already exists in the location...
			if (ei_FileAccess( pc, ei_FILE_ACCS_EXIST, TRUE) == 
															eus_SUCCESS)	{
				//if the user does not wish to overwrite the existing file... 
				sprintf( pc_buf, ppc_OVR_WRT[0], ppc_OVR_WRT[1], pc_extFileNm, 
																ppc_OVR_WRT[2]);
				if (IDNO == MessageBox( eh_mainWnd, pc_buf, epc_APPNM, 
											MB_YESNO | MB_ICONQUESTION))	{
					//safe erase (wipe) the decoded file, then short-circuit
					if (i_err = ei_PgpWipeFile( pc_outFileNmOs))	{
						i_error = remove( pc_outFileNmOs);
						_ASSERTE( i_error == eus_SUCCESS);
					}
					goto errJump;
				//else safe erase the file we're "overwriting"
				}else if (i_err = ei_PgpWipeFile( pc_extFileNm))	{
					int i_error = remove( pc_extFileNm);
					_ASSERTE( i_error == eus_SUCCESS);
					goto errJump;
				} //if (IDNO == MessageBox(
			} //if (ei_FileAccess( pc, ei_FILE_ACCS_EXIST) ==

			//rename the PGP generated file to the filename requested by user
			i_error = rename( pc_outFileNmOs, pc_extFileNm);
			_ASSERTE( i_error == eus_SUCCESS);
		//else if choice was to launch...
		}else if ((i_choice & mi_IDC_RD_LAUNCH) == mi_IDC_RD_LAUNCH)	{
			static char  pc_ITMNM_LAUNCHED[] = "xlt_Launched";

			//note that the decoded file should be deleted once parent 
			//	message closes
			if (!OSTranslate( OS_TRANSLATE_NATIVE_TO_LMBCS, pc_outFileNmOs, 
												(WORD) strlen( pc_outFileNmOs), 
												pc_buf, sizeof( pc_buf)))	{
				us_err = !eus_SUCCESS;
				goto errJump;
			}
			if (us_err = NSFItemAppendTextList( h, pc_ITMNM_LAUNCHED, pc_buf, 
												(WORD) strlen( pc_buf), FALSE))
				goto errJump;

			//if the file fails to launch...
			if ((i_err = (int) ShellExecute( eh_mainWnd, NULL, pc_outFileNmOs, 
										NULL, NULL, SW_SHOWNORMAL)) <= 32)	{
				const char * pc = strrchr( pc_outFileNmOs, ec_PATH_SPECIFIER);

				//inform the user
				sprintf( pc_buf, "%s%s%s", "The PGP decoded file \'", pc ? pc + 
													1 : pc_outFileNmOs, "\'\n"
													"could not be launched.");
				MessageBox( eh_mainWnd, pc_buf, epc_APPNM, MB_OK | 
															MB_ICONEXCLAMATION);

				//safe erase (wipe) the decoded file
				if (ei_PgpWipeFile( pc_outFileNmOs) != eus_SUCCESS)	{
					i_error = remove( pc_outFileNmOs);
					_ASSERTE( i_error == eus_SUCCESS);
				}
			}else
				i_err = NULL;
		} //if (i_choice & mi_MASK_DCRYPTD && strnicmp(
	} //if (i_choice != mi_IDC_RD_EXTR)

errJump:
	if (pc_outFileNmOs)
		e_FreePgpMem( pc_outFileNmOs);
	if (pc_fileNm)
		free( pc_fileNm);
	free( pc_objNm);

	return (short) (us_err ? us_err : i_err);
} //es_SpecialDecryptAttachment(


/** f_DlgProcExtract( ***
Handles Windows messages directed at the "extract" dialog.

--- parameters & return ----
H: handle of dialog window
US: message sent to dialog
ui_W: "word" parameter associated with given message. With WM_INITDIALOG, 
	specifies handle of the control the dialog template has for receipt of 
	start-up focus.
l_L: "long" parameter associated with given message. With WM_INITDIALOG 
	message, specifies dialog-initiation inputs in a DlgIoExtract instance.
RETURN:
	per the Windows API spec: TRUE if message is completely handled by this 
		routine, FALSE otherwise
	EndDialog() returns
		eus_ERR_INVLD_ARG if the necessary initiation inputs were not provided 
			in the kick-off DialogBoxParam() call
		the Windows API error code if available
		!eus_SUCCESS if an unspecified error occurred
		eus_SUCCESS otherwise

--- revision history ------
9/6/02 PR: minor exception-handling adjustment
7/2/01 PR: listing format adjustment
9/16/00 PR: created			*/
static BOOL CALLBACK f_DlgProcExtract( HWND  H, 
										UINT  UI, 
										WPARAM  ui_W, 
										LPARAM  l_L)	{
	//the first time thru, specify some default state for the dialog: the 
	//	radio-button choice should be "Launch," and the "decrypted" should be 
	//	checked
	static DlgStateExtract  t_st = {NULL, mi_IDC_RD_LAUNCH, TRUE};

	DWORD  ul_err = NULL;

	switch (UI)	{
	static DlgIoExtract * pt_io;
	static BOOL  f_fileClear, f_dcryptdChecked_orig;
	static int  i_choice_orig;

	case WM_INITDIALOG:	{
		const HWND h_FOCUS = (HWND) ui_W;

		//if we obvoiously didn't get valid input from the dialog initiator or 
		//	if the focus parameter is invalid, short-circuit with failure
		pt_io = (DlgIoExtract *) l_L;
		if (!( pt_io && pt_io->pc_extFileNm && pt_io->pc_fileNm && 
									*pt_io->pc_fileNm && pt_io->pi_choice))	{
			ul_err = eus_ERR_INVLD_ARG;
			goto errJump;
		}else if (!ui_W)	{
			ul_err = !eus_SUCCESS;
			goto errJump;
		} //if (!( pt_io && pt_io->pc_extFileNm &&

		//set up the dialog and initialize state
		*pt_io->pi_choice = NULL;
		if (ul_err = ul_ExtractDlgInit( H, pt_io, h_FOCUS, &t_st))
			goto errJump;
		f_fileClear = FALSE;
		i_choice_orig = t_st.i_choice;
		f_dcryptdChecked_orig = t_st.f_dcryptdChecked;

		//store the module-scope handle to the font that may have been created 
		//	for the bolding of the static-text filename control so that the 
		//	dynamically allocated resource can be freed when we shut down
		mh_fntFileCaption = t_st.h_fntFileCaption;

		//if needed because no focus has yet been, allow default processing to 
		//	continue (error in function call is harmless, so not handled)
		return GetDlgCtrlID( h_FOCUS) == t_st.i_choice ? FALSE : TRUE;
	} //case WM_INITDIALOG

	case WM_COMMAND:
	switch (LOWORD( ui_W))	{
		char  pc[ MAXPATH];
		size_t  ui;

		case mi_IDC_DECRYPT:
		switch (HIWORD( ui_W))	{
			case BN_CLICKED:	{
				UINT  ui_len;

				//toggle the state variable tracking the current checkbox 
				//	choice and, if appropriate, reset the extended filename 
				//	accordingly
				t_st.f_dcryptdChecked = !t_st.f_dcryptdChecked;
				if (ui_len = GetWindowText( t_st.h_ebxExtFileNm, pc, 
																MAXPATH))	{
					BOOL  f = stricmp( pc + (ui = ui_len - 
											mui_LEN_ARMOR_EXT_ASC), 
											mpc_PGP_ARMOR_EXT_ASC) == ei_SAME;

					if (f && t_st.f_dcryptdChecked)	{
						pc[ ui] = NULL;
						SetWindowText( t_st.h_ebxExtFileNm, pc);
					}else if (!f && !t_st.f_dcryptdChecked && stricmp( 
										pc + ui_len - mui_LEN_ARMOR_EXT_PGP, 
										mpc_PGP_ARMOR_EXT_PGP) != ei_SAME)	{
						strcat( pc, mpc_PGP_ARMOR_EXT_ASC);
						SetWindowText( t_st.h_ebxExtFileNm, pc);
					} //if (f && t_st.f_dcryptdChecked)
				} //if (ui_len = GetWindowText( t_st.h_ebxExtFileNm,
			} //case BN_CLICKED
			} //switch (HIWORD( ui_W))
			break;

		case mi_IDC_BROWSE:
			switch (HIWORD( ui_W))	{
			case BN_CLICKED:	{
				//static for speed
				static const char  pc_FLTR[] = "All Files - *.*\0*.*\0";

				OPENFILENAME  ofn;

				//if the user canceled her navigation to the save-as choice, do 
				//	nothing
				memset( &ofn, NULL, sizeof( OPENFILENAME));
				ofn.lStructSize = sizeof( OPENFILENAME);
				ofn.hwndOwner = H;
				GetWindowText( t_st.h_ebxExtFileNm, pc, MAXPATH);
				ofn.lpstrFilter = pc_FLTR;
				ofn.lpstrFile = pc;
				ofn.nMaxFile = MAXPATH;
				ofn.Flags = OFN_ENABLESIZING | OFN_HIDEREADONLY;
				if (!GetSaveFileName( &ofn))	{
					if (ul_err = CommDlgExtendedError())
						goto errJump;

				//else set the edit-box control accordingly, accounting for 
				//	whether the user has chosen to extract the file in encoded 
				//	or decoded form...
				}else	{
					//if the user wants it extracted encoded and the returned 
					//	filename does not reflect that, tack on a standard PGP 
					//	extension to the filename
					if (!t_st.f_dcryptdChecked && (ui = strlen( pc)) + 
									mui_LEN_ARMOR_EXT_ASC < MAXPATH && 
									stricmp( pc + ui - mui_LEN_ARMOR_EXT_PGP, 
									mpc_PGP_ARMOR_EXT_PGP) != ei_SAME && 
									stricmp( pc + ui - mui_LEN_ARMOR_EXT_ASC, 
									mpc_PGP_ARMOR_EXT_ASC) != ei_SAME)
						strcat( pc, mpc_PGP_ARMOR_EXT_ASC);

					//set the edit-box control
					if (!SetWindowText( t_st.h_ebxExtFileNm, pc))	{
						ul_err = GetLastError();
						goto errJump;
					}
				} //if (!GetSaveFileName( &ofn))
			} //case BN_CLICKED
			} //switch (HIWORD( ui_W))
			break;

		case mi_IDC_EXT_FILENM:
			switch (HIWORD( ui_W))	{
			case EN_CHANGE:	{
				static HWND  h;

				//if there was no text in the edit-box before this change...
				if (f_fileClear)	{
					//enable the OK button
					EnableWindow( h, TRUE);
					f_fileClear = FALSE;
				//else if all the text has been cleared from the edit-box, 
				//	disable the OK button
				}else if (!GetWindowTextLength( (HWND) l_L)) 	{
					if (!( h = GetDlgItem( H, mi_IDOK)))	{
						ul_err = GetLastError();
						goto errJump;
					}
					EnableWindow( h, FALSE);
					f_fileClear = TRUE;
				} //if (f_fileClear)
			} //case EN_CHANGE:
			} //switch (HIWORD( ui_W))
			break;

		case mi_IDOK:
			switch (HIWORD( ui_W))	{
			case BN_CLICKED:
				//output the user's choice
				if ((*pt_io->pi_choice = t_st.i_choice) == mi_IDC_RD_EXTR)	{
					GetWindowText( t_st.h_ebxExtFileNm, pt_io->pc_extFileNm, 
																	MAXPATH);
					*pt_io->pi_choice |= t_st.f_dcryptdChecked ? 
														mi_MASK_DCRYPTD : NULL;
				}
				EndDialog( H, mi_SUCCESS);
				return TRUE;
			} //switch (HIWORD( ui_W))

		case mi_IDCANCEL:
			switch (HIWORD( ui_W))	{
			case BN_CLICKED:
				//reset UI-associated variables back to the state they had when 
				//	dialog was created
				t_st.i_choice = i_choice_orig;
				t_st.f_dcryptdChecked = f_dcryptdChecked_orig;

				//return that user canceled
				EndDialog( H, mi_CANCEL);
				return TRUE;
			} //switch (HIWORD( ui_W))
	} //switch (LOWORD( ui_W))
	break;

	case WM_CLOSE:
	case WM_DESTROY:	{
		//remove the override (subclass) on the radio buttons
		if (t_st.h_rdExtract)	{
			ClearRdOverride( t_st.h_rdExtract);
			t_st.h_rdExtract = NULL;
		}
		if (t_st.h_rdLaunch)	{
			ClearRdOverride( t_st.h_rdLaunch);
			t_st.h_rdLaunch = NULL;
		}
	} //case WM_CLOSE, WM_DESTROY
	} //switch (UI)

	return FALSE;

errJump:
	//remove the override (subclass) on the radio buttons
	if (t_st.h_rdExtract)	{
		ClearRdOverride( t_st.h_rdExtract);
		t_st.h_rdExtract = NULL;
	}
	if (t_st.h_rdLaunch)	{
		ClearRdOverride( t_st.h_rdLaunch);
		t_st.h_rdLaunch = NULL;
	}

	//Return the most descriptive error possible. For GDI errors, only NT will 
	//	provide problem indication.
	if (!ul_err)
		ul_err = mt_Os.dwPlatformId == VER_PLATFORM_WIN32_NT ? GetLastError() : 
																!eus_SUCCESS;

	EndDialog( H, ul_err);
	return TRUE;
} //f_DlgProcExtract(


/** ClearRdOverride( ***
Clears out the override and associated properties set on the specified 
radio-button on the plug-in's special extract dialog.

--- parameter ----------
H: handle to the window (Extract dialog radio-button, presumably) from which 
	the override should be cleared

--- revision history ---
9/16/00 PR: created			*/
static __inline void ClearRdOverride( const HWND  H)	{
	WNDPROC  pr;

	_ASSERTE( H);

	RemoveProp( H, mpc_PROPNM_EXTR_DLG_CHOICE);
	if (pr = RemoveProp( H, mpc_PROPNM_ORIG_WNDPROC))
		SetWindowLong( H, GWL_WNDPROC, (long) pr);
} //ClearRdOverride(


/** e_FreeGdiResources( ***
Free any statically allocated Windows GDI resources.

--- revision history ---
9/16/00 PR: created			*/
void e_FreeGdiResources()	{
	if (mh_fntFileCaption)
		DeleteObject( mh_fntFileCaption);
} //e_FreeGdiResources(


/** ul_ExtractDlgInit( ***
Initializes the extract dialog before display to user.

--- parameters & return ----
H: handle of dialog window
PT: address of structure of dialog-initiation inputs
h_FOCUS: handle of the control the dialog template has for receipt of 
	start-up focus
pt_st: Input & Output. Address of information structure in which dialog state 
	is maintained. If font member (for filename static text control) is null, 
	the appropriate font will be constructed and output.  Handles to the 
	radio-button controls' windows will also be determined and output.
RETURN:
	the Windows API error code if available
	!eus_SUCCESS if an unspecified error occurred
	eus_SUCCESS otherwise

--- revision history ------
7/2/01 PR: listing format adjustment
9/16/00 PR: created			*/
static DWORD ul_ExtractDlgInit( const HWND  H, 
								const DlgIoExtract *const  PT, 
								const HWND  h_FOCUS, 
								DlgStateExtract *const  pt_st)	{
	static BOOL  f_firstTime = TRUE;

	const HWND  h_FILENM = GetDlgItem( H, mi_IDC_FILENM);

	char * pc_extFileNm;
	HWND  h_rdLaunch, h_rdExtract, h;
	WNDPROC  pr;
	size_t  ui, ui_len;
	DWORD  ul_err = NULL;

	_ASSERTE( H && pt_st && PT && PT->pc_extFileNm && PT->pc_fileNm && 
															*PT->pc_fileNm);

	//if the expected filename static-text control isn't found on the dialog, 
	//	short-circuit with failure
	if (!h_FILENM)
		return GetLastError();

	//for safety, nullify the radio-button handle outputs
	pt_st->h_rdExtract = pt_st->h_rdLaunch = NULL;

	//if this is the first time through...
	if (f_firstTime)	{
		HFONT  h_fnt = (HFONT) SendMessage( h_FILENM, WM_GETFONT, NULL, NULL);
		LOGFONT  t_fnt;

		//if we haven't already, gather info on the environment we're running 
		//	(needed for error handling below)
		if (!mt_Os.dwPlatformId)
			if (!GetVersionEx( &mt_Os))
				return GetLastError();

		//Construct the bold font we want to use with the filename static-text 
		//	control. Use the font style already associated with the control as 
		//	a base. If there's not one associated, use the common system style 
		//	instead.
		if (!h_fnt)	{
			const HDC  h_DC = GetDC( h_FILENM);

			if (!( h_DC && (h_fnt = GetCurrentObject( h_DC, OBJ_FONT))))
				goto errJump;
		}
		if (!( GetObject( h_fnt, sizeof( LOGFONT), &t_fnt)))
			goto errJump;
		t_fnt.lfWeight = FW_SEMIBOLD;
		if (!( pt_st->h_fntFileCaption = CreateFontIndirect( &t_fnt)))
			goto errJump;
	} //if (f_firstTime)

	//override (subclass) the message-dispatch procedures of the radio buttons 
	//	on the dialog so we can rect directly to a user's change in activity 
	//	choice
	if (!( h_rdLaunch = GetDlgItem( H, mi_IDC_RD_LAUNCH)))	{
		ul_err = GetLastError();
		goto errJump;
	}
	if (!( h_rdExtract = GetDlgItem( H, mi_IDC_RD_EXTR)))	{
		ul_err = GetLastError();
		goto errJump;
	}
	if (!( pr = (WNDPROC) SetWindowLong( h_rdLaunch, GWL_WNDPROC, 
												(long) WndProcExtractRdo)))	{
		ul_err = GetLastError();
		goto errJump;
	}
	pt_st->h_rdLaunch = h_rdLaunch;
	if (!SetProp( h_rdLaunch, mpc_PROPNM_ORIG_WNDPROC, pr))	{
		ul_err = GetLastError();
		goto errJump;
	}
	if (!SetProp( h_rdLaunch, mpc_PROPNM_EXTR_DLG_CHOICE, &pt_st->i_choice))	{
		ul_err = GetLastError();
		goto errJump;
	}
	if (!( pr = (WNDPROC) SetWindowLong( h_rdExtract, GWL_WNDPROC, 
												(long) WndProcExtractRdo)))	{
		ul_err = GetLastError();
		goto errJump;
	}
	pt_st->h_rdExtract = h_rdExtract;
	if (!SetProp( h_rdExtract, mpc_PROPNM_ORIG_WNDPROC, pr))	{
		ul_err = GetLastError();
		goto errJump;
	}
	if (!SetProp( h_rdExtract, mpc_PROPNM_EXTR_DLG_CHOICE, &pt_st->i_choice))	{
		ul_err = GetLastError();
		goto errJump;
	}

	//set up the dialog with user context: dialog title, target filename, 
	//	default extract path
	if (!SetWindowText( H, epc_APPNM))	{
		ul_err = GetLastError();
		goto errJump;
	}
	if (!SetWindowText( h_FILENM, PT->pc_fileNm))	{
		ul_err = GetLastError();
		goto errJump;
	}

	//emphasize the target filename by bolding it
	SendMessage( h_FILENM, WM_SETFONT, (DWORD) pt_st->h_fntFileCaption, TRUE);

	//if there's no previously entered extract filepath...
	pc_extFileNm = PT->pc_extFileNm;
	if (f_firstTime)	{
		//default it to the Notes data directory
		OSGetDataDirectory( pc_extFileNm);
		strcat( pc_extFileNm, epc_PATH_SPECIFIER);

		//note that we've completed any first-time only stuff done here 
		//	(static local variable)
		f_firstTime = FALSE;
	//else pare the filepath to just the path portion
	}else	{
		char * pc = strrchr( pc_extFileNm, ec_PATH_SPECIFIER);
		if (pc)
			pc[1] = NULL;
		else if (*pc_extFileNm)
			*pc_extFileNm = NULL;
	}//if (f_firstTime)

	//set the extract edit-box to the last path used by the user, if any, or 
	//	to its utter default, adding on the appropriate version of the 
	//	filename
	ui_len = strlen( pc_extFileNm) + strlen( PT->pc_fileNm);
	_ASSERTE( ui_len < MAXPATH);
	strcat( pc_extFileNm, PT->pc_fileNm);
	if (pt_st->f_dcryptdChecked && (stricmp( pc_extFileNm + (ui = ui_len - 
							mui_LEN_ARMOR_EXT_PGP), mpc_PGP_ARMOR_EXT_PGP) == 
							ei_SAME || stricmp( pc_extFileNm + (ui = ui_len - 
							mui_LEN_ARMOR_EXT_ASC), mpc_PGP_ARMOR_EXT_ASC) == 
							ei_SAME))
		pc_extFileNm[ ui] = NULL;
	if (!( h = GetDlgItem( H, mi_IDC_EXT_FILENM)))	{
		ul_err = GetLastError();
		goto errJump;
	}
	if (!SetWindowText( h, pc_extFileNm))
		goto errJump;

	//Set the radio-button choice and the "decrypted" checkbox to the way the 
	//	user last left them, if known, else to their utter defaults. 
	if (!CheckRadioButton( H, mi_IDC_RD_LAUNCH, mi_IDC_RD_EXTR, 
															pt_st->i_choice))
		goto errJump;
	if (!CheckDlgButton( H, mi_IDC_DECRYPT, pt_st->f_dcryptdChecked))
		goto errJump;

	//store handle to the edit-box for ongoing use
	pt_st->h_ebxExtFileNm = h;

	return eus_SUCCESS;

errJump:
	//since an error occurred, back out whatever configuration we've done
	if (pt_st->h_rdExtract)	{
		RemoveProp( h_rdExtract, mpc_PROPNM_EXTR_DLG_CHOICE);
		RemoveProp( h_rdExtract, mpc_PROPNM_ORIG_WNDPROC);
		SetWindowLong( h_rdExtract, GWL_WNDPROC, (long) pr);
		pt_st->h_rdExtract = (HWND) pr = NULL;
	}
	if (pt_st->h_rdLaunch)	{
		const WNDPROC PR = RemoveProp( h_rdLaunch, mpc_PROPNM_ORIG_WNDPROC);

		RemoveProp( h_rdLaunch, mpc_PROPNM_EXTR_DLG_CHOICE);
		if (pr)
			SetWindowLong( h_rdLaunch, GWL_WNDPROC, (long) pr);
		else if (PR)
			SetWindowLong( h_rdLaunch, GWL_WNDPROC, (long) PR);
		pt_st->h_rdLaunch = NULL;
	}
	if (f_firstTime && pt_st->h_fntFileCaption)	{
		DeleteObject( pt_st->h_fntFileCaption);
		pt_st->h_fntFileCaption = NULL;
	}

	//Return the most descriptive error possible. For GDI errors, only NT 
	//	will provide problem indication.
	if (!ul_err)
		ul_err = mt_Os.dwPlatformId == VER_PLATFORM_WIN32_NT ? 
												GetLastError() : !eus_SUCCESS;

	return ul_err;
} //ul_ExtractDlgInit(


/** WndProcExtractRdo( ***
Subclassing window-message procedure for the radio buttons of the Extract 
dialog. Allows direct control of the radio-button "choice change" event, 
which is needed for straightforward enabling/disabling of the controls 
dependant on the "Extract" choice.

--- parameters & return ----
H: handle of the radio button receiving the message
UI: identifier of the message being passed to the procedure
ui_W: the WORD parameter associated with the message
ul_L: the long parameter associated with the message
RETURN: The result of the message. Use of the return varies from message to 
	message.

--- revision history -------
9/16/00 PR: created			*/
static LRESULT CALLBACK WndProcExtractRdo( const HWND  H, 
											const UINT  UI, 
											const WPARAM  ui_W, 
											const LPARAM  ul_L)	{
	static HWND  h_dlg;

	//determine which radio button is receiving the message
	const int  i_ID = GetDlgCtrlID( H);

	WNDPROC  pr;
	DWORD  ul_err;

	//if we failed to determine the recipient radio button, short-circuit 
	//	with failure
	if (!i_ID)	{
		ul_err = GetLastError();
		goto errJump;
	}

	switch (UI)	{
	case BM_SETCHECK:	{
		static HWND  h_extFileNm, h_dcryptd, h_brwse, h_ok;

		HWND  h;

		//if this is a button-check event...
		if (BST_CHECKED == ui_W)	{
			int * pi;

			//if not yet gotten, get handles to the parent dialog and the 
			//	controls associated with the Extract choice
			if (!( h = GetParent( H)))	{
				ul_err = GetLastError();
				goto errJump;
			}
			if (h_dlg != h)	{
				if (!( h_extFileNm = GetDlgItem( h_dlg = h, 
												mi_IDC_EXT_FILENM)))	{
					ul_err = GetLastError();
					goto errJump;
				}
				if (!( h_dcryptd = GetDlgItem( h_dlg, mi_IDC_DECRYPT)))	{
					ul_err = GetLastError();
					goto errJump;
				}
				if (!( h_brwse = GetDlgItem( h_dlg, mi_IDC_BROWSE)))	{
					ul_err = GetLastError();
					goto errJump;
				}
				if (!( h_ok = GetDlgItem( h_dlg, mi_IDOK)))	{
					ul_err = GetLastError();
					goto errJump;
				}
			} //if (h_dlg != h)

			//if the Extract choice has been selected...
			if (mi_IDC_RD_EXTR == i_ID)	{
				//enable the controls associated with the Extract choice
				EnableWindow( h_extFileNm, TRUE);
				EnableWindow( h_dcryptd, TRUE);
				EnableWindow( h_brwse, TRUE);

				//if there's no text in the filename edit-box, disable the OK 
				//	button
				if (!GetWindowTextLength( h_extFileNm))
					EnableWindow( h_ok, FALSE);
			//else...
			}else	{
				//disable the controls associated with the Extract choice
				EnableWindow( h_extFileNm, FALSE);
				EnableWindow( h_dcryptd, FALSE);
				EnableWindow( h_brwse, FALSE);

				//if there's no text in the filename edit-box, enable the OK 
				//	button
				if (!GetWindowTextLength( h_extFileNm))	
					EnableWindow( h_ok, TRUE);
			} //if (mi_IDC_RD_EXTR == i_ID)

			//store the user's latest choice so other code may know of it 
			//	directly
			if (!( pi = GetProp( H, mpc_PROPNM_EXTR_DLG_CHOICE)))
				goto errJump;
			*pi = i_ID;
		} //if (BST_CHECKED == ui_W)
	} //case BM_SETCHECK
	} //switch (UI)

	//pass the current message down the procedure chain
	if (!( pr = GetProp( H, mpc_PROPNM_ORIG_WNDPROC)))
		goto errJump;
	return CallWindowProc( pr, H, UI, ui_W, ul_L);

errJump:
	//Return the most descriptive error possible. For GDI errors, only NT 
	//	will provide problem indication.
	if (!ul_err)
		ul_err = mt_Os.dwPlatformId == VER_PLATFORM_WIN32_NT ? 
												GetLastError() : !eus_SUCCESS;
	if (!h_dlg)
		h_dlg = GetParent( H);
	if (h_dlg)	{
		EndDialog( h_dlg, ul_err);
		h_dlg = NULL;
	}
	return FALSE;		//arbitrary return
} //WndProcExtractRdo(


/** i_PgpEncodeRtfAsAttachment( ***
PGP-encode a copy of the given rich-text field and attach it as an object to 
the specified note. So that the attachment may be hidden from the user, the 
caller is also provided a rich-text hidden paragraph which references the 
attachment and so inhibits its display "below the line" in the Notes 
user-interface.

--- parameters & return ----
pt_EncodeCtx: Input & Output. Address of the PGP information structure being 
	used in the current round of encoding. The opaque structure may be updated 
	if user adjusts presented defaults.
h_NOTE: handle to the note to which the encoded rich-text field will be 
	attached as an attachment
pt_RT_CTX: address of the rich-text context structure describing the rich-text 
	field to be copied, encoded and encapsulated as an attachment
pt_AttchPgrph: Output. Address of the rich-text context structure by which a 
	hidden rich-text paragraph will be constructed which includes a reference 
	to the generated attachment. Caller is responsible for having initialized 
	the structure prior to calling this procedure. If procedure is 
	unsuccessful, it is possible that the structure will have undergone change 
	even so.
pbid_itm: Output. Address of the structure to fill with the Notes BLOCKID 
	associated with the $FILE item that describes the attachment created by 
	this procedure.
RETURN: eus_SUCCESS if successful; the Notes API or PGP SDK error code if 
	unsuccessful

--- revision history -------
9/6/02 PR: token renaming, listing format adjustment, minor logic shortening
5/30/00 PR: documentation adjustment

3/20/00 PR
+ adjustments associated with a change to the eus_CopyRtfIntoBuffer() signature
+ adjusted signature to provide the ultimate attachment name to caller and to 
  allow a PGP error-code return as well as a Notes error-code return
+ documentation adjustment

1/29/99 PR: created			*/
static int i_PgpEncodeRtfAsAttachment( PgpEncodeContext *const  pt_EncodeCtx, 
										NOTEHANDLE  h_NOTE, 
										const RtfContext *const  pt_RT_CTX, 
										RtfContext *const  pt_AttchPgrph, 
										BLOCKID *const  pbid_itm)	{
	DWORD  ul, ul_lenOutp;
	BYTE * puc_inp, * puc_outp = NULL;
	char  pc_objNm[ mi_MAXLEN_ATTCH_OBJNM] = {NULL};
	HANDLE  h;
	int  i_err;
	STATUS  us_err;

	_ASSERTE( pt_EncodeCtx && h_NOTE && pt_RT_CTX && pt_AttchPgrph && pbid_itm);

	//stream the now-complete virtual rich-text content into a buffer
	if (us_err = eus_CopyRtfIntoBuffer( pt_RT_CTX, &h, &ul, &puc_inp))
		return us_err;

	i_err = ei_PgpEncodeBuffer( puc_inp, (long) ul, pt_EncodeCtx, TRUE, 
														&ul_lenOutp, &puc_outp);
	OSUnlockObject( h);
	OSMemFree( h);
	if (i_err)
		return i_err;

	//initialize the rich-text hidden paragraph we will use to store the 
	//	special object
	if (us_err = us_SetupRtfAttachParagraph( h_NOTE, pt_AttchPgrph, pc_objNm))
		goto errJump;

	//Attach the appropriate buffer content (encrypted or not) to the note as a 
	//	file object. The MS-DOS host format is used here at encoding-time to 
	//	ensure that the SMTP MTA will include it if the message goes through 
	//	it. The STREAM format is for some reason ignored and dropped by the MTA 
	//	(in Domino R4 at least).
	if (us_err = eus_AttachBufferAsObject( puc_outp, ul_lenOutp, NULL, h_NOTE, 
													epc_ITEMNM_STANDARD_ATTACH, 
													pc_objNm, HOST_MSDOS))
		goto errJump;

	//retrieve for the caller the item-BLOCKID of the attachment just created
	us_err = eus_getAttachmentInfo( h_NOTE, NULL, pc_objNm, NULL, NULL, NULL, 
														NULL, pbid_itm, NULL);

errJump:
	e_FreePgpMem( puc_outp);

	return us_err;
} //i_PgpEncodeRtfAsAttachment(


/** us_SetupRtfAttachParagraph( ***


--- parameters & return ----

RETURN: eus_SUCCESS if no error occured; the Notes API error code otherwise

--- revision history -------
9/6/02 PR: token renaming, listing format adjustment
1/29/99 PR: created			*/
//DOC!!
static STATUS us_SetupRtfAttachParagraph( NOTEHANDLE  h_NOTE, 
											 RtfContext *const  pt_AttchPgrph, 
											 char *const  pc_objNm)	{
	STATUS  us_err;
//NOTE: might be able the make the module-scope variables in this function 
//	local
	_ASSERTE( pc_objNm && h_NOTE && pt_AttchPgrph);

	if (us_err = eus_AppendRtParagraph( 
0xCBA9, 
													&mt_RT_PGRPH_HIDDEN, 
													ept_InitUtilityRtfContext( 
													pt_AttchPgrph), NULL))
		return us_err;

	//Include an attachment hotspot for the object in the special hidden-
	//	paragraph rich-text item. This will keep the attachment from showing up 
	//	"below-the-line," as it would if not referenced by an attachment 
	//	hotspot in a rich-text field on the document. Since the hotspot always 
	//	remains hidden, there's no need to include a graphic in the hotspot.
	if (!*mpc_fileNmPgpRtf)
		strcat( strcpy( mpc_fileNmPgpRtf, mpc_ROOTFILENM_PGP_RTF), 
														mpc_PGP_ARMOR_EXT_ASC);
	us_err = eus_AppendAttachmentHotspot( h_NOTE, mpc_fileNmPgpRtf, pc_objNm, 
													NULL, NULL, pt_AttchPgrph);

	if (us_err)
		ef_FreeRtfContext( pt_AttchPgrph);

	return us_err;
} //us_SetupRtfAttachParagraph(


/** us_CreateRecipientsArray( ***
Purpose is to convert a string containing the current message recipients into a 
string array, the format expected by the PGP infrastructure. In doing this, the 
procedure resolves the recipients against the Domino Directory using the 
specified lookup configuration. All members of an ambiguous address resolution 
are included in the ultimate list, as are recipients that do not resolve.

--- parameters & return ----
pc_RecipientsList: Input & Output. Pointer to the string containing the 
	message recipients as currently understood, delimited by semi-colons. 
	Output is undefined due to use of strtok() to parse the string.
f_LOOKUP_EXHAUST: Flag telling whether the directories should be searched 
	exhaustively for each name so as to be sure all matches are accounted for.
pc_LookupHow: pointer to the name of the server holding the Domino Directory to 
	be used potentially in the resolution process
pppc_recipients: Output. Pointer to the string array variable to populate with 
	the list of recipients PGP should try to match encryption keys to. CALLER 
	IS RESPONSIBLE for freeing the resources allocated to the array.
pui_recipients: Output. Pointer to the variable in which to store the number of 
	elements placed in the string array (pppc_recipients).
pf_ambiguous: Output. Pointer to the flag variable to set according to whether 
	ambiguous resolutions were encountered.
RETURN: eus_SUCCESS if no error occured; the Notes API error code otherwise

--- revision history -------
9/6/02 PR: listing format adjustment, minor logic shortening

9/12/99 PR
+ logic overhaul to support enhancement of better emulation of Lotus Notes' 
  name lookup behavior
+ documentation adjustment

12/20/98 PR: created		*/
static STATUS us_CreateRecipientsArray( char  pc_RecipientsList[], 
										const BOOL  f_LOOKUP_EXHAUST, 
										char  pc_LookupHow[], 
										char * * *const  pppc_recipients, 
										UINT *const  pui_recipients, 
										BOOL *const  pf_ambiguous)	{
	const char  pc_LIST_DELIMITER[] = ";";

	UINT  ui, ui_entries = 0;
	char * pc_recip, * pc, * * ppc = NULL;
	NameFoundNode * pt_unresolved = NULL, * pt;
	StringNode * pt_resolved = NULL, * pt_nd;
	STATUS  us_err = eus_SUCCESS;
	BOOL  f_fail;

	_ASSERTE( pc_RecipientsList && pppc_recipients && pf_ambiguous);

	*pppc_recipients = (char * *) *pui_recipients = NULL;
	*pf_ambiguous = FALSE;

	//for each entry in the text list...
	pc_recip = strtok( pc_RecipientsList, pc_LIST_DELIMITER);
	_ASSERTE( pc_recip);
	do	{
		//allocate space for a copy of the entry, and copy the entry into it
		if (f_fail = !(pc = malloc( ui = strlen( pc_recip) + 1)))
			goto errJump;
		strcpy( pc, pc_recip);

		//string on a new entry to the unresolved list, transferring ownership 
		//	of the name string to the entry
		pt = pt_unresolved;
		if (f_fail = !ef_AddNameFoundNodeFifo( pc, FALSE, TRUE, &pt))
			goto errJump;
		if (pt && !pt_unresolved)
			pt_unresolved = pt;
	} while (pc_recip = strtok( NULL, pc_LIST_DELIMITER));
	pc = NULL;

	//if there are any unresolved recipient names...
	if (pt_unresolved)	{
		char * pc = epc_strtokStrict( pc_LookupHow, pc_LIST_DELIMITER);

		//attempt to resolve the names against the default Domino Directory set
		if (us_err = eus_ResolveMailAddresses( pt_unresolved, 
									f_LOOKUP_EXHAUST, pc, NULL, &pt_resolved, 
									&ui_entries, pf_ambiguous))
			goto errJump;

		//if a further server to resolve against has been specified...
		if ((pc = epc_strtokStrict( NULL, pc_LIST_DELIMITER)) && *pc)	{
			//attempt to resolve the names against the server's Domino 
			//	Directory set, if necessary
			if (!f_LOOKUP_EXHAUST)	{
				pt = pt_unresolved;
				do
					if (!pt->pt_name->f_found)
						break;
				while (pt = pt->pt_next);
			}
			if (f_LOOKUP_EXHAUST || pt)	{
				if (us_err = eus_ResolveMailAddresses( pt_unresolved, 
										f_LOOKUP_EXHAUST, pc, NULL, 
										&pt_resolved, &ui, 
										*pf_ambiguous ? NULL : pf_ambiguous))
					goto errJump;
				ui_entries += ui;
			}
		} //if (pc = strtok( NULL,

		//Tack any addresses still unresolved onto the "resolved" list so PGP 
		//	may match keys against everything. Internet addresses will often 
		//	not be resolved by the Domino Directory.
		pt = pt_unresolved;
		pt_nd = pt_resolved;
		do	{
			if (pt->pt_name->f_found)
				continue;

			if (f_fail = !ef_AddStringNodeFifo( pt->pt_name->pc_nm, FALSE, 
																		&pt_nd))
				goto errJump;
			pt->pt_name->pc_nm = NULL;
			if (!pt_resolved)
				pt_resolved = pt_nd;
			ui_entries++;
		} while (pt = pt->pt_next);
	} //if (pt_unresolved)

	//transfer control of the resolved list to a string array
	if (ui_entries)	{
		if (f_fail = !( ppc = calloc( ui_entries, sizeof( char *))))
			goto errJump;
		pt_nd = pt_resolved;
		for (ui = 0; ui < ui_entries; ui++)	{
			ppc[ ui] = pt_nd->pc;
			pt_nd = pt_nd->pt_next;
		}
		_ASSERTE( !pt_nd);

		*pppc_recipients = ppc;
		*pui_recipients = ui_entries;
	} //if (ui_entries)

errJump:
	//Free allocated resources no longer needed. If we've been successful, the 
	//	string array now owns the strings in the resolved list.
	if (pt_unresolved)
		e_FreeNameFoundList( &pt_unresolved);
	if (pt_resolved)
		e_FreeList( &pt_resolved, us_err || f_fail);
	if (pc)
		free( pc);

	return us_err + f_fail;
} //us_CreateRecipientsArray(


/** i_PgpEncodeRtfText( ***


--- parameters & return ----

RETURN:
	eus_SUCCESS if no error occured
	the Notes API error code otherwise

--- revision history -------
9/6/02 PR
+ logic enhancement to support high-ASCII (international) characters in message 
  body
+ token renaming, listing format adjustment

12/15/98 PR: created		*/
//DOC!!
static int i_PgpEncodeRtfText( PgpEncodeContext *const  pt_encdCtx, 
								RtfContext *const  pt_rtCtx, 
								char * *const  ppc_outp)	{
	const int  i_PGPERR_OUT_OF_MEM = -11999;	//kPGPError_OutOfMemory

	char * pc_ctnt, * pc_outpOs = NULL;
	ULONG  ul;
	int  i_err;
	BOOL  f_fail = FALSE;

	_ASSERTE( pt_encdCtx && (pt_encdCtx->f_Encrypt || pt_encdCtx->f_Sign) && 
														pt_rtCtx && ppc_outp);

	//get a copy of the rich-text field's textual content
	if (eul_ERR_FAILURE == (ul = eul_GetRtfText( pt_rtCtx, NULL, TRUE, 
																	&pc_ctnt)))
		return !eus_SUCCESS;

	//if the field has content...
	if (ul > 1)	{
		//if the content can be translated to the native character set, PGP 
		//	encode it
		char *const  pc_ctntOs = malloc( ul);
		if (!pc_ctntOs)	{
			i_err = i_PGPERR_OUT_OF_MEM;		
			goto errJump;
		}
		if (OSTranslate( OS_TRANSLATE_LMBCS_TO_NATIVE, pc_ctnt, (WORD) (ul - 
													1), pc_ctntOs, (WORD) ul))
			i_err = ei_PgpEncodeBuffer( pc_ctntOs, strlen( pc_ctntOs), 
										pt_encdCtx, FALSE, NULL, &pc_outpOs);
		else
			i_err = !eus_SUCCESS;
		free( pc_ctntOs);
		if (i_err)
			goto errJump;

		//Translate the result back to the Notes LMBCS character set so it 
		//	displays properly in the Notes client and so the SMTP mailer will 
		//	translate it properly back to the native character set if sending 
		//	the message out. (All this effectively comes into play with SMTP 
		//	mailing of clear-signed content.)
		if (!( *ppc_outp = malloc( ul = strlen( pc_outpOs) + 1)))	{
			i_err = i_PGPERR_OUT_OF_MEM;
			goto errJump;
		}
		f_fail = !OSTranslate( OS_TRANSLATE_NATIVE_TO_LMBCS, pc_outpOs, (WORD) 
												(ul - 1), *ppc_outp, (WORD) ul);
	//else PGP encode the blank content
	}else	{
		if (i_err = ei_PgpEncodeBuffer( pc_ctnt, NULL, pt_encdCtx, FALSE, NULL, 
																	&pc_outpOs))
			goto errJump;
		if (!( *ppc_outp = malloc( ul = strlen( pc_outpOs) + 1)))
			i_err = i_PGPERR_OUT_OF_MEM;
		else
			memcpy( *ppc_outp, pc_outpOs, ul);
	} //if (ul > 1)

errJump:
	if (pc_outpOs)
		e_FreePgpMem( pc_outpOs);
	free( pc_ctnt);

	return i_err + f_fail;
} //i_PgpEncodeRtfText(


/** i_MatchPgpKeys( ***
Match as much as possible recipient addresses with PGP keys, be they local or 
accessible via key servers, and be some addresses Notes names or SMTP 
addresses. In doing so, allow users to gracefully break out of key-server 
searching and get on with regular local PGP processing.

--- parameters & return ----
f_SYNC_UNK_KEYS: Flag telling whether names not found on the user's default 
	keyring should be sought on the PGP key servers she's specified. If false, 
	SMTP addresses won't be processed because no further processing is 
	necessary before sending to the encryption-keys dialog.
ui_RECPTS: the number of members in the list of recipients
ppc_recpts: Input & Output. Address of the string-array list of recipients. 
	Notes names which may be switched to abbreviaed format while preserving a 
	match will be switched. (A Notes name will remain in canonicalized format 
	only if the canonicalized version is the only one for which a matching key 
	can be found.)
pt_encdCtx: Input & Output. Address of the PGP information structure being used 
	in the current round of encoding. If f_SYNC_UNK_KEYS is TRUE, the opaque 
	structure may be updated if names are located via a PGP key-server search.
RETURN:
	!eus_SUCCESS if out-of-memory condition encountered
	ei_USER_ABORT if server-lookup operation was user-canceled
	the PGP error that occurred, if any
	eus_SUCCESS otherwise

--- revision history -------
10/30/02 PR: created		*/
static __inline int i_MatchPgpKeys( const BOOL  f_SYNC_UNK_KEYS, 
									const UINT  ui_RECPTS, 
									char *const *const  ppc_recpts, 
									PgpEncodeContext *const  pt_encdCtx)	{
	int  i_err = 0, i_error = 0;

	//if necessary, seek PGP-key matches for any SMTP recipeint addresses
	if (f_SYNC_UNK_KEYS && (i_err = i_PgpMatchSmtpAddresses( ui_RECPTS, 
													ppc_recpts, pt_encdCtx)) && 
													i_err != ei_USER_ABORT)
		return i_err;

	//seek PGP-key matches for any Notes names in the recipients' list, 
	//	adjusting as needed most canonicalized Notes names to the more 
	//	user-friendly abbreviated format
	if (i_error = i_PgpMatchNotesNames( f_SYNC_UNK_KEYS && !i_err, ui_RECPTS, 
														ppc_recpts, pt_encdCtx))
		return i_error;

	return i_err;
} //i_MatchPgpKeys(


/** i_PgpMatchSmtpAddresses( ***
Match as much as possible recipient SMTP addresses with PGP keys, be they local 
or accessible via key servers.

--- parameters & return ----
ui_RECPTS: the number of members in the list of recipients
ppc_recpts: Input. Address of the string-array list of recipients. Non-SMTP 
	addresses will be ignored.
pt_encdCtx: Input & Output. Address of the PGP information structure being used 
	in the current round of encoding. If f_SYNC_UNK_KEYS is TRUE, the opaque 
	structure may be updated if names are located via a PGP key-server search.
RETURN:
	!eus_SUCCESS if out-of-memory condition encountered
	ei_USER_ABORT if server-lookup operation was user-canceled
	the PGP error that occurred, if any
	eus_SUCCESS otherwise

--- revision history -------
10/30/02 PR: created		*/
static int i_PgpMatchSmtpAddresses( const UINT  ui_RECPTS, 
									char *const *const  ppc_recpts, 
									PgpEncodeContext *const  pt_encdCtx)	{
	static const char  pc_VALID_NON_ALPHANUM_SMTP_ADDR_CHARS[] = "-_"; 

	static char  pc_validSmtpAddrChars[ 62 + sizeof( 
										pc_VALID_NON_ALPHANUM_SMTP_ADDR_CHARS)];

	NameFoundNode * pt_toFind = NULL, * pt_nd = NULL;
	NameFoundInfo * pt;
	UINT  ui;
	char * pc, * pc_;
	BOOL  f_fail;
	int  i_err;

	_ASSERTE( ppc_recpts && pt_encdCtx);

	//the first time through, populate a "constant" string buffer holding 
	//	valid SMTP address characters
	if (!*pc_validSmtpAddrChars)
		strcat( strcpy( pc_validSmtpAddrChars, epc_ALPHANUM), 
										pc_VALID_NON_ALPHANUM_SMTP_ADDR_CHARS);

	//for each name in the full list...
	for (ui = 0; ui < ui_RECPTS; ui++)	{
		//if the name obviously isn't a valid SMTP address, iterate to the next 
		//	name in the full list
		pc_ = strchr( pc = ppc_recpts[ ui], '<');
		if (!( (pc = strchr( pc_ ? ++pc_ : pc, '@')) && strchr( 
								pc_validSmtpAddrChars, *(pc - 1)) && 
								strchr( pc_validSmtpAddrChars, *(pc + 1)) && 
								(pc_ = strrchr( ++pc, '.')) > pc && 
								strchr( pc_validSmtpAddrChars, *(pc_ + 1))))
			continue;

		//create a list node to hold the name information
		if (f_fail = !( pt = calloc( 1, sizeof( NameFoundInfo))))
			goto errJump;
		pt->pc_nm = ppc_recpts[ ui];

		//add the new node to the to-find list
		if (f_fail = !ef_AddListNodeFifo( pt, &pt_nd))	{
			free( pt);
			goto errJump;
		}
		if (!pt_toFind)
			pt_toFind = pt_nd;
	} //for (ui = 0; ui < ui_RECPTS

	//if no SMTP addresses were present, short-circuit with success
	if (!pt_toFind)
		return eus_SUCCESS;

	//check locally within PGP and mark the SMTP addresses it knows of
	if (i_err = ei_PgpLookupEmailLocal( pt_encdCtx, pt_toFind))
		goto errJump;

	//if any name remains not found...
	pt_nd = pt_toFind;
	do
		if (!( !*pt_nd->pt_name->pc_nm || pt_nd->pt_name->f_found))
			break;
	while (pt_nd = pt_nd->pt_next);
	if (pt_nd)	{
		//check servers within PGP and mark accordingly the names found this 
		//	way
		i_err = ei_PgpLookupEmailViaServers( pt_encdCtx, pt_toFind);
	} //if (pt_nd)

errJump:
	if (pt_toFind)
		e_FreeList( &pt_toFind, TRUE);

	return f_fail + i_err;
} //i_PgpMatchSmtpAddresses(


/** i_PgpMatchNotesNames( ***
To accommodate the intuitive UI desired in the context of PGP keys, default any 
Notes names to the user-friendly abbreviated format, albeit while maintaining 
backward compatibility with the canonicalized format used prior to PGP 6.5.1.

--- parameters & return ----
f_SYNC_UNK_KEYS: flag telling whether names not found on the user's default 
	keyring should be sought on the PGP key servers she's specified
ui_RECPTS: the number of members in the list of recipients
ppc_recpts: Input & Output. Address of the string-array list of recipients to 
	check for Notes names which may be switched safely to abbreviaed format. 
	Such names will be switched. A Notes name will remain in canonicalized 
	format only if the canonicalized version is the only one for which a 
	matching key can be found within the PGP infrastructure.
pt_encdCtx: Input & Output. Address of the PGP information structure being used 
	in the current round of encoding. If f_SYNC_UNK_KEYS is TRUE, the opaque 
	structure may be updated if names are located via a PGP key-server search.
RETURN: the PGP error that occurred, if any; out-of-memory condition causes a 
	!eus_SUCCESS return; otherwise eus_SUCCESS is returned

--- revision history -------
10/30/02 PR
+ logic enhancement to support gracefully user-cancelation during key-server 
  searches
+ listing format adjustment, token renaming, documentation improvement

5/30/00 PR: documentation adjustment

3/20/00 PR
+ fixed null-pointer bug that occurred when all names were accounted for in 
  Notes abbreviated format and the user has specified that not-found names 
  should be looked up against her list of PGP key servers
+ minor documentation adjustment

9/12/99 PR: created			*/
static int i_PgpMatchNotesNames( const BOOL  f_SYNC_UNK_KEYS, 
									const UINT  ui_RECPTS, 
									char *const *const  ppc_recpts, 
									PgpEncodeContext *const  pt_encdCtx)	{
	NameFoundNode * pt_abrvd = NULL, * pt_canonicl = NULL, * pt_node = NULL, 
					* pt_nd = NULL;
	NameFoundInfoEx * pt, * pt_nm;
	UINT  ui;
	char * pc = NULL, * pc_src;
	BOOL  f_fail;
	int  i_err;

	_ASSERTE( ppc_recpts && pt_encdCtx);

	//for each name in the full list...
	for (ui = 0; ui < ui_RECPTS; ui++)	{
		//if the name obviously isn't a canonicalized Notes name, iterate to 
		//	the next name in the full list
		if (memcmp( ppc_recpts[ ui], mpc_TKN_NOTES_CANONICAL, 
										mui_LEN_TKN_NOTES_CANONICAL) != ei_SAME)
			continue;

		//Create an abbreviated version of the name. However if it turns out 
		//	that the name is not really a canonicalized name, iterate to the 
		//	next name in the full list.
		if (f_fail = !( pc = malloc( strlen( (pc_src = ppc_recpts[ ui]) + 
												mui_LEN_TKN_NOTES_CANONICAL))))
			goto errJump;
		if (f_fail = !epc_NotesNmAbbreviate( pc_src, pc))	{
			free( pc);
			pc = NULL;
			continue;
		}

		//create a list node to hold the name information, transferring 
		//	ownership of the abbreviated version to it
		if (f_fail = !( pt = calloc( 1, sizeof( NameFoundInfoEx))))
			goto errJump;
		pt->pc_parent = pc_src;
		pt->t.pc_nm = pc;
		pc = NULL;

		//add the new node to the abbreviated-names list
		if (f_fail = !ef_AddListNodeFifo( pt, &pt_node))	{
			free( pt);
			goto errJump;
		}
		if (!pt_abrvd)
			pt_abrvd = pt_node;
	} //for (ui = 0; ui < ui_RECPTS

	//if no Notes names were present, short-circuit with success
	if (!pt_abrvd)
		return eus_SUCCESS;

	//check locally within PGP and mark the abbreviated names it knows of
	if (i_err = ei_PgpLookupEmailLocal( pt_encdCtx, pt_abrvd))
		goto errJump;

	//for each name in the list of abbreviated names...
	pt_node = pt_abrvd;
	do	{
		//If the name has already been found, reset the corresponding name in 
		//	the full list, and then iterate to the next name. Since an 
		//	abbreviated Notes name is always shorter than it's canonicalized 
		//	counterpart, the simple strcpy() is safe.
		if ((pt_nm = (NameFoundInfoEx *) pt_node->pt_name)->t.f_found)	{
			strcpy( pt_nm->pc_parent, pt_nm->t.pc_nm);
			*pt_nm->t.pc_nm = NULL;
			continue;
		}

		//create a list node to hold the name information where the 
		//	canonicalized version is primary (for use if the abbreviated 
		//	version can't be found)
		if (f_fail = !( pt = calloc( 1, sizeof( NameFoundInfoEx))))
			goto errJump;
		pt->pc_parent = pt_nm->t.pc_nm;
		pt->t.pc_nm = pt_nm->pc_parent;
		if (f_fail = !ef_AddListNodeFifo( pt, &pt_nd))	{
			free( pt);
			goto errJump;
		}
		if (!pt_canonicl)
			pt_canonicl = pt_nd;
	} while (pt_node = pt_node->pt_next);

	//if a list was populated...
	if (pt_canonicl)	{
		//check locally within PGP and mark accordingly the canonicalized names 
		//	it knows of
		if (i_err = ei_PgpLookupEmailLocal( pt_encdCtx, pt_canonicl))
			goto errJump;

		//for each of the names found this way, nullify the corresponding name 
		//	entry in the abbreviated list
		pt_nd = pt_canonicl;
		do
			if (pt_nd->pt_name->f_found)
				*((NameFoundInfoEx *) pt_nd->pt_name)->pc_parent = NULL;
		while (pt_nd = pt_nd->pt_next);
	} //if (pt_canonicl)

	//if the user has specified that unknown names should be sought via the PGP 
	//	key servers she's specified...
	if (f_SYNC_UNK_KEYS)	{
		//if any name remains not found...
		pt_nd = pt_abrvd;
		do
			if (!( !*pt_nd->pt_name->pc_nm || pt_nd->pt_name->f_found))
				break;
		while (pt_nd = pt_nd->pt_next);
		if (pt_nd)	{
			//check servers within PGP and mark accordingly the abbreviated 
			//	names it knows of
			if ((i_err = ei_PgpLookupEmailViaServers( pt_encdCtx, pt_abrvd)) && 
														i_err != ei_USER_ABORT)
				goto errJump;

			//for each node in the abbreviated list...
			pt_node = pt_abrvd;
			pt_nd = pt_canonicl;
			do	{
				//if the name wasn't just found, iterate to the next list node
				if (!( *(pt = (NameFoundInfoEx *) 
								pt_node->pt_name)->t.pc_nm && pt->t.f_found))
					continue;

				//flag the associated node in the canonical list that the name 
				//	has been found
				strcpy( pt->pc_parent, pt->t.pc_nm);
				do
					if (pt->t.pc_nm == ((NameFoundInfoEx *) 
												pt_nd->pt_name)->pc_parent)	{
						pt_nd->pt_name->f_found = TRUE;
						break;
					}
				while (pt_nd = pt_nd->pt_next);
				*pt->t.pc_nm = NULL;
			} while (pt_node = pt_node->pt_next);
		} //if (pt_nd)

		//if necessary, determine whether any name in the canonicalized list 
		//	remains not found
		if ((pt_nd = pt_canonicl) && !i_err)
			do
				if (!pt_nd->pt_name->f_found)
					break;
			while (pt_nd = pt_nd->pt_next);

		//if any name remains not found...
		if (pt_nd && !i_err)	{
			//check servers within PGP and mark accordingly the canonicalized 
			//	names it knows of
			if (i_err = ei_PgpLookupEmailViaServers( pt_encdCtx, 
										pt_canonicl) && i_err != ei_USER_ABORT)
				goto errJump;

			//for each of the canonicalized names marked as found, nullify the 
			//	corresponding name entry in the abbreviated list
			pt_nd = pt_canonicl;
			do
				if (pt_nd->pt_name->f_found)
					*((NameFoundInfoEx *) pt_nd->pt_name)->pc_parent = NULL;
			while (pt_nd = pt_nd->pt_next);
		} //if (pt_nd)
	} //if (f_SYNC_UNK_KEYS)

	//for each remaining name in the list of abbreviated names still not found, 
	//	replace the corresponding entry in the full list with the abbreviated 
	//	entry (so it'll be displayed that way in the encryption-keys dialog)
	pt_nd = pt_abrvd;
	do
		if (*pt_nd->pt_name->pc_nm)	{
			strcpy( ((NameFoundInfoEx *) pt_nd->pt_name)->pc_parent, 
														pt_nd->pt_name->pc_nm);
			_ASSERTE( !pt_nd->pt_name->f_found);
		}
	while (pt_nd = pt_nd->pt_next);

errJump:
	if (pt_canonicl)
		e_FreeList( &pt_canonicl, TRUE);
	if (pt_abrvd)
		e_FreeNameFoundList( &pt_abrvd);
	if (pc)
		free( pc);

	return f_fail + i_err;
} //i_PgpMatchNotesNames(


/** xul_R45getHandleIllegit( ***


--- parameter & return ----


--- revision history ------
7/2/01 PR
+ added safety logic
+ documentation adjustment

10/7/00 PR: removed logic to initialize this DLL's PGP connection due to 
	update in the plug-in subform's handling of this issue 
3/20/00 PR: moved from RevealNoteHandle module to nPGPNts module
12/13/98 PR: created		*/
//DOC!!
unsigned long xul_R45getHandleIllegit( 
									  const BYTE *const  puc_NotesDocument)	{
	const int  i_45_OFFSET_TO_HANDLE = 0xEC;

	if (!puc_NotesDocument)
		return NULL;

	return *(unsigned long *) (puc_NotesDocument + i_45_OFFSET_TO_HANDLE);
} //xul_R45getHandleIllegit(


/** xus_ResetToPgpEncodedRtf( ***
Resets the specified note's rich-text field to the PGP-encoded version saved 
in memory via a prior call to es_PreMailSend. This is necessary when the 
Notes R5 mail template is in use. More specifically, the R5 template 
implements the "Save and File" action differently from the R4 template, 
driving the need for the reset functionality. For more specificis, see the 
documentation in the PGP subform concerning this issue.

--- parameters & return ----
h_NOTE: handle to the note whose rich-text field is to be reset to the 
	PGP-encoded version
pc_ITMNM_RTF: pointer to a string telling the name of the rich-text field to 
	be used in the encoding
RETURN: eus_SUCCESS if no error occured; the Notes API error code otherwise

--- revision history -------
3/20/00 PR: created		*/
unsigned short xus_ResetToPgpEncodedRtf( const NOTEHANDLE  h_NOTE, 
											char  pc_ITMNM_RTF[])	{
	RtfContext  t_RtfContext;
	BYTE * puc;
	STATUS  us_err;
//BLOCKID  bid;
	if (!( h_NOTE && pc_ITMNM_RTF))
		return !eus_SUCCESS;

	//if no PGP-encoded rich-text field is ready to be written to the note, 
	//	short-circuit with a general failure
	if (mt_RtfResettable.h_note != h_NOTE)
		return !eus_SUCCESS;
	_ASSERTE( mt_RtfResettable.h_content);

/*if (!FALSE)	{
if (us_err = OSMemAlloc( NULL, mt_RtfResettable.ul_len, &bid.pool))
	goto errJump;
bid.block = NULLBLOCK;
puc = OSLockBlock( BYTE, bid);
memcpy( puc, OSLockObject( mt_RtfResettable.h_content), 
													mt_RtfResettable.ul_len);
OSUnlockBlock( bid);
OSUnlockObject( mt_RtfResettable.h_content);
if (us_err = NSFItemDelete( h_NOTE, pc_ITMNM_RTF, (WORD) strlen( 
															pc_ITMNM_RTF)))
	goto errJump;
us_err = NSFItemAppendByBLOCKID( h_NOTE, ITEM_SIGN | ITEM_SEAL, pc_ITMNM_RTF, 
										(WORD) strlen( pc_ITMNM_RTF), 
										bid, mt_RtfResettable.ul_len, NULL);
if (us_err)
	OSMemFree( bid.pool);
OSMemFree( mt_RtfResettable.h_content);
memset( &mt_RtfResettable, NULL, sizeof( ResettableRtfInfo));
return us_err;
} //CONTROL IF
*/	//initialize resources associated with the rich-text handling we're 
	//	going to do
	if (us_err = eus_InitializeRtfContext( h_NOTE, pc_ITMNM_RTF, NULL, 
															&t_RtfContext))
		return us_err;

	//replace the current rich-text field with the PGP-encoded version 
	//	stored during the "PreMailSend" that must have just before taken 
	//	place
	puc = OSLockObject( mt_RtfResettable.h_content);
	us_err = eus_ReplaceRtfWithCdStream( puc, mt_RtfResettable.ul_len, 
															&t_RtfContext);
	OSUnlockObject( mt_RtfResettable.h_content);
	if (us_err)
		goto errJump;

	//commit the updated rich-text field to the note, overwriting the current 
	//	actual rich-text content
	us_err = eus_CommitChangedRtf( h_NOTE, pc_ITMNM_RTF, &t_RtfContext);

errJump:
	//free resources allocated to the stored version, and flag that this has 
	//	been done
	OSMemFree( mt_RtfResettable.h_content);
	memset( &mt_RtfResettable, NULL, sizeof( ResettableRtfInfo));

	ef_FreeRtfContext( &t_RtfContext);

	return us_err;
} //xus_ResetToPgpEncodedRtf(


/** xs_PrepCheckmarkMenuing( ***
Sets the environment by which the plug-in may monitor and work with Windows 
messages sent to the message's action-bar window.

--- suggestion enhancement ----
8/9/00 PR: use an integer return instead of a "boolean" so that an 
	out-of-memory condition may be indicated

--- return --------------------
RETURN: Visual Basic/LotusScript TRUE if successful, FALSE otherwise 

--- revision history ----
11/14/00 PR: documentation adjustment

9/16/00 PR
+ replaced window-procedure linked-list mechanism with the more 
  straightforward and more loosely coupled SetProp() & GetProp() combination
+ error-handling improvement
+ removed obsolete documentation & code

8/9/00 PR: created			*/
short xs_PrepCheckmarkMenuing()	{
	//static for speed
	static const char  pc_CLASSNM_ACTIONBAR[] = "ActionBar", 
						pc_CLASSNM_RTF_EDITBX[] = "IRIS.tedit", 
						pc_CLASSNM_NOTES_SUBPROG[] = "NotesSubprog", 
						pc_CLASSNM_RTF[] = "NotesRichText";

	HWND  h, h_bar;
	char  pc_strt[ 20];
	WNDPROC  pr;
	DWORD  ul_err;

	//if the cursor focus is starting out in an R5 address edit-box field...
	if (!( h = GetFocus()))
		return ms_VB_FALSE;
	if (!GetClassName( h, pc_strt, sizeof( pc_strt)))
		goto errJump;
	if (!( h = GetParent( h)))
		goto errJump;
	if (!( h = GetParent( h)))
		goto errJump;
	if (strcmp( pc_CLASSNM_RTF_EDITBX, pc_strt) == ei_SAME)	{
		//get the action-bar window handle, 2 up & 1 down
		if (!( h_bar = FindWindowEx( h, NULL, pc_CLASSNM_ACTIONBAR, NULL)))
			goto errJump;
	//else if the focus is starting out in the main window (anywhere in R4 
	//	mail template, Body field in R5 mail template...
	}else if (strcmp( pc_CLASSNM_RTF, pc_strt) == ei_SAME)	{
		HWND  h_prv, h_up1 = h;

		//get the action-bar window handle, 2 up & 2 down, but where the 
		//	first try down may be a dead end
		if (!( h = h_prv = FindWindowEx( h, NULL, pc_CLASSNM_NOTES_SUBPROG, 
																	NULL)))
			goto errJump;
		if (!( h_bar = FindWindowEx( h, NULL, pc_CLASSNM_ACTIONBAR, NULL)))	{
			if (!( h = FindWindowEx( h_up1, h_prv, pc_CLASSNM_NOTES_SUBPROG, 
																	NULL)))
				goto errJump;
			if (!( h_bar = FindWindowEx( h, NULL, pc_CLASSNM_ACTIONBAR, 
																	NULL)))
				goto errJump;
		} //if (!( h_bar = FindWindowEx( 
	//else the focus is in an unsupported place, so short-circuit with failure
	}else
		return ms_VB_FALSE;

	//override (subclass) the action-bar's window procedure with one that we 
	//	control, saving the original procedure so window messages can be 
	//	passed on for regular processing
	if (!( pr = (WNDPROC) SetWindowLong( h_bar, GWL_WNDPROC, 
													(long) ActionBarWndProc)))
		goto errJump;
	if (!SetProp( h_bar, mpc_PROPNM_ORIG_WNDPROC, pr))	{
		SetWindowLong( h_bar, GWL_WNDPROC, (long) pr);
		goto errJump;
	}

	return ms_VB_TRUE;

errJump:
	ul_err = GetLastError();
	return ms_VB_FALSE;
} //xs_PrepCheckmarkMenuing(


/** ActionBarWndProc( ***
Subclassing window-message procedure for the "action bar" of a document in 
the Notes client. The action bar handles all the drop-down menuing done by 
the action buttons on the bar. In the context of the PGP Plug-In, we use this 
procedure to adjust particular menu items hanging off of the PGP action 
button.

--- parameters & return ----
H: handle to the window being called
ui_MSG: identifier of the message being passed to the procedure
ui_W: the WORD parameter associated with the message
ul_L: the long parameter associated with the message
RETURN: The result of the message. Use of the return varies from message to 
	message

--- revision history -------
9/16/00 PR: replaced window-procedure linked-list mechanism with the more 
	straightforward and more loosely coupled SetProp() & GetProp() combination
8/9/00 PR: created			*/
static LRESULT CALLBACK ActionBarWndProc( const HWND  H, 
											const UINT  ui_MSG, 
											const WPARAM  ui_W, 
											const LPARAM  ul_L)	{
	char * pc = NULL;
	WNDPROC  pr;

	switch (ui_MSG)	{
		//if this is the popup menu command we are interested in inspecting...
		case WM_INITMENUPOPUP:	{
			static const char pc_MU_SIGNOFF[] = "Toggle Sign off", 
								pc_MU_ENCRYPTOFF[] = "Toggle Encrypt off", 
								pc_MU_SIGN[] = "Sign", 
								pc_MU_ENCRYPT[] = "Encrypt";
			static UINT  ui_POS_SIGN = 0, ui_POS_ENCRYPT = 1;

			MENUITEMINFO  nii;
			char * pc_tmp;

			//If the PGP-sign menu item needs to be checkmarked, simplify the 
			//	item's name and set the checkmark. The mask member should be 
			//	(and is) set according to whether we're running Win98/Win2000 
			//	or not.
			memset( &nii, NULL, sizeof( MENUITEMINFO));
			nii.cbSize = sizeof( MENUITEMINFO);
			nii.fMask = 
#if WINVER < 0x0500
						MIIM_TYPE;
#else
						MIIM_STRING;
#endif
			if (!GetMenuItemInfo( (HANDLE) ui_W, ui_POS_SIGN, TRUE, &nii))
				goto errJump;
			if (nii.cch == sizeof( pc_MU_SIGNOFF) - 1)	{
				if (!( nii.dwTypeData = pc = malloc( ++nii.cch)))
					goto errJump;
				if (!GetMenuItemInfo( (HANDLE) ui_W, ui_POS_SIGN, TRUE, &nii))
					goto errJump;
				if (stricmp( pc, pc_MU_SIGNOFF) == ei_SAME)	{
					strcpy( pc, pc_MU_SIGN);
					nii.fMask |= MIIM_STATE;
					nii.fState = MFS_CHECKED;
					if (!SetMenuItemInfo( (HANDLE) ui_W, ui_POS_SIGN, TRUE, 
																		&nii))
						goto errJump;
				} //if (stricmp( pc, pc_MU_SIGNOFF
			} //if (nii.cch == sizeof( pc_MU_SIGNOFF)

			//if the PGP-encrypt menu item needs to be checkmarked, simplify 
			//	the item's name and set the checkmark
			nii.dwTypeData = (char *) nii.cch = NULL;
			nii.fMask &= ~MIIM_STATE;
			if (!GetMenuItemInfo( (HANDLE) ui_W, ui_POS_ENCRYPT, TRUE, &nii))
				goto errJump;
			if (nii.cch == sizeof( pc_MU_ENCRYPTOFF) - 1)	{
				if (!( nii.dwTypeData = pc = realloc( pc_tmp = pc, 
															++nii.cch)))	{
					pc = pc_tmp;
					goto errJump;
				}
				if (!GetMenuItemInfo( (HANDLE) ui_W, ui_POS_ENCRYPT, TRUE, 
																		&nii))
					goto errJump;
				if (stricmp( pc, pc_MU_ENCRYPTOFF) == ei_SAME)	{
					strcpy( pc, pc_MU_ENCRYPT);
					nii.fMask |= MIIM_STATE;
					nii.fState = MFS_CHECKED;
					SetMenuItemInfo( (HANDLE) ui_W, ui_POS_ENCRYPT, TRUE, &nii);
				} //if (stricmp( pc, pc_MU_ENCRYPTOFF
			} //if (nii.cch == sizeof( pc_MU_ENCRYPTOFF)
		} //case WM_INITMENUPOPUP
		break;

		//if a window we're monitoring is going away, we have to put things 
		//	back the way we found them...
		case WM_DESTROY:
		case WM_CLOSE:	{
			DWORD  ul_err;

			if (!( pr = RemoveProp( H, mpc_PROPNM_ORIG_WNDPROC)))	{
				ul_err = GetLastError();
				return FALSE;	//should never get here, arbitrary return
			}

			//reset to the original window-procedure configuration, and pass 
			//	the close message down the chain
			if (!SetWindowLong( H, GWL_WNDPROC, (long) pr))
				ul_err = GetLastError();
			return CallWindowProc( pr, H, ui_MSG, ui_W, ul_L);
		} //case WM_CLOSE, WM_DESTROY
	} //switch (ui_MSG)

errJump:
	if (pc)
		free( pc);

	if (!( pr = GetProp( H, mpc_PROPNM_ORIG_WNDPROC)))
		return FALSE;	//should never get here, arbitrary return
	return CallWindowProc( pr, H, ui_MSG, ui_W, ul_L);
} //ActionBarWndProc(


/** x_ClearEncodedRtfCopy( ***
Free resources allocated to the copy of the PGP-encoded rich-text field that 
was stored in memory in case it was needed to reset the message's rich-text 
field with the encoded version (a functionality required due to how Lotus 
implemented its "Save and File" action in the R5 template).

--- parameter ----------
H: handle to the note initiating this procedure call

--- revision history ---
3/20/00 PR: created		*/
void x_ClearEncodedRtfCopy( const NOTEHANDLE  H)	{
	if (!( H && mt_RtfResettable.h_note == H))
		return;

	OSMemFree( mt_RtfResettable.h_content);
	memset( &mt_RtfResettable, NULL, sizeof( ResettableRtfInfo));
} //x_ClearEncodedRtfCopy(


