/*____________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.

	$Id: pgpLicenseNumber.c,v 1.27 2002/11/25 05:45:59 jason Exp $
____________________________________________________________________________*/

#include <ctype.h>
#include <string.h>

/*	We need to include this in order to get the PGP_BETA flag before its use in the license# file	*/
#if PGP_OSX
#include "pgpBuild.h"
#endif

#include "pgpLicenseNumber.h"
#include "pgpLicenseNumberNet.h"
#include "pgpUtilities.h"
#include "pgpPublicKey.h"
#include "pgpMem.h"
#include "pgpHash.h"
#include "pgpSymmetricCipher.h"
#include "pgpKeys.h"
#include "pgpEncode.h"
#include "string.h"
#include "pgpErrors.h"
#include "pgpStrings.h"
#include "pgpEndianConversion.h"
#include "pgpUnicode.h"

#include "pgpLNClientKeys.h" /* keys that are needed for a client */
#ifdef PGP_LICENSE_SERVER
	/* additional private or secret keys that are needed by a server only */
	#include "pgpLNServerKeys.h" 
#endif

#include "stdio.h"

#ifndef PGP_LICENSE_NET
#define PGP_LICENSE_NET 1	/* Enable network LA by default */
#endif

#ifdef PGP_WIN32
#pragma warning (disable: 4057)
#pragma intrinsic(memset)
#define CLEAR_LICENSENUMBER(ppn) memset( ppn, 0, sizeof(LICENSENUMBER) )
#else
#define CLEAR_LICENSENUMBER(ppn) pgpClearMemory( ppn,sizeof(LICENSENUMBER) )
#endif

static const PGPByte beginLicense[]={"-----BEGIN PGP LICENSE AUTHORIZATION-----"};
static const PGPByte endLicense[]={"-----END PGP LICENSE AUTHORIZATION-----"};

static PGPUInt8 base32chars[]={
	"ACDEFGHJKLMNPQRT" "UVWXYZ234679B015"
	"________________" "____________8OIS"	/* similiar looking duplicates */
};

static const PGPByte emptyCustomerName[] = "<blank>";
static const PGPByte * const emptyCompanyName = emptyCustomerName;

// for each field corresponding parameter name is given -- check pgplng and server code
typedef struct _LICENSENUMBER
{
	PGPUInt8 version;		// "version"
	PGPUInt8 productID;		// "prodid"
	PGPUInt16 evalPeriod;	// 0 - is never expires, "evaldays"
	PGPUInt32 numLicenses; // 24 bit, 0 - reserved, "licenses"
	PGPUInt8 flags;			// "flags" = {"eval","enterprise","disk","mail"}
	PGPUInt32 serialNumber;	// "sn"
	PGPUInt8 shaHash[2];
} LICENSENUMBER;

#define LICENSENUMBER_BIN_SIZE 17	/* serialization size, sizeof(LICENSENUMBER) can be more because of alignment */

#define LICENSENUMBER_BASE32_SIZE (LICENSENUMBER_BIN_SIZE*8/5+1) 
		/*upper bound, used for memory allocation only*/

typedef struct PGPLicenseNumber {
	PGPByte ln_encr[LICENSENUMBER_BIN_SIZE];
} PGPLicenseNumber;		// Exported: [17 bytes ln], ln has a version that will tell the size

typedef struct PGPLicenseAuthorization {
	PGPByte		type;
	PGPUInt16	issuanceDate; // days since 9/1/2002
	PGPSize		signatureSize;	
	PGPByte		signature[1];
	// + signatureSize-1 bytes	// Exported: [2 bytes total size] [1 byte type] [2 bytes issuanceDate] [reserved=0] [signature]
} PGPLicenseAuthorization;

static void lnToStream( const LICENSENUMBER *ln, PGPByte ln_stream[LICENSENUMBER_BIN_SIZE] )  {
	PGPUInt32 u;
	PGPByte *p = ln_stream;

	p[0] = ln->version;
	p[1] = ln->productID;

	u = ln->evalPeriod;
	p[2] = (PGPByte)(u >> 8);
	p[3] = (PGPByte)u;

	u = ln->numLicenses;
	p[4] = (PGPByte)(u >> 16);
	p[5] = (PGPByte)(u >> 8);
	p[6] = (PGPByte)u;

	p[7] = ln->flags;

	u = ln->serialNumber;
	p[8] = (PGPByte)(u >> 24);
	p[9] = (PGPByte)(u >> 16);
	p[10] = (PGPByte)(u >> 8);
	p[11] = (PGPByte)u;
	
	p[12] = p[13] = p[14] = 0;
		
	p[15] = ln->shaHash[0];
	p[16] = ln->shaHash[1];
}

static void lnFromStream( const PGPByte ln_stream[LICENSENUMBER_BIN_SIZE], LICENSENUMBER *ln )  {
	const PGPByte *p = ln_stream;

	ln->version = p[0];
	ln->productID = p[1];

	ln->evalPeriod = (PGPUInt16)(p[2] << 8 | p[3]);
	ln->numLicenses = p[4] << 16 | p[5] << 8 | p[6];
	ln->flags = p[7];
	ln->serialNumber = p[8] << 24 | p[9] << 16 | p[10] << 8 | p[11];
	
//	ln->unused[0] = p[12]; ln->unused[1] = p[13]; ln->unused[2] = p[14]; 
		
	ln->shaHash[0] = p[15];
	ln->shaHash[1] = p[16];
}

static void laToStream( const PGPLicenseAuthorization *la, PGPByte *stream )  {
	PGPUInt32 u;

	u = la->signatureSize + 6;
	stream[0] = (PGPByte)(u >> 8);
	stream[1] = (PGPByte)u;

	stream[2] = la->type;

	u = la->issuanceDate;
	stream[3] = (PGPByte)(u >> 8);
	stream[4] = (PGPByte)u;

	stream[5] = 0;	// reserved
	
	pgpCopyMemory( la->signature, stream+6, la->signatureSize );
}

// exact size required to serialize the object
static PGPUInt16 laToStreamSize( const PGPLicenseAuthorization *la )  {
	return (PGPUInt16)(la->signatureSize + 6);
}
// size of the serialized object read from the stream
static PGPUInt16 laFromStreamSize( const PGPByte *stream )  {
	const PGPUInt16 size = (PGPUInt16)(stream[0] << 8 | stream[1]);
	return (PGPUInt16)( size > 128+6 ? 0 : size);
}

/* la should be already allocated, use laPeekSize */
static void laFromStream( const PGPByte *stream, PGPLicenseAuthorization *la )  {
	const PGPUInt16 size = laFromStreamSize( stream );

	if( !size )  {
		la->signatureSize = 0;
		la->type = 0;
		la->issuanceDate = 0;
		return;
	}

	la->signatureSize = size-6;

	la->type = stream[2];
	la->issuanceDate = (PGPUInt16)(stream[3] << 8 | stream[4]);
	pgpCopyMemory( stream+6, la->signature, la->signatureSize );
}


// number of days from the 9/1/02 to the 'curr' time
static const time_t Sep_1_2002_time_t = 0x3d71c901;	// time_t for Sun Sep 01 00:00:01 2002, no DST
static PGPUInt32 daysElapsed( PGPTime curr, PGPBoolean roundUp )  {
	time_t curr_time_t = PGPGetStdTimeFromPGPTime( curr );
	PGPUInt32 days = (PGPUInt32)(-1);

	if( !curr )
		return (PGPUInt32)(-1);

	if( curr_time_t > Sep_1_2002_time_t )  {
		days = ( curr_time_t - Sep_1_2002_time_t ) / 3600;
		days = (days + (roundUp ? 23 : 0))/24;
	}
	else {	
		// 37779 (0x9393) days from the Sep_1_2002_time_t left to overlap 0xffffffff

		// allow one overlap
		curr_time_t /= 3600;
		curr_time_t = (curr_time_t+(roundUp ? 23 : 0)) / 24;
		days = 37779 + curr_time_t;
	}

	return days;
}
#if 0 /* not used, but working */
static PGPTime daysElapsedToTime( PGPUInt32 daysElapsed )  {
	time_t t;
	
	if( daysElapsed > 37779 )	
		return (PGPTime)(-1);	// need to fix after 2037

	t = Sep_1_2002_time_t + daysElapsed*24*3600;
	return PGPGetPGPTimeFromStdTime( t );
}
#endif
#if 0	/* not used, but working */
static void daysElapsedToStr( PGPUInt32 daysElapsed, PGPByte str[16] )  {
	PGPTime t = daysElapsedToTime(daysElapsed);
    struct tm *tm;

	tm = gmtime(&t);

    (void)strftime( str, 16, "%Y%m%d", tm); 
}
#endif
#if 0	/* not used, but working */
static PGPUInt32 strToDaysElapsed( const PGPByte str[16] )  {
	struct tm tm;
	PGPByte s[10] = { '\0', '\0', '\0', '\0', '\0' };
	time_t t;

	pgpClearMemory( &tm, sizeof(tm) );

	s[0] = str[4]; s[1] = str[5];
	tm.tm_mon = atoi(s)-1;

	s[0] = str[6]; s[1] = str[7];
	tm.tm_mday = atoi(s);

	s[0] = str[0]; s[1] = str[1]; s[2] = str[2]; s[3] = str[3];
	tm.tm_year = atoi(s)-1900;

	tm.tm_sec = 1;

	t = mktime( &tm );
	if( t == (time_t)-1 )
		return (PGPUInt32)(-1);

	return daysElapsed( PGPGetPGPTimeFromStdTime( t ), FALSE );
}
#endif

// Base32 encode binary license data
static PGPError pgpEncode32(
			  const PGPByte *inBytes, PGPSize inSize, 
			  PGPByte outBytes[kPGPLicenseNumberMaxSize])
{
	PGPUInt32 bits,bitcnt,value,bytenum,shift,memamt;
	PGPError err;

	err=kPGPError_NoErr;

	bits=inSize*8;

	memamt=((bits+4)/5)+1;  // +1 for convenience NULL terminate
	if( memamt > kPGPLicenseNumberMaxSize )  {
		outBytes[0] = '\0';
		return kPGPError_OutOfMemory;
	}

	pgpClearMemory(outBytes,memamt);
	
	for(bitcnt=0;bitcnt<bits;bitcnt+=5)
	{
		bytenum=bitcnt/8;
		shift=bitcnt%8;

		value = inBytes[bytenum];
		if( bytenum < inSize-1 )
			value |= (inBytes[bytenum+1] << 8);
		value = (value >> shift) & 0x1f;
		
		outBytes[bitcnt/5]=base32chars[value];
	}

	return err;
}

// Base32 decode license data
static PGPError pgpDecode32(const PGPByte *inBytes,
			  PGPByte outBytes[kPGPLicenseNumberMaxSize], PGPSize *outSize )
{
	PGPUInt32 chcnt,numbits,memamt,place,bytenum,shift,value;
	PGPError err;
	PGPSize inSize;
	PGPByte *pc;
	
	if (!inBytes)
		return kPGPError_BadParams;
	
	inSize = strlen( inBytes );

	err=kPGPError_NoErr;
	outBytes[0] = '\0';
	*outSize = 0;

	numbits=inSize*5;
	memamt=(numbits+7)/8;

	if(memamt+1 > kPGPLicenseNumberMaxSize)
		return kPGPError_BufferTooSmall;

	/* check that last base 32 character carries the value no greater then needed */
	pc = strchr( base32chars, inBytes[inSize-1] );
	if( pc==NULL || *pc == '_' )  /* wrong character */
		return kPGPError_CorruptData;
	value=(PGPUInt32)(pc - base32chars) / sizeof(base32chars[0]);
	if( value > 32 )
		value -= 32;		/* we hit duplicate */
	shift = (inSize*5) % 8;	/* extra bits that must be zero */
	if( shift > 5 )
		return kPGPError_CorruptData;
	shift = 5 - shift;		/* now shift is number of bits that last base char encodes  */
	if( shift > 0 && value > ~((PGPUInt32)(-1)<<shift) )
		return kPGPError_CorruptData;
	
	pgpClearMemory( outBytes, memamt+1 );

	for(chcnt=0;chcnt<inSize;chcnt++)
	{
		place=chcnt*5;
		bytenum=place/8;
		shift=place%8;
		
		pc = strchr( base32chars, inBytes[chcnt] );
		if( pc==NULL || *pc == '_' )  /* wrong character */
			return kPGPError_CorruptData;
		value=(PGPUInt32)(pc - base32chars) / sizeof(base32chars[0]);
		if( value > 32 )
			value -= 32;	/* we hit duplicate */
		
		outBytes[bytenum] |= (PGPByte)(value << shift);
		if( shift > 8-5 )  {
			outBytes[bytenum+1] |= (value >> (8-shift));
			*outSize = bytenum + 1;
		}
		else
			*outSize = bytenum;
	}

	return err;
}


#define TC_E_PARSE kPGPError_CorruptData
#define SKIPWS(c) while (isspace ((int) (*(c)))) c++
static const unsigned char BaseChars[] = 
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
#define BaseChar(c) BaseChars[c]
static const unsigned char BaseVals[] = {
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,62, 0xff,0xff,0xff,63,
  52,53,54,55, 56,57,58,59, 60,61,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
  15,16,17,18, 19,20,21,22, 23,24,25,0xff, 0xff,0xff,0xff,0xff,
  0xff,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
  41,42,43,44, 45,46,47,48, 49,50,51,0xff, 0xff,0xff,0xff,0xff
};
#define BaseVal(c) BaseVals[c]

PGPError 
tc_encode_base64 (PGPMemoryMgrRef mem, unsigned char **out, size_t *outlen,
		      const unsigned char *in, size_t inlen )
{
  unsigned char *curpos;
  unsigned char a, b, c;

  /* determine how long the output will be.  if it is not a multple of 3
   * bytes, increase length so that it is so that the pad characters are
   * accounted for.
   */
  if (inlen % 3 != 0)
    *outlen = inlen + 3 - inlen % 3;
  else
    *outlen = inlen;

  /* base64 encoding creates 4 output chars for every 3 input chars */
  *outlen = 4 * (*outlen) / 3;

  if ((*out = PGPNewData(mem, *outlen + 1, 0)) == NULL)
	  return kPGPError_OutOfMemory;

  curpos = *out;

  while (inlen)
  {
    a = *in++;
    inlen--;

    *curpos++ = BaseChar (a >> 2);

    if (inlen)
      b = *in++;
    else
      b = 0;

    *curpos++ = BaseChar (((a & 0x03) << 4) | (b >> 4));

    if (!inlen)
    {
      *curpos++ = '=';
      *curpos++ = '=';
      break;
    }
    inlen--;

    if (inlen)
      c = *in++;
    else
      c = 0;

    *curpos++ = BaseChar (((b & 0x0f) << 2) | (c >> 6));

    if (!inlen)
    {
      *curpos++ = '=';
      break;
    }
    inlen--;

    *curpos++ = BaseChar (c & 0x3f);
  }

  *curpos = 0;

  return kPGPError_NoErr;
}

PGPError
tc_decode_base64 (PGPMemoryMgrRef mem, unsigned char **out,
		  size_t *outlen, const unsigned char *in )
{
  unsigned char *curpos;
  unsigned char a, b;
  PGPError err = kPGPError_NoErr;

  *outlen = (3 * strlen ((PGPByte *) in)) / 4; /* maximum length */
  if ((*out = PGPNewData(mem, *outlen, 0)) == NULL)
	  return kPGPError_OutOfMemory;

  curpos = *out;

  SKIPWS (in);
  while (*in)
  {
    a = *in++;
    SKIPWS (in);
    if (!*in)  {
      err = TC_E_PARSE;
	  break;
	}
    b = *in++;
    SKIPWS (in);
    if (!*in)  {
      err = TC_E_PARSE;
	  break;
	}

    *curpos++ = (BaseVal (a) << 2) | (BaseVal (b) >> 4);

    a = b;
    b = *in++;
    SKIPWS (in);
    if (!*in)  {
      err = TC_E_PARSE;
	  break;
	}

    if (b == '=')
      break;

    *curpos++ = (BaseVal (a) << 4) | (BaseVal (b) >> 2);

    a = b;
    b = *in++;
    SKIPWS (in);

    if (b == '=')
      break;

    *curpos++ = (BaseVal (a) << 6) | BaseVal (b);
  }

  if( IsntPGPError(err) )
	*outlen = curpos - *out; /* real length */
  else  {
	PGPFreeData( *out );
	*out = NULL;
  }

  return err;
}

#ifdef PGP_LICENSE_SERVER

// AES encrypt LN 
static PGPError pgpEncryptLicenseNumber(PGPContextRef context,
										const LICENSENUMBER *ln, PGPLicenseNumberRef ln_ref)
{
	PGPError err;
	PGPSymmetricCipherContextRef AES256context;
	PGPByte *ln_encr = ((PGPLicenseNumber *)ln_ref)->ln_encr;
	
	err=kPGPError_NoErr;
	
	err=PGPNewSymmetricCipherContext(context,kPGPCipherAlgorithm_AES256,
		&AES256context);
	
	if(IsntPGPError(err))
	{
		err=PGPInitSymmetricCipher(AES256context,AES256key);
		
		if(IsntPGPError(err))
		{
			PGPByte ln_decr_buf[LICENSENUMBER_BIN_SIZE];
			lnToStream( ln, ln_decr_buf );
			
			err=PGPSymmetricCipherEncrypt( AES256context,
				ln_decr_buf+1, ln_encr+1 );
			
			ln_encr[0]=ln->version;
		}
		
		PGPFreeSymmetricCipherContext(AES256context);
	}
	
	return err;
}

#endif

// AES decrypt LN and check hash
static PGPError pgpDecryptLicenseNumber(PGPContextRef context,
	PGPLicenseNumberRef ln_ref, LICENSENUMBER *ln)
{
	PGPError err;
	PGPSymmetricCipherContextRef AES256context;
	PGPHashContextRef SHAContext;
	PGPByte HashedLN[20];
	const PGPByte *ln_encr;
	
	if (!context || !ln_ref || !ln)
		return kPGPError_BadParams;
	
	ln_encr = ((PGPLicenseNumber*)ln_ref)->ln_encr;
	
	err=kPGPError_NoErr;
	
	err=PGPNewSymmetricCipherContext(context,kPGPCipherAlgorithm_AES256,
		&AES256context);
	
	if(IsntPGPError(err))
	{
		err=PGPInitSymmetricCipher( AES256context,AES256key);
		
		if(IsntPGPError(err))
		{
			PGPByte ln_decr_buf[LICENSENUMBER_BIN_SIZE];

			err=PGPSymmetricCipherDecrypt( AES256context, ln_encr+1, ln_decr_buf+1);
			ln_decr_buf[0] = ln_encr[0];
			
			lnFromStream( ln_decr_buf, ln );
			
			// Hash the LN
			err=PGPNewHashContext(context,
				kPGPHashAlgorithm_SHA,&SHAContext);
			
			if(IsntPGPError(err))
			{
				// Hash everything except the hash field
 				err=PGPContinueHash(SHAContext, ln_decr_buf, LICENSENUMBER_BIN_SIZE-2);
				
				if(IsntPGPError(err))
				{
					err=PGPFinalizeHash(SHAContext,HashedLN);
					
					if(ln->shaHash[0]==HashedLN[0] && ln->shaHash[1]==HashedLN[1])
						err=kPGPError_NoErr; // Its OK
					else
						err=kPGPError_LNCorrupt;
				}
				
				PGPFreeHashContext(SHAContext);
			}
			
		}
		
		PGPFreeSymmetricCipherContext(AES256context);
	}
	
	return err;
}

#ifdef PGP_LICENSE_SERVER

// Creates a LN based on entry fields (using hardcoded private key).
// For use by web tools.
PGPError PGPLNCreate(PGPContextRef context,
							PGPUInt8 productID,
							PGPUInt16 evalPeriod,
							PGPUInt32 numLicenses, // 24 bit
							PGPUInt8 flags,
							PGPUInt32 serialNumber,
							PGPLicenseNumberRef *buffer)
{
	LICENSENUMBER pf;
	PGPByte HashedLN[20];
	PGPError err;
	PGPHashContextRef SHAContext;

	err=kPGPError_NoErr;
	*buffer = NULL;

	CLEAR_LICENSENUMBER(&pf);
	
	pf.version=kPGPLicenseNumberVersion;
	
	pf.productID=productID;
	pf.evalPeriod =		evalPeriod;
	pf.numLicenses =	numLicenses;
	pf.flags =			flags;
	pf.serialNumber =	serialNumber;
	
	// Hash the LN
	err=PGPNewHashContext(context,
		kPGPHashAlgorithm_SHA,&SHAContext);
	
	if(IsntPGPError(err))
	{
		PGPByte ln_buf[LICENSENUMBER_BIN_SIZE];

		// Hash everything except the hash field
		lnToStream( &pf, ln_buf );
		err=PGPContinueHash(SHAContext, ln_buf, LICENSENUMBER_BIN_SIZE-2);
		
		if(IsntPGPError(err))
		{
			err=PGPFinalizeHash(SHAContext,HashedLN);
			
			pf.shaHash[1]=HashedLN[1];
			pf.shaHash[0]=HashedLN[0];
			
		 	*buffer=(PGPLicenseNumberRef)PGPNewData(
				PGPPeekContextMemoryMgr(context),sizeof(PGPLicenseNumber),0);

			if( *buffer == NULL)
				err = kPGPError_OutOfMemory;

			if( IsntPGPError(err) )
				err=pgpEncryptLicenseNumber(context, &pf, *buffer);

			if( IsPGPError(err) )  {
				PGPFreeData( *buffer );
				*buffer = NULL;
			}
		}
		
		PGPFreeHashContext(SHAContext);
	}
	
	CLEAR_LICENSENUMBER( &pf );
	
	return err;
}

#endif

// Takes a base32 LN and imports it into internal opaque structure
PGPError PGPLNImportBase32(PGPContextRef context,
						const PGPByte *licenseNumber32, PGPLicenseNumberRef *buffer)
{
	PGPError err;
	PGPByte ln_buf[kPGPLicenseNumberMaxSize];
	PGPSize ln_buf_size;

	err=kPGPError_NoErr;

	if (!licenseNumber32 || !buffer)
		return kPGPError_BadParams;

	*buffer = NULL;

	/* remove '-' */
	{
		PGPByte ln_buf1[kPGPLicenseNumberMaxSize];
		PGPByte *p = ln_buf1;
		PGPUInt32 i;

		for( i=0; i<strlen(licenseNumber32) && i<sizeof(ln_buf1)-1; i++ )  {
			if( licenseNumber32[i] != '-' )
				*(p++) = licenseNumber32[i];
		}
		*p = '\0';
		
		err=pgpDecode32(ln_buf1, ln_buf, &ln_buf_size);

#ifdef PGP_DEBUG
	{
		PGPByte ln_buf2[kPGPLicenseNumberMaxSize];
		pgpEncode32( ln_buf, ln_buf_size, ln_buf2 );
		pgpAssert( strcmp(ln_buf2, ln_buf1)==0 );
	}
#endif
	}
	
	if(IsntPGPError(err))
	{
		LICENSENUMBER pn;
		PGPLicenseNumber ln_encr;
		
		if(ln_buf_size < LICENSENUMBER_BIN_SIZE) 
			return kPGPError_LNCorrupt;

		/* build LN opaque structure */
		pgpAssert( ln_buf_size <= 0xffff );
		memcpy( ln_encr.ln_encr, ln_buf, LICENSENUMBER_BIN_SIZE );
		
		err=pgpDecryptLicenseNumber(context, &ln_encr, &pn);
		
		if(IsntPGPError(err))
		{
			CLEAR_LICENSENUMBER( &pn );

			*buffer=(PGPLicenseNumberRef)PGPNewData( PGPPeekContextMemoryMgr(context), 
				sizeof(ln_encr), 0 );
			if( *buffer )  
				memcpy( *buffer, &ln_encr, sizeof(ln_encr) );
			else
				err = kPGPError_OutOfMemory;
		}
	}

	return err;
}

// Exports internal opaque structure to base32
PGPError PGPLNExportBase32( PGPContextRef context,
										  const PGPLicenseNumberRef licenseNumber,
										  const PGPByte *templ, 
										  PGPByte base32[kPGPLicenseNumberMaxSize] )  
{
	PGPError err;
	const PGPLicenseNumber *ln = (PGPLicenseNumber *)licenseNumber;
	PGPByte b32[kPGPLicenseNumberMaxSize];
	PGPByte *pb32 = b32;
	PGPUInt32 i;

	(void)context;

	if( templ == NULL )
		return pgpEncode32(ln->ln_encr, LICENSENUMBER_BIN_SIZE, base32);

	err=pgpEncode32(ln->ln_encr, LICENSENUMBER_BIN_SIZE, b32);

	if( IsPGPError(err) )
		return err;

	if( strlen(templ) >= sizeof(b32) )  {
		base32[0] = '\0';
		return kPGPError_BufferTooSmall;
	}

	for( i=0; i<strlen( templ ); i++ )  {
		if( templ[i]=='x' || templ[i]=='X' )
			base32[i] = *(pb32++);
		else
			base32[i] = templ[i];
	}
	base32[i] = '\0';

	return err;
}

#ifdef PGP_LICENSE_SERVER // [
// Exports internal opaque structure into application/form-urlencoded format
PGPError pgpLNExportUrlEncoded(PGPContextRef context,
	const PGPLicenseNumberRef ln_ref, PGPByte urlEncoded[kPGPLicenseNumberMaxSize])
{
	PGPError err = kPGPError_NoErr;
	LICENSENUMBER ln;
	PGPByte flags_str[80] = { '\0', '\0' };
	PGPByte *f = flags_str+1;

	*urlEncoded = '\0';

	err=pgpDecryptLicenseNumber(context,ln_ref,&ln);
	if( IsPGPError(err) )
		return err;

	/* we know that all fields consist of valid urlencoded characters, so we don't use
	   addFormUrlEncodedParam 
	*/

	if( ln.flags & kPGPLicenseNumberFlag_Evaluation )
		strcat( flags_str, ",eval" );
	if( ln.flags & kPGPLicenseNumberFlag_Enterprise )
		strcat( flags_str, ",enterprise" );
	if( ln.flags & kPGPLicenseNumberFlag_Disk )
		strcat( flags_str, ",disk");
	if( ln.flags & kPGPLicenseNumberFlag_Mail )
		strcat( flags_str, ",mail" );
	if( ln.flags & kPGPLicenseNumberFlag_Desktop )
		strcat( flags_str, ",desktop" );

	sprintf( urlEncoded, "version=%d&prodid=%d&evaldays=%d&licenses=%d&flags=%s&sn=%d",
		ln.version, ln.productID, ln.evalPeriod, ln.numLicenses, f, ln.serialNumber );

	CLEAR_LICENSENUMBER(&ln);

	return err;
}
#endif // PGP_LICENSE_SERVER ]


// Functions for retrieving internal values
PGPError PGPLNGetProductID(PGPContextRef context,PGPLicenseNumberRef licenseNumber,
							PGPUInt8 *productID)
{
	PGPError err;
	LICENSENUMBER pn;

	err=pgpDecryptLicenseNumber(context,licenseNumber,&pn);

	if(IsntPGPError(err))
	{
		*productID=pn.productID;

		CLEAR_LICENSENUMBER(&pn);
	}

	return err;
}


PGPError			PGPLNGetEvalPeriod(PGPContextRef context,PGPLicenseNumberRef licenseNumber,
							PGPUInt16 *evalPeriod)
{
	PGPError err;
	LICENSENUMBER pn;

	err=pgpDecryptLicenseNumber(context,licenseNumber,&pn);

	if(IsntPGPError(err))  {
		*evalPeriod=pn.evalPeriod;
		CLEAR_LICENSENUMBER(&pn);
	}
	return err;
}

PGPError			PGPLNGetNumLicenses(PGPContextRef context,PGPLicenseNumberRef licenseNumber,
							PGPUInt32 *numLicenses)
{
	PGPError err;
	LICENSENUMBER pn;

	err=pgpDecryptLicenseNumber(context,licenseNumber,&pn);

	if(IsntPGPError(err))  {
		*numLicenses=pn.numLicenses;
		CLEAR_LICENSENUMBER(&pn);
	}

	return err;
}

PGPError			PGPLNGetFlags(PGPContextRef context,PGPLicenseNumberRef licenseNumber,
							PGPUInt8 *flags)
{
	PGPError err;
	LICENSENUMBER pn;

	err=pgpDecryptLicenseNumber(context,licenseNumber,&pn);

	if(IsntPGPError(err))
	{
		*flags=pn.flags;

		CLEAR_LICENSENUMBER(&pn);
	}

	return err;
}

PGPError			PGPLNGetSN(PGPContextRef context,PGPLicenseNumberRef licenseNumber,
							PGPUInt32 *serialNumber)
{
	PGPError err;
	LICENSENUMBER pn;

	err=pgpDecryptLicenseNumber(context,licenseNumber,&pn);

	if(IsntPGPError(err))  {
		*serialNumber=pn.serialNumber;
		CLEAR_LICENSENUMBER(&pn);
	}

	return err;
}
							
// Free internal opaque structure
void				PGPLNFree(PGPLicenseNumberRef licenseNumber)
{
	pgpClearMemory(licenseNumber,sizeof(PGPLicenseNumber));
	PGPFreeData(licenseNumber);
}

static PGPError calculateNameHash( PGPContextRef context, 
		const PGPByte *customerName, const PGPByte *companyName, PGPHashContextRef hashContext )
{
	PGPError err = kPGPError_NoErr;
	const PGPByte nullStr[] = "";
	PGPSize ucs2BufSize;
	PGPUInt16 *ucs2Buf = NULL; 
	PGPSize ucs2Size;
	PGPSize utf8charSize;
	PGPByte utf8charBuf[16];
	PGPUInt32 i;

	if( customerName==NULL && companyName==NULL )
		return kPGPError_LazyProgrammer;
		
	ucs2BufSize = strlen(customerName);
	if( ucs2BufSize < strlen(companyName) )
		ucs2BufSize = strlen(companyName);

	/* allocate large enough buffer to hold customerName or companyName */
	ucs2Buf = (PGPUInt16*)PGPNewData(
				PGPPeekContextMemoryMgr(context), (ucs2BufSize+1)*sizeof(PGPUInt16), 0);
	if( ucs2Buf == NULL )
		return kPGPError_OutOfMemory;

	/* hash customerName */
	if(customerName!=NULL && IsntPGPError(err))   {
     	pgpUTF8StringToUCS2( customerName, kPGPUnicodeNullTerminated, ucs2Buf, ucs2BufSize+1, (PGPUInt32 *)&ucs2Size );
		if( ucs2Size==0 )
			return kPGPError_BadParams;

		pgpUCS2StringToUpper( ucs2Buf, ucs2Size );
		for( i=0; i<ucs2Size; i++ )  {
			if( !pgpUnicodeCharIsWhitespace( ucs2Buf[i] ) && !pgpUnicodeCharIsPunctuation( ucs2Buf[i] ) )  
			{
				pgpUCS2StringToUTF8( ucs2Buf+i, 1, utf8charBuf, sizeof(utf8charBuf), (PGPUInt32 *)&utf8charSize );
				if( IsntPGPError(err) )
					err=PGPContinueHash(hashContext, utf8charBuf, utf8charSize);
			}
		}
	}
	if( IsntPGPError(err) )	/* terminator */
		err=PGPContinueHash(hashContext, nullStr, 1);
			
	/* hash companyName, similiar to customerName */
	if(companyName!=NULL && IsntPGPError(err))   {
     	pgpUTF8StringToUCS2( companyName, kPGPUnicodeNullTerminated, ucs2Buf, ucs2BufSize+1, (PGPUInt32 *)&ucs2Size );
		if( ucs2Size==0 )
			return kPGPError_BadParams;

		pgpUCS2StringToUpper( ucs2Buf, ucs2Size );
		for( i=0; i<ucs2Size; i++ )  {
			if( !pgpUnicodeCharIsWhitespace( ucs2Buf[i] ) && !pgpUnicodeCharIsPunctuation( ucs2Buf[i] ) )  
			{
				pgpUCS2StringToUTF8( ucs2Buf+i, 1, utf8charBuf, sizeof(utf8charBuf), (PGPUInt32 *)&utf8charSize );
				if( IsntPGPError(err) )
					err=PGPContinueHash(hashContext, utf8charBuf, utf8charSize);
			}
		}
	}
	if( IsntPGPError(err) )	/* terminator */
		err=PGPContinueHash(hashContext, nullStr, 1);

	/* free UCS2 buffer */
	if( ucs2Buf != NULL )
		PGPFreeData( ucs2Buf );

	return err;
}

/* single entry for calculating the hash of the license number to be used in 
   signature generation (license authorization) and verification (license verification). 
*/ 
static PGPError calculateLNHash( 
		PGPContextRef context, const PGPLicenseNumberRef licenseNumber,
		const PGPByte *customerName, const PGPByte *companyName, PGPUInt16 issuanceDate, 
		PGPHashContextRef hashContext )
{
	PGPError err;
	LICENSENUMBER ln;
	PGPByte ln_buf[LICENSENUMBER_BIN_SIZE];

	err=pgpDecryptLicenseNumber(context,licenseNumber,&ln);
	lnToStream( &ln, ln_buf );

	if(IsPGPError(err))
		return err;

	/* hash LN */
	err=PGPContinueHash(hashContext, ln_buf, LICENSENUMBER_BIN_SIZE);
	CLEAR_LICENSENUMBER( &ln );
	pgpClearMemory( ln_buf, sizeof(ln_buf) );

	/* hash customerName and companyName */
	if( IsntPGPError(err) )
		err = calculateNameHash( context, customerName, companyName, hashContext );
	
	/* hash date */
	if(IsntPGPError(err))   {
		const PGPByte two[2] = { (PGPByte)(issuanceDate>>8), (PGPByte)issuanceDate };
		err=PGPContinueHash(hashContext, two, sizeof(two));
	}
				
	return err;
}

/* Creates license number based on LN and entry fields (using
 hard coded private key). For use by web tools.

 Creates an output buffer and writes PGP DSA signature in binary form, sets size. 
 */
PGPError pgpLACreateLocal(PGPContextRef context, const PGPLicenseNumberRef licenseNumber,
		const PGPByte *customerName, const PGPByte *companyName, PGPLicenseAuthorizationRef *buffer)
{
#ifndef PGP_LICENSE_SERVER // Must have the private key
	(void)context;
	(void)licenseNumber;
	(void)customerName;
	(void)companyName;
	(void)buffer;

	return kPGPError_LazyProgrammer;
#else //PGP_LICENSE_SERVER
	PGPError err;
	PGPKeyDBRef		keyDB = kInvalidPGPKeyDBRef;
	PGPKeyDBObjRef	key = kInvalidPGPKeyDBObjRef;
	PGPHashContextRef shaContext = kInvalidPGPHashContextRef;
	PGPPrivateKeyContextRef privContext = kInvalidPGPPrivateKeyContextRef;
	PGPSize sigSize=0;
	PGPLicenseAuthorization *la = NULL;
	PGPUInt32 issuanceDate;

	*buffer = NULL;

	err=PGPNewHashContext(context, kPGPHashAlgorithm_SHA, &shaContext);
	if(IsPGPError(err))  
		return err;

	issuanceDate = daysElapsed( PGPGetTime(), TRUE );	// since PGP epoch

	err = calculateLNHash( context, licenseNumber, 
			customerName, companyName, (PGPUInt16)issuanceDate, shaContext );

	if(IsntPGPError(err))  {
		PGPKeyIterRef	keyIter = kInvalidPGPKeyIterRef;
								
		if(IsntPGPError(err)) 
			err = PGPImport (context, &keyDB, 							
				PGPOInputBuffer (context, DSSPrivKey, sizeof(DSSPrivKey)),
				PGPOLastOption (context));
									
		if(IsntPGPError(err))
			err=PGPNewKeyIterFromKeyDB (keyDB, &keyIter);
										
		if(IsntPGPError(err)) 
			err = PGPKeyIterNextKeyDBObj (keyIter, kPGPKeyDBObjType_Key, &key);

		if( PGPKeyIterRefIsValid(keyIter) )
			PGPFreeKeyIter(keyIter);
	}

	if(IsntPGPError(err)) 
		err = PGPNewPrivateKeyContext( key, kPGPPublicKeyMessageFormat_PGP, 
			&privContext, PGPOLastOption(context) );

	if(IsntPGPError(err)) 
		err = PGPGetPrivateKeyOperationSizes( privContext, NULL, NULL, &sigSize );

	if( sigSize )  {
		*buffer = (PGPLicenseAuthorizationRef)PGPNewData(PGPPeekContextMemoryMgr(context), 
			sigSize + sizeof(PGPLicenseAuthorization) - 1, 0 );
		la = (PGPLicenseAuthorization*)(*buffer);
	}

	if( *buffer )  {
		PGPSize size=0;
		err = PGPPrivateKeySign( privContext, shaContext, la->signature, &size );
		la->signatureSize = size;
		la->type = 0;			/* first DSA key */
		la->issuanceDate = (PGPUInt16)issuanceDate;
		shaContext = NULL;		/* hash context is freed in PGPPrivateKeySign */
	}

	if( PGPHashContextRefIsValid(shaContext) )
		PGPFreeHashContext( shaContext );

	if( PGPPrivateKeyContextRefIsValid(privContext) )
		PGPFreePrivateKeyContext( privContext );

	if( PGPKeyDBRefIsValid(keyDB) )
		PGPFreeKeyDB (keyDB);
			
	return err;
#endif // PGP_LICENSE_SERVER
}


static PGPBoolean isEmptyString( const PGPByte *s )  {
	const PGPByte *whiteChars = " \t\n\r";
	PGPUInt32 i;
	if( s==NULL )
		return TRUE;

	i = strlen(s);

	while( i-- )  {
		if( strchr(whiteChars, *(s++))==NULL )
			return FALSE;
	}

	return TRUE;
}

PGPError PGPLACreate(PGPContextRef context, const PGPByte *url, 
		 const PGPLicenseNumberRef licenseNumber,
		const PGPByte *customerName, const PGPByte *companyName, PGPLicenseAuthorizationRef *buffer)
{
	if( isEmptyString(customerName) )
		customerName = emptyCustomerName;
	if( isEmptyString(companyName) )
		companyName = emptyCompanyName;

	if( isEmptyString( licenseNumber ) )
		return kPGPError_LazyProgrammer;
	
#if PGP_LICENSE_NET
	return (url == NULL || url[0]=='\0') ? 
		pgpLACreateLocal(context, licenseNumber, customerName, companyName, buffer) :
		pgpLACreateNet(context, url, licenseNumber, customerName, companyName, buffer);
#else
	return pgpLACreateLocal(context, licenseNumber, customerName, companyName, buffer);
#endif
}

/* exports LA into base64 format (serialization). This is analog to base32 for LN format */
PGPError PGPLAExportBase64( PGPContextRef context, 							
		const PGPLicenseAuthorizationRef licenseAuthorization, PGPByte **b64 )
{
	PGPSize b64_size;
	PGPError err;
	PGPLicenseAuthorization *la = (PGPLicenseAuthorization*)licenseAuthorization;
	const PGPSize la_buf_size =  laToStreamSize( la );
	PGPByte *la_buf = PGPNewData( PGPPeekContextMemoryMgr(context), la_buf_size, 0 );

	if( la_buf==NULL )
		return kPGPError_OutOfMemory;

	laToStream( la, la_buf );
	err = tc_encode_base64( PGPPeekContextMemoryMgr(context), b64, &b64_size, 
		la_buf, la_buf_size );

	PGPFreeData( la_buf );

	return err;
}
/* imports LA from base64 format */
PGPError PGPLAImportBase64( PGPContextRef context, 							
		const PGPByte *b64, PGPLicenseAuthorizationRef *licenseAuthorization )
{
	PGPError err;
	PGPLicenseAuthorization *la = NULL;

	PGPByte *la_buf = NULL;
	PGPSize la_buf_size;
	
	if (!b64 || !licenseAuthorization)
		return kPGPError_BadParams;
	
	*licenseAuthorization = NULL;

	err = tc_decode_base64( PGPPeekContextMemoryMgr(context), &la_buf, &la_buf_size, b64 );
	if( IsPGPError(err) )
		return kPGPError_LACorrupt;

	if( la_buf_size <= sizeof(PGPLicenseAuthorization) || laFromStreamSize( la_buf ) != la_buf_size )	 {
		 PGPFreeData( la_buf );
		 return kPGPError_LACorrupt;
	}
	
	la = (PGPLicenseAuthorization*)PGPNewData( PGPPeekContextMemoryMgr(context), 
		la_buf_size + sizeof(PGPLicenseAuthorization)-1 /* count packed header twice */, 0 );
	if( la )  {
		laFromStream( la_buf, la );
		if( la->type != 0 )  {
			PGPFreeData( la_buf );
			PGPFreeData( la );
			return kPGPError_LACorrupt;
		}
	}

	if( la_buf != NULL )
		PGPFreeData( la_buf );

	*licenseAuthorization = la;

	return err;
}

// Takes a blocked license auth. number and imports it into internal opaque structure
// Returns kPGPError_CorruptData for the error in the ascii armor and 
// kPGPError_CorruptLicenseAuth for the error in LA, among others
PGPError PGPLAImportBlocked(PGPContextRef context,
							const PGPByte *licenseAuthorizationBlocked,
							PGPLicenseAuthorizationRef *buffer)
{
	PGPError err;
	PGPByte *pStart,*pEnd;
	static const PGPByte mailWhiteSpace[] = "\t \r\n>:-";

	err=kPGPError_NoErr;

	pStart=strstr(licenseAuthorizationBlocked,beginLicense);	// assume the header of armor is not broken

	if(pStart==0)
		return kPGPError_CorruptData;

	pStart+=sizeof(beginLicense)-1;
	if( *pStart=='\0' )
		return kPGPError_CorruptData;

	/* advance to the begining of base64 LA */
	while( strchr( mailWhiteSpace, *pStart ) )
		pStart++;

	{
		PGPByte b[256];
		strncpy( b, endLicense, sizeof(endLicense)/2 );	
		b[sizeof(endLicense)/2] = '\0';

		pEnd = strstr( pStart, b );	/* look for only half of an armor end for reliability */
		if( pEnd )  {
			PGPSize i, j=0, size = pEnd-pStart;
			if( sizeof(b)-1 < size )
				size = sizeof(b)-1;
			for( i=0; i<size; i++ )  {
				if( strchr( mailWhiteSpace, pStart[i] ) )
					continue;
				b[j++] = pStart[i];
			}
			b[j] ='\0';

			err = PGPLAImportBase64( context, b, buffer );
		}
		else 
			err = kPGPError_CorruptData;
	}

	return err;
}

// Exports internal opaque structure to base64. To be used on the server. 
// Client should use PGPLAExportBase64
PGPError PGPLAExportBlocked(PGPContextRef context,
							PGPLicenseAuthorizationRef licenseAuthorization,
							PGPByte **blocked )
{
	PGPError err;
	PGPUInt32 memamt;
	PGPByte *b64 = NULL;

	err=kPGPError_NoErr;

	err = PGPLAExportBase64( context, licenseAuthorization, &b64 );
	if( IsPGPError(err) )
		return err;

	memamt=strlen(beginLicense)+sizeof('\n')+
		strlen( b64 )+sizeof('\n') + 
		strlen(endLicense)+1;

	*blocked=(PGPByte *)PGPNewData(PGPPeekContextMemoryMgr(context),memamt,0);
			
	if(*blocked==0)
	{
		err=kPGPError_OutOfMemory;
	}
	else
	{
		strcpy(*blocked,beginLicense);
		strcat(*blocked,"\n");

		strcat(*blocked,b64);
		strcat(*blocked,"\n");

		strcat(*blocked,endLicense);
	}

	PGPFreeData( b64 );

	return err;
}

// Free internal opaque structure
void PGPLAFree(PGPLicenseAuthorizationRef licenseAuthorization)
{
	pgpClearMemory(licenseAuthorization,laToStreamSize((PGPLicenseAuthorization*)licenseAuthorization));
	PGPFreeData(licenseAuthorization);
}

/* Compares imported LN and AN and determines if valid. Error if not.
*/
PGPError PGPLAVerify(PGPContextRef context,
							PGPLicenseNumberRef licenseNumber,
							PGPLicenseAuthorizationRef licenseAuthorization,
							const PGPByte *customerName,
							const PGPByte *companyName)
{
	PGPError err;
	PGPKeyDBRef		keyDB = kInvalidPGPKeyDBRef;
	PGPKeyDBObjRef	key = kInvalidPGPKeyDBObjRef;
	PGPHashContextRef shaContext = kInvalidPGPHashContextRef;
	PGPPublicKeyContextRef pubContext = kInvalidPGPPublicKeyContextRef;
	PGPLicenseAuthorization *la = (PGPLicenseAuthorization*)licenseAuthorization;
	PGPInt32 daysLeft;

	if( isEmptyString(customerName) )
		customerName = emptyCustomerName;
	if( isEmptyString(companyName) )
		companyName = emptyCompanyName;

	err=PGPNewHashContext(context, kPGPHashAlgorithm_SHA, &shaContext);
	if(IsPGPError(err))  
		return err;

	err = calculateLNHash( context, licenseNumber, 
			customerName, companyName, la->issuanceDate, shaContext );

	if(IsPGPError(err))  {
		PGPFreeHashContext( shaContext );
		return err;
	}

	if(IsntPGPError(err))  {
		PGPKeyIterRef	keyIter = kInvalidPGPKeyIterRef;
								
		if(IsntPGPError(err)) 
			err = PGPImport (context, &keyDB, 							
				PGPOInputBuffer (context, DSSPubKey, sizeof(DSSPubKey)),
				PGPOLastOption (context));
									
		if(IsntPGPError(err))
			err=PGPNewKeyIterFromKeyDB (keyDB, &keyIter);
										
		if(IsntPGPError(err)) 
			err = PGPKeyIterNextKeyDBObj (keyIter, kPGPKeyDBObjType_Key, &key);

		if( PGPKeyIterRefIsValid(keyIter) )
			PGPFreeKeyIter(keyIter);
	}

	if(IsntPGPError(err)) 
		err = PGPNewPublicKeyContext( key, kPGPPublicKeyMessageFormat_PGP, &pubContext );

	if(IsntPGPError(err))  {
		err = PGPPublicKeyVerifySignature( pubContext, shaContext, la->signature, la->signatureSize );
		shaContext = kInvalidPGPHashContextRef;
	}

	if(IsntPGPError(err)) 
		err = PGPLADaysLeft( context, licenseNumber, licenseAuthorization, &daysLeft );

	if(IsntPGPError(err)) 
	{
		if (daysLeft <= 0)
			err = kPGPError_KeyExpired;
	}

	if( PGPHashContextRefIsValid(shaContext) )
		PGPFreeHashContext( shaContext );

	if( PGPPublicKeyContextRefIsValid(pubContext) )
		PGPFreePublicKeyContext( pubContext );

	if( PGPKeyDBRefIsValid(keyDB) )
		PGPFreeKeyDB (keyDB);

	return err;
}

/* 
   Sets daysLeft to huge positive number (0x7fffffff) to indicate unexpiring product. 
   Sets to huge negative number (-0x7fffffff) in case the error is returned. 
   Otherwise, sets to number of days left, possibly negative number. 
 */
PGPError PGPLADaysLeft(PGPContextRef context, PGPLicenseNumberRef ln_ref,
	PGPLicenseAuthorizationRef la_ref, PGPInt32 *daysLeft)
{
	PGPError err = kPGPError_NoErr;
	PGPUInt16 evalPeriod;
	PGPUInt32 daysUsed;
	PGPLicenseAuthorization *la = (PGPLicenseAuthorization*)la_ref;

	err = PGPLNGetEvalPeriod( context, ln_ref, &evalPeriod );
	if( IsPGPError(err) )  {
		*daysLeft = -(0x7fffffff);	
		return err;
	}

	if( evalPeriod==0 )  {
		*daysLeft = 0x7fffffff;	// plus infinity
		return err;	
	}

	daysUsed = daysElapsed( PGPGetTime(), TRUE );	// since PGP epoch
	if( daysUsed == (PGPUInt32)(-1) || daysUsed < la->issuanceDate  )  {
		*daysLeft = -(0x7fffffff);	// error
		return kPGPError_UnknownError;
	}

	daysUsed -= la->issuanceDate;		// days product used
	*daysLeft = evalPeriod - daysUsed;

	return err;
}

/* Returns hash of customerName and companyName in the form of null-terminated string in base64 encoding. 
 * Application that needs to compare two pairs of (companyName,customerName) must compare these hashes.
 * For the purpose of LA it is insufficient to compare customer/company name by other methods, such as strcmp, 
 * to determine that the strings are the same
 */
PGPError PGPLAGetNameHash( PGPContextRef context, const PGPByte *customerName, const PGPByte *companyName, PGPByte **hashb64 )  
{
	PGPError err = kPGPError_NoErr;
	PGPHashContextRef shaContext = kInvalidPGPHashContextRef;
	PGPByte binaryHash[80];
	PGPSize binaryHashSize;
		
	*hashb64 = NULL;
	
	err=PGPNewHashContext(context, kPGPHashAlgorithm_SHA, &shaContext);
	if(IsPGPError(err))  
		return err;
		
	PGPGetHashSize( shaContext, &binaryHashSize );

	err = calculateNameHash( context, customerName, companyName, shaContext );

	/* base64 encoding of the hash must fit into static buffer */
	pgpAssert( sizeof(binaryHash) > (binaryHashSize + 1)*4/3 );

	if( IsntPGPError(err) )
		err = PGPFinalizeHash( shaContext, binaryHash );

	if( IsntPGPError(err) )  {
		PGPSize hashb64_size;
		err = tc_encode_base64( PGPPeekContextMemoryMgr(context), 
				hashb64, &hashb64_size, binaryHash, binaryHashSize );
	}
	
	if( PGPHashContextRefIsValid(shaContext) ) 
		PGPFreeHashContext( shaContext );

	return err;
}

/*__Editor_settings____

	Local Variables:
	tab-width: 4
	End:
	vi: ts=4 sw=4
	vim: si
_____________________*/

